/**
 * 
 */
package org.seatiger.pso;

import java.util.*;


/**
 * ANPSO: Defines neighbourhoods by speciation. Particles are put into
 * neighbourhoods (species) based on their proximity in the decision 
 * space.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class ANPSO implements SpeciesClassifier
{
   /** The minimum number of steps two particles have to have been close
    * in order to be considered as the same species */
   private int minimumSteps = 3; 
   /** The minimum number of particles in order to form a species. Note that
    * values below 2 are meaningless for this method */
   private int minimumParticles = 2;
   /** The number of iterations that particle A has been "near" particle B.
    * Only after they have been "near" for {@link #minimumSteps} steps will they
    * be put in the same species
    */
   private HashMap<Particle, HashMap<Particle, MutableInteger>> 
         stepsNear = new HashMap<Particle, HashMap<Particle, MutableInteger>>();
   
   /**
    * Creates a new ParticleLinkingClassifier with the default 
    * minimum steps setting
    */
   public ANPSO() { }

   /**
    * Creates a new ParticleLinkingClassifier which waits the given 
    * number of steps before creating a species
    * 
    * @param minimumSteps The number of steps two particles need to be 
    * together before forming a species
    */
   public ANPSO(int minimumSteps)
   {
      this.minimumSteps = minimumSteps;
   }
   
   /* (non-Javadoc)
    * @see org.seatiger.pso.SpeciesClassifier#classify(java.util.Collection)
    */
   public Map<Particle, Integer> classify(Collection<Particle> particles)
   {
      List<Particle> particlesList;
      
      if (particles instanceof List)
         particlesList = (List<Particle>) particles;
      else
         particlesList = new ArrayList<Particle>(particles);
      
      // Remove any particles that aren't in the particles Collection
      stepsNear.keySet().retainAll(particlesList);
      
      double[][] particleDists = calculateDistances(particlesList);
      double limit = calculateAvgClosestDistance(particleDists);
      
      return buildSpecies(particlesList, particleDists, limit);
   }
   
   /**
    * Returns the minimum number of steps a particle must be near 
    * another before they form a species
    * 
    * @return Returns the minimum number of steps.
    */
   public int getMinimumSteps()
   {
      return minimumSteps;
   }

   /**
    * Sets the minimum number of steps a particle must be near 
    * another before they form a species
    * 
    * @param minimumSteps The minimum number of steps to set.
    */
   public void setMinimumSteps(int minimumSteps)
   {
      this.minimumSteps = minimumSteps;
   }

   /**
    * Returns the minimum number of particles required to form a species
    * 
    * @return Returns the minimum particles required in a species.
    */
   public int getMinimumParticles()
   {
      return minimumParticles;
   }

   /**
    * Sets the minimum number of particles required to form a species
    * 
    * @param minimumParticles Returns the minimum particles required in a species.
    */
   public void setMinimumParticles(int minimumParticles)
   {
      this.minimumParticles = minimumParticles;
   }

   /**
    * Builds the species based on the distances between particles
    * 
    * @param particles The particles to build into species
    * @param distances The distances between each particle and every
    * other particle
    * @param limit The maximum distance to join species
    * @return A Set of species. Each species is represented as a Set
    * of Integers which correspond to the indexes in <code>distances</code> 
    */
   private Map<Particle, Integer> buildSpecies(List<Particle> particles, 
         double[][] distances, double limit)
   {
      MutableInteger[][] stepsTogether = buildStepsNearArray(particles);
      
      // This function finds which particles are "reachable" from every
      // other particle and creates a graph. The groups are then defined
      // as the particles in each unconnected area of the graph
      ArrayList<ArrayList<Integer>> edges = 
         new ArrayList<ArrayList<Integer>>(distances.length);
      
      // This is an optimisation to reduce Java object creation
      Integer[] integers = new Integer[distances.length];
      for (int i = integers.length - 1; i >= 0; i--)
         integers[i] = new Integer(i);
      
      for (int i = 0; i < distances.length; i++)
      {
         ArrayList<Integer> currNode = new ArrayList<Integer>();
         edges.add(currNode);
         Particle p = particles.get(i);
         HashMap<Particle, MutableInteger> pTimeNear = stepsNear.get(p);
         
         for (int j = i + 1; j < distances.length; j++)
         {
            Particle other = particles.get(j);
            
            if (distances[i][j] < limit)
            {
               if (stepsTogether[i][j] != null)
               {
                  // Limit the time to minimumSteps * 2. This means
                  // that if minimumSteps is 3, it will take a particle
                  // being near another 3 steps to form a species. It 
                  // also means that a particle must be removed from a 
                  // species for 3 steps before it will be free
                  if (stepsTogether[i][j].getValue() < minimumSteps * 2)
                     stepsTogether[i][j].increment();
               }
               else
               {
                  stepsTogether[i][j] = new MutableInteger(1);
                  
                  if (pTimeNear == null)
                  {
                     pTimeNear = new HashMap<Particle, MutableInteger>();
                     stepsNear.put(p, pTimeNear);
                  }
                  pTimeNear.put(other, stepsTogether[i][j]);
               }
               
            }
            else if (stepsTogether[i][j] != null)
            {
               stepsTogether[i][j].decrement();
               if (stepsTogether[i][j].getValue() == 0)
               {
                  stepsTogether[i][j] = null;
                  pTimeNear.remove(other);
               }
            }
            
            if (pTimeNear != null && pTimeNear.size() == 0)
               stepsNear.remove(p);
            
            if (stepsTogether[i][j] != null &&
                  stepsTogether[i][j].getValue() >= minimumSteps)
               currNode.add(integers[j]);
         }
      }
      
      boolean[] visitedNodes = new boolean[distances.length];

      HashMap<Particle, Integer> species = new HashMap<Particle, Integer>();
      
      // Build the groups by connecting the edges
      for (int i = 0; i < distances.length; i++)
      {
         if (edges.get(i) == null || edges.get(i).isEmpty())
         {
            // This particle has already been added to a species
            // or it is not near any other particles
            continue;
         }
         
         Stack<Integer> stack = new Stack<Integer>();
         stack.push(integers[i]);
         
         do
         {
            Integer particleID = stack.pop();
            species.put(particles.get(particleID), integers[i]);
            visitedNodes[particleID] = true;
            
            for (Integer connectedNode : edges.get(particleID))
            {
               if (!visitedNodes[connectedNode])
               {
                  stack.push(connectedNode);
                  visitedNodes[connectedNode] = true;
               }
            }
            
            edges.set(particleID, null);
         }
         while (!stack.isEmpty());
      }

      
      // Remove all the species that are smaller than minimumParticles
      int[] speciesSizes = new int[distances.length];
      for (int speciesNum : species.values())
         speciesSizes[speciesNum]++;
      
      Iterator<Integer> iterator = species.values().iterator();
      while (iterator.hasNext())
      {
         if (speciesSizes[iterator.next()] < minimumParticles)
            iterator.remove();
      }
           
      return species;
   }

   /**
    * Returns an array containing how many steps each particle has been 
    * near each other particle. The array is upper-triangular - all elements 
    * below the diagonal are null. Elements where the particles are not
    * near each other will also be null (ie all elements where the result
    * is 0 are null). The particles in the array have the same index as 
    * the ones in <code>particles</code>. The MutableIntegers in this array
    * are references to the originals in {@link #stepsNear}, so modifying 
    * their values will modify the value of the MutableIntegers in 
    * {@link #stepsNear}
    * 
    * @param particles The list of particles to build the array from
    * @return An array of how many steps each particle has been near each 
    * other particle.  
    */
   private MutableInteger[][] buildStepsNearArray(List<Particle> particles)
   {
      MutableInteger[][] stepsTogether = 
         new MutableInteger[particles.size()][particles.size()];
      
      for (int particleNum = 0; particleNum < particles.size() - 1; particleNum++)
      {
         Particle particle = particles.get(particleNum);
         HashMap<Particle, MutableInteger> particleSteps = stepsNear.get(particle);
         
         if (particleSteps == null)
            continue;
         
         for (int otherParticleNum = particleNum + 1; 
               otherParticleNum < particles.size(); otherParticleNum++)
         {
            Particle other = particles.get(otherParticleNum);
            MutableInteger steps = particleSteps.get(other);
            
            if (steps == null)
               continue;
            else
               stepsTogether[particleNum][otherParticleNum] = steps;
         }
      }
      
      return stepsTogether;
   }
   
   /**
    * Calculates the average distance between each particle and its closest
    * neighbour 
    * 
    * @param particles The particles to calculate the distance of
    * @return The average distance between each particle and its closest
    * neighbour
    */
   public static double calculateAvgClosestDistance(Collection<Particle> particles)
   {
      List<Particle> particlesList;
      
      if (particles instanceof List)
         particlesList = (List<Particle>) particles;
      else
         particlesList = new ArrayList<Particle>(particles);
      
      return calculateAvgClosestDistance(calculateDistances(particlesList));
   }

   /**
    * Calculates the average distance between each particle and its closest
    * neighbour 
    * 
    * @param distances The distances between each particle. This array must
    * be square
    * @return The average distance between each particle and its closest
    * neighbour
    */
   public static double calculateAvgClosestDistance(double[][] distances)
   {
      double closestDistanceSum = 0;
      
      for (int i = 0; i < distances.length; i++)
      {
         double closestDistance = Double.MAX_VALUE;
         
         for (int j = 0; j < distances.length; j++)
         {
            if (i == j)
               continue;
            
            if (distances[i][j] < closestDistance)
               closestDistance = distances[i][j];
         }
         
         closestDistanceSum += closestDistance;
      }
      
      return closestDistanceSum / distances.length;
   }
   
   
   /**
    * Calculates the distance between each particle and every other
    * 
    * @param particles The particles to calculate the distances for
    * @return An array of distances indexed by the particle index
    */
   public static double[][] calculateDistances(List<Particle> particles)
   {
      double[][] particleDists = new double[particles.size()][particles.size()];
      
      for (int particleNum = particles.size() - 1; particleNum >= 0; 
            particleNum--)
      {
         Particle p = particles.get(particleNum);
         
         for (int otherParticleNum = particles.size() - 1; 
               otherParticleNum >= 0; otherParticleNum--)
         {
            Particle other = particles.get(otherParticleNum);
            
            if (p == other)
               continue;
            
            double distance = VectorMath.length(VectorMath.subtract(
                  p.getLocation(), other.getLocation()));
            
            particleDists[particleNum][otherParticleNum] = distance;
         }
      }
      
      return particleDists;
   }
}
