package org.seatiger.pso;

import java.util.*;

/**
 * Represents a brange in the tree
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class BranchNode extends TreeNode
{
   /** The child nodes of this node */
   private TreeNode[] children = null;
   /** The extents in each dimension of this node */
   private Range[] extents = null;
   /** Optimisation: cache the minimum of each extent range */
   private double[] extentMins = null;
   /** Optimisation: cache the maximum of each extent range */
   private double[] extentMaxs = null;
   
  
   /**
    * Creates a new TreeNode representing the given point
    *
    * @param extents The range of this node in each dimension
    * @throws IllegalArgumentException if extents has more than 
    * 32 elements (an implementation limitation)
    */
   public BranchNode(Range[] extents)
   {
      super(extents.length);
      
      if (extents.length > 32)
         throw new IllegalArgumentException("Cannot represent > 32 dimensions");
      
      this.extents = new Range[extents.length];
      System.arraycopy(extents, 0, this.extents, 0, extents.length);
      
      extentMins = new double[extents.length];
      extentMaxs = new double[extents.length];
      
      for (int i = 0; i < extents.length; i++)
      {
         extentMins[i] = extents[i].getMin();
         extentMaxs[i] = extents[i].getMax();
      }
      
      this.children = new TreeNode[(int) Math.pow(2, extents.length)];
   }

   /**
    * Creates a new BranchNode with a specified child. This is the only
    * way to add a BranchNode (as opposed to adding individual LeafNodes)
    *
    * @param extents The range of this node in each dimension
    * @param firstChild A child to add to the branch
    * @throws IllegalArgumentException if <code>firstChild</code> does not 
    * have the same number of dimensions as <code>extents</code> or 
    * <code>firstChild</code> does not lie within <code>extents</code>
    */
   private BranchNode(Range[] extents, BranchNode firstChild)
   {
      this(extents);
      
      if (firstChild.getNumDimensions() != getNumDimensions())
         throw new IllegalArgumentException("Wrong number of dimensions");
      
      Range[] childExtents = firstChild.getExtents();
      double[] centrePoint = new double[extents.length];
      
      for (int i = 0; i < centrePoint.length; i++)
      {
         centrePoint[i] = (childExtents[i].getMin() + childExtents[i].getMax()) 
               / 2;
      }
      
      int cellIndex = getCellIndex(centrePoint);
      
      if (cellIndex == -1)
         throw new IllegalArgumentException("Child not within extents");
      
      children[cellIndex] = firstChild;
      firstChild.setParent(this);
   }
   
   /**
    * Returns the child specified by <code>childToGet</code>
    * 
    * @param childToGet An array specifying which child to get. Each array 
    * element represents the respective dimension. For each of these, 
    * a false indicates a position towards the top-corner, a true indicates
    * a position towards the bottom-right corner. 
    * 
    * <br><br>For example, in a 2D tree:
    * <li>getChild({false, false}) would return the top left corner node, 
    * <li>getChild({false, true}) would return the bottom left corner node,
    * <li>getChild({true, false}) would return the top right corner node,
    * <li>getChild({true, true}) would return the bottom right corner node.
    * @return The requested child node, or null if there is no node there
    * @throws IllegalArgumentException if childToGet has a different length
    * to the length of the data elements in the tree
    */
   public TreeNode getChild(boolean[] childToGet)
   {
      if (childToGet.length != getNumDimensions())
         throw new IllegalArgumentException("Wrong number of dimensions");
      
      return getChild(convertToCellIndex(childToGet));
   }
   
   /**
    * Returns the child node at index <code>childIndex</code>, where 0
    * represents the top-left corner and 2^{@link #getNumDimensions()} - 1
    * represents the bottom-right corner. 
    * 
    * @param childIndex The index of the child node to get
    * @return The requested child, or null if there is no child in that cell
    */
   public TreeNode getChild(int childIndex)
   {
      if (children != null)
         return children[childIndex];
      else 
         return null;
   }
   
   
   /**
    * Adds the node to this branch
    * 
    * @param node The node to add
    * @throws IllegalArgumentException if the node's point is not within
    * this BranchNode's extents
    */
   public void add(LeafNode node)
   {
      double[] point = node.getPoint();
      
      if (node.getNumDimensions() != getNumDimensions())
         throw new IllegalArgumentException("Wrong number of dimensions");

      int cellIndex = getCellIndex(point);
         
      if (cellIndex == -1)
         throw new IllegalArgumentException("Point is not within node extents");
      
      if (children[cellIndex] == null)
      {
         children[cellIndex] = node;
         node.setParent(this);
      }
      else if (children[cellIndex] instanceof BranchNode &&
            ((BranchNode) children[cellIndex]).contains(point))
         ((BranchNode) children[cellIndex]).add(node);
      else if (!children[cellIndex].equals(node))
         createNewBranch(node, cellIndex);
   }

   /**
    * Adds the given point to this tree
    * 
    * @param point The point to add
    * @throws IllegalArgumentException if point does not have the same number
    * of dimensions as the other points in the tree or the point being added 
    * is outside the extents of this node
    */
   public void add(double[] point)
   {
      add(new LeafNode(point));
   }
   
   /**
    * Returns the index of the cell the given point belongs in
    * 
    * @param point The point to get the cell index of
    * @return The cell that point would belong to, or -1 if point is
    * outside of the range of this node
    * @throws IllegalArgumentException if point does not have the same number
    * of dimensions as this BranchNode
    */
   public int getCellIndex(double[] point)
   {
      if (point.length != getNumDimensions())
         throw new IllegalArgumentException("Wrong number of dimensions");

      int cellRef = 0;
      int lastIndex = point.length - 1;
      
      for (int i = 0; i <= lastIndex; i++)
      {
         if (point[i] < extentMins[i] || point[i] > extentMaxs[i])
            return -1;
         
         double midPoint = (extentMaxs[i] + extentMins[i]) / 2;
         
         if (point[i] > midPoint)
            cellRef |= 1 << (lastIndex - i);
      }
      
      return cellRef;
   }
   
   /**
    * Converts an array specifying a cell into the cell index
    *  
    * @param cell The array detailing the cell location in each dimension
    * @return The corresponding cell index
    * @see #getChild(boolean[])
    */
   protected int convertToCellIndex(boolean[] cell)
   {
      int requestedCell = 0;
      
      for (boolean temp : cell)
      {
         requestedCell *= 2;
         
         if (temp)
            requestedCell++;
      }
      
      return requestedCell;
   }
   
   /**
    * Returns the extents of this node. All contained points lie within
    * this range
    * 
    * @return The node's extents
    */
   public Range[] getExtents()
   {
      Range[] result = new Range[extents.length];
      System.arraycopy(extents, 0, result, 0, extents.length);
      
      return result;
   }
   
   /**
    * Converts
    * @param newNode
    * @param cellIndex
    * @throws IllegalArgumentException if point does not have the same number
    * of dimensions as this BranchNode
    */
   private void createNewBranch(LeafNode newNode, int cellIndex)
   {
      TreeNode existing = children[cellIndex];
      double[] point = newNode.getPoint();
      double[] existingPoint = null;
      Range[] existingExtents = null;
      
      if (existing instanceof LeafNode)
         existingPoint = ((LeafNode) existing).getPoint();
      else if (existing instanceof BranchNode)
         existingExtents = ((BranchNode) existing).getExtents();
      
      int dimensions = getNumDimensions();
      
      int divisions = Integer.MAX_VALUE;
      
      for (int i = 0; i < dimensions; i++)
      {
         double min = extentMins[i];
         double max = extentMaxs[i];
      
         double pointValue = point[i];
         double existingValue;
         
         if (existingPoint != null)
            existingValue = existingPoint[i];
         else
         {
            if (pointValue < existingExtents[i].getMin())
               existingValue = existingExtents[i].getMin();
            else if (pointValue > existingExtents[i].getMax())
               existingValue = existingExtents[i].getMax();
            else
            {
               // pointValue is within the range of existing in
               // this dimension, thus we won't be able to find
               // a line that separates them (in this dimension)
               continue;
            }
         }
         
         for (int j = 0; j < divisions; j++)
         {
            double midPoint = (min + max) / 2;
            
            // The first if statement handles the boundary case of
            //
            // 0--------|--------1--------|--------2
            //          |existing|
            //        * <- new
            
            if (existingExtents != null && existingExtents[i].getMin() == midPoint
                  && pointValue <= midPoint)
               divisions = j;
            else if (existingValue <= midPoint && pointValue <= midPoint)
               max = midPoint;
            else if (existingValue > midPoint && pointValue > midPoint)
               min = midPoint;
            else
               divisions = j;
         }            
      }
      
      Range[] newNodeExtents = new Range[extents.length];
     
      // We only need to look at pointValue for this loop because if it
      // were possible to create a dividing line between newNode and 
      // existing at a lower number of divisions, we would have detected
      // it in the loop above. 
      for (int i = 0; i < point.length; i++)
      {
         double min = extentMins[i];
         double max = extentMaxs[i];
      
         double pointValue = point[i];
         
         for (int j = 0; j < divisions; j++)
         {
            double midPoint = (min + max) / 2;
            
            if (pointValue <= midPoint)
               max = midPoint;
            else if (pointValue > midPoint)
               min = midPoint;
         }

         newNodeExtents[i] = new Range(min, max);
      }
      
      BranchNode newBranch;
      
      if (existing instanceof BranchNode)
      {
         newBranch = new BranchNode(newNodeExtents, (BranchNode) existing);
      }
      else
      {
         newBranch = new BranchNode(newNodeExtents);
         newBranch.add((LeafNode) existing);
      }
      
      newBranch.add(newNode);
      
      children[cellIndex] = newBranch;  
      newBranch.setParent(this);
   }

   /**
    * Returns true if the area handled by this BranchNode contains
    * <code>point</code>
    * @param point The point to check
    * @return True if <code>point</code> is within the area of this BranchNode
    * @throws IllegalArgumentException if point does not have the same number
    * of dimensions as this BranchNode
    */
   public boolean contains(double[] point)
   {
      if (point.length != extents.length)
         throw new IllegalArgumentException("Wrong number of dimensions");
      
      for (int i = 0; i < point.length; i++)
      {
         if (!extents[i].contains(point[i]))
            return false;
      }
      
      return true;
   }
   
   /**
    * Returns all of the LeafNodes that exist within this BranchNode
    * 
    * @return A set containing all of the LeafNodes
    */
   public Set<LeafNode> getAllLeafNodes()
   {
      LinkedHashSet<LeafNode> result = new LinkedHashSet<LeafNode>();
      
      addAllLeafNodesTo(result);
      return result;
   }
   
   /**
    * Returns the LeafNodes that are within <code>radius</code> of 
    * <code>location</code>
    * 
    * @param location The point the returned LeafNodes are near
    * @param radius The maximum distance allowed from <code>location</code>
    * @return The set of LeafNodes that are within <code>radius</code>
    * of <code>location</code>
    * @throws IllegalArgumentException if location has a different number
    * of dimensions to the points stored within the tree. 
    */
   public Set<LeafNode> getAllLeafNodesNear(double[] location, double radius)
   {
      if (location.length != extents.length)
         throw new IllegalArgumentException("Location has wrong number of dimensions");
      
      LinkedHashSet<LeafNode> result = new LinkedHashSet<LeafNode>();
      
      Range[] boundingBox = new Range[extents.length];
      for (int i = 0; i < location.length; i++)
         boundingBox[i] = new Range(location[i] - radius, location[i] + radius);

      getAllLeafNodesNear(result, boundingBox, location, radius);
      
      return result;
   }
   
   /**
    * Adds the LeafNodes that are within <code>radius</code> of 
    * <code>location</code> to <code>resultSet</code>
    * 
    * @param resultSet The set to add the LeafNodes to. 
    * @param boundingBox The bounding box of [location - radius, location + radius]
    * in every dimension 
    * @param location The point the returned LeafNodes are near
    * @param radius The maximum distance allowed from <code>location</code>
    */
   private void getAllLeafNodesNear(Set<LeafNode> resultSet, Range[] boundingBox, 
         double[] location, double radius)
   {
      if (getParent() == null)
      {
         getChildLeafNodesNear(resultSet, location, radius);
         return;
      }
      
      boolean fullySurrounds = true;
      
      for (int i = 0; fullySurrounds && i < extents.length; i++)
      {
         if (!extents[i].contains(boundingBox[i]))
            fullySurrounds = false;
      }
     
      if (!fullySurrounds)
         getParent().getAllLeafNodesNear(resultSet, boundingBox, location, radius);
      else
         getChildLeafNodesNear(resultSet, location, radius);
   }
   
   /**
    * Adds all the LeafNodes that are within <code>radius</code> of 
    * <code>location</code> to <code>resultSet</code>. Note that this
    * function does not traverse up the tree. 
    * 
    * @param resultSet The set to add the LeafNodes to. 
    * @param location The point the returned LeafNodes are near
    * @param radius The maximum distance allowed from <code>location</code>
    */
   private void getChildLeafNodesNear(Set<LeafNode> resultSet,  
         double[] location, double radius)
   {
      // If we aren't within radius of location, just return
      double closestDistance = VectorMath.length(
            VectorMath.subtract(location, getClosestPointTo(location)));
      if (closestDistance > radius)
         return;
            
      for (int i = 0; i < children.length; i++)
      {
         if (children[i] instanceof BranchNode)
         {
            ((BranchNode) children[i]).getChildLeafNodesNear(resultSet,
                  location, radius);
         }
         else if (children[i] instanceof LeafNode)
         {
            LeafNode child = (LeafNode) children[i];
            
            if (VectorMath.length(VectorMath.subtract(location, child.getPoint()))
                  <= radius)
               resultSet.add(child);
         }
      }
   }
   
   /**
    * Returns the centre point of this BranchNode's extents
    * 
    * @return The centre point of this BranchNode's extents
    */
   public double[] getCentrePoint()
   {
      double[] result = new double[extents.length];
      
      for (int i = 0; i < result.length; i++)
         result[i] = (extentMins[i] + extentMaxs[i]) / 2;
      
      return result;
   }
   
   /**
    * Returns the location of the closest corner to location. 
    * 
    * @param location The point to find the closest corner to
    * @return The location of the closest corner
    */
   public double[] getClosestCornerTo(double[] location)
   {
      double[] result = new double[extents.length];
      
      for (int i = 0; i < extents.length; i++)
      {
         double minDelta = Math.abs(location[i] - extentMins[i]);
         double maxDelta = Math.abs(location[i] - extentMaxs[i]);
         
         if (minDelta < maxDelta)
            result[i] = extentMins[i];
         else
            result[i] = extentMaxs[i];
      }
      
      return result;
   }
   
   @Override
   public double[] getClosestPointTo(double[] location)
   {
      double[] result = new double[extents.length];
      
      for (int i = 0; i < extents.length; i++)
      {
         if (location[i] < extentMins[i])
            result[i] = extentMins[i];
         else if (location[i] > extentMaxs[i])
            result[i] = extentMaxs[i];
         else
            result[i] = location[i];
      }
      
      return result;
   }
   
   /**
    * Returns the furthest point from <code>location</code> that is
    * still within the extents of this BranchNode
    * 
    * @param location The location to find the furthest point from
    * @return The furthest point from <code>location</code>
    */
   public double[] getFurthestPointFrom(double[] location)
   {
      double[] result = new double[extents.length];
      
      for (int i = 0; i < extents.length; i++)
      {
         if (Math.abs(location[i] - extentMins[i]) > 
            Math.abs(location[i] - extentMaxs[i]))
            result[i] = extentMins[i];
         else
            result[i] = extentMaxs[i];
      }
      
      return result;
   }
   
   /**
    * Adds all of the LeafNodes to <code>set</code>
    * 
    * @param set The Set to add the LeafNodes to
    */
   private void addAllLeafNodesTo(Set<LeafNode> set)
   {
      for (TreeNode child : children)
      {
         if (child == null)
            continue;
         else if (child instanceof LeafNode)
            set.add((LeafNode) child);
         else if (child instanceof BranchNode)
            ((BranchNode) child).addAllLeafNodesTo(set);
      }
   }
   
   /**
    * Removes the given node from this branch. If this branch is no
    * longer needed, it will be removed from the tree, possibly causing
    * an existing child node to be promoted one level. 
    * 
    * @param node The node to remove
    * @return True if the node was removed, false if it is not a 
    * child of this node. 
    */
   public boolean remove(TreeNode node)
   {
      return replace(node, null);
   }
   
   /**
    * Replaces a node
    * 
    * @param oldNode The node to replace
    * @param newNode The node to replace <code>oldNode</code> with
    * @return True if the node was replaced, false if it is not a
    * child of this node. 
    */
   private boolean replace(TreeNode oldNode, TreeNode newNode)
   {
      int remainingNodes = 0;
      int lastNodeFound = 0;
      int indexToRemove = -1;
      
      for (int i = 0; i < children.length; i++)
      {
         if (children[i] == oldNode)
            indexToRemove = i;
         else if (children[i] != null)
         {
            lastNodeFound = i;
            remainingNodes++;
         }
      }
      
      if (indexToRemove == -1)
         return false;
      
      children[indexToRemove].setParent(null);
      children[indexToRemove] = newNode;
      
      if (newNode != null)
         newNode.setParent(this);
      
      if (remainingNodes <= 1 && getParent() != null && newNode == null)
      {
         if (remainingNodes == 1)
            getParent().replace(this, children[lastNodeFound]);
         else
            getParent().remove(this);
      }
      
      return true;
   }
}
