package org.seatiger.pso;

import java.io.*;
import java.util.*;

/**
 * Reads a CSV-formatted file as a regression source
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class CSVSource implements RegressionSource
{
   /** A mapping of the location to value of each test point */
   private HashMap<List<Double>, Double> pointValues = new HashMap<List<Double>, Double>();
   /** The list of test points that are known about */
   private double[][] testPoints;
   
   /**
    * Creates a new CSVSource, reading the data from <code>input</code>
    * 
    * @param input The reader to read the data from
    * @throws IOException If there is a problem reading the data
    */
   public CSVSource(Reader input) throws IOException
   {
      loadData(input);
   }
   
   /**
    * Creates a new CSVSource, reading the data from <code>inputFile</code>
    * 
    * @param inputFile The file to read the data from
    * @throws IOException If there is a problem reading the data
    */   
   public CSVSource(File inputFile) throws IOException
   {
      this(new FileReader(inputFile));
   }
   
   /**
    * Loads the data from <code>inputFile</code>
    * 
    * @param input The reader to load the data from
    * @throws IOException If there is a problem reading the data
    */
   private void loadData(Reader input) throws IOException
   {
      int numVariables = 0;
      boolean firstLine = true;
      String currLine;
      BufferedReader fileIn = new BufferedReader(input);
      ArrayList<double[]> pointBuffer = new ArrayList<double[]>();

      int lineNum = 0;
      
      do 
      {
         currLine = fileIn.readLine();
         if (currLine == null)
            break;
         
         lineNum++;
         StringTokenizer strtok = new StringTokenizer(currLine, ", \t");
         
         String trimmedLine = currLine.trim();
         
         // Skip blank lines or lines starting with #
         if (trimmedLine.length() == 0 || trimmedLine.charAt(0) == '#' || strtok.countTokens() == 0)
            continue;
         
         if (firstLine)
         {
            numVariables = strtok.countTokens() - 1;
            firstLine = false;
         }
         else if (strtok.countTokens() - 1 < numVariables)
            throw new NoSuchElementException("Insufficient tokens reading line " 
                  + lineNum);
         
         ArrayList<Double> location = new ArrayList<Double>(numVariables);
         double result = 0;
         
         for (int i = 0; i <= numVariables; i++)
         {
            String currToken = strtok.nextToken();
            
            try
            {
               if (i < numVariables)
                  location.add(new Double(currToken));
               else 
                  result = Double.parseDouble(currToken);
            }
            catch (NumberFormatException nfe)
            {
               throw new NumberFormatException("Row " + lineNum + 
                     " column " + (i + 1) + " (" + currToken + ")");
            }
         }

         pointValues.put(location, result);
         double[] tmp = new double[location.size()];
         
         for (int i = 0; i < tmp.length; i++)
            tmp[i] = location.get(i);
         pointBuffer.add(tmp);
      }
      while (currLine != null);

      fileIn.close();
      testPoints = pointBuffer.toArray(new double[0][]);
   }
   
   public double calculate(double[] location)
   {
      // This is necessary because calling .hashCode() on identical arrays 
      // will not return the same result, therefore we can't directly use
      // the array as a HashMap key
      ArrayList<Double> tmp = new ArrayList<Double>(location.length);
      for (double element : location)
         tmp.add(element);
      
      Double result = pointValues.get(tmp);
      
      if (result == null)
         throw new IllegalArgumentException("Value for " 
               + Arrays.toString(location) + " is unknown");
      
      return result.doubleValue();
      
   }

   public double[][] getTestPoints()
   {
      double[][] result = new double[testPoints.length][];
      
      for (int i = 0; i < result.length; i++)
      {
         result[i] = new double[testPoints[i].length];
         System.arraycopy(testPoints[i], 0, result[i], 0, testPoints[i].length);
      }
      
      return result;
   }
}
