package org.seatiger.pso;

import java.util.*;

/**
 * Represents the circle neighbourhood topology
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class CircleNeighborhood implements Neighborhood
{
   /** The number of neighbours each particle has */
   private int neighbors;

   /** The particles being managed by this neighbourhood */
   private ArrayList<Particle> particles = new ArrayList<Particle>();
   /** A map of each particle to its index in {@link #particles} */
   private LinkedHashMap<Particle, Integer> particleNums
         = new LinkedHashMap<Particle, Integer>();
   
   /**
    * Creates a new CircleNeighborhood where each particle has
    * <code>neighbors</code> neighbours. The neighborhoods are built
    * in the same order as the particles were registered. If 
    * <code>neighbors</code> is an odd number, each particle will have 
    * one more neighbour to its right than to its left. Otherwise the number
    * of neighbours on both sides will be equal  
    *
    * @param neighbors The nubmer of neighbours each particle has
    */
   public CircleNeighborhood(int neighbors)
   {
      this.neighbors = neighbors;
   }
   
   /**
    * Returns the number of neighbors for each particle 
    *
    * @return The number of neighbors for each particle.
    */
   public int getNeighbors()
   {
      return this.neighbors;
   }

   /**
    * Sets the number of neighbours for each particle 
    *
    * @param neighbors The number of neighbours for each particle
    */
   public void setNeighbors(int neighbors)
   {
      this.neighbors = neighbors;
   }

   /**
    * Returns the particles in the neighbourhood of <code>p</code>
    * 
    * @see org.seatiger.pso.Neighborhood#getNeighborhoodOf(org.seatiger.pso.Particle)
    */
   public Set<Particle> getNeighborhoodOf(Particle p)
   {
      Integer particleNum = particleNums.get(p);
      int numParticles = particles.size();
      
      if (particleNum == null)
         return null;
      
      int currParticleNum = particleNum - neighbors / 2;
      LinkedHashSet<Particle> result = new LinkedHashSet<Particle>();
      
      for (int i = 0; i < neighbors; i++)
      {
         // Get the particles, wrapping around the ends of the list
         result.add(particles.get(
               (numParticles + currParticleNum++) % numParticles));
      }
      
      return result;
   }

   /**
    * @see org.seatiger.pso.Neighborhood#register(org.seatiger.pso.Particle)
    */
   public void register(Particle p)
   {
      if (!particleNums.containsKey(p))
      {
         int particleNum = particles.size();
         particles.add(p);
         particleNums.put(p, particleNum);
      }
   }

   /**
    * @see org.seatiger.pso.Neighborhood#deregister(org.seatiger.pso.Particle)
    */
   public void deregister(Particle p)
   {
      Integer particleNum = particleNums.get(p);
      
      if (particleNum == null)
         return;

      particleNums.remove(p);
      particles.remove(particleNum.intValue());
      
      // Adjust the index numbers
      for (int i = particleNum; i < particles.size(); i++)
         particleNums.put(particles.get(i), i);
   }
   
   /**
    * @see Neighborhood#deregisterAll()
    */
   public void deregisterAll()
   {
      particles.clear();
      particleNums.clear();
   }
}
