package org.seatiger.pso;

import java.util.Arrays;

/**
 * DeJong's test function 1 - a parabolic surface of the given 
 * dimensionality. This function is designed to test what
 * the GA does in optimum conditions
 * 
 * The equation for this function is: -sum(x[i]^2) over i, where x is the 
 * location. The global optimum is at (0, 0, ... , 0) which has a fitness
 * of 0. All other locations have a negative fitness.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class DeJongF1 implements FitnessFunction
{
   /** The number of parameters (dimensions) in this problem */
   private int parameters;
   /** The preferred range of this fitness function */
   private final static Range range = new Range(-5.12, 5.12); 
   
   /**
    * Creates a DF1 fitness function 
    *
    * @param parameters How many parameters (dimensions) the problem
    * should be in 
    */
   public DeJongF1(int parameters)
   {
      this.parameters = parameters;
   }

   /**
    * Returns the fitness of the given location. 
    * 
    * @param location The location to test
    * @return The fitness value of that location
    * @see org.seatiger.pso.FitnessFunction#getFitness(double[])
    */
   public double getFitness(double[] location)
   {
      double sum = 0;
      
      for (int i = 0; i < parameters; i++)
         sum -= location[i] * location[i];
      
      return sum;
   }

   /**
    * Returns the number of parameters expected by this fitness
    * function. 
    * 
    * @return The number of parameters specified in the constructor
    * @see org.seatiger.pso.FitnessFunction#getNumVariables()
    */
   public int getNumVariables()
   {
      return parameters;
   }

   /**
    * Returns the preferred range of this fitness function, 
    * which is [-5.12, 5.12] for all parameters
    * 
    * @return The preferred range
    * @see org.seatiger.pso.FitnessFunction#getPreferredRanges()
    */
   public Range[] getPreferredRanges()
   {
      Range[] result = new Range[parameters];
      Arrays.fill(result, range);
      
      return result;
   }
   
   
}
