package org.seatiger.pso;

import java.util.Arrays;
import uchicago.src.sim.util.Random;

/**
 * De Jong's test function 4. This  shows how well a GA copes with 
 * noisy fitness functions. A large component of the fitness 
 * function is a gaussian random number - two particles visiting 
 * the same point will have different fitnesses.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class DeJongF4 implements FitnessFunction
{
   /** The number of parameters (dimensions) in this problem */
   private int parameters;
   
   /** The preferred range of this fitness function */
   private final static Range range = new Range(-1.28, 1.28); 
   
   /**
    * Creates a DF4 fitness function 
    *
    * @param parameters How many parameters (dimensions) the problem
    * should be in 
    */
   public DeJongF4(int parameters)
   {
      this.parameters = parameters;
      
      if (Random.normal == null)
         Random.createNormal(0, 1);
   }
   
   /**
    * Returns the number of parameters expected by this fitness
    * function. 
    * 
    * @return The number of parameters specified in the constructor
    * @see org.seatiger.pso.FitnessFunction#getNumVariables()
    */
   public int getNumVariables()
   {
      return parameters;
   }

   /**
    * Returns the fitness of the given location. 
    * 
    * @param location The location to test
    * @return The fitness value of that location
    * @see org.seatiger.pso.FitnessFunction#getFitness(double[])
    */
   public double getFitness(double[] location)
   {
      double sum = 0;
      
      for (int i = location.length - 1; i >= 0; i--)
      {
         double x = location[i];
         sum += (i + 1) * x * x * x * x + Random.normal.nextDouble();
      }
      
      return sum;
   }

   /**
    * Returns the preferred range of this fitness function, 
    * which is [-1.28, 1.28] for all parameters
    * 
    * @return The preferred range
    * @see org.seatiger.pso.FitnessFunction#getPreferredRanges()
    */
   public Range[] getPreferredRanges()
   {
      Range[] result = new Range[parameters];
      Arrays.fill(result, range);
      
      return result;
   }

}
