package org.seatiger.pso;

import java.awt.Color;
import java.util.*;

import uchicago.src.sim.util.Random;

public class ESPSO implements SpeciesClassifier
{
   /** The default age at which a natural species is converted to an artificial
    * one (this stops it catching surrounding particles) */
   private final static int DEFAULT_CONVERSION_AGE = 3;
   
   /** A velocity of {0, 0, ..., 0}. The value is never changed, but
    * we don't necessarily know how many dimensions the array should be
    * until {@link #classify(Collection)} is called */
   protected double[] STATIONARY = null;
   
   /** The SPSO we are enhancing */
   private SPSO spso;
   
   /** The Swarm we are working with */
   private Swarm swarm;
   
   /** The age at which a natural species will be converted to an 
    * artificial one */
   private int conversionAge = DEFAULT_CONVERSION_AGE;
   
   /** The number of particles to put in newly created species */
   private int maxNewSpeciesSize = 8;
   
   /** The minimum number of particles to put in newly created species */
   private int minNewSpeciesSize = 8;
   
   /** The maximum allowable population. */
   private int maxPopulation = Integer.MIN_VALUE;

   /** The species that have been created by this instance */
   protected ArrayList<ArrayList<Particle>> createdSpecies =
      new ArrayList<ArrayList<Particle>>(); 
   
   /** The minimum distance two species seeds can be apart while being
    * considered different species */ 
   private double epsilon = 0.01;
   
   /** The age of each species. The key is the location of the species seed,
    * the value is the number of ticks there has been a seed at that point */
   private HashMap<double[], Integer> speciesAges = new HashMap<double[], Integer>();
   
   /** The MovementModel for created particles */
   private MovementModel movementModel;
   
   /** The species allocations */
   private LinkedHashMap<Particle, ArrayList<Particle>> speciesMap = 
      new LinkedHashMap<Particle, ArrayList<Particle>>();
   
   /** If set, this will be used to display various information about 
    * the species
    */
   private LabelDisplay labelDisplay = null;
   
   public ESPSO(Swarm swarm, double radius)
   {
      this(swarm, new SPSO(radius, true));
   }
   
   public ESPSO(Swarm swarm, SPSO classifier)
   {
      this(swarm, classifier, swarm.getDefaultMovementModel());
   }
   
   public ESPSO(Swarm swarm, SPSO classifier, 
         MovementModel movementModel)
   {
      spso = classifier;
      spso.setUsePBest(true);
      this.swarm = swarm;
      this.movementModel = movementModel;
      swarm.addSwarmListener(new StepNotifier());
   }
   
   public SPSO getBaseSPSO()
   {
      return spso;
   }
   
  
   /**
    * @return Returns the labelDisplay.
    */
   public LabelDisplay getLabelDisplay()
   {
      return labelDisplay;
   }

   /**
    * @param labelDisplay The labelDisplay to set.
    */
   public void setLabelDisplay(LabelDisplay labelDisplay)
   {
      this.labelDisplay = labelDisplay;
   }

   /**
    * Returns the maximum population limit. By default this will be set to
    * twice the number of particles in the swarm when 
    * {@link #classify(Collection)} is first called. 
    * 
    * @return Returns the maximum population limit
    */
   public int getMaxPopulation()
   {
      if (maxPopulation == Integer.MIN_VALUE)
         return swarm.getNumParticles();
      else
         return maxPopulation;
   }

 
   
   /**
    * Sets the maximum population limit. 
    * 
    * @param maxPopulation The maximum population limitt.
    */
   public void setMaxPopulation(int maxPopulation)
   {
      this.maxPopulation = maxPopulation;
   }
   


   /**
    * @return Returns the conversionAge.
    */
   public int getConversionAge()
   {
      return conversionAge;
   }

   /**
    * @param conversionAge The conversionAge to set.
    */
   public void setConversionAge(int conversionAge)
   {
      this.conversionAge = conversionAge;
   }

   /**
    * @return Returns the newSpeciesSize.
    */
   public int getMaxNewSpeciesSize()
   {
      return maxNewSpeciesSize;
   }

   /**
    * @param newSpeciesSize The newSpeciesSize to set.
    */
   public void setMaxNewSpeciesSize(int newSpeciesSize)
   {
      this.maxNewSpeciesSize = newSpeciesSize;
   }
   
   

   /**
    * @return Returns the epsilon.
    */
   public double getEpsilon()
   {
      return epsilon;
   }

   /**
    * @param epsilon The epsilon to set.
    */
   public void setEpsilon(double epsilon)
   {
      this.epsilon = epsilon;
   }

   /**
    * @return Returns the minNewSpeciesSize.
    */
   public int getMinNewSpeciesSize()
   {
      return minNewSpeciesSize;
   }

   /**
    * @param minNewSpeciesSize The minNewSpeciesSize to set.
    */
   public void setMinNewSpeciesSize(int minNewSpeciesSize)
   {
      this.minNewSpeciesSize = minNewSpeciesSize;
   }

   /**
    * Updates the age of each species
    */
   private void updateAges()
   {
      ArrayList<Particle> seeds = 
         new ArrayList<Particle>(spso.getSpeciesSeeds());
      seeds.addAll(getCreatedSpeciesSeeds());
      
      if (seeds == null || seeds.size() == 0)
      {
         speciesAges.clear();
         return;
      }

      ArrayList<double[]> currPositions = new ArrayList<double[]>();
      
      for (Particle p : seeds)
      {
         double[] position;
         
         if (spso.isUsePBest())
            position = p.getPersonalBestLocation();
         else
            position = p.getLocation();
         
         currPositions.add(position);
         
         if (!speciesAges.containsKey(position))
            speciesAges.put(position, 1);
         else
            speciesAges.put(position, speciesAges.get(position) + 1);
      }
      
      speciesAges.keySet().retainAll(currPositions);
   }

   /**
    * Returns the seed particle for the created species
    * 
    * @return A collection containing the species seeds of the created
    * populations
    */
   private Collection<Particle> getCreatedSpeciesSeeds()
   {
      ArrayList<Particle> seeds = new ArrayList<Particle>();
      FitnessComparator comparator = new FitnessComparator(spso.isUsePBest());
      
      for (ArrayList<Particle> species : createdSpecies)
         seeds.add(Collections.max(species, comparator));
      
      return seeds;
   }
   
   /**
    * The original classify method (unsimplified). Converted species members
    * are created randomly instead of from the originalk species members. 
    * 
    * @see org.seatiger.pso.SpeciesClassifier#classify(java.util.Collection)
    */
   public Map<Particle, Integer> classify(Collection<Particle> particles)
   {
      return convertSpeciesMap(speciesMap);
   }

   private void buildSpeciesMap()
   {
      if (labelDisplay != null)
         labelDisplay.clearLabels();
      if (STATIONARY == null)
         STATIONARY = new double[swarm.getNumVariables()];
      
      // Initialise the variables from the getter methods, which 
      // calculate from the number of current particles in the swarm
      // if the value is not yet set. 
      if (maxPopulation == Integer.MIN_VALUE)
         maxPopulation = getMaxPopulation();
//      if (populationTarget == Integer.MIN_VALUE)
//         populationTarget = getPopulationTarget();
//      if (desiredNormals == Integer.MIN_VALUE)
//         desiredNormals = getDesiredNormals();
      
      if (movementModel == null)
         movementModel = swarm.getDefaultMovementModel();
      
      
      LinkedHashSet<Particle> mainPopulation = new LinkedHashSet<Particle>(swarm.getParticles());
      
      for (ArrayList<Particle> species : createdSpecies)
         mainPopulation.removeAll(species);
      
      Map<Particle, Integer> spsoClassification = spso.classify(mainPopulation);
      
      HashMap<Integer, ArrayList<Particle>> reverseMap = 
         new HashMap<Integer, ArrayList<Particle>>();
      speciesMap.clear();
      
      for (Particle p : spsoClassification.keySet())
      {
         Integer speciesNum = spsoClassification.get(p);
         
         if (!reverseMap.containsKey(speciesNum))
            reverseMap.put(speciesNum, new ArrayList<Particle>());
         
         reverseMap.get(speciesNum).add(p);
         speciesMap.put(p, reverseMap.get(speciesNum));
      }
      
      for (ArrayList<Particle> species : createdSpecies)
      {
         for (Particle p : species)
            speciesMap.put(p, species);
      }
      
      updateAges();
      
      // Build a list of the species seeds, sorted in order of their
      // fitness. We will use this to kill underperforming species
      ArrayList<Particle> seeds = new ArrayList<Particle>(spso.getSpeciesSeeds());
      Collection<Particle> createdSeeds = getCreatedSpeciesSeeds(); 
      seeds.addAll(createdSeeds);
      Collections.sort(seeds, new FitnessComparator(true));
      Collections.reverse(seeds);
      
      int populationSoFar = 0;
      ArrayList<Particle> seedsToRemove = new ArrayList<Particle>();
      
      for (int i = 0; i < seeds.size(); i++)
      {
         Particle seed = seeds.get(i);
         double[] point = seed.getPersonalBestLocation();;
         
         if (labelDisplay != null)
         {
            String label = speciesAges.get(point) + " (" + i + ", " + populationSoFar + ")";
            
            if (createdSeeds.contains(seed))
               label += " C";

            labelDisplay.addLabel(point, label, Color.BLUE);
         }
         
         boolean shouldDelete = false;
         
         if (speciesAges.get(point) >= conversionAge)
         {
            // Make sure the seed is not too close to some other seed
            for (int j = 0; j < i && !shouldDelete; j++)
            {
               Particle other = seeds.get(j);
               double[] otherPoint = other.getPersonalBestLocation();
               
               if (VectorMath.length(VectorMath.subtract(point, otherPoint))
                     < epsilon)
                  shouldDelete = true;
            }
         
            // If it is a duplicate species, delete it
            if (shouldDelete)
            {
               seedsToRemove.add(seed);
               
               Iterator<ArrayList<Particle>> iterator = createdSpecies.iterator();
               while (iterator.hasNext())
               {
                  if (iterator.next().contains(seed))
                  {
                     iterator.remove();
                     break;
                  }
               }

               for (Particle p : speciesMap.get(seed))
               {
                  speciesMap.remove(p);
                  swarm.removeParticle(p);
               }
               
               continue;
            }
         }
         
         // Change it to an artificial species so it doesn't trap particles
         // investigating a nearby optimum
         if (speciesAges.get(point) == conversionAge
               && !createdSeeds.contains(seed))
         {
            convertSpecies(seed, speciesMap);
         }

         populationSoFar += speciesMap.get(seed).size();
      }
      
      seeds.removeAll(seedsToRemove);

      int particlesNeeded = maxPopulation - swarm.getNumParticles();
      
      swarm.addParticles(particlesNeeded, movementModel);
   }
   
   /**
    * Converts a particle -> particle set species map to a particle -> int
    * species map
    * 
    * @param speciesMap The map to convert
    * @return The converted map
    */
   private LinkedHashMap<Particle, Integer> convertSpeciesMap(LinkedHashMap<Particle, ArrayList<Particle>> speciesMap)
   {
      int nextSpeciesNum = 0;
      LinkedHashMap<Particle, Integer> result = 
         new LinkedHashMap<Particle, Integer>();
      
      for (ArrayList<Particle> species : speciesMap.values())
      {
         if (result.containsKey(species.get(0)))
            continue;
         
         Integer speciesNum = new Integer(nextSpeciesNum++);
         
         for (Particle p : species)
            result.put(p, speciesNum);
      }
      return result;
   }      

   /** 
    * Converts a species so it no longer influences the rest of the swarm
    *  
    * @param seed The seed particle of the species to convert
    * @param speciesMap The map of particles to their species
    */
   private void convertSpecies(Particle seed, 
         Map<Particle, ArrayList<Particle>> speciesMap)
   {
      ArrayList<Particle> speciesParticles = speciesMap.get(seed);
      Collections.sort(speciesParticles, new FitnessComparator(true));
      ArrayList<Particle> newSpeciesParticles = 
         new ArrayList<Particle>();
      
      int firstToKeep;
      if (maxNewSpeciesSize == 0)
         firstToKeep = 0;
      else 
         firstToKeep = speciesParticles.size() - maxNewSpeciesSize;
      
      double maxDist = 0;
      
//      System.out.println("ftk:" + firstToKeep + " size: " + speciesParticles.size());
      
      // Handle the case where there's only one particle.
      if (speciesParticles.size() == 1)
         maxDist = spso.getRadius();
      
      for (int i = 0; i < speciesParticles.size(); i++)
      {
         Particle p = speciesParticles.get(i);
         
         if (i < firstToKeep)
         {
            // Make sure we don't accidently remove the seed
            // This can happen if there are many particles with equal 
            // fitness. Instead we will delete the particle in last
            // position (which would not normally be deleted) and put the
            // seed particle in its place.  
            if (seed == p)
            {
               p = speciesParticles.get(speciesParticles.size() - 1);
               speciesParticles.set(speciesParticles.size() - 1, seed);
            }

            speciesMap.remove(p);
            swarm.removeParticle(p);
         }
         else
         {
            newSpeciesParticles.add(speciesParticles.get(i));
            speciesMap.put(p, newSpeciesParticles);
            
            double distFromSeed = VectorMath.length(
                  VectorMath.subtract(p.getPersonalBestLocation(), 
                        seed.getPersonalBestLocation()));
            if (distFromSeed > maxDist)
               maxDist = distFromSeed;
         }
      }
      
      int desiredParticles = speciesParticles.size();
      
      if (maxNewSpeciesSize != 0 && desiredParticles > maxNewSpeciesSize)
         desiredParticles = maxNewSpeciesSize;
      
      if (minNewSpeciesSize != 0 && desiredParticles < minNewSpeciesSize)
         desiredParticles = minNewSpeciesSize;
      
      int numToAdd = desiredParticles - newSpeciesParticles.size();
      for (int j = 0; j < numToAdd; j++)
      {
         double[] location = swarm.sanitizeLocation(
               generateRandomPoint(seed.getPersonalBestLocation(),
                     maxDist));
         double[] velocity = generateRandomPoint(STATIONARY, maxDist);

         Particle temp = new Particle(location, velocity, swarm, 
               movementModel);
         swarm.addParticle(temp);
         newSpeciesParticles.add(temp);
         speciesMap.put(temp, newSpeciesParticles);
      }
      
      createdSpecies.add(newSpeciesParticles);
   }
   
   /**
    * Returns a random point within maxDistance of seedPoint
    * 
    * @param seedPoint The point to generate a point near
    * @param maxDistance The maximum allowable distance from seedPoint
    * @return A poing within maxDistance of seedPoint
    */
   protected double[] generateRandomPoint(double[] seedPoint, double maxDistance)
   {
      double r = Random.uniform.nextDouble() * maxDistance;
      double[] point = new double[seedPoint.length];
      Arrays.fill(point, r);
      
      for (int i = 1; i < point.length; i++)
      {
         double angle = Random.uniform.nextDouble() * Math.PI * 2;
         
         for (int j = 0; j < i; j++)
            point[j] *= Math.cos(angle);
         
         point[i] *= Math.sin(angle);
      }
      
      for (int i = 0; i < point.length; i++)
         point[i] += seedPoint[i];
      
      return point;
   }
   
   
   /**
    * Listens for SwarmEvents and notifies the outer class
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private class StepNotifier implements SwarmListener
   {
      public void initialized(SwarmEvent e)
      {
         step(e);
      }

      public void step(SwarmEvent e)
      {
         buildSpeciesMap();
      }
   }
}
