package org.seatiger.pso;

import java.util.*;


public class ESPSO2 implements SpeciesClassifier, SwarmListener
{
   /** The swarm we are working for */
   private Swarm swarm;
   /** Used to track the species */
   private SpeciesTracker tracker;
   /** The movement model that's using this class */
   private MovementModel model = null;
   
   private SPSO spso;
   /** How many ticks between each redistribution */
   private int redistInterval = 10;

   private int minSpeciesSize = 1;
   /** The maximum number of particles in a species */
   private int pMax = 10;
   
   /** Stores information for each species */
   private HashMap<Species, SpeciesInfo> speciesInfo = 
      new HashMap<Species, SpeciesInfo>();
   
   public ESPSO2(Swarm swarm, double radius)
   {
      this.swarm = swarm;
      spso = new SPSO(radius, true);
      spso.setPMax(pMax);
      
      tracker = new SpeciesTracker(new SpeciesClassifier()
      {
         public Map<Particle, Integer> classify(Collection<Particle> particles)
         {
            return spso.classify(particles);
         }
      }, swarm);
      
      swarm.addSwarmListener(this);
   }
   
   public Map<Particle, Integer> classify(Collection<Particle> particles)
   {
      return tracker.classify(particles);
   }

   private void redistributeParticles()
   {
      List<SpeciesInfo> speciesToRedist = getSpeciesToRedist();
      
      setSpeciesPerf(speciesToRedist);
      setSpeciesAlloc(speciesToRedist);
      allocateParticles(speciesToRedist);
      updateSpeciesInfos();
   }
   
   private List<SpeciesInfo> getSpeciesToRedist()
   {
      ArrayList<SpeciesInfo> speciesToRedist = new ArrayList<SpeciesInfo>();
      
      for (Species species : tracker.getAllSpecies())
      {
         SpeciesInfo info = speciesInfo.get(species);
         if (info != null)
            speciesToRedist.add(info);
      }
      
      return speciesToRedist;
   }
   
   private void setSpeciesPerf(List<SpeciesInfo> speciesInfos)
   {
      for (SpeciesInfo info : speciesInfos)
      {
         double fitnessDelta = info.species.getSeedP().getPersonalBestFitness()[0] - 
               info.lastFitness;
            
         if (fitnessDelta < 0)
            fitnessDelta = 0;
         
         info.performance = fitnessDelta;
      }
   }
   
   private void setSpeciesAlloc(Collection<SpeciesInfo> speciesInfos)
   {
      setSpeciesAllocFitness(speciesInfos);
   }
   
   private void setSpeciesAllocFitness(Collection<SpeciesInfo> speciesInfos)
   {
      ArrayList<SpeciesInfo> sortedInfos = new ArrayList<SpeciesInfo>();
      sortedInfos.addAll(speciesInfos);
      Collections.sort(sortedInfos, Collections.reverseOrder(
            new Comparator<SpeciesInfo>()
                  {
                     public int compare(SpeciesInfo a, SpeciesInfo b)
                     {
                        double aFitness = a.species.getSeedFitness()[0];
                        double bFitness = b.species.getSeedFitness()[0];
                        
                        if (aFitness < bFitness)
                           return -1;
                        else if (aFitness > bFitness)
                           return 1;
                        else
                           return 0;
                     }
                  }));
      
      
      double minFitness = Double.POSITIVE_INFINITY;
      double maxFitness = Double.NEGATIVE_INFINITY;
      double totalFitness = 0;
      int availableParticles = 0;
      
      for (SpeciesInfo info : sortedInfos)
      {
         if (info.species.getSeedFitness()[0] < minFitness)
            minFitness = info.species.getSeedFitness()[0];
         if (info.species.getSeedFitness()[0] > maxFitness)
            maxFitness = info.species.getSeedFitness()[0];
         
         totalFitness += info.species.getSeedFitness()[0];
         availableParticles += info.species.getParticles().size() - minSpeciesSize;
      }
      
      totalFitness -= minFitness * speciesInfos.size();
      double excess = 0;      
      
      int totalAllocated = 0;
      
      for (SpeciesInfo info : sortedInfos)
      {
         double fitness = info.species.getSeedFitness()[0];
         double share = excess + availableParticles * 
               (fitness - minFitness) / totalFitness;
         
         info.allocation = share > pMax ? pMax : (int) Math.floor(share);
         excess = share - info.allocation;         
         
         totalAllocated += info.allocation;
         
         info.allocation += minSpeciesSize;
      }
      
      System.out.println("Allocated: " + totalAllocated + " available: " + availableParticles);
   }
   
   private void setSpeciesAllocNonDom(Collection<SpeciesInfo> speciesInfos)
   {
      List<Collection<SpeciesInfo>> sortedInfos = nonDomSort(speciesInfos);
      
      int availableParticles = 0;
      int totalParticles = 0;
      for (SpeciesInfo info : speciesInfos)
      {
         availableParticles += info.species.getParticles().size() - minSpeciesSize;
         totalParticles += info.species.getParticles().size();
      }
      
      if (availableParticles < 0)
         availableParticles = totalParticles / minSpeciesSize;
      
      double multiplier = 1;
      double total = 0;
      
      for (Collection<SpeciesInfo> level : sortedInfos)
      {
         total += multiplier * level.size();
         multiplier /= 2;
      }
      
      double rank1Particles = availableParticles / total;
      if (rank1Particles > pMax)
         availableParticles = (int) (pMax / rank1Particles * availableParticles);
      
      int rank = 1;
      multiplier = 1;
      
      int totalAllocated = 0;
      double remainder = 0;
      
      for (Collection<SpeciesInfo> level : sortedInfos)
      {
         double exactProportion = minSpeciesSize + availableParticles * multiplier / total;
         int proportion = (int) exactProportion;
         double proportionFPart = exactProportion - proportion;
         
//         if (multiplier != 1)
//            System.out.print(", ");
//         System.out.print(proportion);
         
         for (SpeciesInfo info : level)
         {
            if (proportion > totalParticles - totalAllocated)
               proportion = totalParticles - totalAllocated;
            
            info.allocation = proportion;
            
            remainder += proportionFPart;
            if (remainder >= 1 && totalAllocated + proportion < totalParticles)
            {
               info.allocation++;
               remainder--;
            }
            
//            System.out.print(info.species.hashCode() + " R: " + rank);
//            System.out.print(" P: " + info.performance);
//            System.out.print(" F: " + info.species.getSeedFitness()[0]);
//            System.out.println(" A: " + info.allocation);
            
            totalAllocated += info.allocation;
         }
            
         multiplier /= 2;   
         rank++;   
      }
      
      System.out.println("Tick " + swarm.getCurrTick() + ": " + totalAllocated + "/" + totalParticles);
   }
   
   private void allocateParticles(Collection<SpeciesInfo> speciesToRedist)
   {
      int particleBalance = 0;
      ArrayList<Species> needExtra = new ArrayList<Species>();
      
      for (SpeciesInfo info : speciesToRedist)
      {
         Species species = info.species;
         
         if (info.allocation > species.getParticles().size())
            needExtra.add(species);
         else if (info.allocation < species.getParticles().size())
         {
            ArrayList<Particle> speciesParticles = new ArrayList<Particle>();
            speciesParticles.addAll(species.getParticles());
            Collections.sort(speciesParticles, new FitnessComparator(true));
            
            ArrayList<Particle> particlesToRemove = new ArrayList<Particle>();
            particlesToRemove.addAll(speciesParticles.subList(0, 
                  speciesParticles.size() - info.allocation));
            
            if (model == null)
               model = particlesToRemove.get(0).getMovementModel();
            
            speciesParticles.removeAll(particlesToRemove);
            for (Particle p : particlesToRemove)
               swarm.removeParticle(p);
            
            particleBalance -= particlesToRemove.size();
            
            if (speciesParticles.isEmpty())
               species.die();
            else   
               species.setParticles(speciesParticles);
         }
      }
      
      for (Species species : needExtra)
      {
         SpeciesInfo info = speciesInfo.get(species); 
         int needed = info.allocation - species.getParticles().size();
         //double maxDist = VectorMath.distance(info.lastSeedPos, species.getSeed());
         double maxDist =0.1;
         addToSpecies(species, maxDist, needed);
         
         particleBalance += needed;
      }

      swarm.addParticles(-particleBalance, model);
//      for (Particle p : pool)
//      {
//         p.setLocation(swarm.getRandomLocation());
//         p.setVelocity(swarm.getRandomVelocity());
//         p.setPBestToCurrent();
//         p.setStepsUntilActive(1);
//      }
   }
   
   private void updateSpeciesInfos()
   {
      speciesInfo.clear();
      for (Species species : tracker.getAllSpecies())
         speciesInfo.put(species, new SpeciesInfo(species));
   }
 

//   private void killUnderperformingSpecies(Collection<Species> allSpecies, int needed)
//   {
//      ArrayList<Species> killable = new ArrayList<Species>();
//      
//      for (Species s : allSpecies)
//      {
//         if (s.getLastMovement() < swarm.getCurrTick() - stagnationThreshold)
//            killable.add(s);
//      }
//      
//      Collections.sort(killable, new Comparator<Species>()
//      {
//         public int compare(Species a, Species b)
//         {
//            double delta = a.getSeedP().getPersonalBestFitness()[0] -
//                  b.getSeedP().getPersonalBestFitness()[0];
//            
//            if (delta < 0)
//               return -1;
//            else if (delta > 0)
//               return 1;
//            else
//               return 0;
//         }
//      });
//      
//      Iterator<Species> iterator = killable.iterator();
//      
//      while (needed > 0 && iterator.hasNext())
//      {
//         Species species = iterator.next();
//         SpeciesInfo info = speciesInfo.get(species);
//         
//         if (info != null)
//         {
//            needed -= species.getParticles().size();
//            info.kill = true;
//         }
//      }
//   }
   
   
   
   /**
    * Adds the needed number of particles to the species, taking them from pool
    * 
    * @param species The species to add the particles to
    * @param maxDist The maximum distance away from the seed the new particles can be
    * @param needed The number of extra particles needed
    */
   private void addToSpecies(Species species, double maxDist, int needed)
   {
      final double[] STATIONARY = new double[swarm.getNumVariables()];
      
      ArrayList<Particle> particles = new ArrayList<Particle>();
      particles.addAll(species.getParticles());
      
      for (int i = 0; i < needed; i++)
      {
         double[] location = swarm.getRandomLocation(species.getSeed(), maxDist);
         
         Particle p = new Particle(location, STATIONARY, swarm, model);
//         System.out.println(Arrays.toString(p.getVelocity()));
//         System.out.println(VectorMath.distance(p.getLocation(), species.getSeed()));
         swarm.addParticle(p);
      }
   }


   public void step(SwarmEvent e)
   {
      if (swarm.getCurrTick() % redistInterval == 0)
         redistributeParticles();
   }
   
   public void initialized(SwarmEvent e) { }
   
   /**
    * Sorts the SpeciesInfos into non-dominated levels, with the fittest/fastest 
    * improving species in level 0
    * 
    * @param infos The infos to sort
    * @return The sorted infos
    */
   private List<Collection<SpeciesInfo>> nonDomSort(Collection<SpeciesInfo> infos)
   {
      HashMap<double[], SpeciesInfo> mapping = new HashMap<double[], SpeciesInfo>();
      for (SpeciesInfo currInfo : infos)
      {
         double[] fitness = new double[2];
         fitness[0] = currInfo.performance;
         fitness[1] = currInfo.species.getSeedFitness()[0];

         mapping.put(fitness, currInfo);
      }
      
      List<Collection<double[]>> sortedFitnesses = Swarm.nonDomSort(mapping.keySet());
      ArrayList<Collection<SpeciesInfo>> result = 
         new ArrayList<Collection<SpeciesInfo>>(sortedFitnesses.size());
      
      for (Collection<double[]> fitnessLevel : sortedFitnesses)
      {
         ArrayList<SpeciesInfo> currLevelInfos = new ArrayList<SpeciesInfo>();
         result.add(currLevelInfos);
         
         for (double[] fitness : fitnessLevel)
            currLevelInfos.add(mapping.get(fitness));
      }
         
      return result;
   }
   
   private boolean dominates(SpeciesInfo a, SpeciesInfo b)
   {
      if (a.species.getSeedFitness()[0] > b.species.getSeedFitness()[0] &&
            a.performance >= b.performance)
         return true;
      else if (a.species.getSeedFitness()[0] >= b.species.getSeedFitness()[0] &&
            a.performance > b.performance)
         return true;
      else
         return false;
   }
   
   private class SpeciesInfoDom
   {
      public int dominatingSpecies = 0;
      public ArrayList<SpeciesInfo> dominatedSpecies = new ArrayList<SpeciesInfo>();
   }
   
   private class SpeciesInfo
   {
      public double lastFitness;
      public double[] lastSeedPos;
      public double performance;
      public int allocation;
      public Species species;
      
      public SpeciesInfo(Species species)
      {
         lastFitness = species.getSeedP().getPersonalBestFitness()[0];
         lastSeedPos = species.getSeed();
         this.species = species;
      }
   }
}
