package org.seatiger.pso;

import java.util.*;

/**
 * Watches for changes in particle fitness - indicates that the fitness landscape has
 * changed
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class FitnessChangeListener extends Observable implements SwarmListener
{
   /** Whether or not to use the species best points */
   private final static boolean USE_SPECIES_BESTS = true;
   
   /** The swarm controlling things */
	private Swarm swarm;
	/** The function we're checking */
	private FitnessFunction fitnessFunc;
	/** The number of points to check */
	private int numToCheck = 5;
   /** The points to check */
   private double[][] points;
   /** The known fitnesses of the points */
   private double[] fitnesses;

	
	/**
	 * Creates a new FitnessChangeListener
	 *
	 * @param swarm The swarm to monitor
	 * @param fitnessFunc The fitness function to monitor
	 */
	public FitnessChangeListener(Swarm swarm, FitnessFunction fitnessFunc)
	{
		this.swarm = swarm;
		this.fitnessFunc = fitnessFunc;
      swarm.addSwarmListener(this);
	}

	public void step(SwarmEvent e)
	{
      boolean changeDetected;
    
      if (USE_SPECIES_BESTS)
         changeDetected = checkUsingSpeciesBest();
      else 
         changeDetected = checkUsingRandomPoints(); 
      
      if (changeDetected)
      {
         swarm.clearPersonalBests();
         setChanged();
         notifyObservers();
      }
      
   }
   
   /**
    * Checks for changes using the species bests 
    * 
    * @return True if a change was detected, false otherwise
    */
   private boolean checkUsingSpeciesBest()
   {
      ArrayList<Particle> particlesToCheck = new ArrayList<Particle>(numToCheck);
      Neighborhood neighborhood = swarm.getDefaultMovementModel().getNeighborhood();
      
      if (neighborhood instanceof SpeciesNeighborhood)
      {
   		SpeciesNeighborhood speciesNeighborhood = (SpeciesNeighborhood) neighborhood;
   		
   		List<org.seatiger.pso.Species> species = speciesNeighborhood.getAllSpecies();
   		if (species == null)
   			return false;
   		
   		ArrayList<Species> speciesList = new ArrayList<Species>(species.size());
   		for (org.seatiger.pso.Species currSpecies : species)
   			speciesList.add(new Species(currSpecies.getParticles()));
   		
   		Collections.sort(speciesList);
   
   		for (int count = 0; count < numToCheck && count < speciesList.size(); count++)
   		{
   			particlesToCheck.add(
   			         speciesList.get(speciesList.size() - count - 1).getBest());
   		}
   	}
      else
      {
         ArrayList<Particle> allParticles = new ArrayList<Particle>();
         allParticles.addAll(swarm.getParticles());
         
         particlesToCheck.add(
               Collections.max(allParticles, new FitnessComparator(true)));

         Collections.shuffle(allParticles);
         Iterator<Particle> iterator = allParticles.iterator();
         
         for (int i = 1; i < numToCheck && iterator.hasNext(); i++)
            particlesToCheck.add(iterator.next());
      }
      
      for (Particle p : particlesToCheck)
      {
         if (Math.abs(fitnessFunc.getFitness(p.getPersonalBestLocation()) -
               p.getPersonalBestFitness()[0]) > Swarm.EPSILON)
         {
            return true;
         }
      }
      
      return false;
   }
	
   /**
    * Checks using random points
    * 
    * @return True if a change was detected, false otherwise
    */
   private boolean checkUsingRandomPoints()
   {
      if (points == null)
      {
         Range[] ranges = swarm.getVariableRange();
         
         points = new double[numToCheck][swarm.getNumVariables()];
         fitnesses = new double[numToCheck];
         
         for (int i = 0; i < points.length; i++)
         {
            for (int j = 0; j < points[i].length; j++)
               points[i][j] = ranges[j].getMin() + Math.random() * ranges[j].getSize();
            
            fitnesses[i] = fitnessFunc.getFitness(points[i]);
         }
      }
      
      boolean changeDetected = false;
      
      // We could exit on change detected, but then we'd have to recalculate
      // the new fitnesses anyway. Easier just to finish the loop
      for (int i = 0; i < points.length; i++)
      {
         double currentFitness = fitnessFunc.getFitness(points[i]);
         
         if (Math.abs(currentFitness - fitnesses[i]) > Swarm.EPSILON)
         {
            fitnesses[i] = currentFitness;
            changeDetected = true;
         }
      }
      
      return changeDetected;
   }
   
	public void initialized(SwarmEvent e) { }
	
	/**
	 * Represents a single species. 
	 *  
	 * @author Stefan Bird <stbird@seatiger.org>
	 */
	private static class Species implements Comparable<Species>
	{
		/** The member particles of this species */
		private Collection<Particle> members;
		/** The fittest particle of this species */
		private Particle best;
		
		/**
		 * Creates a new Species with the given members
		 *
		 * @param members The members of this species
		 */
		public Species(Collection<Particle> members)
		{
			this.members = members;
			recalcBest();
		}
		
		/**
		 * Returns the fittest member of the species
		 * 
		 * @return The fittest member of the species
		 */
		public Particle getBest()
		{
			return best;
		}
		
		/**
		 * Recalculates the fittest member of the species
		 */
		public void recalcBest()
		{
			best = Collections.max(members, new FitnessComparator(true));
		}

		/**
		 * Compares this species to another bast on the personal best fitness
		 * of the best particle. 
		 */
		public int compareTo(Species other)
		{
			double delta = best.getPersonalBestFitness()[0] - 
				other.getBest().getPersonalBestFitness()[0];
			
			if (delta < 0)
				return -1;
			else if (delta > 0)
				return 1;
			else
				return 0;
		}
	}
}
