package org.seatiger.pso;

import java.util.Comparator;

/**
 * Compares two particles based on either their current or personal 
 * best fitnesses. Note this class only tests on a single fitness - 
 * it does not return whether one particle dominates another. Domination
 * is not transitive (ie x dominates y, y dominates z does not imply 
 * x dominates y). Non-transitive comparisons breaks the contract for 
 * the Comparator interface. 
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class FitnessComparator implements Comparator<Particle>
{
   /** Whether to use the current or PBest fitness in the comparison */
   private boolean usePBestFitness = false;
   /** The fitness function number that is compared on */
   private int fitnessNum = 0;
   
   /**
    * Creates a new FitnessComparitor using the first fitness
    * 
    * @param usePBestFitness Whether to use the personal best fitness
    * in the comparison. If true, the PBest will be used, otherwise the
    * current fitness will be used
    */ 
   public FitnessComparator(boolean usePBestFitness)
   {
      this(usePBestFitness, 0);
   }
   
   /**
    * Creates a new FitnessComparitor using the first fitness
    * 
    * @param usePBestFitness Whether to use the personal best fitness
    * in the comparison. If true, the PBest will be used, otherwise the
    * current fitness will be used
    * @param fitnessNum The index of the fitness function to use for the
    * comparison
    */ 
   public FitnessComparator(boolean usePBestFitness, int fitnessNum)
   {
      this.usePBestFitness = usePBestFitness;
      this.fitnessNum = fitnessNum;
   }

   /**
    * Compares the personal best fitness of two particles. 
    *  
    * @see java.util.Comparator#compare(Object, Object)
    */
   public int compare(Particle o1, Particle o2)
   {
      double[] fitness1;
      double[] fitness2; 
      
      if (usePBestFitness)
      {
         fitness1 = o1.getPersonalBestFitness();
         fitness2 = o2.getPersonalBestFitness();
      }
      else
      {
         fitness1 = o1.getFitness();
         fitness2 = o2.getFitness();
      }

      if (fitness1[fitnessNum] > fitness2[fitnessNum])
         return 1;
      else if (fitness1[fitnessNum] < fitness2[fitnessNum])
         return -1;
      else 
         return 0;
   }

   /**
    * Returns true if this comparitor is using the personal best fitness 
    * 
    * @return True if this comparitor is using the personal best fitness,
    * false if it is using the current fitness.  
    */
   public boolean isUsingPBestFitness()
   {
      return usePBestFitness;
   }
   
   /**
    * Returns the index of the fitness function being used to 
    * compare the particles. 
    *  
    * @return The fitness function number
    */
   public int getFitnessNum()
   {
      return fitnessNum;
   }
}