package org.seatiger.pso;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;

import uchicago.src.sim.engine.CustomProbeable;
import uchicago.src.sim.gui.*;

/**
 * Displays a 2D fitness landscape from a top-down view
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class FitnessDisplay2D implements Displayable, Probeable
{
   /** The Swarm to display */
   protected Swarm swarm;
   
   /** The image of the fitness landscape */ 
   protected BufferedImage fitnessLandscape;
   
   /** The display surface being drawn on */   
   protected DisplaySurface displaySurface = null;   
   
   /** The decision space range to display on the screen */
   protected Range[] displayRange = null;
   
   /** The width of the display in pixels */
   protected int width = 0;
   /** The height of the display in pixels */
   protected int height = 0;
      
   /**
    * Creates a new FitnessDisplay2D
    *
    * @param swarm The Swarm to display the fitness landscape of
    * @param displaySurface The DisplaySurface to display on 
    */
   public FitnessDisplay2D(Swarm swarm, DisplaySurface displaySurface)
   {
      this.swarm = swarm;
      this.displaySurface = displaySurface;
      fitnessLandscape = null;
   }
   
   /**
    * Paints the display onto the given Graphics instance
    * 
    * @param simGraphics The SimGraphics to paint on 
    */
   public void drawDisplay(SimGraphics simGraphics)
   {
      if (!swarm.isRunning())
         return;
      
      if (displayRange == null)
         displayRange = swarm.getVariableRange();
      
      width = simGraphics.getDisplayWidth();
      height = simGraphics.getDisplayHeight();
      
      Graphics2D g = simGraphics.getGraphics();
      
      // If we don't have a fitness landscape or it's the wrong size,
      // and we are not already generating a landscape for this size, 
      // generate a new landscape
      if (fitnessLandscape == null || fitnessLandscape.getWidth() != width
            || fitnessLandscape.getHeight() != height)
         generateLandscape();
      
      g.drawImage(fitnessLandscape, 0, 0, width - 1, height - 1, 0, 0,
            fitnessLandscape.getWidth() - 1, fitnessLandscape.getHeight() - 1, 
            displaySurface);
   }
   
 
   /**
    * @see uchicago.src.sim.gui.Probeable#getObjectsAt(int, int)
    */
   public ArrayList getObjectsAt(int x, int y)
   {
      ArrayList<Object> result = new ArrayList<Object>();
      
      Point2D temp = convertToModel(x, y);
      double[] modelLoc = {temp.getX(), temp.getY()};
      result.add(new FitnessValue(modelLoc, swarm.calculateFitness(modelLoc)));
      
      return result;
   }

   /* (non-Javadoc)
    * @see uchicago.src.sim.gui.Probeable#setMoveableXY(uchicago.src.sim.gui.Moveable, int, int)
    */
   public void setMoveableXY(Moveable moveable, int x, int y) { }

   /* (non-Javadoc)
    * @see uchicago.src.sim.gui.Displayable#getDisplayableInfo()
    */
   public ArrayList getDisplayableInfo() 
   {
      return new ArrayList();
   }

   /* (non-Javadoc)
    * @see uchicago.src.sim.gui.Displayable#viewEventPerformed(uchicago.src.sim.gui.ViewEvent)
    */
   public void viewEventPerformed(ViewEvent evt) { }

   /**
    * @see uchicago.src.sim.gui.Displayable#getSize()
    */
   public Dimension getSize()
   {
      return new Dimension(400, 400);
   }
   
   /**
    * Converts a model point to screen coordinates
    * 
    * @param x1 The point in the model in dimension 1
    * @param x2 The point in the model in dimension 2
    * @return The location on screen of the specified model coordinate, or 
    * null if {@link #drawDisplay(SimGraphics)} has not been called
    */
   protected Point convertToScreen(double x1, double x2)
   {
      if (displayRange == null)
         return null;
      
      double[] coordinates = {x1, x2};
      Range[] visualRange = {new Range(0, width), new Range(0, height)};
      
      coordinates = convert(displayRange, coordinates, visualRange);
      
      return new Point((int) coordinates[0], (int) coordinates[1]);
   }
   
   /** 
    * Converts a screen point to model coordinates
    * 
    * @param x The x screen coordinate
    * @param y The y screen coordinate
    * @return The location in the model of the specified coordinate, or
    * null if {@link #drawDisplay(SimGraphics)} has not been called
    */
   protected Point2D.Double convertToModel(int x, int y)
   {
      if (displayRange == null)
         return null;
      
      double[] coordinates = {x, y};
      Range[] visualRanges = {new Range(0, width), new Range(0, height)};
      
      coordinates = convert(visualRanges, coordinates, displayRange);
      
      return new Point2D.Double(coordinates[0], coordinates[1]);
   }
   
   /**
    * Converts a set of coordinates from one range to another
    * 
    * @param inputRanges The range of <code>coordinates</code>
    * @param coordinates The coordinates to convert
    * @param outputRanges The range to convert <code>coordinates</code> into
    * @return <code>coordinates</code> converted from <code>inputRanges</code>
    * to <code>outputRanges</code>
    * 
    * @throws IllegalArgumentException if <code>inputRanges</code>, 
    * <code>coordinates</code> and <code>outputRanges</code> do not have
    * the same length
    */
   public static double[] convert(Range[] inputRanges, 
         double[] coordinates, Range[] outputRanges)
   {
      if (inputRanges.length != coordinates.length || 
            outputRanges.length != coordinates.length)
         throw new IllegalArgumentException("All parameter arrays must have" +
                " the same length");
      
      double[] result = new double[coordinates.length];
      for (int i = coordinates.length - 1; i >= 0; i--)
      {
         double inputRange = inputRanges[i].getMax() - inputRanges[i].getMin();
         double outputRange = outputRanges[i].getMax() - 
               outputRanges[i].getMin();
         result[i] = (coordinates[i] - inputRanges[i].getMin()) / inputRange
               * outputRange + outputRanges[i].getMin();
      }
      
      return result;
   }
   
   /**
    * Generates a new BufferedImage showing the fitness landscape
    */
   private void generateLandscape()
   {
      double bestFitness = Double.NEGATIVE_INFINITY;
      double worstFitness = Double.POSITIVE_INFINITY;
      double[][] fitnessValues = new double[width][height];
      
      Range x1Range = swarm.getVariableRange()[0];
      Range x2Range = swarm.getVariableRange()[1];
      
      double x1Step = (x1Range.getMax() - x1Range.getMin()) 
            / width;
      double x2Step = (x2Range.getMax() - x2Range.getMin()) 
            / height;
      double x1 = x1Range.getMin();
      double[] location = new double[2];
      double[] fitness;
      
      for (int x = 0; x < width; x++)
      {
         location[0] = x1;
         double x2 = x2Range.getMin();
         
         for (int y = 0; y < height; y++)
         {
            location[1] = x2;
            fitness = swarm.calculateFitness(location);

            fitnessValues[x][y] = fitness[0];
            if (fitness[0] < worstFitness)
               worstFitness = fitness[0];
            if (fitness[0] > bestFitness)
               bestFitness = fitness[0];
            
            x2 += x2Step;
         }
         
         x1 += x1Step;
      }
      
      int[] imageBuffer = new int[width * height];
      double fitnessRange = bestFitness - worstFitness;
      
      for (int x = 0; x < width; x++)
      {
         for (int y = 0; y < height; y++)
         {
            int pixelValue = (int) (255 * 
                  (fitnessValues[x][y] - worstFitness) / fitnessRange);
            
            // By multiplying pixel value by 0x010101, we copy it into
            // the R and G channels as well. By ORing with 0xff000000
            // we make the pixel fully opaque. Combining these two 
            // operations gives us an opaque pixel that ranges from 
            // black to white
            imageBuffer[y * width + x] = 0xff000000 | pixelValue * 0x010101;
         }
      }
      
      fitnessLandscape = new BufferedImage(width, height,
            BufferedImage.TYPE_INT_ARGB);
      fitnessLandscape.setRGB(0, 0, width, height, imageBuffer, 0, 
            width);
   }

   /**
    * Allows the user to see the fitness of a particular point. This
    * class is only needed because Repast doesn't know how to probe
    * Strings and double[]s
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   public static class FitnessValue implements CustomProbeable
   {
      /** The fitness value as a string */
      private final String fitness;
      /** The location as a string */
      private final String location;
      
      /** 
       * Creates a new FitnessValue for the given fitness
       *
       * @param location The location of the point being queried
       * @param fitness The fitness of the point being queried
       */
      public FitnessValue(double[] location, double[] fitness)
      {
         this.location = Swarm.toString(location);
         this.fitness = Swarm.toString(fitness);
      }
      
      /**
       * @see uchicago.src.sim.engine.CustomProbeable#getProbedProperties()
       */
      public String[] getProbedProperties()
      {
         return new String[] {"Location", "Fitness"};
      }
      
      /**
       * Returns the fitness of the point as a String
       * 
       * @return The fitness value as a String
       */
      public String getFitness()
      {
         return fitness;
      }
      
      /**
       * Returns the location of the point as a String
       * 
       * @return The location of the point as a String
       */
      public String getLocation()
      {
         return location;
      }
   }
}
