package org.seatiger.pso;

/**
 * Interface for returning the fitness of a certain point in the 
 * search space. Fitness functions must be thread-safe.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public interface FitnessFunction
{
   /**
    * Returns the number of variables in the decision space 
    * 
    * @return The number of input variables required by this
    * function
    */
   public int getNumVariables();
   
   /**
    * Returns the fitness of the given point. This function may be asked
    * to evaluate points outside the ranges given by {@link #getPreferredRanges()}.
    * This function must return a valid value in these situations (ie it
    * cannot throw an Exception). If a point is out of bounds, this function
    * should return {@link Double#NEGATIVE_INFINITY}
    * 
    * @param location The location of the point to test. This will have
    * the same number of elements as {@link #getNumVariables()}
    * @return The fitness of the given point - larger numbers indicate fitter
    * points
    * @see FitnessFunction#getNumVariables()
    */
   public double getFitness(double[] location);

   /**
    * Returns the preferred range of each dimension for this fitness function.
    * The preferred range is the range of values that the search should be
    * constrained by. The optimum solution should normally exist within the
    * preferred range, although this is not a requirement.
    *  
    * <p>Note: This function returns the <i>preferred</i> range only. At its
    * discression the search system may ask the fitness function to evaluate
    * positions outside of this range. {@link FitnessFunction#getFitness(double[])} must
    * return a valid value in this case. 
    * @return An array of {@link Range}s, one for each dimension of the fitness
    * function
    */
   public Range[] getPreferredRanges();
}
