package org.seatiger.pso;

import java.util.*;

/**
 * Implements GCPSO adapted for local neighbourhoods as described in Peer E. et. al,
 * "Using Neighbourhoods with Guaranteed Convergence PSO". The original GCPSO 
 * algorithm is described in Bergh, F. et. al "A New Locally Convergent Particle
 * Swarm Optimiser"
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class GCPSO extends ConstrictionPSO
{
   /**
    * The current p value for each particle
    */
   private HashMap<Particle, PInfo> pValues = new HashMap<Particle, PInfo>();
   
   /**
    * Creates a new GCPSO
    *
    * @param swarm The swarm to control
    * @param neighborhood The neighbourhood model to use
    */
   public GCPSO(Swarm swarm, Neighborhood neighborhood)
   {
      super(swarm, neighborhood);
   }

   @Override
   public double[] getNextVelocity(Particle particle)
   {
      ArrayList<Particle> neighbors = new ArrayList<Particle>();
      neighbors.addAll(getNeighborhood().getNeighborhoodOf(particle));
      Collections.sort(neighbors, new FitnessComparator(true));
      
      if (neighbors.get(neighbors.size() - 1) == particle)
      {
//         System.out.println("Active - " + neighbors.size());
         // This particle is the best in its neighborhood, so activate the GCPSO
         // extensions
         
         PInfo pInfo = pValues.get(particle);
         pInfo.check();
         
         // Constrict the particle's original velocity
         double[] dampenedVel = VectorMath.multiply(getConstriction(), particle.getVelocity());
         
         // This velocity will return the particle to its pBest location
         double[] velocity = VectorMath.subtract(particle.getPersonalBestLocation(), 
               particle.getLocation());

         // Get a random vector whose size is controlled by the particle's P value 
         double[] random = getRandomVector(pInfo.getP());
         
         return VectorMath.add(dampenedVel, velocity, random);
      }
      else
         return super.getNextVelocity(particle);
   }

   @Override
   public void deregister(Particle particle)
   {
      pValues.remove(particle);
      super.deregister(particle);
   }

   @Override
   public void register(Particle particle)
   {
      pValues.put(particle, new PInfo(particle));
      super.register(particle);
   }
   
   /**
    * Returns a random vector where each component is between -max and max
    * 
    * @param max The largest size for each dimension
    * @return A random vector
    */
   protected double[] getRandomVector(double max)
   {
      double[] result = new double[swarm.getNumVariables()];
      
      // Fill result with values from -max -> max
      for (int i = 0; i < result.length; i++)
         result[i] = uchicago.src.sim.util.Random.uniform.nextDouble() * max * 2 - max;
      
      return result;
   }

   
   /**
    * Maintains the current P value for a particular particle
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   protected static class PInfo
   {
      /** The number of failures before P is decreased */
      public static int FC = 5;
      /** The number of successes before P is increased */
      public static int SC = 15;
      
      /** The size of the space to try jumping to */
      private double p = 1;
      /** The count of successes we've had */
      private int successes = 0;
      /** The count of failures we've had */
      private int failures = 0;
      /** The particle we're monitoring */
      private Particle particle;
      /** The last pBest value */
      private double lastPBest;
      
      /**
       * Creates a new PInfo
       *
       * @param particle The particle to monitor
       */
      public PInfo(Particle particle)
      {
         this.particle = particle;
         if (particle.getPersonalBestFitness() != null)
            lastPBest = particle.getPersonalBestFitness()[0];
         else
            lastPBest = Double.NEGATIVE_INFINITY;
      }
      
      /**
       * Checks whether there has been an improvement or not in the particle's pBest 
       */
      public void check()
      {
         double pBest = particle.getPersonalBestFitness()[0];
         
         if (pBest > lastPBest)
            success();
         else
            failure();
         
         lastPBest = pBest;
      }
      
      /**
       * Registers a success
       */
      private void success()
      {
         successes++;
         failures = 0;
         
         if (successes > SC)
         {
            successes = 0;
            p *= 2;
         }
      }

      /**
       * Registers a failure 
       */
      private void failure()
      {
         failures++;
         successes = 0;
         
         if (failures > FC)
         {
            failures = 0;
            p /= 2;
         }
      }

      /**
       * Returns the current value of P
       * 
       * @return P
       */
      public double getP()
      {
         return p;
      }
   }
}
