package org.seatiger.pso;

import java.awt.Color;
import java.util.*;

import uchicago.src.sim.analysis.Plot;
import uchicago.src.sim.engine.SimModel;

/**
 * A regression evaluator that graphs the best regressions found so far
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class GraphingRegressionEvaluator extends RegressionEvaluator
{
   /** The instance used to compare entries in {@link #regressions} */
   private static final RegArrayComparitor COMPARITOR = 
      new RegArrayComparitor();
   
   /** The best regressions, sorted from worst-to-best. The first element
    * of each double[] is the fitness of the regression, the rest of the
    * elements are the constants (or location) of that regression */
   private ArrayList<double[]> regressions = new ArrayList<double[]>();
   
   /** The number of regressions to show on the graph */
   private int regressionsToShow = 5;
   
   /** The graph being displayed */
   private Plot graph; 
   
   /** The Repast model used for the plot */
   private SimModel model;
   
   /** The number of points to plot for the regression */
   private int pointsToPlot = 100;
   
   /**
    * Creates a new GraphingRegressionEvaluator 
    *  
    * @param regression The regression to tune
    * @param source The source data to tune the regression against
    * @param model The model for the graph
    */
   public GraphingRegressionEvaluator(Regression regression, 
         RegressionSource source, SimModel model)
   {
      super(regression, source);
      this.model = model;
      initialiseGraph();
      updateGraph();
   }

   /** 
    * Initialises the graph
    */
   private void initialiseGraph()
   {
      if (graph != null)
         graph.dispose();
      
      graph = new Plot("Top regressions", model);
      graph.addLegend(0, "Target points", Color.BLUE);
      
      for (int i = 1; i <= regressionsToShow; i++)
      {
         int value = 255 - (int) (255 * (i * 1.0 - 1) / (regressionsToShow - 1));
         
         graph.addLegend(i, "Rank " + i, new Color(value, 0, 0));
      }
      
      graph.display();
   }
   
   @Override
   public double getFitness(double[] location)
   {
      double fitness = super.getFitness(location);
      
      if (regressions.size() < regressionsToShow || 
            regressions.get(0)[0] < fitness)
      {
         double[] newEntry = new double[location.length + 1];
         newEntry[0] = fitness;
         System.arraycopy(location, 0, newEntry, 1, location.length);
         
         if (regressions.size() < regressionsToShow)
            regressions.add(newEntry);
         else
            regressions.set(0, newEntry);
         
         Collections.sort(regressions, COMPARITOR);
      }
      
      return fitness;
   }
   
   /** 
    * Updates the regression graph
    */
   public void updateGraph()
   {
      RegressionSource source = getRegressionSource();
      Regression regression = getRegression();
      
      double[][] testPoints = source.getTestPoints();
      double xMax = Double.NEGATIVE_INFINITY;
      double xMin = Double.POSITIVE_INFINITY;
      double yMax = Double.NEGATIVE_INFINITY;
      double yMin = Double.POSITIVE_INFINITY;
      
      graph.setConnected(false, 0);
      for (int i = 0; i <= regressionsToShow; i++)
         graph.clear(i);
      
      for (int i = 0; i < testPoints.length; i++)
      {
         double y = source.calculate(testPoints[i]);
         
         if (testPoints[i][0] > xMax)
            xMax = testPoints[i][0];
         if (testPoints[i][0] < xMin)
            xMin = testPoints[i][0];
         
         if (y > yMax)
            yMax = y;
         if (y < yMin)
            yMin = y;
                     
         graph.plotPoint(testPoints[i][0], y, 0);
      }
      
      double step = (xMax - xMin) / pointsToPlot;
      double[] pos = new double[1];

      for (int i = 0; i < regressions.size(); i++)
      {
         graph.setConnected(true, i);
         
         double[] constants = new double[regression.getNumConstants()];
         double[] regDetails = regressions.get(i);
         System.arraycopy(regDetails, 1, constants, 0, constants.length);
         regression.setConstants(constants);
         
         for (double x = xMin; x <= xMax; x += step)
         {
            pos[0] = x;
            double y = regression.calculate(pos);
            
            if (y > yMax)
               yMax = y;
            if (y < yMin)
               yMin = y;
            
            graph.plotPoint(x, y, regressions.size() - i);
         }
      }
      
      graph.setXRange(xMin, xMax);
      graph.setYRange(yMin, yMax);
      graph.updateGraph();
   }

   /**
    * Returns the number of regressions being shown on the graph
    * 
    * @return Returns the regressions to show.
    */
   public int getRegressionsToShow()
   {
      return regressionsToShow;
   }

   /**
    * Sets the number of regressions to show on the graph
    * 
    * @param regressionsToShow The number of regressions to show
    * on the graph
    */
   public void setRegressionsToShow(int regressionsToShow)
   {
      if (this.regressionsToShow == regressionsToShow)
         return;
      
      this.regressionsToShow = regressionsToShow;
      
      int excess = regressions.size() - regressionsToShow;
      
      if (excess > 0)
      {
         for (int i = 0; i < regressionsToShow; i++)
            regressions.set(i, regressions.get(i + excess));
      }
      
      initialiseGraph();
   }

   /**
    * Compares the first element of a double array, in which is stored
    * the fitness of the regression. 
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private static class RegArrayComparitor implements Comparator<double[]>
   {
      public int compare(double[] o1, double[] o2)
      {
         if (o1[0] < o2[0])
            return -1;
         else if (o1[0] > o2[0])
            return 1;
         else
            return 0;
      }
   }

   @Override
   protected void finalize() throws Throwable
   {
      if (graph != null)
         graph.dispose();
      
      super.finalize();
   }



}
