package org.seatiger.pso;

import java.util.*;

import uchicago.src.sim.util.Random;

public class GravityPSO implements MovementModel, SwarmListener
{
   /** The particles handled by this MovementModel */
   protected LinkedHashSet<Particle> particles = new LinkedHashSet<Particle>();

   protected HashMap<Particle, double[]> nextVels = new HashMap<Particle, double[]>();
   
   protected HashMap<Particle, Double> dupWeights = new HashMap<Particle, Double>();
   
   protected Neighborhood neighborhood = new VonNeumannNeighborhood();
   
   /** The total phi */
   protected double phi = 4.10;
   /** The constriction */
   protected double constriction = ConstrictionPSO.calculateConstriction(phi);
   
   protected double bestEverFitness = Double.NEGATIVE_INFINITY;
   
   protected double worstEverFitness = Double.POSITIVE_INFINITY;
   
   protected Swarm swarm;
   
   protected Range[] decisionRange = null;
   
   protected boolean useGravityModel = false;
   
   protected final static double G = 1;
   
   public GravityPSO(Swarm swarm)
   {
      this.swarm = swarm;
      swarm.addSwarmListener(this);
   }
   
   public void deregister(Particle particle)
   {
      particles.remove(particle);
      dupWeights.remove(particle);
      nextVels.remove(particle);
      neighborhood.deregister(particle);
   }

   public double[] getNextVelocity(Particle particle)
   {
      return nextVels.get(particle);
   }
   
   private double[] calcNextVelGrav(Particle particle)
   {
      double[] velocity = particle.getVelocity();
      double mass = calcRelativeFitness(particle.getPersonalBestFitness()[0]);
      
      for (Particle other : particles)
      {
         double otherMass = calcRelativeFitness(other.getPersonalBestFitness()[0]);
         double attraction = otherMass;// - mass;
         
         double[] delta = 
            VectorMath.subtract(other.getPersonalBestLocation(), particle.getLocation());
         double distance = VectorMath.length(delta);
         
         if (distance == 0)// || fitnessDelta < 0)
            continue;
       
         double[] force = VectorMath.multiply(G * attraction / (distance * distance), 
               VectorMath.unitVector(delta));
         
         velocity = VectorMath.add(velocity, force);
      }
      
      return velocity;
   }
   
   public Neighborhood getNeighborhood()
   {
      return neighborhood;
   }
   
   /**
    * Calulates the relative fitness where 0 is the fitness of the current worst 
    * known point and 1 is the fitness of the current best known point
    * 
    * @param fitness The fitness to calculate
    * @return The relative fitness on a scale of 0 to 1
    */
   protected double calcRelativeFitness(double fitness)
   {
      return (fitness - worstEverFitness) / (bestEverFitness - worstEverFitness);
   }
   
   private double[] calcNextVelConst(Particle particle)
   {

      
      double[] velocity = VectorMath.multiply(constriction, 
            VectorMath.add(particle.getVelocity(), getForce(particle)));
      
      // Restrict the velocity
      for (int i = 0; i < velocity.length; i++)
         velocity[i] = Math.min(velocity[i], decisionRange[i].getSize());
      
      return velocity;
   }
   
   private void calcDupWeights()
   {
      for (Particle p : particles)
      {
         double weight = 0;
         
         for (Particle other : particles)
         {
            double distance = VectorMath.length(VectorMath.subtract(
                  p.getPersonalBestLocation(), other.getPersonalBestLocation()));
            
            weight += 1 / (distance + 1);
         }
         
         dupWeights.put(p, weight);
      }
   }

   protected double[] getRandomVector(double max)
   {
      double[] result = new double[swarm.getNumVariables()];
      
      for (int i = 0; i < result.length; i++)
         result[i] = Random.uniform.nextDouble() * max;
      
      return result;
   }
   
   public double[] getForce(Particle particle, Particle other)
   {
      double[] delta = 
         VectorMath.subtract(other.getPersonalBestLocation(), particle.getLocation());
      double distance = VectorMath.length(delta);
      double fitnessDelta = other.getPersonalBestFitness()[0] - particle.getFitness()[0];
      
      double force = 0;

      if (distance != 0 && fitnessDelta > 0)
         force = phi * fitnessDelta / (distance * distance);
      
      return VectorMath.multiply(force, delta);
   }
   
   public double[] getForce(Particle particle)
   {
      double[] velocity = particle.getVelocity();
      double[] numerator = new double[velocity.length];
      double[] denominator = new double[velocity.length];
      
      Set<Particle> neighbors = neighborhood.getNeighborhoodOf(particle);
      double randMax = phi / neighbors.size();
      
      List<Particle> distanceRanks = rankDistances(particle);
      int rank = 1;
      
      double closestDist = VectorMath.length(VectorMath.subtract(
            particle.getLocation(), distanceRanks.get(0).getPersonalBestLocation()));
      double furthestDist = VectorMath.length(VectorMath.subtract(particle.getLocation(), 
            distanceRanks.get(distanceRanks.size() - 1).getPersonalBestLocation()));
      
      
//      for (Particle other : particles)
      for (Particle other : distanceRanks)
      {
//         if (other.getPersonalBestFitness()[0] < particle.getPersonalBestFitness()[0])
//            continue;
         
         double absDist = VectorMath.length(VectorMath.subtract(
               particle.getLocation(), other.getPersonalBestLocation()));
 
//         double distance = VectorMath.length(VectorMath.subtract(
//               other.getPersonalBestLocation(), particle.getLocation()));
//         double distance = rank++;
         double distance = (absDist - closestDist) / (furthestDist - closestDist); 
         
     
         if (distance == 0)// || fitnessDelta < 0)
            continue;
         
         // FDR would be fitnessDelta / distance
         // Gravity is fitnessDelta / distance^2   
//         double weight = fitnessDelta / (distance * distance);
//         double weight = calcRelativeFitness(other.getPersonalBestFitness()[0]) / 
//               (distance * distance * dupWeights.get(other)/* * distance */);
         
         double weight = calcRelativeFitness(other.getPersonalBestFitness()[0])/ 
               (distance * distance);// * dupWeights.get(other));
         
         double[] random = getRandomVector(randMax * weight);
         
         numerator = VectorMath.add(numerator, VectorMath.multiply(
               random, other.getPersonalBestLocation()));
         denominator = VectorMath.add(denominator, random);
      }
      
      double[] attractor = VectorMath.divide(numerator, denominator);
      double[] delta = VectorMath.subtract(attractor, particle.getLocation());
      
      return VectorMath.multiply(phi, delta);
   }
   
   private List<Particle> rankDistances(Particle from)
   {
      Set<Particle> neighbours = neighborhood.getNeighborhoodOf(from);
      
      ArrayList<PDist> pDists = new ArrayList<PDist>(neighbours.size());
      for (Particle p : neighbours)
         pDists.add(new PDist(p, from));
      
      Collections.sort(pDists);
      
      ArrayList<Particle> result = new ArrayList<Particle>(particles.size());
      for (PDist pDist : pDists)
         result.add(pDist.getParticle());
      
      return result;
   }
   
   public void register(Particle particle)
   {
      particles.add(particle);
      neighborhood.register(particle);
   }

   public void initialized(SwarmEvent e)
   {
      step(e);
   }

   public void step(SwarmEvent e)
   {
      for (Particle other : particles)
      {
         if (other.getFitness()[0] < worstEverFitness)
            worstEverFitness = other.getFitness()[0];
         
         if (other.getFitness()[0] > bestEverFitness)
            bestEverFitness = other.getFitness()[0];
      }
      
      if (decisionRange == null)
         decisionRange = swarm.getVariableRange();
      
      calcDupWeights();
      
      for (Particle p : particles)
      {
         if (useGravityModel)
            nextVels.put(p, calcNextVelGrav(p));
         else
            nextVels.put(p, calcNextVelConst(p));
      }
   }
   
   private static class PDist implements Comparable<PDist>
   {
      private Particle p;
      private double distance;
      
      public PDist(Particle p, Particle from)
      {
         this.p = p;
         distance = VectorMath.length(VectorMath.subtract(
               p.getPersonalBestLocation(), from.getLocation()));
      }
      
      public Particle getParticle()
      {
         return p;
      }
      
      public int compareTo(PDist other)
      {
         if (distance < other.distance)
            return -1;
         else if (distance > other.distance)
            return 1;
         else 
            return 0;
      }
   }
}
