package org.seatiger.pso;

import java.util.Arrays;

/**
 * Hump problem, as described in Gulshan Singh, et al, 
 * "Comparison of Multi-Modal Optimization Algorithms 
 * Based on Evolutionary Algorithms"
 * 
 * A number of varying-sloped peaks with a flat basis function
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class HumpProblem implements FitnessFunction
{
   /** Indicates the value should be chosen randomly between 0 and 1 */
   public final static double RANDOM = -1;  
   
   /** The number of parameters (dimensions) in this problem */
   private int parameters;
   /** The peaks */
   private Peak[] peaks;
   /** Whether to have a basis function that provides a flat floor of fitness = 0 */
   private boolean useBasisFunction = false;
   
   /** The preferred range of this fitness function */
   private final static Range range = new Range(0, 1); 
   
   /**
    * Initialises the fitness function with random peak heights, widths and alphas
    *
    * @param parameters How many parameters (dimensions) the problem
    * should be in 
    * @param numPeaks The number of peaks the problem should have
    */
   public HumpProblem(int parameters, int numPeaks)
   {
      this(parameters, numPeaks, RANDOM, RANDOM, RANDOM);
   }
   
   /**
    * Initialises the fitness function 
    *
    * @param parameters How many parameters (dimensions) the problem
    * should be in 
    * @param numPeaks The number of peaks the problem should have
    * @param radius The radius of the peaks (can be {@link #RANDOM})
    * @param height The radius of the peaks (can be {@link #RANDOM})
    * @param alpha The radius of the peaks (can be {@link #RANDOM})
    */
   public HumpProblem(int parameters, int numPeaks, 
         double radius, double height, double alpha)
   {
      double recMax = 0.5 * Math.pow(numPeaks, -1 / (double) parameters);
      this.parameters = parameters;
      
      peaks = new Peak[numPeaks];
      
      if (radius > recMax)
      {
         System.out.println("WARNING: Peak radius=" + radius + 
               " is larger than the recommended maximum (" + recMax + ")");
         System.out.println("Trying to place peaks...");
      }
      
      for (int i = 0; i < numPeaks; i++)
         peaks[i] = new Peak(radius, height, alpha);
      
      if (radius > recMax)
         System.out.println("Successfully placed all peaks.");
   }

   /**
    * Returns the fitness of the given location. 
    * 
    * @param location The location to test
    * @return The fitness value of that location
    */
   public double getFitness(double[] location)
   {
      double max = peaks[0].getHeight(location);
      
      for (int i = 1; i < peaks.length; i++)
      {
         double height = peaks[i].getHeight(location);
         if (height > max)
            max = height;
      }
      
      return max;
   }

   /**
    * Returns the number of parameters expected by this fitness
    * function. 
    * 
    * @return The number of parameters specified in the constructor
    * @see org.seatiger.pso.FitnessFunction#getNumVariables()
    */
   public int getNumVariables()
   {
      return parameters;
   }

   /**
    * Returns the preferred range of this fitness function, 
    * which is [0, 1] for all parameters
    * 
    * @return The preferred range
    * @see org.seatiger.pso.FitnessFunction#getPreferredRanges()
    */
   public Range[] getPreferredRanges()
   {
      Range[] result = new Range[parameters];
      Arrays.fill(result, range);
      
      return result;
   }
   
   /**
    * Returns the distance to the edge of the nearest peak. If the
    * point is within the radius of a peak, the returned value will
    * be negative
    * 
    * @param location The location to test
    * @return The distance to the nearest peak. 
    */
   private double distToNearestPeakEdge(double[] location)
   {
      double min = Double.POSITIVE_INFINITY;
      
      for (int i = 0; i < peaks.length && peaks[i] != null; i++)
      {
         double dist = peaks[i].getDistanceToEdge(location);
         if (dist < min)
            min = dist;
      }

      return min;
   }
   
   /**
    * Returns true if the basis function is being used (default)
    * 
    * @return True if the basis function is being used
    */
   public boolean isUseBasisFunction()
   {
      return useBasisFunction;
   }

   /**
    * Sets whether the basis function should be used or not. If true, then
    * the fitness of all points not within a peak is 0, otherwise it is the 
    * fitness of the highest peak at that point. 
    * 
    * @param useBasisFunction Whether to use the basis function or not. 
    */
   public void setUseBasisFunction(boolean useBasisFunction)
   {
      this.useBasisFunction = useBasisFunction;
   }
   
   /**
    * Represents a single peak
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private class Peak
   {
      /** The location of the peak */
      public double[] location;
      
      /** The height of the peak */
      public double h;
      
      /** The radius of the peak */
      public double r;
      
      /** The steepness of the peak */
      public double alpha;
      
      /** 
       * Initialises the peak
       * 
       * @param radius The radius of the peak (can be {@link #RANDOM})
       * @param height The height of the peak (can be {@link #RANDOM})
       * @param alpha The alpha of the peak (can be {@link #RANDOM})
       */
      public Peak(double radius, double height, double alpha)
      {
         if (height == RANDOM)
            h = Math.random();
         else
            h = height;
         
         if (alpha == RANDOM)
            this.alpha = Math.random();
         else
            this.alpha = alpha;
         
         location = new double[getNumVariables()];
         double distToNearest;
         
         do 
         {
            for (int i = 0; i < getNumVariables(); i++)
               location[i] = Math.random() * range.getSize() + range.getMin();
            
            distToNearest = distToNearestPeakEdge(location);
            
            // If the radius has been set, we need to make sure there is
            // >= 2r between the peaks
            if (radius != RANDOM)
               distToNearest -= radius;
            
            // Loop until we find a point outside all existing peaks
         } while (distToNearest <= 0);
         
         // We use 1 - random() because getHeight() will fail if r = 0.00
         if (radius == RANDOM)
            r = distToNearest * (1 - Math.random());
         else 
            r = radius;
      }
      
      /** 
       * Gets the height of the peak at point x
       * 
       * @param x The point to test
       * @return The height of the peak at x
       */
      public double getHeight(double[] x)
      {
         double distance = VectorMath.distance(location, x);
         
         if (useBasisFunction && distance > r)
            return 0;
         
         return h * (1 - Math.pow(distance / r, alpha));
      }

      /**
       * Returns the distance from x to the edge of the peak
       * 
       * @param x The point to test
       * @return The distance from the edge of this peak
       */
      public double getDistanceToEdge(double[] x)
      {
         return VectorMath.distance(location, x) - r;
      }
   }
}
