package org.seatiger.pso;

/**
 * The Knapsack problem. Depending on the {@link KnapsackRepresentation} used, 
 * this can be either a 0/1, bounded or unbounded problem.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class KnapsackProblem implements FitnessFunction
{
   /** Converts between a location in the decision space and how many of each item */
   private KnapsackRepresentation representation;
   
   /** The number of different item types */
   private int numItemTypes;
   
   /** The weight of each item */
   private double[] weights;
   
   /** THe value of each item */
   private double[] values;
   
   /** The maximum weight the knapsack can carry */
   private double capacity;
   
   /** THe maximum number of items in the knapsack */
   private int maxItems;
   
   /**
    * Creates a new Knapsack problem. The weights and values will be the set of
    * integers in the range of [0, numItemTypes). Capacity and maxItems = numItemTypes. The
    * representation will be a {@link DefaultKnapsackRepresentation}
    * 
    * @param numItemTypes The number of different items
    */
   public KnapsackProblem(int numItemTypes)
   {
      this(new DefaultKnapsackRepresentation(), numItemTypes);
   }
   
   /**
    * Creates a new Knapsack problem. The weights and values will be the set of
    * integers in the range of [0, numItemTypes). Capacity and maxItems = numItemTypes
    * 
    * @param representation The representation to use
    * @param numItemTypes The number of different items
    */
   public KnapsackProblem(KnapsackRepresentation representation, int numItemTypes)
   {
      this.representation = representation;
      this.numItemTypes = numItemTypes;
      capacity = numItemTypes;
      maxItems = numItemTypes;
      
      weights = new double[numItemTypes];
      values = new double[numItemTypes];
      for (int i = 0; i < weights.length; i++)
      {
         weights[i] = i;
         values[i] = i;
      }
   }
   
   public double getFitness(double[] location)
   {
      int[] itemQuantities = representation.decode(location, this);
      
      if (itemQuantities == null)
         return Double.NEGATIVE_INFINITY;
      
      double totalValue = 0;
      double totalWeight = 0;
      
      for (int i = 0; i < itemQuantities.length; i++)
      {
         totalValue += itemQuantities[i] * values[i];
         totalWeight += itemQuantities[i] * weights[i];
      }

      if (totalWeight > capacity)
         return capacity - totalWeight;
      else
         return totalValue;
   }

   public int getNumVariables()
   {
      return representation.getNumVariables(this);
   }

   public Range[] getPreferredRanges()
   {
      return representation.getPreferredRanges(this);
   }
   
   /**
    * Returns the maximum capacity of the knapsack
    * 
    * @return The maximum capacity of the knapsack
    */
   public double getCapacity()
   {
      return capacity;
   }

   /**
    * Sets the maximum capacity of the knapsack
    * 
    * @param capacity The maximum capacity of the knapsack
    */
   public void setCapacity(double capacity)
   {
      this.capacity = capacity;
   }

   /**
    * Gets the value of each item type
    * 
    * @return The value of each item type
    */
   public double[] getValues()
   {
      return values;
   }

   /**
    * Sets the value of each item type
    * 
    * @param values The value of each item type
    */
   public void setValues(double[] values)
   {
      if (this.values != null && values.length != this.values.length)
         throw new IllegalArgumentException("Cannot change the number of item types");
         
      this.values = values;
   }

   /**
    * Gets the weight of each item type
    * 
    * @return The weight  of each item type
    */
   public double[] getWeights()
   {
      return weights;
   }

   /**
    * Sets the weight of each item type
    * 
    * @param weights The weight of each item type
    */
   public void setWeights(double[] weights)
   {
      if (this.weights != null && weights.length != this.weights.length)
         throw new IllegalArgumentException("Cannot change the number of item types");
      
      this.weights = weights;
   }

   /**
    * Returns the maximum number of items in the knapsack
    * 
    * @return The maximum number of items in the knapsack
    */
   public int getMaxItems()
   {
      return maxItems;
   }

   /**
    * Sets the maximum number of itesm in the knapsack
    * 
    * @param maxItems The maximum number of items in the knapsack
    */
   public void setMaxItems(int maxItems)
   {
      this.maxItems = maxItems;
   }

   /**
    * Returns the number of item types 
    * 
    * @return The number of item types
    */
   public int getNumItemTypes()
   {
      return numItemTypes;
   }

   /**
    * Sets the number of item types 
    * 
    * @param numItemTypes The number of item types
    */
   public void setNumItemTypes(int numItemTypes)
   {
      this.numItemTypes = numItemTypes;
   }
   
   
}
