package org.seatiger.pso;

/**
 * Tells each particle how to move. Concrete classes will probably also
 * want to implement {@link SwarmListener} to be notified of initialisation
 * and step events. It is the MovementModel's responsibility to register
 * the Particle with any Neighborhoods it may be using.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 * @see SwarmListener
 */
public interface MovementModel
{
   /**
    * Returns the next velocity the particle should have. The particle's 
    * position will be immediately adjusted by the velocity after this method
    * returns. If the returned velocity would cause the particle to move outside
    * the decision space it will automatically be sanitised using whatever
    * method the swarm uses.  
    * 
    * @param particle The particle to move
    * @return The location in the search space the particle should move to
    * next. This array must be the same length as the number of dimensions
    * in the particle
    */
   public double[] getNextVelocity(Particle particle);
   
   /**
    * Registers the particle with this movement model. This method may be called
    * at any point during the simulation
    * 
    * @param particle The particle to register
    */
   public void register(Particle particle);

   /**
    * Deregisteres the particle from the movement model. This method may be 
    * called at any point during the simulation
    * 
    * @param particle The particle to deregister
    */
   public void deregister(Particle particle);
   
   
   /**
    * Returns the neighbourhood this movement model is using
    * 
    * @return The neighbourhood. 
    */
   public Neighborhood getNeighborhood();
}
