package org.seatiger.pso;

import java.util.*;

import peaks.*;

/**
 * Represents Juergen Branke's Moving Peaks implementation as a FitnessFunction.
 * 
 * Note: Due to the implementation of the {@link movpeaks} class, only one instance
 * of this class should be created at a time
 *  
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class MovingPeaks extends movpeaks implements FitnessFunction
{
   /** Indicates the given peak is currently covered by other peaks */
   public static final double HIDDEN = -1;
   /** Indicates there are no particles with a pBest on this peak */
   public static final double NO_PARTICLES = Double.POSITIVE_INFINITY;
   
	/** Which peaks are currently hidden */
   private boolean[] hiddenPeaks = null;
   /** Used to notify listeners when the peaks have changed */
   private Observable2 peakChangeObservable = new Observable2();
   
   /** Used to notify listeners when an evaluation occured. 
    * This field is lazily created by {@link #getEvalObservable()} */
   private Observable2 evalObservable = null;
   
	/**
	 * Creates a new MovingPeaks
	 *
	 * @param numPeaks The required number of peaks
    * @param dimensions The number of dimensions in the decision space
	 */
	public MovingPeaks(int numPeaks, int dimensions)
	{
      super(numPeaks);
      movpeaks.geno_size = dimensions;
		init_peaks();
	}
	
	public double getFitness(double[] location)
	{
      if (evalObservable != null)
      {
         evalObservable.setChanged();
         evalObservable.notifyObservers(this);
      }
         
      return eval_movpeaks(location); 
	}

	public int getNumVariables()
	{
		return movpeaks.geno_size;
	}

	public Range[] getPreferredRanges()
	{
		Range range = new Range(getMinCoordinate(), 
				getMaxCoordinate());
		
		Range[] result = new Range[getNumVariables()];
		Arrays.fill(result, range);
		
		return result;
	}

	/**
	 * Returns the number of peaks in the simulation 
	 * 
	 * @return The number of peaks
	 */
	public int getNumPeaks()
	{
		return movpeaks.peak.length;
	}
   
   /**
    * Returns the Observable which will notify immediately before the peaks have moved. Observers
    * will be notified with this MovingPeaks instance as the argument
    * 
    * @return The observable which will notifiy when the peaks have moved
    */
   public Observable getPeakChangeObservable()
   {
      return peakChangeObservable;
   }

   /**
    * Returns the Observable which will notify immediately before an evaluation occurs. 
    * Observers will be notified with this MovingPeaks instance as the argument
    * 
    * @return The observable which will notifiy when the peaks have moved
    */
   public Observable getEvalObservable()
   {
      if (evalObservable == null)
         evalObservable = new Observable2();
      
      return evalObservable;
   }
	
   /**
    * Returns the euclidean distance between each peak and the closest pBest
    * of any of the particles. Note that because this function only looks at
    * the peak whose catchment area each particle is in, it is possible in 
    * some cases for a larger result to be returned than is in reality. In these
    * cases it is unlikely the particle would converge on the closer peak because
    * it not in that peak's catchment area. 
    * 
    * @param particles The particles to test
    * @return An array with an element per peak. Each element will either contain
    * the distance or be set to {@link #HIDDEN} or {@link #NO_PARTICLES} if no
    * particles know about that peak. 
    */
   public double[] getDistancesToPeaks(Collection<Particle> particles)
   {
      double[] result = new double[getNumPeaks()];
      Arrays.fill(result, NO_PARTICLES);
      
      for (int i = 0; i < getNumPeaks(); i++)
      {
         if (isPeakHidden(i))
            result[i] = HIDDEN;
      }
      
      // Note that because getPeakAt can return a peak that is currently hidden
      // (if that peak is not *entirely* hidden), only particles that are in
      // a peak's actual catchment area are considered
      
      for (Particle p : particles)
      {
         int peakNum = getPeakAt(p.getPersonalBestLocation());
         double distance = VectorMath.length(VectorMath.subtract(
               p.getPersonalBestLocation(), getPeakPositions()[peakNum]));
         
         if (distance < result[peakNum])
            result[peakNum] = distance;
      }

      return result;
   }
   
   /**
    * Returns the fitness delta between each peak and the fittest pBest
    * of any of the particles. Note that because this function only looks at
    * the peak whose catchment area each particle is in, it is possible in 
    * some cases for a larger result to be returned than is in reality. In these
    * cases it is unlikely the particle would converge on the closer peak because
    * it not in that peak's catchment area. 
    * 
    * @param particles The particles to test
    * @return An array with an element per peak. Each element will either contain
    * the fitness delta or be set to {@link #HIDDEN} or {@link #NO_PARTICLES} if no
    * particles know about that peak. 
    */
   public double[] getFitnessDeltasToPeaks(Collection<Particle> particles)
   {
      double[] peakHeights = getPeakHeights();
      double[] result = new double[getNumPeaks()];
      Arrays.fill(result, NO_PARTICLES);
      
      for (int i = 0; i < getNumPeaks(); i++)
      {
         if (isPeakHidden(i))
            result[i] = HIDDEN;
      }
      
      // Note that because getPeakAt can return a peak that is currently hidden
      // (if that peak is not *entirely* hidden), only particles that are in
      // a peak's actual catchment area are considered
      
      for (Particle p : particles)
      {
         int peakNum = getPeakAt(p.getPersonalBestLocation());
         double fitness = peakHeights[peakNum] - dummy_eval(p.getPersonalBestLocation());
         
         if (fitness < result[peakNum])
            result[peakNum] = fitness;
      }

      return result;
   }
      
   /**
    * Returns the particle that is closest to the best peak. Note that the 
    * particle is not necessarily within that peak's catchment area.
    * 
    * @param particles The particles to compare
    * @return The particle that has a personal best closest to the best peak
    */
   public Particle getClosestToBestPeak(Collection<Particle> particles)
   {
      return getClosestToPeak(getMaximumPeak(), particles);
   }
   
   /**
    * Returns the particle that is closest to the given peak. Note that the 
    * particle is not necessarily within that peak's catchment area.
    * 
    * @param peakNum The number of the peak to find the closest particle to
    * @param particles The particles to compare
    * @return The particle that has a personal best closest to the given peak
    */
   public Particle getClosestToPeak(int peakNum, Collection<Particle> particles)
   {
      Particle best = null;
      double bestDist = Double.POSITIVE_INFINITY;
      double[] peakPos = getPeakPositions()[peakNum];
      
      for (Particle p : particles)
      {
         double dist = VectorMath.distance(peakPos, p.getPersonalBestLocation());
         
         if (dist < bestDist)
         {
            bestDist = dist;
            best = p;
         }
      }
      
      return best;
   }
   
   /** 
    * Returns whether the given peak is currently covered by another peak
    * 
    * @param peakNum The number of the peak to test
    * @return True if the peak is hidden, false otherwise
    */
   public boolean isPeakHidden(int peakNum)
   {
      if (hiddenPeaks == null)
      {
         hiddenPeaks = new boolean[getNumPeaks()];
         
         for (int i = 0; i < getNumPeaks(); i++)
            hiddenPeaks[i] = peakNum != getPeakAt(getPeakPositions()[peakNum]); 
      }
      
      return hiddenPeaks[peakNum];
   }
   
   
   /**
    * Returns the number of the peak at the given point.
    *  
    * @param point The point to test
    * @return The index of the representitive peak at that point, 
    * or -1 if none of the peaks are higher than the basis function at
    * the given point. 
    */
   public int getPeakAt(double[] point)
   {
      double dummy;
      double maximum = Double.NEGATIVE_INFINITY;
      int heighestPeak = 0;
      
      for (int i = 0; i < number_of_peaks; i++) {
          dummy = this.pf.calculate(point, i);
          if (dummy > maximum)
          {
             maximum = dummy;
             heighestPeak = i;
          }
      }

      if (this.use_basis_function) {

          dummy = this.bf.calculate(point);
          /* If value of basis function is higher return it */
          if (maximum < dummy)
             heighestPeak = -1;
      }
      
      return heighestPeak;
   }
   
   /**
    * Sets the peak function to use
    * 
    * @param pf The peak function to use
    */
   public void setPeakFunction(Peak_Function pf)
   {
      this.pf = pf;
   }
   
   /**
    * Returns the peak function being used
    * 
    * @return The peak function being used
    */
   public Peak_Function getPeakFunction()
   {
      return pf;
   }
   
   @Override
   public void change_peaks()
   {
      hiddenPeaks = null;
      peakChangeObservable.setChanged();
      peakChangeObservable.notifyObservers(this);
      super.change_peaks();
   }
   
   /**
    * Returns how far the peaks will move each time
    * 
    * @return How far the peaks will move
    */
   public double getSeverity()
   {
      return vlength;
   }
   
   /**
    * Sets how far the peaks will move each time
    * 
    * @param severity How far the peaks will move
    */
   public void setSeverity(double severity)
   {
      vlength = severity;
   }
   
   

   /**
    * Makes the setChanged() method visible
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private static class Observable2 extends Observable 
   { 
      @Override
      protected void setChanged() { super.setChanged(); }
   }  
}

