package org.seatiger.pso;

import java.util.*;

/**
 * Implements several metrics for the MovingPeaks fitness function
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class MovingPeaksMetric
{
  
   /** The moving peaks instance */
   private MovingPeaks peaks;
   /** The swarm being analysed */
   private Swarm swarm;

   /** Whether BKPE should return fitness error (false = distance error) */
   private boolean useFitness = false;
   /** The cumulative percentage of known, non-hidden peaks */
   private double totalPercentKnownPeaks = 0;
   /** The amount of error to each peak since the last move */
   private double[] errsSinceLastMove;
   /** Whether each peak was known at every tick since the last move */
   private boolean[] knownPeaks;
//   /** The rank of each peak, with 0 being the highest peak */
//   private Integer[] peakRanks;
   /** The cumulative error for the best known peak error metric */
   private double bestKnownPeakTotalErr = 0;
   /** The distance between the best peak and the nearest particle to it */
   private double offlineDistErr = 0;
   /** The tick that the peaks last moved on */
   private int lastMoveTick = 0;
   /** The number of ticks for which at least one peak was known for the
    * entire time between movements */
   private int numTicks = 0;
   /** A record of the current errors */
   private ArrayList<Double> currentErrorHistory = new ArrayList<Double>();
   
   /**
    * Creates a new MovingPeaksMetric
    *
    * @param swarm The swarm being analysed
    * @param peaks The moving peaks instance
    */
   public MovingPeaksMetric(Swarm swarm, MovingPeaks peaks)
   {
      this.swarm = swarm;
      this.peaks = peaks;
      
      Listener listener = new Listener();
      swarm.addSwarmListener(listener);
      peaks.getPeakChangeObservable().addObserver(listener);
      
      errsSinceLastMove = new double[peaks.getNumPeaks()];
      knownPeaks = new boolean[peaks.getNumPeaks()];
//      peakRanks = new Integer[peaks.getNumPeaks()];
      
      Arrays.fill(knownPeaks, true);
      
//      for (int i = 0; i < peakRanks.length; i++)
//         peakRanks[i] = new Integer(i);
//      calculatePeakRanks();
   }
   
   
   

   /**
    * Returns whether the BKPE metric uses the fitness error instead of the positional
    * error
    * 
    * @return True if the BKPE metric is using the fitness error
    */
   public boolean isUseFitness()
   {
      return useFitness;
   }

   /** 
    * Controls whether the BKPE metric uses the fitness error instead of the positional
    * error
    * 
    * @param useFitness True to use the fitness error instead of the positional error
    */ 
   public void setUseFitness(boolean useFitness)
   {
      this.useFitness = useFitness;
   }

   /**
    * Records the current error. This function records only the current error
    * at the time it was called, not the generation or number of evaluations. 
    * It is assumed that whoever calls this function makes note of when it was called.  
    */
   public void recordCurrentError()
   {
      currentErrorHistory.add(peaks.get_current_error());
   }

   /**
    * Returns the current error history
    * 
    * @return The current error history
    */
   public double[] getCurrentErrorHistory()
   {
      double[] result = new double[currentErrorHistory.size()];
      
      for (int i = currentErrorHistory.size() - 1; i >=0 ; i--)
         result[i] = currentErrorHistory.get(i);
      
      return result;
   }

   /**
    * Returns the offline error
    * 
    * @return The offline error
    */
   public double getOfflineError()
   {
      return peaks.get_offline_error();
   }
   
   /**
    * Returns the current error
    * 
    * @return The current error
    */
   public double getCurrentError()
   {
      return peaks.get_current_error();
   }
   
   /**
    * Returns the offline distance error, that is the average distance
    * between the best peak and the closest particle to it
    * 
    * @return The offline distance error
    */
   public double getOfflineDistanceError()
   {
      return offlineDistErr / swarm.getCurrTick();
   }
   
   /**
    * Returns the average number of peaks known about
    * 
    * @return THe average number of peaks known about
    */
   public double getPCP()
   {
      return totalPercentKnownPeaks / swarm.getCurrTick();
   }
   
   /**
    * Returns the average error for the best known peak. This measures the
    * convergence speed of the algorithm
    * 
    * @return The average error for the best known peak
    */
   public double getBKPE()
   {
      return bestKnownPeakTotalErr / numTicks;
   }
   
   /**
    * Updates the {@link #totalPercentKnownPeaks} variable
    * 
    * @param distances The distance between each peak and the closest pBest in its
    * catchment area
    */
   private void updatePercentKnownPeaks(double[] distances)
   {
      int numKnownPeaks = 0;
      int numVisiblePeaks = 0;
      
      for (double dist : distances)
      {
         if (dist == MovingPeaks.HIDDEN)
            continue;
         
         numVisiblePeaks++;
         if (dist != MovingPeaks.NO_PARTICLES)
            numKnownPeaks++;
      }
      
      if (numVisiblePeaks == 0)
         totalPercentKnownPeaks++;
      else
         totalPercentKnownPeaks += numKnownPeaks / (double) numVisiblePeaks;
   }
   
//   /**
//    * Recalculates the height rank of each peak
//    */
//   private void calculatePeakRanks()
//   {
//      final double[] peakHeights = peaks.getPeakHeights();
//      
//      Arrays.sort(peakRanks, new Comparator<Integer>() {
//         public int compare(Integer a, Integer b)
//         {
//            // We want the highest peaks ranked first, so reverse
//            // the normal comparison order
//            if (peakHeights[a] > peakHeights[b])
//               return -1;
//            else if (peakHeights[a] < peakHeights[b])
//               return 1;
//            else
//               return 0;
//         }  
//      });
//   }

   /**
    * Updates the metrics
    */
   private void updateMetrics()
   {
      double[] peakErrors;
      
      if (useFitness)
         peakErrors = peaks.getFitnessDeltasToPeaks(swarm.getParticles());
      else
         peakErrors = peaks.getDistancesToPeaks(swarm.getParticles());
      
      updatePercentKnownPeaks(peakErrors);
      for (int i = 0; i < peakErrors.length; i++)
      {
         if (peakErrors[i] == MovingPeaks.NO_PARTICLES)
            knownPeaks[i] = false;
         else
            errsSinceLastMove[i] += peakErrors[i];
      }
      
      Particle closestParticle = peaks.getClosestToBestPeak(swarm.getParticles());
      offlineDistErr += VectorMath.distance(closestParticle.getPersonalBestLocation(), 
            peaks.getPeakPositions()[peaks.getMaximumPeak()]);
   }

   /**
    * Returns the current error to the given peak
    * 
    * @param peakNum The index of the peak to get the error of
    * @return The error to the given peak
    */
   public double getCurrError(int peakNum)
   {
      if (useFitness)
         return peaks.getFitnessDeltasToPeaks(swarm.getParticles())[peakNum];
      else
         return peaks.getDistancesToPeaks(swarm.getParticles())[peakNum];
   }

   /** 
    * Notification that the peaks have moved
    */
   private void peaksMoved()
   {
      int bestPeak = getBestKnownPeak();
      
      if (bestPeak != -1)
      {
         bestKnownPeakTotalErr += errsSinceLastMove[bestPeak];
         numTicks += swarm.getCurrTick() - lastMoveTick;
      }
      
      lastMoveTick = swarm.getCurrTick();
       
      Arrays.fill(errsSinceLastMove, 0);
      Arrays.fill(knownPeaks, true);
      //calculatePeakRanks();
   }

   /**
    * Returns the highest peak continually covered since the last peak movement
    * 
    * @return The index of the highest known peak, or -1 if there were no continually
    * covered peaks
    */
   public int getBestKnownPeak()
   {
      int bestPeak = -1;
      double bestFitness = Double.NEGATIVE_INFINITY;
//      double[] peakHeights = peaks.getPeakHeights();
//      for (int i = 0; i < peakHeights.length; i++)
//      {
//         if (knownPeaks[i] && !peaks.isPeakHidden(i) && 
//               (bestPeak == -1 || peakHeights[i] > peakHeights[bestPeak]))
//            bestPeak = i;
//      }
      for (Particle p : swarm.getParticles())
      {
         int peakNum = peaks.getPeakAt(p.getPersonalBestLocation());
         if (!knownPeaks[peakNum] && peaks.isPeakHidden(peakNum))
            continue;
         
         double fitness = peaks.dummy_eval(p.getPersonalBestLocation());
         
         if (fitness > bestFitness)
         {
            bestFitness = fitness;
            bestPeak = peakNum;
         }
      }
      
      return bestPeak;
   }

   /**
    * Listens for external events
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private class Listener implements Observer, SwarmListener
   {
      public void update(Observable o, Object arg)
      {
         peaksMoved();
      }
      
      public void step(SwarmEvent e)
      {
         updateMetrics();
      }

      public void initialized(SwarmEvent e) { }
   }
}
