package org.seatiger.pso;

import java.util.*;


/**
 * Finds the nearest points to the given location. This uses the method
 * described in Hjaltason, G., Samet, H., "Ranking in Spatial Databases".
 * It is slightly simplified from the one described in the paper because
 * it is guaranteed that all BranchNodes are disjoint.  
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class NearestPointQuery implements Iterator
{
   /** The location that we are finding the closest points to */ 
   private double[] target;
   
   /** The queue to put the TreeNodes on */
   // 11 is the default initial capacity for a priority queue
   private PriorityQueue<TreeNode> queue = 
      new PriorityQueue<TreeNode>(11, new DistanceComparitor());
   
   /** The next result to be returned */
   private LeafNode next;
   
   /** Whether the next result has been computed */
   private boolean nextComputed = false;
   
   /** The number of children each BranchNode can have */
   private int numChildren;
   
   /**
    * Creates a new NearestPointQuery
    * 
    * @param target The location to find the nearest points to
    * @param rootNode The root node of the tree
    */
   public NearestPointQuery(double[] target, BranchNode rootNode)
   {
      numChildren = (int) Math.pow(2, rootNode.getNumDimensions());
      this.target = new double[target.length];
      System.arraycopy(target, 0, this.target, 0, target.length);
      queue.add(rootNode);
   }
   
   /**
    * Computes the next closest LeafNode to {@link #target}
    */
   private void computeNextNode()
   {
      next = null;
      while (next == null && !queue.isEmpty())
      {
         TreeNode element = queue.remove();
         
         if (element instanceof LeafNode)
            next = (LeafNode) element;
         else if (element instanceof BranchNode)
         {
            BranchNode branch = (BranchNode) element;
            
            for (int i = 0; i < numChildren; i++)
            {
               TreeNode child = branch.getChild(i);
               if (child != null)
                  queue.add(child);
            }
         }
      }
      
      nextComputed = true;
   }
   
   /**
    * Compares two TreeNodes based on the minimum distance to the target
    * point of this query. If a LeafNode and a BranchNode are the same 
    * distance away, the BranchNode is considered to be first. 
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private class DistanceComparitor implements Comparator
   {
      public int compare(Object obj1, Object obj2)
      {
         TreeNode node1 = (TreeNode) obj1;
         TreeNode node2 = (TreeNode) obj2;
         
         double[] node1Point = node1.getClosestPointTo(target);
         double[] node2Point = node2.getClosestPointTo(target);
         
         double node1Dist = 
            VectorMath.length(VectorMath.subtract(node1Point, target));
         double node2Dist = 
            VectorMath.length(VectorMath.subtract(node2Point, target));

         if (node1Dist < node2Dist)
            return -1;
         else if (node1Dist > node2Dist)
            return 1;
         else if (node1 instanceof BranchNode && node2 instanceof LeafNode)
            return -1;
         else if (node1 instanceof LeafNode && node2 instanceof BranchNode)
            return 1;
         else
            return 0;
      }
   }
   
   public boolean hasNext()
   {
      if (!nextComputed)
         computeNextNode();
         
      return next != null;
   }

   public LeafNode next()
   {
      if (!nextComputed)
         computeNextNode();
      
      if (next == null)
         throw new NoSuchElementException();

      nextComputed = false;
      
      return next;
   }

   /**
    * @throws UnsupportedOperationException
    */
   public void remove()
   {
      throw new UnsupportedOperationException();
   }
}
