package org.seatiger.pso;

import java.util.HashMap;
import java.util.LinkedHashSet;

import uchicago.src.sim.util.Random;

/**
 * Implements a constriction PSO
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class OldConstrictionPSO implements MovementModel, SwarmListener
{
   /** The swarm that the particles belong to */
   protected Swarm swarm;
   /** The best fitness of any particle registered to this MovementModel */
   protected double gBestFitness = Double.NEGATIVE_INFINITY;
   /** The location of the fittest particle registered to this MovementModel */
   protected double[] gBestLocation = null;
   /** The weighting factor for the personal best */
   protected double phi1 = 2.05;
   /** The weighting factor for the neighbourhood best */
   protected double phi2 = 2.05;
   /** The amount of inertia for the particle */
   protected double alpha = 0.1;
   /** Modifies the constriction factor */
   protected double kappa = 1;
   /** The constriction factor. This is calculated by the 
    * {@link #calculateConstriction()} method */
   protected double constriction;
   /** The range of the decision space */
   protected Range[] decisionRange = null;
   /** The amount to change each particle's velocity by */
   protected HashMap<Particle, double[]> velocityDeltas
         = new HashMap<Particle, double[]>();
   /** The neighborhood to use */
   protected Neighborhood neighborhood;
   
   /** The particles handled by this MovementModel */
   protected LinkedHashSet<Particle> particles = new LinkedHashSet<Particle>();

  
   /**
    * Creates a new ConstrictionMovementModel for the specified Swarm. phi1
    * and phi2 will both be set to 2.05. A {@link GlobalNeighborhood} 
    * will be used as the neighbourhood model
    *
    * @param swarm The Swarm that owns the particles to control
    * @see #ConstrictionPSO(Swarm, double, double)
    */
   public OldConstrictionPSO(Swarm swarm)
   {
      this(swarm, null);
   }

   /**
    * Creates a new ConstrictionMovementModel for the specified Swarm. phi1
    * and phi2 will both be set to 2.05. 
    *
    * @param swarm The Swarm that owns the particles to control
    * @param neighborhood The neighborhood model to use. If null, a 
    * {@link GlobalNeighborhood} will be used
    * @see #ConstrictionPSO(Swarm, double, double)
    */
   public OldConstrictionPSO(Swarm swarm, Neighborhood neighborhood)
   {
      this(swarm, 2.05, 2.05, neighborhood);
   }
   
   /**
    * Creates a new ConstrictionMovementModel for the specified Swarm. A
    * {@link GlobalNeighborhood} will be used as the neighbourhood model
    *
    * @param swarm The Swarm that owns the particles to control
    * @param phi1 The weighting factor for the personal best
    * @param phi2 The weighting factor for the neighbourhood best
    * @throws IllegalArgumentException if phi1 + phi2 &lt; 4
    */
   public OldConstrictionPSO(Swarm swarm, double phi1, double phi2)
   {
      this(swarm, phi1, phi2, null);
   }
   
   /**
    * Creates a new ConstrictionMovementModel for the specified Swarm
    *
    * @param swarm The Swarm that owns the particles to control
    * @param phi1 The weighting factor for the personal best
    * @param phi2 The weighting factor for the neighbourhood best
    * @param neighborhood The neighborhood model for the particles. If 
    * null, a {@link GlobalNeighborhood} will be used
    * @throws IllegalArgumentException if phi1 + phi2 &lt; 4
    */
   public OldConstrictionPSO(Swarm swarm, double phi1, double phi2, 
         Neighborhood neighborhood)
   {
      this.swarm = swarm;
      swarm.addSwarmListener(this);
      setPhi(phi1, phi2);
      
      if (neighborhood == null)
         neighborhood = new GlobalNeighborhood();
      this.neighborhood = neighborhood;
   }
   
   /**
    * Returns the neighbourhood in use
    * 
    * @return The neighbourhood in use
    */
   public Neighborhood getNeighborhood()
   {
      return this.neighborhood;
   }
   
   /**
    * Returns the location that the particle should move to 
    * 
    * @see org.seatiger.pso.MovementModel#getNextVelocity(org.seatiger.pso.Particle)
    */
   public double[] getNextVelocity(Particle particle)
   {
      if (decisionRange == null)
         decisionRange = swarm.getVariableRange();
      
      double[] nBestLocation;
//      if (neighborhood instanceof GlobalNeighborhood)
//         nBestLocation = gBestLocation;
//      else
         nBestLocation = calculateNBestLoc(particle);
      
      double[] pBestLocation = particle.getPersonalBestLocation();
      double[] currentLocation = particle.getLocation();
      double[] velocity = particle.getVelocity();
      
      double multiplier1; 
      double multiplier2; 
      
      do
      {
         multiplier1 = phi1 * Random.uniform.nextDouble();
         multiplier2 = phi2 * Random.uniform.nextDouble();
      } 
      while (multiplier1 == 0 && multiplier2 == 0);
      
      
      
      // v = constriction * (v + phi1(pBest - x) + phi2(nBest - x))
      velocity = VectorMath.multiply(constriction, VectorMath.add(
            velocity, 
            VectorMath.multiply(multiplier1, 
                  VectorMath.subtract(pBestLocation, currentLocation)),
            VectorMath.multiply(multiplier2, 
                  VectorMath.subtract(nBestLocation, currentLocation))));
      
      // Restrict the velocity
      for (int i = 0; i < velocity.length; i++)
         velocity[i] = Math.min(velocity[i], decisionRange[i].getSize());

      return velocity;
   }
   
   /**
    * Calculates the neighborhood best location for <code>p</code>
    * 
    * @param p The particle to calculate the neighborhood best for
    * @return The neighborhood best location for <code>p</code>
    */
   protected double[] calculateNBestLoc(Particle p)
   {
      Particle bestParticle = null;
      
      for (Particle particle : neighborhood.getNeighborhoodOf(p))
      {
         if (bestParticle == null ||
               particle.getPersonalBestFitness()[0] > 
               bestParticle.getPersonalBestFitness()[0])
            bestParticle = particle;
      }

      return bestParticle.getLocation();
   }
     

   /**
    * Notification that the swarm has completed its current step and is about
    * to start the next. 
    */
   public void step(SwarmEvent e)
   {
      // Update the global best fitness and location
      Particle bestParticle = null;
      
      for (Particle particle : particles)
      {
         if (bestParticle == null ||
               particle.getFitness()[0] > bestParticle.getFitness()[0])
            bestParticle = particle;
      }

      if (bestParticle.getFitness()[0] > gBestFitness)
      {
         gBestFitness = bestParticle.getFitness()[0];
         System.arraycopy(bestParticle.getLocation(), 0, gBestLocation, 0, 
               gBestLocation.length);
         
//         System.out.println("Tick " + swarm.getCurrTick() + ": New best fitness: " + gBestFitness);
      }
   }
   
   public void initialized(SwarmEvent e)
   {
      if (swarm.getNumObjectives() != 1)
         throw new RuntimeException("This movement model only supports" +
               " single objective optimisation");
      
      gBestLocation = new double[swarm.getNumVariables()];
   }

   /**
    * Registers a particle with the movement model
    * 
    * @param particle The particle to register
    * @see org.seatiger.pso.MovementModel#register(org.seatiger.pso.Particle)
    */
   public void register(Particle particle)
   {
      particles.add(particle);
      neighborhood.register(particle);
   }
   
   /**
    * Removes a particle from the movement model
    * 
    * @see org.seatiger.pso.MovementModel#deregister(org.seatiger.pso.Particle)
    */
   public void deregister(Particle particle)
   {
      particles.remove(particle);
      neighborhood.deregister(particle);
   }

   /**
    * Returns the best fitness found by any Particle in this MovementModel
    * 
    * @return The best fitness found so far
    * @see #getBestLocation()
    */
   public double getBestFitness()
   {
      return gBestFitness;
   }
   
   /**
    * Returns the location of the fittest point found so far
    * 
    * @return The location of the fittest point found so far
    * @see #getBestFitness()
    * 
    */
   public double[] getBestLocation()
   {
      return gBestLocation;
   }
   
   /**
    * Sets the weighting factor for the personal and neighbourhood bests 
    * 
    * @param phi1 The weighting factor for the personal best
    * @param phi2 The weighting factor for the neighbourhood best
    * @throws IllegalArgumentException if phi1 + phi2 &lt; 4
    */
   public void setPhi(double phi1, double phi2)
   {
      if (phi1 + phi2 < 4)
         throw new IllegalArgumentException("phi1 + phi2 must be >= 4");
      
      this.phi1 = phi1;
      this.phi2 = phi2;
      
      constriction = calculateConstriction();
   }
   
   /**
    * Calculates the contriction factor chi. 
    * Ref Kennedy, J., "Swarm Intelligence", pp 338
    * 
    * @return The constriction factor chi
    */
   protected double calculateConstriction()
   {
      double phi = phi1 + phi2;
      
      //                  2 * kappa
      //    chi = ---------------------------------
      //          |2 - phi - sqrt(phi^2 - 4 * phi)|
      return 2 * kappa / Math.abs(2 - phi - Math.sqrt(phi * phi - 4 * phi)); 
   }
   
   /**
    * Holds information about a single attractor
    *  
    * @author Stefan Bird <stbird@seatiger.org>
    */
   private static class Attractor
   {
      /** The attractor */
      public ParticleAttractor attractor;
      /** The weight of the attractor */
      public double weight;
   }
}
