package org.seatiger.pso;

import java.awt.Color;
import java.io.*;
import java.util.*;

import uchicago.src.reflector.ListPropertyDescriptor;
import uchicago.src.sim.analysis.*;
import uchicago.src.sim.engine.*;
import uchicago.src.sim.gui.DisplaySurface;
import uchicago.src.sim.util.Random;

/**
 * A model of a particle swarm for use with Repast
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class PSOModel extends SimModelImpl
{
   private static final String[] PARAMETERS = { "anpsoRedistDesiredParticles",
         "anpsoSpeciesMinParticles", "anpsoStepsBeforeSpeciesForm",
         "espsoConversionAge", "espsoEpsilon", "espsoMinNewSpeciesSize",
         "espsoNewSpeciesSize", "FitnessFunction", "humpPeaks",
         "humpPeaksRadius", "humpUseBasis", "movementModel",
         "movPeaksCERecordFreq", "movPeaksCustMaxPower",
         "movPeaksCustMinPower", "movPeaksCustMaxAmp", "movPeaksCustMaxFreq",
         "movPeaksNumPeaks", "movPeaksSeverity", "movPeaksUseCustomPeakFunc",
         "Neighborhood", "particles", "RecordFile", "RecordIntervalTicks",
         "RecordIntervalEvals", "regMaxExcess", "regressionDataFile",
         "RequiredFitnessAccuracy", "ShowDisplays", "smallestOptimaSeparation",
         "spsoPMax", "spsoRadius", "spsoUsePBest", "StopEvals", "StopOptima",
         "StopTick", "variables" };

   /** True if this is a batch run (PF file specified on the commandline */
   private static boolean batch;

   /** A list of the available fitness functions */
   private static final String[] AVAIL_FITNESS_FUNCS = { "DeJong_F1",
         "DeJong_F2", "DeJong_F3", "DeJong_F4", "DeJong_F5", "Rastrigin",
         "Himmelblau", "Shubert", "Debs_First", "Moving_Peaks",
         "Six_Hump_Camel_Back", "Brainin_RCOS", "SigmoidRegression",
         "Knapsack", "Humps" };

   /** A list of the available neighbourhoods */
   private static final String[] AVAIL_NEIGHBORHOODS = { "Global", "Star",
         "Circle", "VonNeumann", "SPSO", "rSPSO", "RegESPSO2", "ANPSO", "rANPSO",
         "ESPSO", "ESPSO2" };

   private static final String[] AVAIL_MOVEMENT_MODELS = { "Constriction",
         "GCPSO", "Gravity_PSO" };

   /** The number of particles in the simulation */
   private int particles = 100;

   /** The number of decision variables */
   private int variables = 5;

   /** The number of peaks for the moving peaks problem */
   private int movPeaksNumPeaks = 10;

   /** How far the peaks move each time */
   private double movPeaksSeverity = 1;

   /** Whether or not to use the custom (non conic) peak function */
   private boolean movPeaksUseCustomPeakFunc = true;

   /** The maximum amplitude of the local undulations */
   private double movPeaksCustMaxAmp = 0;

   /** The maximum frequency of the local undulations */
   private double movPeaksCustMaxFreq = 0;

   /**
    * The minumum power of the peak's sides 1 = d^1, where d is the distance
    * from the peak
    */
   private double movPeaksCustMinPower = 1;

   /**
    * The maximum power of the peak's sides 3 = d^3, where d is the distance
    * from the peak
    */
   private double movPeaksCustMaxPower = 1;

   /** How many evaluations the OE is recorded */
   private int movPeaksCERecordFreq = 0;

   /** The name of the fitness function to use */
   private String fitnessFuncName = "Moving_Peaks";

   /** The name of the neighborhood to use */
   private String neighborhoodName = "rSPSO";

   /** The model's schedule */
   private Schedule schedule = null;

   /** The surface to draw the swarm on */
   private DisplaySurface display = null;

   /** The graph of the best fitness over time */
   private OpenSequenceGraph particlesOverTime;

   /** The graph of the species distance limit over time */
   private OpenSequenceGraph limitOverTime;

   /** The graph of the number of found optima over time */
   private OpenSequenceGraph speciesFoundOptimaOverTime;

   /** The graph of the number of found optima over time */
   private OpenSequenceGraph foundOptimaOverTime;

   /** Graphs the moving peaks metrics */
   private OpenSequenceGraph movingPeaksMetrics;

   /** A histogram of how fit each species is */
   private Plot speciesFitness;

   /** A plot showing how well each peak is represented */
   private Plot peakRepresentation;

   /** The last time the tick was announced in batch mode */
   private long lastTickAnnouncement = 0;

   /** The Swarm */
   private Swarm swarm;

   /** The MovementModel used by the swarm */
   protected ConstrictionPSO model;

   /** The Neighborhood used by the swarm */
   private Neighborhood neighborhood;

   /** The classifier used to determine species */
   private SpeciesClassifier classifier;

   /** The tracker used to track the species */
   private SpeciesTracker tracker;

   /** Monitors the neighborhood */
   private SpeciesNeighborhoodMonitor monitor;

   /** Tracks performance on the Moving Peaks test function */
   MovingPeaksMetric movPeakMetrics = null;

   /** The fitness function being used */
   private FitnessFunction fitnessFunc;

   /** The object that listens for fitness movements */
   private FitnessChangeListener fitnessListener;

   /** Steps before a particle can join a species */
   private int anpsoStepsBeforeSpeciesForm = 2;

   /** The radius to use for the species seed classifier */
   private double spsoRadius = 30;

   /** The maximum number of particles in an SPSO species */
   private int spsoPMax = 10;

   /** The distance between the closest two global optima */
   private double smallestOptimaSeparation = 0.1;

   // /** The minimum number of particles a species must have */
   // private int anpsoSpeciesMinParticles = 2;
   /** Min particles a species can have for a redist */
   private int anpsoRedistDesiredParticles = 8;

   // /** Max fitness difference between best and worst individuals
   // * in a species for a redist to be triggered */
   // private double anpsoMaxFitnessDeltaForRedist = -1000;
   // /** If true, particles will be deleted instead of redistributed */
   // private boolean anpsoDeleteParticles = false;
   /** The fitness of the best solution found so far */
   private double bestFitness;

   /** The required fitness accuracy */
   private double requiredFitnessAccuracy = 0.0001;

   /** The number of optima that have been found */
   private int foundOptima;

   /** The log recorder of the simulation */
   private DataRecorder recorder;

   /** Whether the graphs and swarm displays are shown */
   private boolean showDisplays = true;

   /** The tick to stop on */
   private int stopTick = 0;

   /** The number of evaluations to stop at */
   private int stopEvals = 500000;

   /** Stop when this number of optima have been found */
   private int stopOptima = 0;

   /** The file to record to */
   private String recordFile = "data.txt";

   /** The number of ticks between recording */
   private int recordIntervalTicks = 0;

   /** The number of evaluations between recording */
   private int recordIntervalEvals = 0;

   /** THe number of evals at which we last recorded to disk */
   private int lastRecordEvals = 0;

   /**
    * Whether the pBest location should be used instead of the current location
    * for determining species with SPSO
    */
   private boolean spsoUsePBest = true;

   /** The evaluation counter */
   private EvaluationCounter counter;

   /**
    * True if {@link #getEvaluations()} should report the final evaluations
    * rather than the current evaluations
    */
   private boolean reportFinalEvaluations = false;

   /** The age at which natural species are converted to artificial ones */
   private int espsoConversionAge = 3;

   /** The target population as a fraction of the maximum population */
   private double espsoTargetPopFrac = .75;

   /** The number of particles to place in the new species */
   private int espsoNewSpeciesSize = 8;

   // /** The minimum number of particles to place in the new species */
   // private int espsoMinNewSpeciesSize = 8;
   /** Species closer than this are considered duplicates */
   private double espsoEpsilon = 0.1;

   /** The file to load the regression source points from */
   private String regressionSourceFile = "SalesData.csv";

   private String movementModel = "GCPSO";

   /** The maximum number of excess points for the regression */
   private int regMaxExcess = 10;

   /** The number of peaks for the hump problem */
   private int humpPeaks = 20;

   /** The radius of each peak */
   private double humpPeaksRadius = 0.1;

   /** Whether the humps problem uses the basis function */
   private boolean humpUseBasis = true;

   /** A Plot to display the fitness for each Iteration * */
   protected Plot aPlot;

   /**
    * Creates a new PSOModel
    */
   @SuppressWarnings("unchecked")
   public PSOModel()
   {
      if (!batch)
      {
         ListPropertyDescriptor desc = new ListPropertyDescriptor(
               "FitnessFunction", AVAIL_FITNESS_FUNCS);
         descriptors.put("FitnessFunction", desc);

         desc = new ListPropertyDescriptor("Neighborhood", AVAIL_NEIGHBORHOODS);
         descriptors.put("Neighborhood", desc);

         desc = new ListPropertyDescriptor("MovementModel",
               AVAIL_MOVEMENT_MODELS);
         descriptors.put("MovementModel", desc);
      }

      this.schedule = new Schedule(1);
   }

   /**
    * Returns an array of all the parameters that should be settable on the
    * RePast control panel
    * 
    * @see uchicago.src.sim.engine.SimModel#getInitParam()
    */
   public String[] getInitParam()
   {
      return PARAMETERS;
   }

   /**
    * Begins the simulation
    * 
    * @see uchicago.src.sim.engine.SimModel#begin()
    */
   public void begin()
   {
      foundOptima = 0;
      buildModel();

      System.out.println("Writing to " + recordFile); 
      recorder = new DataRecorder(recordFile, this);
      // recorder.createNumericDataSource("numOptimaFound", this,
      // "getNumOptimaFound");
      recorder.createNumericDataSource("evaluations", this, "getEvaluations");
      recorder.createNumericDataSource("TotalPopulation", swarm,
            "getNumParticles");
      recorder.createNumericDataSource("gBest", this, "getGBest");

      if (monitor != null && !(fitnessFunc instanceof MovingPeaks))
      {
         recorder.addObjectDataSource("OptimaFound", new DataSource() {
            public Integer execute()
            {
               return monitor.findSpeciesOnOptima(smallestOptimaSeparation)
                     .size();
            }
         });
      }

      if (fitnessFunc instanceof MovingPeaks)
      {
         movPeakMetrics = new MovingPeaksMetric(swarm,
               (MovingPeaks) fitnessFunc);
         movPeakMetrics.setUseFitness(false);

         recorder.createNumericDataSource("movpeak Evals", fitnessFunc,
               "get_number_of_evals");
         recorder.createNumericDataSource("OfflineError", movPeakMetrics,
               "getOfflineError");
         recorder.createNumericDataSource("OfflineDistError", movPeakMetrics,
               "getOfflineDistanceError");
         recorder.createNumericDataSource("PCP", movPeakMetrics, "getPCP");
         // recorder.createNumericDataSource("AllKnownPeaksAvgErr", metric,
         // "getAllKnownPeaksAvgErr");
         recorder.createNumericDataSource("BKPE Distance", movPeakMetrics,
               "getBKPE");

         movPeakMetrics = new MovingPeaksMetric(swarm,
               (MovingPeaks) fitnessFunc);
         movPeakMetrics.setUseFitness(true);
         recorder.createNumericDataSource("BKPE Fitness", movPeakMetrics,
               "getBKPE");

         if (movPeaksCERecordFreq > 0)
         {
            final MovingPeaks peaks = (MovingPeaks) fitnessFunc;

            if (getController().getRunCount() == 1)
            {
               try
               {
                  FileWriter writer = new FileWriter("CE_" + recordFile, false);
                  StringBuffer strbuf = new StringBuffer("run");

                  for (int i = 0; i < peaks.getChangeFrequency(); i += movPeaksCERecordFreq)
                  {
                     strbuf.append(",");
                     strbuf.append(i);
                  }

                  strbuf.append("\n");

                  writer.append(strbuf.toString());
                  writer.close();
               }
               catch (IOException ioe)
               {
                  throw new RuntimeException(ioe);
               }
            }

            if (peaks.getChangeFrequency() % movPeaksCERecordFreq != 0)
            {
               throw new IllegalArgumentException(
                     "Peak change frequency must be"
                           + "a multiple of movPeaksCERecordFreq");
            }

            peaks.getEvalObservable().addObserver(new Observer() {
               public void update(Observable o, Object arg)
               {
                  int numEvals = peaks.get_number_of_evals();

                  // Don't record anything after we were supposed to stop
                  if (stopEvals > 0 && numEvals > stopEvals)
                     return;

                  if (numEvals % movPeaksCERecordFreq == 0)
                     movPeakMetrics.recordCurrentError();
               }
            });
         }
      }

      // if (neighborhood instanceof SpeciesNeighborhood)
      // recorder.createNumericDataSource("UnspeciedPopulation", this,
      // "getNumUnspeciedParticles");
      buildSchedule();

      if (showDisplays)
      {
         buildDisplay();
         if (particlesOverTime != null)
            particlesOverTime.display();
         if (foundOptimaOverTime != null)
            foundOptimaOverTime.display();
         if (movingPeaksMetrics != null)
            movingPeaksMetrics.display();
         if (limitOverTime != null)
            limitOverTime.display();
         if (speciesFoundOptimaOverTime != null)
            speciesFoundOptimaOverTime.display();

         if (display != null)
            display.display();
         if (aPlot != null)
            aPlot.display();
      }
   }

   /**
    * Stops the simulation
    */
   @Override
   public void stop()
   {
      recorder.record();

      if (movPeakMetrics != null && movPeaksCERecordFreq > 0)
      {
         StringBuffer strbuf = new StringBuffer();
         strbuf.append(this.getController().getRunCount());

         MovingPeaks peaks = (MovingPeaks) fitnessFunc;

         double[] errorHist = movPeakMetrics.getCurrentErrorHistory();
//         System.out.println(Arrays.toString(errorHist));
//         double[] results = new double[peaks.getChangeFrequency()
//               / movPeaksCERecordFreq];
//         int[] sampleCounts = new int[results.length];
//
//         for (int i = 0; i < errorHist.length; i++)
//         {
//            int temp = i % results.length;
//            results[temp] += errorHist[i];
//            sampleCounts[temp]++;
//         }
//
//         for (int i = 0; i < results.length; i++)
//            results[i] /= sampleCounts[i];

         for (double currError : errorHist)//results)
         {
            strbuf.append(",");
            strbuf.append(currError);
         }

         strbuf.append("\n");

         try
         {
            FileWriter writer = new FileWriter("CE_" + recordFile, true);
            writer.append(strbuf.toString());
            writer.close();
         }
         catch (IOException ioe)
         {
            throw new RuntimeException(ioe);
         }
      }

      // Particle bestParticle = getBestParticle();
      //      
      // System.out.print("Best solution found: ");
      // System.out.print(Arrays.toString(
      // bestParticle.getPersonalBestLocation()));
      // System.out.print(" (Fitness: " +
      // bestParticle.getPersonalBestFitness()[0]);
      // System.out.println(")");

      // if (batch)
      // System.out.println();

      super.stop();
   }

   /**
    * Returns the movPeaksNumPeaks.
    * 
    * @return Returns the movPeaksNumPeaks.
    */
   public int getMovPeaksNumPeaks()
   {
      return movPeaksNumPeaks;
   }

   /**
    * Sets the movPeaksNumPeaks
    * 
    * @param movPeaksNumPeaks
    *           The movPeaksNumPeaks to set.
    */
   public void setMovPeaksNumPeaks(int movPeaksNumPeaks)
   {
      this.movPeaksNumPeaks = movPeaksNumPeaks;
   }

   public int getRegMaxExcess()
   {
      return regMaxExcess;
   }

   public void setRegMaxExcess(int regMaxExcess)
   {
      this.regMaxExcess = regMaxExcess;
   }

   /**
    * @return the movPeaksCERecordFreq
    */
   public int getMovPeaksCERecordFreq()
   {
      return movPeaksCERecordFreq;
   }

   /**
    * @param movPeaksCERecordFreq
    *           the movPeaksCERecordFreq to set
    */
   public void setMovPeaksCERecordFreq(int movPeaksOERecordFreq)
   {
      this.movPeaksCERecordFreq = movPeaksOERecordFreq;
   }

   /**
    * @return the movPeaksCustMaxAmp
    */
   public double getMovPeaksCustMaxAmp()
   {
      return movPeaksCustMaxAmp;
   }

   /**
    * @param movPeaksCustMaxAmp
    *           the movPeaksCustMaxAmp to set
    */
   public void setMovPeaksCustMaxAmp(double movPeaksCustMaxAmp)
   {
      this.movPeaksCustMaxAmp = movPeaksCustMaxAmp;
   }

   /**
    * @return the movPeaksCustMaxFreq
    */
   public double getMovPeaksCustMaxFreq()
   {
      return movPeaksCustMaxFreq;
   }

   /**
    * @param movPeaksCustMaxFreq
    *           the movPeaksCustMaxFreq to set
    */
   public void setMovPeaksCustMaxFreq(double movPeaksCustMaxFreq)
   {
      this.movPeaksCustMaxFreq = movPeaksCustMaxFreq;
   }

   /**
    * @return the movPeaksCustMinPower
    */
   public double getMovPeaksCustMinPower()
   {
      return movPeaksCustMinPower;
   }

   /**
    * @param movPeaksCustMinPower
    *           the movPeaksCustMinPower to set
    */
   public void setMovPeaksCustMinPower(double movPeaksCustMinPower)
   {
      this.movPeaksCustMinPower = movPeaksCustMinPower;
   }

   /**
    * @return the movPeaksMaxPower
    */
   public double getMovPeaksCustMaxPower()
   {
      return movPeaksCustMaxPower;
   }

   /**
    * @param movPeaksMaxPower
    *           the movPeaksMaxPower to set
    */
   public void setMovPeaksCustMaxPower(double movPeaksMaxPower)
   {
      this.movPeaksCustMaxPower = movPeaksMaxPower;
   }

   /**
    * @return the movPeaksUseCustomPeakFunc
    */
   public boolean isMovPeaksUseCustomPeakFunc()
   {
      return movPeaksUseCustomPeakFunc;
   }

   /**
    * @param movPeaksUseCustomPeakFunc
    *           the movPeaksUseCustomPeakFunc to set
    */
   public void setMovPeaksUseCustomPeakFunc(boolean movPeaksUseCustomPeakFunc)
   {
      this.movPeaksUseCustomPeakFunc = movPeaksUseCustomPeakFunc;
   }

   public double getMovPeaksSeverity()
   {
      return movPeaksSeverity;
   }

   public void setMovPeaksSeverity(double movPeaksSeverity)
   {
      this.movPeaksSeverity = movPeaksSeverity;
   }

   /**
    * Returns the max particles in an SPSO species
    * 
    * @return The max number of particles
    */
   public int getSpsoPMax()
   {
      return spsoPMax;
   }

   /**
    * Sets the max number of particles in an SPSO species
    * 
    * @param spsoPMax
    *           The max number of particles in an SPSO species
    */
   public void setSpsoPMax(int spsoPMax)
   {
      this.spsoPMax = spsoPMax;
   }

   /**
    * Builds the PSO model
    */
   public void buildModel()
   {
      Random.createUniform();
      swarm = new Swarm();
      swarm.setMinParticles(particles);
      fitnessFunc = createFitnessFunction();
      counter = new EvaluationCounter(fitnessFunc);

      if (fitnessFunc instanceof MovingPeaks)
         fitnessListener = new FitnessChangeListener(swarm, counter);
      
      neighborhood = createNeighborhood();

      if (neighborhood instanceof SpeciesNeighborhood)
      {
         if (fitnessFunc.getNumVariables() < 32)
         {
            monitor = new SpeciesNeighborhoodMonitor(swarm,
                  (SpeciesNeighborhood) neighborhood);
            monitor.setRequiredFitnessAccuracy(requiredFitnessAccuracy);
            if (fitnessFuncName.equalsIgnoreCase("Humps"))
               monitor.setOptimalFitness(1);
         }
         else
            monitor = null;

         if (neighborhoodName.toLowerCase().contains("anpso"))
         {
            System.out.println("Using redistrib");
            ParticleRedistributor redistrib = new ParticleRedistributor(swarm,
                  (SpeciesNeighborhood) neighborhood,
                  anpsoRedistDesiredParticles);
            // redistrib.setDeleteParticles(anpsoDeleteParticles);

            swarm.addSwarmListener(redistrib);
         }
         // else if (((SpeciesNeighborhood) neighborhood).getClassifier()
         // instanceof SPSO && spsoPMax > 0)
         // {
         // ParticleRedistributor redistrib = new ParticleRedistributor(swarm,
         // (SpeciesNeighborhood) neighborhood, spsoPMax);
         // redistrib.setPBestFitnessDelta(Double.POSITIVE_INFINITY);
         // // redistrib.setDeleteParticles(anpsoDeleteParticles);
         //             
         // swarm.addSwarmListener(redistrib);
         // }
      }
      else
         monitor = null;

      if (movementModel.equals("Constriction"))
         swarm
               .setDefaultMovementModel(new ConstrictionPSO(swarm, neighborhood));
      // else if (movementModel.equals("FDR_PSO"))
      // swarm.setDefaultMovementModel(FDR_Attractor.buildConstrictionFDRPSO(swarm));
      else if (movementModel.equals("Gravity_PSO"))
         swarm.setDefaultMovementModel(new GravityPSO(swarm));
      else if (movementModel.equals("GCPSO"))
         swarm.setDefaultMovementModel(new GCPSO(swarm, neighborhood));

      swarm.addFitnessFunction(counter);
      swarm.addParticles(getParticles());

      if (fitnessFunc instanceof MovingPeaks)
         swarm.addSwarmListener(fitnessListener);
   }

   /**
    * Creates and initialises the Schedule object
    */
   public void buildSchedule()
   {
      this.schedule = new Schedule(1);
      this.schedule.scheduleActionBeginning(0, new TickListener());
      schedule.scheduleActionAtEnd(recorder, "writeToFile");
      if (stopTick > 0)
         schedule.scheduleActionAt(stopTick, this, "stop", ScheduleBase.LAST);
   }

   /**
    * Creates the user display
    */
   public void buildDisplay()
   {
      if (!showDisplays)
         return;

      display = new DisplaySurface(this, "Swarm");
      registerDisplaySurface("Main", display);

      if (neighborhood instanceof SpeciesNeighborhood)
      {
         // foundOptimaOverTime = new OpenSequenceGraph("Found Optima Over
         // Time", this);
         // foundOptimaOverTime.setAxisTitles("Iteration", "Number");
         // foundOptimaOverTime.addSequence("Found Optima", new Sequence() {
         // public double getSValue() {
         // return foundOptima;
         // }
         // });
         // foundOptimaOverTime.addSequence("Species", new Sequence()
         // {
         // public double getSValue()
         // {
         // SpeciesNeighborhood sNeighborhood =
         // (SpeciesNeighborhood) neighborhood;
         //
         // return sNeighborhood.getAllSpecies().size();
         // }
         // });
         //          
         // particlesOverTime = new OpenSequenceGraph("Particles Over Time",
         // this);
         // particlesOverTime.setAxisTitles("Iteration", "Particles");
         // particlesOverTime.addSequence("Unspecied Particles", new Sequence()
         // {
         // public double getSValue() {
         // SpeciesNeighborhood sNeighborhood =
         // (SpeciesNeighborhood) neighborhood;
         //
         // return sNeighborhood.getUnspeciedParticles().size();
         // }
         // });
         // particlesOverTime.addSequence("Total Particles", new Sequence()
         // {
         // public double getSValue() {
         // return swarm.getNumParticles();
         // }
         // });

         // speciesFitness = new Plot("Species Fitness", this);
         // speciesFitness.addLegend(0, "Species On An Optimum", Color.BLUE,
         // OpenGraph.CROSS);
         // speciesFitness.addLegend(1, "Species", Color.RED,
         // OpenGraph.CROSS);
         // speciesFitness.setAxisTitles("Fitness", "Species Rank");
         // speciesFitness.setConnected(true);
         // speciesFitness.setXRange(Double.MIN_VALUE, Double.MAX_VALUE);
         // speciesFitness.setYRange(0, 10);
         // speciesFitness.setYIncrement(1);
         // speciesFitness.display();
      }

      if (fitnessFunc instanceof MovingPeaks)
      {
         int numPeaks = ((MovingPeaks) fitnessFunc).getNumPeaks();
         double maxRange = Math.sqrt(fitnessFunc.getPreferredRanges()[0]
               .getSize()
               * fitnessFunc.getPreferredRanges()[0].getSize() * numPeaks);

         peakRepresentation = new Plot("Particles on each peak", this);
         peakRepresentation.setAxisTitles("Distance from peak", "Peak number");
         peakRepresentation.setYRange(0, numPeaks + 1);
         peakRepresentation.setXRange(0, 10);
         peakRepresentation.setYIncrement(1);
         peakRepresentation.setConnected(false);
         peakRepresentation.addLegend(0, "Particle", Color.BLACK);
         peakRepresentation.addLegend(1, "Particle", Color.RED);
         peakRepresentation.addLegend(2, "Hidden peak", Color.GRAY);

         peakRepresentation.display();

         final MovingPeaksMetric metrics = new MovingPeaksMetric(swarm,
               (MovingPeaks) fitnessFunc);

         movingPeaksMetrics = new OpenSequenceGraph("Moving Peaks Metrics",
               this);
         movingPeaksMetrics.setAxisTitles("Time", "Value");
         movingPeaksMetrics.addSequence("Offline error", new Sequence() {
            public double getSValue()
            {
               if (swarm.getCurrTick() % 10 == 0)
               {
                  System.out.print("Tick " + swarm.getCurrTick() + ": ");
                  System.out.print(" Current: " + metrics.getCurrentError());
                  System.out.print(" Offline: " + metrics.getOfflineError());
                  System.out.print(" Peak cover: " + metrics.getPCP());
                  System.out.println(" BKPE: " + metrics.getBKPE());
               }
               return metrics.getOfflineError();
            }
         });
         movingPeaksMetrics.addSequence("Offline distance error",
               new Sequence() {
                  public double getSValue()
                  {
                     return metrics.getOfflineDistanceError();
                  }
               });
         movingPeaksMetrics.addSequence("BKPE", new Sequence() {
            public double getSValue()
            {
               return metrics.getBKPE();
            }
         });
         // movingPeaksMetrics.addSequence("All known peaks error", new
         // Sequence() {
         // public double getSValue() {
         // return metrics.getAllKnownPeaksAvgErr();
         // }
         // });
         movingPeaksMetrics.addSequence("PCP", new Sequence() {
            public double getSValue()
            {
               return 100 * metrics.getPCP();
            }
         });
         movingPeaksMetrics.addSequence("Current BKPE", new Sequence() {
            public double getSValue()
            {
               double value = metrics.getCurrError(metrics.getBestKnownPeak());
               if (value == MovingPeaks.NO_PARTICLES)
                  return -1;
               else
                  return value;
            }
         });
         movingPeaksMetrics.addSequence("Current Error", new Sequence() {
            public double getSValue()
            {
               return metrics.getCurrentError();
            }
         });
      }

      if (swarm.getNumVariables() == 2)
      {
         SwarmDisplay2D swarmDisplay = new SwarmDisplay2D(swarm);
         if (neighborhood instanceof StarNeighborhood)
            swarmDisplay.setParticleColor(((StarNeighborhood) neighborhood)
                  .getFocalParticle(), Color.RED);

         FitnessDisplay2D fitnessDisplay = new FitnessDisplay2D(swarm, display);

         display.addDisplayableProbeable(fitnessDisplay, "Fitness");

         if (neighborhood != null)
         {
            NeighborhoodDisplay2D neighborhoodDisplay = new NeighborhoodDisplay2D(
                  swarm, neighborhood);
            display.addDisplayable(neighborhoodDisplay, "Neighbourhoods");
         }
         display.addDisplayableProbeable(swarmDisplay, "Swarm");
         addSimEventListener(display);

         if (monitor != null)
            monitor.setDisplay(swarmDisplay);

         if (classifier instanceof ESPSO)
         {
            // display.addDisplayable(new SpeciesAgeDisplay(swarm,
            // (SpeciesNeighborhood) neighborhood,
            // ((ESPSO) classifier).getBaseSPSO()), "Species Ages");

            // LabelDisplay labelDisp = new LabelDisplay(swarm);
            // ((ESPSO) classifier).setLabelDisplay(labelDisp);
            // display.addDisplayable(labelDisp, "Labels");
         }

         if (neighborhood instanceof SpeciesNeighborhood)
            display.addDisplayable(new SpeciesDisplay2D(tracker, swarmDisplay),
                  "Species");
      }
      else
         display = null;

      // evaluationsOverTime = new OpenSequenceGraph("Evaluations", this);
      // evaluationsOverTime.setAxisTitles("Iteration", "Number of Species");
      // evaluationsOverTime.addSequence("Evaluations", new Sequence() {
      // public double getSValue() {
      // return evaluations;
      // }
      // });

      // if (neighborhood instanceof SpeciesNeighborhood)
      // {
      // speciesFitness = new Plot("Species Fitness", this);
      // speciesFitness.addLegend(0, "Fitness", Color.RED, OpenGraph.CROSS);
      // speciesFitness.setAxisTitles("Fitness", "Species Rank");
      // speciesFitness.setConnected(true);
      // speciesFitness.setXRange(Double.MIN_VALUE, Double.MAX_VALUE);
      // speciesFitness.setYRange(0, 10);
      // speciesFitness.setYIncrement(1);
      // speciesFitness.display();
      // }
      // else if (speciesFitness != null)
      // {
      // speciesFitness.dispose();
      // speciesFitness = null;
      // }
      //       
      // if (neighborhood instanceof SpeciesNeighborhood)
      // {
      // particlesOverTime.addSequence("Unspecied Particles", new Sequence() {
      // public double getSValue() {
      // return getNumUnspeciedParticles();
      //
      // }
      // });
      //
      //          
      // limitOverTime = new OpenSequenceGraph("Species Distance Limit Over
      // Time", this);
      // limitOverTime.setAxisTitles("Iteration", "Distance Limit");
      // limitOverTime.addSequence("Distance Limit", new Sequence() {
      // public double getSValue() {
      // return ((SpeciesNeighborhood) neighborhood).getLimit();
      // }
      // });
      //
      // monitor = new SpeciesNeighborhoodMonitor(swarm,
      // (SpeciesNeighborhood) neighborhood, swarmDisplay);
      //          
      // // monitor.setOptimas(optimas);
      //
      //          
      // foundOptimaOverTime.addSequence("Species on Optima", new Sequence() {
      // public double getSValue() {
      // return monitor.countFoundOptimas();
      // }
      // });
      // foundOptimaOverTime.addSequence("Num Species", new Sequence() {
      // public double getSValue() {
      // return ((SpeciesNeighborhood) neighborhood).getSpeciesCount();
      // }
      // });
      // }
      // else
      // monitor = null;
      // Create the plot
      if (fitnessFunc instanceof MovingPeaks)
      {
         aPlot = new Plot("Decision space plot");
         aPlot.addLegend(1, "Seed", Color.RED, aPlot.CROSS);
         aPlot.addLegend(2, "Peak", Color.BLUE, aPlot.SQUARE);
         aPlot.addLegend(3, "Maximum Peak", Color.GREEN, aPlot.CROSS);
         aPlot.addLegend(4, "Quantum", Color.BLACK, aPlot.CIRCLE);

         aPlot.clear(0);
         // Display a plot to for all the particles at the 1st iteration
         aPlot.setConnected(false);
         for (Particle p : swarm.getParticles())
            aPlot.plotPoint(p.getLocation()[0], p.getLocation()[1], 0);
         // for (int i = 0; i < swarm.getNumParticles(); i++) { //step through
         // each particle
         //             
         // PSOAgent p = (PSOAgent)agentList.get(i);
         // aPlot.plotPoint(p.present[0], p.present[1], 0);
         // }
         aPlot.fillPlot();
         aPlot.updateGraph();
      }
   }

   /**
    * Returns the number of particles that don't belong to any species
    * 
    * @return The number of particles that don't belong to any species
    */
   public int getNumUnspeciedParticles()
   {
      SpeciesNeighborhood sNeighborhood = (SpeciesNeighborhood) neighborhood;

      List<Particle> unspeciedParticles = new ArrayList<Particle>(swarm
            .getParticles());
      for (Species species : sNeighborhood.getAllSpecies())
      {
         for (Particle p : species.getParticles())
            unspeciedParticles.remove(p);
      }

      return unspeciedParticles.size();
   }

   /**
    * Sets up the simulation for a run
    * 
    * @see uchicago.src.sim.engine.SimModel#setup()
    */
   public void setup()
   {
      if (this.display != null)
         this.display.dispose();
      if (particlesOverTime != null)
         particlesOverTime.dispose();
      if (foundOptimaOverTime != null)
         foundOptimaOverTime.dispose();
      if (limitOverTime != null)
         limitOverTime.dispose();
      if (speciesFoundOptimaOverTime != null)
         speciesFoundOptimaOverTime.dispose();
      if (peakRepresentation != null)
         peakRepresentation.dispose();
      if (movingPeaksMetrics != null)
         movingPeaksMetrics.dispose();
      if (speciesFitness != null)
         speciesFitness.dispose();
      if (aPlot != null)
         aPlot.dispose();

      display = null;
      particlesOverTime = null;
      foundOptimaOverTime = null;
      limitOverTime = null;
      speciesFoundOptimaOverTime = null;
      peakRepresentation = null;
      movingPeaksMetrics = null;
      speciesFitness = null;
      aPlot = null;
   }

   /**
    * Returns the number of optima found so far
    * 
    * @return The number of optima found so far
    */
   public int getNumOptimaFound()
   {
      return foundOptima;
   }

   /**
    * Returns the global best fitness found
    * 
    * @return The global best fitness found
    */
   public double getGBest()
   {
      return bestFitness;
   }

   /**
    * Returns the schedule object
    * 
    * @see uchicago.src.sim.engine.SimModel#getSchedule()
    */
   public Schedule getSchedule()
   {
      return this.schedule;
   }

   /**
    * Returns the name of the model
    * 
    * @see uchicago.src.sim.engine.SimModel#getName()
    */
   public String getName()
   {
      return "Particle Swarm Model";
   }

   /**
    * Returns the number of particles in the simulation
    * 
    * @return The number of particles.
    */
   public int getParticles()
   {
      return this.particles;
   }

   /**
    * Sets the number of particles in the simulation
    * 
    * @param numParticles
    *           The number of particles to have
    */
   public void setParticles(int numParticles)
   {
      this.particles = numParticles;
   }

   public void setMovementModel(String movementModel)
   {
      this.movementModel = movementModel;
   }

   public String getMovementModel()
   {
      return this.movementModel;
   }

   /**
    * Returns the best particle in a single-objective function
    * 
    * @return The best particle
    */
   public Particle getBestParticle()
   {
      ArrayList<Particle> particles = new ArrayList<Particle>();
      particles.addAll(swarm.getParticles());
      Collections.sort(particles, new FitnessComparator(true));

      return particles.get(particles.size() - 1);
   }

   /**
    * Sets the name of the data file to load the regression target points from
    * 
    * @param filename
    *           The file to load the regression target points from
    */
   public void setRegressionDataFile(String filename)
   {
      regressionSourceFile = filename;
   }

   /**
    * Returns the name of the data file used to load the regresion target points
    * from
    * 
    * @return The name of the data file
    */
   public String getRegressionDataFile()
   {
      return regressionSourceFile;
   }

   /**
    * @return Returns the closestOptimaDistance.
    */
   public double getSmallestOptimaSeparation()
   {
      return smallestOptimaSeparation;
   }

   /**
    * @param closestOptimaDistance
    *           The closestOptimaDistance to set.
    */
   public void setSmallestOptimaSeparation(double closestOptimaDistance)
   {
      this.smallestOptimaSeparation = closestOptimaDistance;
   }

   /**
    * Returns the fitnessFuncName
    * 
    * @return The fitnessFuncName.
    */
   public String getFitnessFunction()
   {
      return this.fitnessFuncName;
   }

   /**
    * Sets the fitnessFuncName
    * 
    * @param fitnessFuncName
    *           The fitnessFuncName to set.
    */
   public void setFitnessFunction(String fitnessFuncName)
   {
      this.fitnessFuncName = fitnessFuncName;
   }

   /**
    * Returns the neighborhoodName
    * 
    * @return The neighborhoodName.
    */
   public String getNeighborhood()
   {
      return this.neighborhoodName;
   }

   /**
    * Sets the neighborhoodName
    * 
    * @param neighborhoodName
    *           The neighborhoodName to set.
    */
   public void setNeighborhood(String neighborhoodName)
   {
      this.neighborhoodName = neighborhoodName;
   }

   /**
    * Returns the variables
    * 
    * @return The variables.
    */
   public int getVariables()
   {
      return this.variables;
   }

   /**
    * Sets the variables
    * 
    * @param variables
    *           The variables to set.
    */
   public void setVariables(int variables)
   {
      this.variables = variables;
   }

   /**
    * Returns the showDisplays
    * 
    * @return The showDisplays.
    */
   public boolean isShowDisplays()
   {
      return this.showDisplays;
   }

   /**
    * Sets the showDisplays
    * 
    * @param showDisplays
    *           The showDisplays to set.
    */
   public void setShowDisplays(boolean showDisplays)
   {
      this.showDisplays = showDisplays;
   }

   /**
    * Returns the stopTick
    * 
    * @return The stopTick.
    */
   public int getStopTick()
   {
      return this.stopTick;
   }

   /**
    * Sets the stopTick
    * 
    * @param stopTick
    *           The stopTick to set.
    */
   public void setStopTick(int stopTick)
   {
      this.stopTick = stopTick;
   }

   /**
    * Returns the stopEvals.
    * 
    * @return Returns the stopEvals.
    */
   public int getStopEvals()
   {
      return stopEvals;
   }

   /**
    * Sets the stopEvals
    * 
    * @param stopEvals
    *           The stopEvals to set.
    */
   public void setStopEvals(int stopEvals)
   {
      this.stopEvals = stopEvals;
   }

   /**
    * Returns the evaluations
    * 
    * @return The evaluations.
    */
   public int getEvaluations()
   {
      return counter.getEvaluations();
   }

   /**
    * Returns the recordgFile
    * 
    * @return The recordgFile.
    */
   public String getRecordFile()
   {
      return this.recordFile;
   }

   /**
    * Sets the recordFile
    * 
    * @param recordFile
    *           The recordgFile to set.
    */
   public void setRecordFile(String recordFile)
   {
      this.recordFile = recordFile;
   }

   public boolean isSpsoUsePBest()
   {
      return spsoUsePBest;
   }

   public void setSpsoUsePBest(boolean usePBest)
   {
      this.spsoUsePBest = usePBest;
   }

   /**
    * @return the humpPeaks
    */
   public int getHumpPeaks()
   {
      return humpPeaks;
   }

   /**
    * @param humpPeaks
    *           the humpPeaks to set
    */
   public void setHumpPeaks(int humpPeaks)
   {
      this.humpPeaks = humpPeaks;
   }

   public double getHumpPeaksRadius()
   {
      return humpPeaksRadius;
   }

   public void setHumpPeaksRadius(double humpPeakRadius)
   {
      this.humpPeaksRadius = humpPeakRadius;
   }

   public boolean isHumpUseBasis()
   {
      return humpUseBasis;
   }

   public void setHumpUseBasis(boolean humpUseBasis)
   {
      this.humpUseBasis = humpUseBasis;
   }

   /**
    * @return Returns the espsoConversionAge.
    */
   public int getEspsoConversionAge()
   {
      return espsoConversionAge;
   }

   /**
    * @param espsoConversionAge
    *           The espsoConversionAge to set.
    */
   public void setEspsoConversionAge(int espsoConversionAge)
   {
      this.espsoConversionAge = espsoConversionAge;
   }

   /**
    * @return Returns the espsoNewSpeciesSize.
    */
   public int getEspsoNewSpeciesSize()
   {
      return espsoNewSpeciesSize;
   }

   /**
    * @param espsoNewSpeciesSize
    *           The espsoNewSpeciesSize to set.
    */
   public void setEspsoNewSpeciesSize(int espsoNewSpeciesSize)
   {
      this.espsoNewSpeciesSize = espsoNewSpeciesSize;
   }

   // /**
   // * @return Returns the espsoMinNewSpeciesSize.
   // */
   // public int getEspsoMinNewSpeciesSize()
   // {
   // return espsoMinNewSpeciesSize;
   // }
   //
   // /**
   // * @param espsoMinNewSpeciesSize The espsoMinNewSpeciesSize to set.
   // */
   // public void setEspsoMinNewSpeciesSize(int espsoMinNewSpeciesSize)
   // {
   // this.espsoMinNewSpeciesSize = espsoMinNewSpeciesSize;
   // }

   /**
    * @return Returns the espsoTargetPopFrac.
    */
   public double getEspsoTargetPopFrac()
   {
      return espsoTargetPopFrac;
   }

   /**
    * @param espsoTargetPopFrac
    *           The espsoTargetPopFrac to set.
    */
   public void setEspsoTargetPopFrac(double espsoTargetPopFrac)
   {
      this.espsoTargetPopFrac = espsoTargetPopFrac;
   }

   /**
    * @return Returns the espsoEpsilon.
    */
   public double getEspsoEpsilon()
   {
      return espsoEpsilon;
   }

   /**
    * @param espsoEpsilon
    *           The espsoEpsilon to set.
    */
   public void setEspsoEpsilon(double espsoEpsilon)
   {
      this.espsoEpsilon = espsoEpsilon;
   }

   /**
    * Called by the TickListener whenever its execute method is called
    */
   protected void tick()
   {
      // if (batch && System.currentTimeMillis() - lastTickAnnouncement > 2000)
      // {
      // DateFormat df = DateFormat.getTimeInstance(DateFormat.SHORT);
      //         
      // System.out.print("\r(" + df.format(new Date()) + ") tick " +
      // swarm.getCurrTick() + " ");
      // lastTickAnnouncement = System.currentTimeMillis();
      // }

      if (neighborhood instanceof SpeciesNeighborhood
            && recordIntervalTicks > 0
            && swarm.getCurrTick() % recordIntervalTicks == 0)
      {
         try
         {
            StringBuffer filename = new StringBuffer(recordFile);
            filename.append(getFitnessFunction());
            filename.append("-");
            filename.append(getVariables());
            filename.append("D_");
            filename.append(getNeighborhood());
            filename.append(".csv");

            PrintWriter writer = new PrintWriter(new BufferedWriter(
                  new FileWriter(new File(filename.toString()), true)));

            writer.print(swarm.getCurrTick() + ", ");
            writer.print(ANPSO
                  .calculateAvgClosestDistance(swarm.getParticles()));

            ArrayList<Particle> seeds = new ArrayList<Particle>();
            FitnessComparator comparator = new FitnessComparator(true);

            for (Set<Particle> species : monitor
                  .findSpeciesOnOptima(smallestOptimaSeparation))
               seeds.add(Collections.max(species, comparator));

            // ArrayList<Double> distances = new
            // ArrayList<Double>(seeds.size());

            Collections.sort(seeds, comparator);
            Collections.reverse(seeds);

            // for (Particle p : seeds)
            // {
            // double minDistance = Double.MAX_VALUE;
            // int minDistIdx = -1;
            // int index = 0;
            //               
            // for (Particle q : seeds)
            // {
            // if (p == q)
            // continue;
            //                  
            // double distance = VectorMath.length(VectorMath.subtract(
            // p.getPersonalBestLocation(), q.getPersonalBestLocation()));
            //                  
            // if (distance < minDistance)
            // {
            // minDistance = distance;
            // minDistIdx = index;
            // }
            //                  
            // index++;
            // }
            //               
            //               
            //               
            //               
            // distances.add(minDistance);
            // }

            // for (int i = 0 ; i < seeds.size(); i++)
            // {
            // writer.print(", " +
            // Arrays.toString(seeds.get(i).getPersonalBestLocation()) + ", " +
            // distances.get(i));
            // }

            for (int i = 0; i < seeds.size(); i++)
            {
               writer.print(", " + seeds.get(i).getPersonalBestFitness()[0]);
            }

            writer.println();
            writer.close();
         }
         catch (IOException ioe)
         {

         }
      }

      swarm.step();

      if (fitnessFunc instanceof KnapsackProblem)
      {
         System.out.println("---");
         int bestFitness = (int) Collections.max(swarm.getParticles(),
               new FitnessComparator(true)).getPersonalBestFitness()[0];
         ArrayList<int[]> uniqueSolutions = new ArrayList<int[]>();

         System.out.println("Best fitness: " + bestFitness);

         for (Particle p : swarm.getParticles())
         {
            if (p.getPersonalBestFitness()[0] == bestFitness)
            {
               // int[] sol = new int[swarm.getNumVariables()];
               // for (int i = 0; i < swarm.getNumVariables(); i++)
               // {
               // sol[i] = (int) p.getPersonalBestLocation()[i];
               // }
               //               
               // boolean found = false;
               //               
               // for (int[] testSol : uniqueSolutions)
               // {
               // boolean diff = false;
               //                  
               // for (int j = 0; j < testSol.length && !diff ; j++)
               // {
               // if (testSol[j] != sol[j])
               // diff = true;
               // }
               //                  
               // if (!diff)
               //                     
               // }

               // for (int i = 0; i < swarm.getNumVariables(); i++)
               // System.out.print((int) p.getPersonalBestLocation()[i] + ", ");

               // System.out.println();
               // total++;
            }
         }

      }
      // evaluations += swarm.getNumParticles();
      // if (dynamicAdding)
      // addParticlesIfNeeded();

      // if (stepsBetweenAdd != 0 && (int) getTickCount() % stepsBetweenAdd ==
      // 0)
      // swarm.addParticles(particlesPerAdd, model);
      //     
      // if ((int) getTickCount() % 20 == 0)
      // System.out.println("Tick " + ((int) getTickCount()) + ": "
      // + evaluations + " evaluations");

      if (showDisplays)
      {
         this.updateGraphs();

         if (display != null)
         {
            if (monitor != null)
               monitor.updateDisplay();

            display.updateDisplay();
         }
      }

      ArrayList<Particle> sortedParticles = new ArrayList<Particle>(swarm
            .getParticles());
      Collections.sort(sortedParticles, new FitnessComparator(true));
      Particle best = sortedParticles.get(sortedParticles.size() - 1);
      // double lastGBest = bestFitness;
      bestFitness = best.getPersonalBestFitness()[0];

      if (monitor != null)
         foundOptima = monitor.findSpeciesOnOptima(smallestOptimaSeparation)
               .size();

      if (stopOptima > 0 && foundOptima >= stopOptima)
         stop();
      // if (bestFitness - lastGBest > 1E-9)
      // {
      // System.out.println("Tick " + swarm.getCurrTick() +
      // " gBest fitness: " + bestFitness);
      // }

      if ((recordIntervalTicks > 0 && swarm.getCurrTick() % recordIntervalTicks == 0)
            || (recordIntervalEvals > 0 && counter.getEvaluations()
                  / recordIntervalEvals != lastRecordEvals
                  / recordIntervalEvals))
      {
         recorder.record();
         recorder.writeToFile();

         lastRecordEvals = counter.getEvaluations();
      }

      if (fitnessFunc instanceof GraphingRegressionEvaluator)
         ((GraphingRegressionEvaluator) fitnessFunc).updateGraph();

      if (stopEvals > 0 && counter.getEvaluations() >= stopEvals)
         stop();
   }

   // private void addParticlesIfNeeded()
   // {
   // if (!(neighborhood instanceof SpeciesNeighborhood))
   // return;
   //      
   // SpeciesNeighborhood sNeighborhood = (SpeciesNeighborhood) neighborhood;
   //      
   // List<Particle> unSpeciedParticles =
   // new ArrayList<Particle>(swarm.getParticles());
   //      
   // for (HashSet<Particle> species : sNeighborhood.getAllSpecies())
   // {
   // for (Particle p : species)
   // unSpeciedParticles.remove(p);
   // }
   //      
   // double bestFitness = Double.NEGATIVE_INFINITY;
   //      
   // for (Particle p : unSpeciedParticles)
   // {
   // if (p.getPersonalBestFitness()[0] > bestFitness)
   // bestFitness = p.getPersonalBestFitness()[0];
   // }
   //         
   // if (bestFitness != lastBestFitness)
   // {
   // lastBestFitnessChange = (int) getTickCount();
   // lastBestFitness = bestFitness;
   // }
   //      
   // if ((int) getTickCount() - lastBestFitnessChange >
   // unSpeciedParticles.size())
   // {
   // for (int i = 0; i < particlesPerAdd; i++)
   // swarm.addParticle(model);
   //         
   // lastBestFitnessChange = (int) getTickCount();
   // }
   // }

   /**
    * Updates the graphs being displayed to the user
    */
   protected void updateGraphs()
   {
      if (particlesOverTime != null)
         particlesOverTime.step();
      if (foundOptimaOverTime != null)
         foundOptimaOverTime.step();
      if (limitOverTime != null)
         limitOverTime.step();
      if (speciesFoundOptimaOverTime != null)
         speciesFoundOptimaOverTime.step();
      if (peakRepresentation != null)
      {
         peakRepresentation.clear(0);
         peakRepresentation.clear(1);
         peakRepresentation.clear(2);

         double[][] peakLocations = ((MovingPeaks) fitnessFunc)
               .getPeakPositions();
         double[] peakHeights = ((MovingPeaks) fitnessFunc).getPeakHeights();

         int tallestPeak = 0;
         for (int i = 1; i < peakHeights.length; i++)
         {
            if (peakHeights[i] > peakHeights[tallestPeak])
               tallestPeak = i;
         }

         for (Particle p : swarm.getParticles())
         {
            int closest = ((MovingPeaks) fitnessFunc).getPeakAt(p
                  .getPersonalBestLocation());
            double dist = VectorMath.length(VectorMath.subtract(
                  peakLocations[closest], p.getPersonalBestLocation()));

            int plotNum = closest == tallestPeak ? 1 : 0;
            peakRepresentation.plotPoint(dist, closest + 1, plotNum);
         }

         for (int i = 0; i < ((MovingPeaks) fitnessFunc).getNumPeaks(); i++)
         {
            if (((MovingPeaks) fitnessFunc).isPeakHidden(i))
               peakRepresentation.plotPoint(0, i + 1, 2);
         }

         peakRepresentation.updateGraph();
      }

      if (movingPeaksMetrics != null)
         movingPeaksMetrics.step();

      if (speciesFitness != null)
      {
         SpeciesNeighborhood speciesModel = (SpeciesNeighborhood) neighborhood;

         speciesFitness.clear(0);
         speciesFitness.clear(1);

         List<Species> species = speciesModel.getAllSpecies();

         if (species.size() > 0)
         {

            double[] fitnesses = new double[species.size()];
            double minFitness = Double.MAX_VALUE;
            double maxFitness = Double.MIN_VALUE;
            int speciesNum = 0;

            for (Species currSpecies : species)
            {
               Particle best = Collections.max(currSpecies.getParticles(),
                     new FitnessComparator(true));
               double fitness = best.getPersonalBestFitness()[0];
               fitnesses[speciesNum++] = fitness;
               if (fitness > maxFitness)
                  maxFitness = fitness;
               if (fitness < minFitness)
                  minFitness = fitness;
            }

            Arrays.sort(fitnesses);

            for (speciesNum = 1; speciesNum <= fitnesses.length; speciesNum++)
            {
               int plotID = 0;
               double fitness = fitnesses[fitnesses.length - speciesNum];

               if (bestFitness - fitness > monitor.getRequiredFitnessAccuracy())
                  plotID = 1;

               speciesFitness.plotPoint(fitness, speciesNum, plotID);
            }

            if (minFitness == maxFitness)
            {
               minFitness--;
               maxFitness++;
            }

            double[] xRange = speciesFitness.getXRange();

            if (xRange[0] != Double.MIN_VALUE || xRange[1] != Double.MAX_VALUE)
            {
               if (xRange[0] < minFitness)
                  minFitness = xRange[0];
               if (xRange[1] > maxFitness)
                  maxFitness = xRange[1];
            }

            speciesFitness.setXRange(minFitness, maxFitness);

            if (fitnesses.length > speciesFitness.getYRange()[1])
               speciesFitness.setYRange(0, fitnesses.length);

            speciesFitness.step();
         }
      }

      if (aPlot != null)
      {
         // Clear the plot first
         aPlot.clear(0);
         aPlot.clear(1);
         aPlot.clear(2);
         aPlot.clear(3);
         aPlot.clear(4);

         // Display a plot for all the particles at each Iteration
         aPlot.setConnected(false);
         for (Particle p : swarm.getParticles())
            aPlot.plotPoint(p.getLocation()[0], p.getLocation()[1], 0);

         if (tracker != null)
         {
            // Display a plot for all the seeds
            for (Species species : tracker.getAllSpecies())
               aPlot.plotPoint(species.getSeed()[0], species.getSeed()[1], 1);
         }

         // SpeciesNeighborhood sn = (SpeciesNeighborhood) neighborhood;
         // for (Set<Particle> species : sn.getAllSpecies())
         // {
         // ArrayList<Particle> specList = new
         // ArrayList<Particle>(species.size());
         // specList.addAll(species);
         // Collections.sort(specList, new FitnessComparator(true));
         //            
         // Particle seed = specList.get(specList.size() - 1);
         // aPlot.plotPoint(seed.getPersonalBestLocation()[0],
         // seed.getPersonalBestLocation()[1], 1);
         //            
         // }

         // for (int i=0; i < seedList.size(); i++)
         // {
         // PSOAgent p = (PSOAgent)seedList.get(i);
         // //aPlot.plotPoint(p.present[0], p.present[1], 1);
         // aPlot.plotPoint(p.pbest[0], p.pbest[1], 1);
         // }

         aPlot.setConnected(true);
         // Display all the peaks on the plot
         int maxPosition = ((MovingPeaks) fitnessFunc).getMaximumPeak();
         double[][] positions = ((MovingPeaks) fitnessFunc).getPeakPositions();
         for (int i = 0; i < ((MovingPeaks) fitnessFunc).getNumPeaks(); i++)
         {
            for (int j = 0; j < 2; j++)
            {
               aPlot.plotPoint(positions[i][j], positions[i][j + 1], 2);
               j++;
            }
         }

         aPlot.setConnected(false);
         for (int i = 0; i < ((MovingPeaks) fitnessFunc).getNumPeaks(); i++)
         {
            if (i == maxPosition)
               for (int j = 0; j < 2; j++)
               {
                  aPlot.plotPoint(positions[i][j], positions[i][j + 1], 3);
                  j++;
               }
         }

         aPlot.fillPlot();
         aPlot.updateGraph();
      }
   }

   /**
    * Starts up the RePast simulation
    * 
    * @param args
    *           This parameter is ignored.
    */
   public static void main(String[] args)
   {
      batch = args.length > 0;
      String parameterFile = "";

      if (args.length > 0)
         parameterFile = args[0];

      SimInit init = new SimInit();
      PSOModel model = new PSOModel();

      if (parameterFile.length() > 0)
         model.recordFile = new File(parameterFile + ".csv").getName();

      init.loadModel(model, parameterFile, batch);
   }

   /**
    * This class informs the PSOModel that a tick happened without exposing the
    * execute() method to the outside world.
    * 
    * @author Stefan Bird <stbird@seatiger.org>
    */
   protected class TickListener extends BasicAction
   {
      /**
       * Tells the PSO model that a "tick" happened
       * 
       * @see uchicago.src.sim.engine.BasicAction#execute()
       */
      @Override
      public void execute()
      {
         PSOModel.this.tick();
      }
   }

   /**
    * @return Returns the requiredFitnessAccuracy.
    */
   public double getRequiredFitnessAccuracy()
   {
      return requiredFitnessAccuracy;
   }

   /**
    * @param requiredFitnessAccuracy
    *           The requiredFitnessAccuracy to set.
    */
   public void setRequiredFitnessAccuracy(double requiredFitnessAccuracy)
   {
      this.requiredFitnessAccuracy = requiredFitnessAccuracy;
   }

   /**
    * @return Returns the radius.
    */
   public double getSpsoRadius()
   {
      return spsoRadius;
   }

   /**
    * @param radius
    *           The radius to set.
    */
   public void setSpsoRadius(double radius)
   {
      this.spsoRadius = radius;
   }

   //   
   // /**
   // * @return Returns the deleteParticles.
   // */
   // public boolean isAnpsoDeleteParticles()
   // {
   // return anpsoDeleteParticles;
   // }
   //
   // /**
   // * @param deleteParticles The deleteParticles to set.
   // */
   // public void setAnpsoDeleteParticles(boolean deleteParticles)
   // {
   // this.anpsoDeleteParticles = deleteParticles;
   // }

   /**
    * @return Returns the recordInterval.
    */
   public int getRecordIntervalTicks()
   {
      return recordIntervalTicks;
   }

   /**
    * @param recordInterval
    *           The recordInterval to set.
    */
   public void setRecordIntervalTicks(int recordInterval)
   {
      this.recordIntervalTicks = recordInterval;
   }

   /**
    * Returns the recordIntervalEvals.
    * 
    * @return Returns the recordIntervalEvals.
    */
   public int getRecordIntervalEvals()
   {
      return recordIntervalEvals;
   }

   /**
    * Sets the recordIntervalEvals
    * 
    * @param recordIntervalEvals
    *           The recordIntervalEvals to set.
    */
   public void setRecordIntervalEvals(int recordIntervalEvals)
   {
      this.recordIntervalEvals = recordIntervalEvals;
   }

   /**
    * Returns the redistMinParticles
    * 
    * @return The redistMinParticles.
    */
   public int getAnpsoRedistDesiredParticles()
   {
      return this.anpsoRedistDesiredParticles;
   }

   /**
    * Sets the redistMinParticles
    * 
    * @param redistMinParticles
    *           The redistMinParticles to set.
    */
   public void setAnpsoRedistDesiredParticles(int redistMinParticles)
   {
      this.anpsoRedistDesiredParticles = redistMinParticles;
   }

   // /**
   // * Returns the redistTrigger
   // *
   // * @return The redistTrigger.
   // */
   // public double getAnpsoMaxFitnessDeltaForRedist()
   // {
   // return this.anpsoMaxFitnessDeltaForRedist;
   // }
   //
   // /**
   // * Sets the redistTrigger
   // *
   // * @param redistTrigger The redistTrigger to set.
   // */
   // public void setAnpsoMaxFitnessDeltaForRedist(double redistTrigger)
   // {
   // this.anpsoMaxFitnessDeltaForRedist = redistTrigger;
   // }

   /**
    * Returns the stepsBeforeSpecies
    * 
    * @return The stepsBeforeSpecies.
    */
   public int getAnpsoStepsBeforeSpeciesForm()
   {
      return this.anpsoStepsBeforeSpeciesForm;
   }

   /**
    * Sets the stepsBeforeSpecies
    * 
    * @param stepsBeforeSpecies
    *           The stepsBeforeSpecies to set.
    */
   public void setAnpsoStepsBeforeSpeciesForm(int stepsBeforeSpecies)
   {
      this.anpsoStepsBeforeSpeciesForm = stepsBeforeSpecies;
   }

   /**
    * @return Returns the stopOptima.
    */
   public int getStopOptima()
   {
      return stopOptima;
   }

   /**
    * @param stopOptima
    *           The stopOptima to set.
    */
   public void setStopOptima(int stopOptima)
   {
      this.stopOptima = stopOptima;
   }

   /**
    * Returns an instance of the fitness function
    * 
    * @return The an instance of the selected fitness function, or null if no
    *         fitness function of that name is known
    */
   private FitnessFunction createFitnessFunction()
   {
      if (fitnessFuncName.equalsIgnoreCase("DeJong_F1"))
         return new DeJongF1(variables);
      else if (fitnessFuncName.equalsIgnoreCase("DeJong_F2"))
         return new DeJongF2();
      else if (fitnessFuncName.equalsIgnoreCase("DeJong_F3"))
         return new DeJongF3(variables);
      else if (fitnessFuncName.equalsIgnoreCase("DeJong_F4"))
         return new DeJongF4(variables);
      else if (fitnessFuncName.equalsIgnoreCase("DeJong_F5"))
         return new DeJongF5();
      else if (fitnessFuncName.equalsIgnoreCase("Rastrigin"))
         return new Rastrigin(variables);
      else if (fitnessFuncName.equalsIgnoreCase("Himmelblau"))
         return new Himmelblau();
      else if (fitnessFuncName.equalsIgnoreCase("Shubert"))
         return new Shubert(variables);
      else if (fitnessFuncName.equalsIgnoreCase("Debs_First"))
         return new DebsFirstFunction();
      else if (fitnessFuncName.equalsIgnoreCase("Six_Hump_Camel_Back"))
         return new SixHumpCamelBack();
      else if (fitnessFuncName.equalsIgnoreCase("Brainin_RCOS"))
         return new BraininRCOS();
      else if (fitnessFuncName.equalsIgnoreCase("Knapsack"))
      {
         KnapsackProblem knapsack = new KnapsackProblem(100);
         knapsack.setMaxItems(100);
         return knapsack;
      }
      else if (fitnessFuncName.equalsIgnoreCase("Moving_Peaks"))
      {
         MovingPeaks peaks = new MovingPeaks(movPeaksNumPeaks, variables);

         if (movPeaksUseCustomPeakFunc)
            peaks.setPeakFunction(new Peak_Function_Custom(
                  movPeaksCustMinPower, movPeaksCustMaxPower,
                  movPeaksCustMaxFreq, movPeaksCustMaxAmp));

         peaks.setSeverity(movPeaksSeverity);
         return peaks;
      }
      else if (fitnessFuncName.equalsIgnoreCase("SigmoidRegression"))
      {
         try
         {
            CSVSource source = new CSVSource(new File(regressionSourceFile));
            Regression regression = new SigmoidRegression();

            if (showDisplays)
               return new GraphingRegressionEvaluator(regression, source, this);
            else
               return new RegressionEvaluator(regression, source);
         }
         catch (IOException ioe)
         {
            ioe.printStackTrace();
            return null;
         }
      }
      else if (fitnessFuncName.equalsIgnoreCase("Humps"))
      {
         // The max radius (assuming the peaks are ideally placed)
         // is 0.5 * (peaks ^ (-1/dimensions))
         HumpProblem hp = new HumpProblem(variables, humpPeaks,
               humpPeaksRadius, 1, 1);
         hp.setUseBasisFunction(humpUseBasis);

         return hp;
      }
      else
         return null;
   }

   /**
    * Returns an instance of the neighbourhood
    * 
    * @return The an instance of the selected neighbourhood, or null if no
    *         neighbourhood of that name is known
    */
   private Neighborhood createNeighborhood()
   {
      return this.createNeighborhood(neighborhoodName);
   }

   /**
    * Returns an instance of the specified neighbourhood
    * 
    * @param name
    *           The name of the neighborhood
    * @return The an instance of the selected neighbourhood, or null if no
    *         neighbourhood of that name is known
    */

   private Neighborhood createNeighborhood(String name)
   {
      // Backwards compatibility for old .pf files
      if (name.equalsIgnoreCase("LinkingSpecies"))
         name = "ASPSO";
      else if (name.equalsIgnoreCase("RadiusSpecies"))
         name = "SPSO";
      else if (name.equalsIgnoreCase("RegSPSO"))
         name = "rSPSO";

      classifier = null;

      if (name.equalsIgnoreCase("Star"))
         return new StarNeighborhood();
      else if (name.equalsIgnoreCase("Global"))
         return new GlobalNeighborhood();
      else if (name.equalsIgnoreCase("Circle"))
         return new CircleNeighborhood(2);
      else if (name.equalsIgnoreCase("VonNeumann"))
         return new VonNeumannNeighborhood();
      else if (name.indexOf("Species") != -1 || name.indexOf("SPSO") != -1
            || name.indexOf("ANPSO") != -1)
      {
         if (name.equalsIgnoreCase("ANPSO"))
         {
            ANPSO tmp = new ANPSO(anpsoStepsBeforeSpeciesForm);
            // tmp.setMinimumParticles(anpsoSpeciesMinParticles);
            classifier = tmp;
         }
         else if (name.equalsIgnoreCase("rANPSO"))
         {
            ANPSO tmp = new ANPSO(anpsoStepsBeforeSpeciesForm);
            // tmp.setMinimumParticles(anpsoSpeciesMinParticles);
            
            classifier = new RegressionHeuristic(tmp, swarm, spsoRadius,
                  fitnessListener);
            ((RegressionHeuristic) classifier).setMaxExcessPoints(regMaxExcess);            
         }
         else if (name.equalsIgnoreCase("SPSO"))
         {
            classifier = new SPSO(spsoRadius, spsoUsePBest);
            ((SPSO) classifier).setPMax(spsoPMax);
         }
         else if (name.equalsIgnoreCase("rSPSO"))
         {
            SPSO spso = new SPSO(spsoRadius, spsoUsePBest);
            spso.setPMax(spsoPMax);

            classifier = new RegressionHeuristic(spso, swarm, spsoRadius,
                  fitnessListener);
            ((RegressionHeuristic) classifier).setMaxExcessPoints(regMaxExcess);

         }
         else if (name.equalsIgnoreCase("RegESPSO2"))
         {
            ESPSO2 espso2 = new ESPSO2(swarm, spsoRadius);
            classifier = new RegressionHeuristic(espso2, swarm, spsoRadius,
                  fitnessListener);
         }
         else if (name.equalsIgnoreCase("ESPSO"))
         {
            ESPSO espso = new ESPSO(swarm, spsoRadius);

            int maxPop = particles;
            espso.setEpsilon(espsoEpsilon);

            // NOTE: Min tied to max, ignored by ESPSO2
            espso.setMaxNewSpeciesSize(espsoNewSpeciesSize);
            espso.setMinNewSpeciesSize(espsoNewSpeciesSize);

            // Ignored by ESPSO2
            espso.setMaxPopulation(maxPop);

            espso.setConversionAge(espsoConversionAge);

            classifier = espso;
         }
         else if (name.equalsIgnoreCase("ESPSO2"))
         {
            ESPSO2 espso2 = new ESPSO2(swarm, spsoRadius);

            classifier = espso2;
         }
         else if (name.equalsIgnoreCase("VaryingRadiusSPSO"))
            classifier = new VaryingRadiusSeedClassifier();
         else
            return null;

         SpeciesNeighborhood temp = new SpeciesNeighborhood(classifier, swarm,
               new VonNeumannNeighborhood());

         tracker = temp.getTracker();
         return temp;
      }
      // else if (name.equalsIgnoreCase("FDR_PSO"))
      // // FDR acts like an implicit speciation, so it's listed as a
      // neighbourhood
      // // model even though it's not implemented that way
      // return new GlobalNeighborhood();
      else
         return null;
   }

}
