package org.seatiger.pso;

import java.util.Arrays;

import uchicago.src.sim.engine.CustomProbeable;


/**
 * Represents a single particle within the swarm model.
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class Particle implements Cloneable, CustomProbeable
{
   /** Indicates the particle has been set inactive */
   private static final int INACTIVE = -1;
   
   /** The personal best location for this particle */
   private double[] personalBestLocation;
   
   /** The fitness at the personal best location */
   private double[] personalBestFitness;
   
   /** The particle's current location */
   private double[] location;
   
   /** The particle's current velocity */
   private double[] velocity;
   
   /** The fitness of the current location */
   private double[] fitness;
   
   /** The swarm this particle belongs to */
   private Swarm swarm;
   
   /** The model that controls how the particle moves */
   private MovementModel movementModel;
   
   /** Number of steps until the particle is active in the simulation. 
    * An active particle is one that will be moved by the Swarm on its 
    * next turn. Inactive particles remain at their current location.
    */
   private int stepsUntilActive = 0;
   
   /**
    * Creates a new Particle at the specified location
    *
    * @param location The initial location for the particle
    * @param velocity The initial velocity for the particle
    * @param swarm The Swarm this particle belongs to
    * @param movementModel The MovementModel that controls how the particle 
    * moves
    */
   public Particle(double[] location, double[] velocity, Swarm swarm, 
         MovementModel movementModel)
   {
      this.velocity = velocity;
      this.swarm = swarm;
      this.movementModel = movementModel;
      
      movementModel.register(this);
      setLocation(location);
   }

   /**
    * Returns the current location of the particle 
    *
    * @return The current location of the particle
    */
   public double[] getLocation()
   {
      return location;
   }

   /**
    * Returns the current velocity of the particle
    * 
    * @return The current velocity of the particle
    */
   public double[] getVelocity()
   {
      return velocity;
   }
   
   /**
    * Sets the location 
    *
    * @param location The location to set.
    */
   public void setLocation(double[] location)
   {
      this.location = location;
      fitness = swarm.calculateFitness(location);
      
      if (personalBestFitness == null || fitness[0] > personalBestFitness[0])
      {
         if (personalBestFitness == null)
         {
            personalBestFitness = new double[fitness.length];
            personalBestLocation = new double[location.length];
         }
         
         setPBestToCurrent();
      }
   }
   
   /**
    * Sets the velocity of the particle. The particle is not actually moved
    * by this call. 
    * 
    * @param velocity The velocity of the particle
    */
   public void setVelocity(double[] velocity)
   {
      if (velocity.length != this.velocity.length)
         throw new IllegalArgumentException("Array is wrong size");
      
      System.arraycopy(velocity, 0, this.velocity, 0, velocity.length);
   }

   /**
    * Returns the Swarm that this particle belongs to
    * 
    * @return the Swarm that this particle belongs to
    */
   public Swarm getSwarm()
   {
      return swarm;
   }
   
   /**
    * Returns the fitness of the personal best location 
    *
    * @return The fitness of the personal best location 
    */
   public double[] getPersonalBestFitness()
   {
      return this.personalBestFitness;
   }

   /**
    * Returns the best location found by this particle 
    *
    * @return The the best location found by this particle 
    */
   public double[] getPersonalBestLocation()
   {
      return this.personalBestLocation;
   }

   /**
    * Tells the particle to move to the next location as determined by the
    * velocity given by the MovementModel. This function will do nothing if the 
    * particle is inactive and not recalculating 
    */
   public void move()
   {
      if (isActive())
      {
         // If the velocity would place the particle out of bounds, calculate
         // where it should be (using bouncing or clipping etc) and set the velocity to 
         // make it go there instead. 
         double[] validLocation = swarm.sanitizeLocation(
               VectorMath.add(location, movementModel.getNextVelocity(this)));
         velocity = VectorMath.subtract(validLocation, location);
         
         this.setLocation(VectorMath.add(location, velocity));
      }
      else if (stepsUntilActive > 0)
         stepsUntilActive--;
   }
   
   /**
    * Clears the personal best memory and recalculates the current fitness 
    */
   public void clearPBestMemory()
   {
	   Arrays.fill(personalBestFitness, Double.NEGATIVE_INFINITY);
	   setLocation(location);
   }

   /**
    * Sets the pBest location to be the current location 
    */
   public void setPBestToCurrent()
   {
      System.arraycopy(fitness, 0, personalBestFitness, 0, fitness.length);
      System.arraycopy(location, 0, personalBestLocation, 0, location.length);
   }
   
   /**
    * Returns the movement model being used by this particle
    * 
    * @return the movement model being used by this particle
    */
   public MovementModel getMovementModel()
   {
      return movementModel;
   }
   
   /**
    * Returns the fitness of this particle's current location
    *
    * @return The fitness the particle's current location
    */
   public double[] getFitness()
   {
      return fitness;
   }
   
   /**
    * Returns true if this particle dominates <code>other</code>. A particle
    * dominates another if at least one fitness is better and there are no
    * fitnesses that are worse. 
    * 
    * @param other The particle to test
    * @return true if this particle dominates the other particle
    * @throws InvalidComparisonException if this particle uses a different
    * number of fitness functions to <code>other</code>
    */
   public boolean dominates(Particle other)
   {
      boolean dominates = false;
      
      double[] ourFitness = getFitness();
      double[] otherFitness = other.getFitness();
      
      if (otherFitness.length != ourFitness.length)
         throw new InvalidComparisonException("Cannot compare a particle with "
               + ourFitness.length + " fitness(es) to a particle with "
               + otherFitness.length + " fitness(es)");
      
      for (int i = 0; i < ourFitness.length; i++)
      {
         if (ourFitness[i] > otherFitness[i])
            dominates = true;
         else if (ourFitness[i] < otherFitness[i])
            return false; 
      }
      
      return dominates;
   }

   /**
    * @param fitness1
    * @param fitness2
    * @return
    */
   public static boolean dominates(double[] fitness1, double[] fitness2)
   {
      boolean dominates = false;
      
      if (fitness2.length != fitness1.length)
         throw new InvalidComparisonException("fitness1 and fitness2 must have " +
               "the same length");
      
      for (int i = 0; i < fitness1.length; i++)
      {
         if (fitness1[i] > fitness2[i])
            dominates = true;
         else if (fitness1[i] < fitness2[i])
            return false; 
      }
      
      return dominates;
   }
   
   /** 
    * True if the particle is active in the simulation. An active particle
    * is one that will be moved by the Swarm on its next turn. Inactive 
    * particles remain at their current location.
    * 
    * @return true if this particle is currently active in the simulation
    */
   public boolean isActive()
   {
      return stepsUntilActive == 0;
   }

   /**
    * Sets whether the particle is active
    *
    * @param active True if the particle should be moved on its next turn
    * @see #isActive()
    */
   public void setActive(boolean active)
   {
      this.stepsUntilActive = active ? 0 : INACTIVE;
   }

   /**
    * Sets the number of steps before the particle is active
    * 
    * @param stepsUntilActive The number of steps until the particle is active
    */
   public void setStepsUntilActive(int stepsUntilActive)
   {
      this.stepsUntilActive = stepsUntilActive;
   }
   
   /**
    * Returns the number of steps before the particle is inactive
    * 
    * @return The number of steps until the particle is active, or -1 if 
    * setActive(false) has been called. 
    */
   public int getStepsUntilActive()
   {
      return stepsUntilActive;
   }
   
   /**
    * Returns a copy of this particle. This method does not register the 
    * returned particle with its MovementMethod
    * 
    * @return A copy of this particle
    */
   @Override
   public Particle clone()
   {
      try
      {
         return (Particle) super.clone();
      }
      catch (CloneNotSupportedException cnse)
      {
         // This shouldn't happen
         throw new InternalError("Got impossible CloneNotSupportedException");
      }
   }

   /**
    * @see uchicago.src.sim.engine.CustomProbeable#getProbedProperties()
    */
   public String[] getProbedProperties()
   {
      return new String[] {"CurrLocation", "CurrVelocity", "CurrFitness", 
            "pBestLocation", "pBestFitness"};
   }
   
   /**
    * Returns the current location as a String
    * 
    * @return The current location as a String
    */
   public String getCurrLocation()
   {
      return Swarm.toString(location);
   }

   /**
    * Returns the current velocity as a String
    * 
    * @return The current velocity as a String
    */
   public String getCurrVelocity()
   {
      return Swarm.toString(velocity);
   }
   
   
   /**
    * Returns the current fitness as a String
    * 
    * @return The current fitness as a String
    */
   public String getCurrFitness()
   {
      return Swarm.toString(fitness);
   }
   
   /**
    * Returns the best location as a String
    * 
    * @return The bestlocation as a String
    */
   public String getPBestLocation()
   {
      return Swarm.toString(personalBestLocation);
   }
   
   /**
    * Returns the best fitness as a String
    * 
    * @return The best fitness as a String
    */
   public String getPBestFitness()
   {
      return Swarm.toString(personalBestFitness);
   }
   
   
}
