/**
 * 
 */
package org.seatiger.pso;

import java.util.*;

/**
 * Prevents overcrowding of species by redistributing particles from 
 * species that have too many members. The excess members will be killed
 * and new particles will be created to replace them. This class implements
 * SwarmListener - to cause the species to be checked every generation, if
 * added to a Swarm as a listener, it will redistribute the particles whenever
 * it receives a {@link SwarmListener#step(SwarmEvent)} event  
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class ParticleRedistributor implements SwarmListener
{
   /** The desired number of members a population should have */
   private int desiredSize = 3;
   
   /** The maximum amount the pBest fitness may differ from the best
    * to the worst particle of a species for redistribution to occur */
   private double pBestFitnessDelta = Double.POSITIVE_INFINITY;
   
   /** If true, particles will be deleted, not redistributed */
   private boolean deleteParticles = false;
   
   /** The model to query for the species */
   private SpeciesNeighborhood model;
   
   /** The swarm the model maintains the species for */
   private Swarm swarm;
   
   /**
    * Creates a new ParticleRedistributor that redistributes after each step
    * of <code>swarm</code>
    * 
    * @param model The model to get species information from
    * @param swarm The Swarm the model belongs to
    */
   public ParticleRedistributor(Swarm swarm, SpeciesNeighborhood model)
   {
      this.model = model;
      this.swarm = swarm;
   }

   /**
    * Creates a new ParticleRedistributor that redistributes after each step
    * of <code>swarm</code>
    * 
    * @param model The model to get species information from
    * @param swarm The Swarm the model belongs to
    * @param desiredSize The desired species size
    * fitness of the best and worst particle for a redistribution to occur
    */
   public ParticleRedistributor(Swarm swarm, SpeciesNeighborhood model,
         int desiredSize)
   {
      this(swarm, model);
      this.desiredSize = desiredSize;
   }   
   
   /**
    * Redistributes particles 
    */
   public void redistribute()
   {
      ArrayList<Particle> speciesList = new ArrayList<Particle>();
      ArrayList<MovementModel> killedParticleMMs = new ArrayList<MovementModel>();
      
      for (Species species : model.getAllSpecies())
      {
         if (species.getParticles().size() < desiredSize)
            continue;
         
         speciesList.clear();
         speciesList.addAll(species.getParticles());
         Collections.sort(speciesList, new FitnessComparator(true));
         
         Particle best = speciesList.get(speciesList.size() - 1);
         Particle worst = speciesList.get(0);
         double delta = 
            best.getPersonalBestFitness()[0] - worst.getPersonalBestFitness()[0];
         
         if (delta < pBestFitnessDelta)
         {
            for (int i = 0; i < speciesList.size() - desiredSize; i++)
            {
               Particle p = speciesList.get(i);
               Swarm pSwarm = p.getSwarm();
               p.setLocation(pSwarm.getRandomLocation());
               p.setVelocity(pSwarm.getRandomVelocity());
               p.setPBestToCurrent();
               p.setStepsUntilActive(1);
            }
         }
      }
   }
   
   
   /**
    * Returns true if particles are deleted instead of being redistributed
    * 
    * @return True if particles are deleted instead of being redistributed
    */
   public boolean isDeleteParticles()
   {
      return deleteParticles;
   }

   /**
    * Sets whether particles should be deleted instead of being redistributed
    * 
    * @param deleteParticles True if particles should be deleted
    */
   public void setDeleteParticles(boolean deleteParticles)
   {
      this.deleteParticles = deleteParticles;
   }

   /**
    * Returns the desired number of particles per species
    * 
    * @return Returns the desired number of particles per species
    */
   public int getDesiredSize()
   {
      return desiredSize;
   }

   /**
    * Sets the desired number of particles per species
    * 
    * @param desiredSize The desired number of particles per species
    */
   public void setDesiredSize(int desiredSize)
   {
      this.desiredSize = desiredSize;
   }

   /**
    * Returns the maximum allowed difference in fitness between the best and
    * and worst particles in a species before its members may be redistributed 
    * 
    * @return The personal best fitness delta
    */
   public double getPBestFitnessDelta()
   {
      return pBestFitnessDelta;
   }

   /**
    * Sets the maximum allowed difference in fitness between the best and
    * and worst particles in a species before its members may be redistributed 
    * 
    * @param bestFitnessDelta The personal best fitness delta to set.
    */
   public void setPBestFitnessDelta(double bestFitnessDelta)
   {
      pBestFitnessDelta = bestFitnessDelta;
   }

   public void step(SwarmEvent e) 
   { 
      redistribute(); 
   }
   
   public void initialized(SwarmEvent e) {}
}
