package org.seatiger.pso;

import peaks.*;



/**
 * Custom peak function 
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */

public class Peak_Function_Custom implements Peak_Function 
{
   /** The array index of the wave's frequency component */
   public final static int FREQUENCY = 0;
   /** The array index of the wave's amplitude component */
   public final static int AMPLITUDE = 1;

   /** The array index of the power component for the negative side */
   public final static int NEGATIVE = 0;
   /** The array index of the power component for the positive side */
   public final static int POSITIVE = 1;

   /** The individual peaks */
   private Peak[] peaks = new Peak[0];

   /** The minimum power of a peak's side */
   private double minPower;
   /** The maximum power of a peak's side */
   private double maxPower;
   /** The maximum frequency of a peak's local undulations */
   private double maxWaveFreq;
   /** The maximum amplitude of a peak's local undulations */
   private double maxWaveAmp;

   /**
    * Initialises the peak function with random powers and wave characteristics
    * 
    * @param minPower The minimum power of a peak's side
    * @param maxPower The maximum power of a peak's side
    * @param maxWaveFreq The maximum wave freqeuency
    * @param maxWaveAmp The maximum wave amplitude
    */
   public Peak_Function_Custom(double minPower, double maxPower, 
         double maxWaveFreq, double maxWaveAmp)
   {
      this.minPower = minPower;
      this.maxPower = maxPower;
      this.maxWaveFreq = maxWaveFreq;
      this.maxWaveAmp = maxWaveAmp;
   }

   public double calculate(double[] gen, int peak_number)
   {
      // We lazily create peaks as they are requested
      if (peaks.length <= peak_number)
      {
         Peak[] temp = new Peak[peak_number + 1];
         System.arraycopy(peaks, 0, temp, 0, peaks.length);
         peaks = temp;
      }

      if (peaks[peak_number] == null)
         peaks[peak_number] = new Peak();


      // Calculate the height of the requested peak
      double peakWidth = movpeaks.peak[peak_number][movpeaks.geno_size];
      double peakHeight = movpeaks.peak[peak_number][movpeaks.geno_size + 1]; 

      return peaks[peak_number].calculate(gen, movpeaks.peak[peak_number], 
            peakWidth, peakHeight);
   }

   /** 
    * Holds the information about a single peak
    */
   private class Peak
   {
      /** The power components */
      private double[][] powers;
      /** The wave components */
      private double[][] waves;

      /**
       * Initialises the peak
       */
      public Peak()
      {
         powers = new double[movpeaks.geno_size][2];
         waves = new double[movpeaks.geno_size][2];

         for (int i = 0; i < powers.length; i++)
         {
            powers[i][NEGATIVE] = minPower + Math.random() * (maxPower - minPower); 
            powers[i][POSITIVE] = minPower + Math.random() * (maxPower - minPower);

            waves[i][FREQUENCY] = Math.random() * maxWaveFreq;
            waves[i][AMPLITUDE] = Math.random() * maxWaveAmp;
         }
      }

      /**
       * Calculates the peak's height at the given point
       * 
       * @param gen The point to calculate the height of
       * @param peakLoc The location of the peak's maximum
       * @param peakHeight The height of the peak
       * @param peakWidth The width of the peak
       * @return The height of the peak at the given point
       */
      public double calculate(double[] gen, double[] peakLoc, 
            double peakWidth, double peakHeight) 
      {
         double sum = 0;

         for (int i = 0; i < movpeaks.geno_size; i++)
         {
            double relativeLoc = gen[i] - peakLoc[i];

            double power = powers[i][(relativeLoc < 0) ? NEGATIVE : POSITIVE];

            double temp = Math.pow(Math.abs(relativeLoc), power);
            temp += waves[i][AMPLITUDE] * 
            (Math.cos(waves[i][FREQUENCY] * relativeLoc) - 1);

            sum += temp * temp;
         }

         if (sum < 0)
            throw new RuntimeException("BUG: Calculated height exceeds peak height");
         
         return peakHeight - peakWidth * Math.sqrt(sum);
      }
   }
}
