package org.seatiger.pso;

/**
 * Represents a range of numbers
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class Range
{
   /** The minimum number in this range */
   private double min;
   /** The maximum number in this range */
   private double max;
   
   /**
    * Creates a new Range with the given minimum and maximum values
    *
    * @param min The smallest number that is within the range
    * @param max The largest number that is within the range
    */
   public Range(double min, double max)
   {
      this.min = min;
      this.max = max;
   }

   /**
    * Returns the minimum value of the range
    * 
    * @return The smallest number that is still within the range
    */
   public double getMin()
   {
      return this.min;
   }

   /**
    * Returns the maximum value of the range
    * 
    * @return The smallest number that is still within the range
    */
   public double getMax()
   {
      return this.max;
   }
   
   /**
    * Returns the size of the Range, that is max - min
    * 
    * @return {@link #getMax()} - {@link #getMin()}
    */
   public double getSize()
   {
      return max - min;
   }
   
   /**
    * Returns true if the point is within this range  
    * 
    * @param point The point to check
    * @return True if the point is within this range
    */
   public boolean contains(double point)
   {
      return point >= this.min && point <= this.max;
   }
   
   /**
    * Returns true if this range fully contains <code>other</code>
    * 
    * @param other The Range to test
    * @return True if <code>other</code> lies completely within this
    * range
    */
   public boolean contains(Range other) 
   {
      return getMin() <= other.getMin() && getMax() >= other.getMax();
   }
   
   /**
    * Returns the intersection of this Range and other
    * 
    * @param other The Range to calculate the intersection with
    * @return A new Range with the intersection of this and other, 
    * or null if the two Ranges do not intersect. 
    */
   public Range intersection(Range other)
   {
      if (other.getMin() < getMin())
         return other.intersection(this);
      else if (getMax() > other.getMin())
         return new Range(other.getMin(), Math.min(getMax(), other.getMax()));
      else
         return null;
   }
   
   /**
    * Returns whether this Range is equal to another object. 
    * A Range is equal to another Range when its maximum and 
    * minimum values are the same. 
    * 
    * @param other The Object to test against
    */
   @Override
   public boolean equals(Object other)
   {
      if (other.getClass() != getClass())
         return false;
      
      Range otherRange = (Range) other;
      
      return getMin() == otherRange.getMin() && getMax() == otherRange.getMax();
   }

   @Override
   public int hashCode()
   {
      return (int) Double.doubleToRawLongBits(getMin() * getMax());
   }
   
   @Override
   public String toString()
   {
      return super.toString() + "[" + getMin() + ", " + getMax() + "]"; 
   }
}
