package org.seatiger.pso;

import java.util.Arrays;

/**
 * Implements the {@link #getPreferredRanges()} and 
 * {@link #setPreferredRanges(Range[])} functions
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public abstract class RegressionAdapter implements Regression
{
   /** The global default range for each constant */
   public final static Range DEFAULT_RANGE = new Range(-100, 100);
   
   /** The preferred ranges of this regression */
   private Range[] preferredRanges;
   /** The default range for this Regression */
   private Range defaultRange = DEFAULT_RANGE;
   /** The constants being used */
   private double[] constants;

   /** 
    * Initialises the preferred ranges to {@link #DEFAULT_RANGE}
    */
   protected RegressionAdapter()
   {
      this(null);
   }
   
   /**
    * Initialises the preferred ranges to the Range given
    * 
    * @param defaultRange The default preferred range for each constant
    */
   protected RegressionAdapter(Range defaultRange)
   {
      setDefaultRange(defaultRange);
      setPreferredRanges(null);
      constants = new double[getNumConstants()];
   }
   
   /**
    * Sets the default range for each constant. To make this
    * call have effect, <code>setPreferredRanges(null)</code> must be
    * called after. 
    * 
    * @param defaultRange The default range for each constant unless
    * otherwise set by {@link #setPreferredRanges(Range[])}.
    */
   protected void setDefaultRange(Range defaultRange)
   {
      if (defaultRange == null)
         this.defaultRange = DEFAULT_RANGE;
      else
         this.defaultRange = defaultRange;
   }
   
   /**
    * Returns the default range for the constants
    * 
    * @return The default range for the constants
    */
   protected Range getDefaultRange()
   {
      return defaultRange;
   }
   
   /**
    * Returns the preferred ranges.
    * 
    * @return The preferred ranges for the constants
    */
   public Range[] getPreferredRanges()
   {
      Range[] result = new Range[preferredRanges.length];
      System.arraycopy(preferredRanges, 0, result, 0, result.length);
      
      return result;
   }

   /**
    * Sets the preferred range for each constant. If set to null,
    * {@link #DEFAULT_RANGE} will be used for each constant
    * 
    * @param preferredRanges An array of ranges. 
    * <code>preferredRanges[0]</code> is the range for the constant controlling
    * the highest power.
    * @throws IllegalArgumentException if the length of this array does not
    * equal the number returned by {@link #getNumConstants()}
    */
   protected void setPreferredRanges(Range[] preferredRanges)
   {
      if (preferredRanges != null && preferredRanges.length != getNumConstants())
         throw new IllegalArgumentException("Range array is wrong length");
      
      if (this.preferredRanges == null)
         this.preferredRanges = new Range[getNumConstants()];
      
      if (preferredRanges != null)
      {
         System.arraycopy(preferredRanges, 0, this.preferredRanges, 0, 
               preferredRanges.length);
      }
      else
         Arrays.fill(this.preferredRanges, defaultRange);
   }
   
   public void setConstants(double[] constants)
   {
      if (constants.length != getNumConstants())
         throw new IllegalArgumentException("Constants array must have "
               + getNumConstants() + " elements");
      
      if (this.constants.length != getNumConstants())
         this.constants = new double[getNumConstants()];
      
      System.arraycopy(constants, 0, this.constants, 0, constants.length);
   }
   
   /**
    * Returns the current constant values
    * 
    * @return The constants array
    */
   protected double[] getConstants()
   {
      double[] result = new double[constants.length];
      System.arraycopy(constants, 0, result, 0, constants.length);
      
      return result;
   }
}
