package org.seatiger.pso;

/**
 * Tunes a regression to fit a dataset
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class RegressionEvaluator implements FitnessFunction
{
   /** The source of the data to compare the regression's results to */
   private RegressionSource source;
   /** The regression */
   private Regression regression;

   /** The preferred ranges of the regression, if set */
   private Range[] preferredRanges = null;
   
   /**
    * Creates a new regression evaluator, comparing <code>regression</code>
    * to <code>source</code>
    * 
    * @param regression The regression to test
    * @param source The source of the data to compare the regression to
    */
   public RegressionEvaluator(Regression regression, RegressionSource source)
   {
      if (regression.getNumVariables() != source.getTestPoints()[0].length)
         throw new IllegalArgumentException("Regression function requires " +
               "a different number of variables to the source function");
      
      this.source = source;
      this.regression = regression;
   }
   
   /**
    * Returns the number of decision-space variables, which is the
    * number of constants used by the Regression
    * 
    * @see org.seatiger.pso.FitnessFunction#getNumVariables()
    */
   public int getNumVariables()
   {
      return regression.getNumConstants();
   }

   /**
    * Sets the constants of the regression to <code>location</code>
    * and returns -sqrt(sum of the error squared for each point).
    * 
    * @see org.seatiger.pso.FitnessFunction#getFitness(double[])
    */
   public double getFitness(double[] location)
   {
      regression.setConstants(location);
      
      double error = 0;
      
      for (double[] testPoint : source.getTestPoints())
      {
         double expected = source.calculate(testPoint);
         double actual = regression.calculate(testPoint);
         
         double difference = expected - actual;
         
         error += difference * difference;
      }
      
      return -Math.sqrt(error);
   }

   /**
    * Returns the preferred ranges of the regression. If 
    * {@link #setPreferredRanges(Range[])} has not been called or the 
    * last call was with a null argument, this will return the 
    * preferred ranges of the regression
    *
    * @return The preferred ranges of the regression
    * @see org.seatiger.pso.FitnessFunction#getPreferredRanges()
    */
   public Range[] getPreferredRanges()
   {
      if (preferredRanges == null)
         return regression.getPreferredRanges();
      else
         return preferredRanges;
   }

   /**
    * Sets the preferred ranges of the decision space. If set to 
    * null (or not set at all) this will default to the preferred
    * range returned by the Regression
    * 
    * @param preferredRanges The preferred ranges of the regression
    */
   public void setPreferredRanges(Range[] preferredRanges)
   {
      if (preferredRanges == null)
         this.preferredRanges = null;
      else
      {
         if (preferredRanges.length != getNumVariables())
            throw new IllegalArgumentException("Argument has wrong number of elements");
         
         this.preferredRanges = new Range[getNumVariables()];
         System.arraycopy(preferredRanges, 0, this.preferredRanges, 0, getNumVariables());
      }
   }
   
   /**
    * Returns the regression being used
    * 
    * @return The Regression instance being used
    */
   public Regression getRegression()
   {
      return regression;
   }
   
   /**
    * Returns the regression source being used
    * 
    * @return The RegressionSource being used
    */
   public RegressionSource getRegressionSource()
   {
      return source;
   }
}
