package org.seatiger.pso;
import java.util.*;

import Jama.Matrix;

/**
 * Sits on top of a species and tries to improve convergence by estimating
 * the actual location of the peak. It does that by attempting to fit a 
 * surface to the known points and moving the current worst particle to that
 * location. 
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class RegressionHeuristic implements SpeciesClassifier
{
   /** How often we clean up extinct species */
   private static final int CLEANUP_INTERVAL = 50;
   
   /** The maximum number of points to use in the regression, in excess
    * of the number of points actually *needed* */
   private int maxExcessPoints = 10;
   
   /** The swarm we are performing the regressions for */
   private Swarm swarm;

   /** Allows us to track species between iterations */
   private SpeciesTracker tracker;
   
   /** The maximum distance from the specie's best particle that the regression
    * still considers as a "sane" answer */
   private double maxDist;

   /** The points that have been sampled for each species */
   private HashMap<Species, List<double[]>> sampledPoints;
   
   /** The power of the regression - either 2 = quadratic or 3 = cubic */
   private int power = 2;
   
   /** How long until we clean up extinct species */   
   private int timeToCleanup = CLEANUP_INTERVAL;
   
   /** If true, the worst individual will be moved to check the guessed point.
    *  If false, the point will be checked first, the worst individual will only
    *  be moved if if the guessed point is better than the species seed
    */
   private boolean useWorstForCheck = true;
   
   /** The number of hits since the last peak movement */
   private int hits = 0;
   /** The number of attempts since the last peak movement */
   private int attempts = 0;
   
   /**
    * Creates a new regression heuristic
    * 
    * @param delegate The delegate classifier
    * @param swarm The swarm we are performing the heuristic for
    * @param maxDist The maximum "sane" distance - if the regression estimates
    * a point further than this distance from the species' best point, it is discarded
    */
   public RegressionHeuristic(SpeciesClassifier delegate, Swarm swarm, double maxDist)
   {
      this.swarm = swarm;
      sampledPoints = new HashMap<Species, List<double[]>>();
      tracker = new SpeciesTracker(delegate, swarm);
      this.maxDist = maxDist;
   }
   
   /**
    * Creates a new regression heuristic
    * 
    * @param delegate The delegate classifier
    * @param swarm The swarm we are performing the heuristic for
    * @param maxDist The maximum "sane" distance - if the regression estimates
    * a point further than this distance from the species' best point, it is discarded
    * @param changeListener Notifies of changes to the landscape
    */
   public RegressionHeuristic(SpeciesClassifier delegate, Swarm swarm, double maxDist, 
         FitnessChangeListener changeListener)
   {
      this(delegate, swarm, maxDist);
      
      if (changeListener != null)
      {
         changeListener.addObserver(new Observer()
         {
            public void update(Observable o, Object arg)
            {
               clearMemory();
            }
         });
      }
   }
   
   
   
   public Map<Particle, Integer> classify(Collection<Particle> particles)
   {
      Map<Particle, Integer> result = tracker.classify(particles);
      
      for (Species species : tracker.getAllSpecies())
      {
         double[] peak = guessPeakLocation(species);
         
         if (peak == null)
            continue;
         
         
         if (VectorMath.distance(species.getSeed(), peak) < maxDist)
         {
            peak = swarm.sanitizeLocation(peak);
            ArrayList<Particle> speciesP = new ArrayList<Particle>();
            speciesP.addAll(species.getParticles());
            
            if (useWorstForCheck)
            {
               Particle worst = Collections.min(speciesP, new FitnessComparator(true));
   
               worst.setLocation(peak);
               worst.setStepsUntilActive(1);
               
               double newFitness = worst.getFitness()[0];
               double oldFitness = species.getSeedFitness()[0];
               
               if (newFitness > oldFitness)
                  hits++;
            }
            else
            {
               double oldBest = species.getSeedFitness()[0];
               double guessedFitness = swarm.calculateFitness(peak)[0];
               
               if (guessedFitness > oldBest)
               {
                  Particle worst = Collections.min(speciesP, new FitnessComparator(true));
                  worst.setLocation(peak);
                  worst.setStepsUntilActive(1);
                  
                  hits++;
               }
            }
            
            attempts++;
//            if (attempts % 10 == 0)
//               System.out.println("Hit rate: " + (double) hits / attempts);

//            double newFitness = worst.getFitness()[0];
//            double oldFitness = species.getSeedFitness()[0];
//            
//            System.out.print(newFitness > oldFitness ? "HIT" : "MISS");
//            System.out.print(" seed:" + oldFitness);
//            System.out.print(" new:" + newFitness);
//            System.out.print(" particles: " + speciesP.size());
//            System.out.println(" points: " + sampledPoints.get(species).size());
               
         }
      }
      
      return result;
   }



   /**
    * Guesses the peak location for the given species
    * 
    * @param species The species to guess the peak's location of
    * @return The estimated location
    */
   public double[] guessPeakLocation(Species species)
   {
      if (--timeToCleanup < 0)
         cleanUp();
      
      updatePoints(species);
      return getPeakLoc(species);
   }

   /**
    * Updates the points stored for species
    * 
    * @param species The species the points are being stored for
    */
   private void updatePoints(Species species)
   {
      int variables = swarm.getNumVariables();
      int pointsNeeded = numPointsNeeded(variables);
      
      List<double[]> points = sampledPoints.get(species);
      if (points == null)
      {
         points = new ArrayList<double[]>();
         sampledPoints.put(species, points);
      }
      
      for (Particle p : species.getParticles())
      {
         double[] newPoint = new double[variables + 1];
         System.arraycopy(p.getLocation(), 0, newPoint, 0, variables);
         newPoint[newPoint.length - 1] = p.getFitness()[0];
                                        
         points.add(newPoint);
      }
      
      Collections.sort(points, Collections.reverseOrder(new Comparator<double[]>() 
         {
            public int compare(double[] a, double[] b)
            {
               if (a[a.length - 1] < b[b.length - 1])
                  return -1;
               else if (a[a.length - 1] > b[b.length - 1])
                  return 1;
               else
                  return 0;
            }
         }));
      
      for (int i = points.size() - 1; i >= maxExcessPoints + pointsNeeded; i--)
         points.remove(i);
   }
   
   /**
    * Returns the estimated location of the peak occupied by species 
    * 
    * @param species The species to determine the peak of
    * @return The estimated location of the peak, or null if the peak
    * could not be estimated (reasons for this include 
    * insufficient or poorly distributed points)
    */
   private double[] getPeakLoc(Species species)
   {
      int numDimensions = swarm.getNumVariables();
      
      List<double[]> knownPoints = sampledPoints.get(species);
      if (knownPoints == null || knownPoints.size() < numPointsNeeded(numDimensions))
         return null;
      
      double[] fitnesses = new double[knownPoints.size()];
      for (int i = 0; i < fitnesses.length; i++)
         fitnesses[i] = knownPoints.get(i)[numDimensions];
      
      double[][] points = new double[knownPoints.size()][];
      for (int i = 0; i < points.length; i++)
      {
         for (int j = 0; j < numDimensions; j++)
         {
           points[i] = new double[numDimensions];
           System.arraycopy(knownPoints.get(i), 0, points[i], 0, numDimensions);
         }
      }
      
      return findSurfaceMaximum(power, points, fitnesses);
   }

   /**
    * Returns the maximum number of excess points
    * 
    * @return The maximum number of excess points
    */
   
   public int getMaxExcessPoints()
   {
      return maxExcessPoints;
   }

   /**
    * Sets the maximum number of excess points. This allows the regression
    * to be overspecified, possibly improving its accuracy where the peak does 
    * not follow a polynomial curve
    * 
    * @param maxExcessPoints the maximum number of excess points
    */
   public void setMaxExcessPoints(int maxExcessPoints)
   {
      this.maxExcessPoints = maxExcessPoints;
   }
   
   /**
    * Returns the number of successful regression guesses since the last peak movement
    * 
    * @return The number of successful regression guesses since the last peak movement
    */
   public int getHits()
   {
      return hits;
   }
   
   /**
    * Returns the number of regression guesses since the last peak movement
    *  
    * @return The number of regression guesses since the last peak movement
    */
   public int getAttempts()
   {
      return attempts;
   }
   

   /**
    * Returns true if the regression moves the worst particle to 
    * check its guess
    * 
    * @return Whether the regression will move the worst particle
    */
   public boolean isUseWorstForCheck()
   {
      return useWorstForCheck;
   }

   /**
    * Sets whether the regression moves the worst particle to 
    * check its guess. 
    * 
    * @param useWorstForCheck the useWorstForCheck to set
    */
   public void setUseWorstForCheck(boolean useWorstForCheck)
   {
      this.useWorstForCheck = useWorstForCheck;
   }

   /**
    * Returns the highest point on the given surface
    * 
    * @param power The power of the polynomial defining the surface, either 2 or 3
    * @param points The points that define the surface
    * @param values The height of each point
    * @return The highest point on the surface if there is one, otherwise null
    */
   public static double[] findSurfaceMaximum(int power, double[][] points, double[] values)
   {
      if (points.length != values.length)
         throw new IllegalArgumentException("points and values array must be of same length");
      
      int numDimensions = points[0].length;
      double[] constantsArray = fitSurface(power, points, values);
      
      if (constantsArray == null)
         return null;
      
      double[] result = new double[numDimensions];
      double[] temp = new double[power + 1];
      
      for (int i = 0; i < numDimensions; i++)
      {
         // temp[0] is the constant, which isn't used by findMaximum
         // The constant is stored in constantsArray[constantsArray.length - 1] 
         System.arraycopy(constantsArray, i * power, temp, 1, temp.length - 1);
         double max = findMaximum(temp);
         
         if (Double.isNaN(max))
            return null;
         else
            result[i] = max;
      }
      
      return result;
   }
   
   /**
    * Fits a surface to the given points
    * 
    * @param power The power of the polynomials defining the surface
    * @param points The points to regress
    * @param values The value of the surface at each point
    * @return An array of the coeffecients for the surface equations, in the form:
    * {x, x^2, x^3, y, y^2, y^3, z, ..., c}
    */
   public static double[] fitSurface(int power, double[][] points, double[] values)
   {
      int numDimensions = points[0].length;
      int pointsNeeded = numPointsNeeded(power, numDimensions);
      
      double[][] coeffArray = new double[points.length][pointsNeeded];
      for (int i = 0; i < coeffArray.length; i++)
      {
         for (int j = 0; j < numDimensions; j++)
         {
            double[] point = points[i];
            
            for (int k = 0; k < power; k++)
               coeffArray[i][j * power + k] = Math.pow(point[j], k + 1);
         }
         
         coeffArray[i][pointsNeeded - 1] = 1;
      }
      
      double[][] fitnessArray = new double[values.length][1];
      for (int i = 0; i < fitnessArray.length; i++)
         fitnessArray[i][0] = values[i];
      
      Matrix fitnesses = new Matrix(fitnessArray);
      Matrix coeffs = new Matrix(coeffArray);
      
      // fitnesses = coeffs * constants
      // inv(coeffs) * fitnesses = inv(coeffs) * coeffs * constants
      // inv(coeffs) * fitnesses = identity * constants = constants
      Matrix constants;
      
      try
      {
         Matrix coeffInv = coeffs.inverse(); 
         constants = coeffInv.times(fitnesses);
      }
      catch (RuntimeException e)
      {
         // Catch the occasional case where coeffInv is singular
         return null;
      }
      
      double[] constantsArray = new double[pointsNeeded];
      for (int i = 0; i < constantsArray.length; i++)
         constantsArray[i] = constants.get(i, 0);
      
      return constantsArray;
   }
   
   /**
    * Returns the local maximum for the given polynomial
    * 
    * @param coeffs The polynomial's coeffecients, in [0] + [1]x + [2]x^2 + ... form
    * @return The x of the local maximum, or NaN if there isn't one
    * @throws IllegalArgumentException if coeffs doesn't represent a quadratic or
    * cubic polynomial
    */
   public static double findMaximum(double[] coeffs)
   {
      if (coeffs.length < 3 || coeffs.length > 4)
         throw new IllegalArgumentException("Can only solve quadratic and cubic polynomials");

      double[] derivative = getDerivative(coeffs);
      double[] secondDeriv = getDerivative(derivative);
      
      // coeffs is cubic, deriv quadratic
      if (derivative.length == 3)
      {
         double a = derivative[2];
         double b = derivative[1];
         double c = derivative[0];
         double determinent = b * b - 4 * a * c;

         if (a == 0 || determinent <= 0)
            return Double.NaN;
         
         double solnA = (-b + Math.sqrt(determinent)) / (2 * a);
         double solnB = (-b - Math.sqrt(determinent)) / (2 * a);
         
         if (evaluate(secondDeriv, solnA) < 0)
            return solnA;
         else 
            return solnB;
      }
      else
      {
         // coeffs is quadratic, deriv linear
         if (derivative[1] == 0)
            return Double.NaN;

         double turningPoint = -derivative[0] / derivative[1];
         
         if (evaluate(secondDeriv, turningPoint) > 0)
            return Double.NaN;
         else
            return turningPoint;
      }
   }
   
   /**
    * Evaluates the value of coeffs[0] + coeffs[1]x + coeffs[2]x^2 + ...
    *  
    * @param x The point to evaluate
    * @param coeffs The coeffecients of the polynomial
    * @return The function value at x
    */
   public static double evaluate(double[] coeffs, double x)
   {
      double result = 0;
      
      for (int i = 0; i < coeffs.length; i++)
         result += coeffs[i] * Math.pow(x, i);
      
      return result;
   }
   
   /**
    * Returns the derivative of the given polynomial
    * 
    * @param coeffs The coefficients of the polynomial
    * @return The coefficients of the polynomial's derivative
    */
   public static double[] getDerivative(double[] coeffs)
   {
      double[] result = new double[coeffs.length - 1];
      
      for (int i = 0; i < result.length; i++)
         result[i] = (i + 1) * coeffs[i + 1];
      
      return result;
   }
                  
   
   /**
    * Cleans up extinct speciesFitness
    */
   private void cleanUp()
   {
      Iterator<Species> iterator = sampledPoints.keySet().iterator();
      while (iterator.hasNext())
      {
         if (iterator.next().isExtinct())
            iterator.remove();
      }
   }
   
   /**
    * Clears all memory of sampled points
    */
   private void clearMemory()
   {
      sampledPoints.clear();
      
      attempts = 0;
      hits = 0;
   }
   
   /**
    * Returns the minimum number of points required to solve the regression of
    * the given dimensionality
    * 
    * @param numVariables The number of variables
    * @return The number of points needed
    */
   public int numPointsNeeded(int numVariables)
   {
      return numPointsNeeded(power, numVariables);
   }
   
   /**
    * Returns the minimum number of points required to solve the regression of
    * the given dimensionality
    * 
    * @param numVariables The dimensionality
    * @param power The power of the regression
    * @return The minimum number of points required
    */
   public static int numPointsNeeded(int power, int numVariables)
   {
      return power * numVariables + 1;
   }
}
