/**
 * 
 */
package org.seatiger.pso;

import java.util.*;

/**
 * Implements a radius-based species classifier using the method presented
 * in "Efficient Differential Evolution using Speciation for Multimodal
 * Function Optimisation" by Xiaodong Li.  
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class SPSO implements SpeciesClassifier
{
   /** The species radius. Only particles within this distance of a seed
    * will be added to that seed's species */
   private double radius;
   
   /** If true, pBest will be used to determine species seed and which 
    * particles belong to which species. If false, the current fitness 
    * and location will be used. */
   private boolean usePBest;
   
   /** The seed particle for each species */
   private ArrayList<Particle> speciesSeeds = new ArrayList<Particle>();
   
   /** The maximum number of particles per species */
   private int pMax = 10;
   
   /**
    * Creates a new RadiusBasedSpeciesClassifier using the specified radius.
    * The particle's current location will be used to determine species. 
    * 
    * @param radius The maximum distance a particle can be from a species 
    * seed and still be considered as part of that species. 
    */
   public SPSO(double radius)
   {
      this(radius, false);
   }
   
   /**
    * Creates a new RadiusBasedSpeciesClassifier using the specified radius
    * 
    * @param radius The maximum distance a particle can be from a species 
    * seed and still be considered as part of that species.
    * @param usePBest If true, the particle's PBest will be used to determine
    * species. If false, the current location will be used.  
    */
   public SPSO(double radius, boolean usePBest)
   {
      this.radius = radius;
      this.usePBest = usePBest;
   }
   
   
   /**
    * Returns the maximum distance a particle can be from a species 
    * seed and still be considered as part of that species. 
    * 
    * @return Returns the radius.
    */
   public double getRadius()
   {
      return radius;
   }
   
   /**
    * Returns whether the personal best is being used to determine 
    * species  
    * 
    * @return True if the personal best is being used
    */
   public boolean isUsePBest()
   {
      return usePBest;
   }

   /**
    * Sets whether the personal best should be used to determine 
    * species  
    * 
    * @param usePBest True if the personal best should be used
    */
   public void setUsePBest(boolean usePBest)
   {
      this.usePBest = usePBest;
   }

   /**
    * Sets the maximum distance a particle can be from a species 
    * seed and still be considered as part of that species. 
    * 
    * @param radius The radius to set.
    */
   public void setRadius(double radius)
   {
      this.radius = radius;
   }

   /* (non-Javadoc)
    * @see org.seatiger.pso.SpeciesClassifier#classify(java.util.Collection)
    */
   public Map<Particle, Integer> classify(Collection<Particle> particles)
   {
      // Create an array of particles reverse sorted according to fitness
      // (ie the fittest particle will be first)
      ArrayList<Particle> sortedParticles = new ArrayList<Particle>(particles);
      Collections.sort(sortedParticles, new FitnessComparator(usePBest));
      Collections.reverse(sortedParticles);
      
      HashMap<Particle, Integer> result = new HashMap<Particle, Integer>();
      HashMap<Integer, Integer> speciesCounts = new HashMap<Integer, Integer>();
      ArrayList<Particle> particlesToReinit = new ArrayList<Particle>();
      
      speciesSeeds = new ArrayList<Particle>();
      
      for (Particle p : sortedParticles)
      {
         boolean foundSpecies = false;
         double[] pLocation;
         
         if (usePBest)
            pLocation = p.getPersonalBestLocation();
         else
            pLocation = p.getLocation();

         // Loop through all of the species seeds to see if this particle
         // is within the required distance of any of them. 
         for (int speciesNum = 0; speciesNum < speciesSeeds.size()
               && !foundSpecies; speciesNum++)
         {
            double[] seedLocation;
            
            if (usePBest)
               seedLocation = speciesSeeds.get(speciesNum).getPersonalBestLocation();
            else
               seedLocation = speciesSeeds.get(speciesNum).getLocation();
            
            if (VectorMath.length(VectorMath.subtract(pLocation, seedLocation))
                  < radius)
            {
               foundSpecies = true;
               
               if (pMax == 0 || speciesCounts.get(speciesNum) < pMax)
               {
                  result.put(p, speciesNum);
                  speciesCounts.put(speciesNum, speciesCounts.get(speciesNum) + 1);
               }
               else
                  particlesToReinit.add(p);
            }
         }
         
         if (!foundSpecies)
         {
            speciesCounts.put(speciesSeeds.size(), 1);
            result.put(p, speciesSeeds.size());
            speciesSeeds.add(p);
         }
      }
      
      for (Particle p : particlesToReinit)
      {
         Swarm swarm = p.getSwarm();
         p.setLocation(swarm.getRandomLocation());
         p.setVelocity(swarm.getRandomVelocity());
         p.setPBestToCurrent();
         p.setStepsUntilActive(1);
      }
      
      return result;
   }
   
   /**
    * Returns the maximum number of particles per species
    * 
    * @return The maximum number of particles per species
    */
   public int getPMax()
   {
      return pMax;
   }

   /**
    * Sets the maximum number of particles per species
    * 
    * @param max The maximum number of particles per species
    */
   public void setPMax(int max)
   {
      pMax = max;
   }

   /**
    * Returns a list of all the species seeds, in descending order
    * of fitness
    * 
    * @return A list of the all the species seeds. 
    */
   public List<Particle> getSpeciesSeeds()
   {
      return Collections.unmodifiableList(speciesSeeds);
   }
}
