package org.seatiger.pso;

import java.util.Arrays;

/**
 * The Shubert fitness function. For the 2-dimensional version, this
 * has 9 sets of optima pairs. The proximity of each optima in the pair
 * makes this function a good test for niching 
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class Shubert implements FitnessFunction
{
   /** The number of parameters (dimensions) in this problem */
   private int parameters;
   /** The preferred range of this fitness function */
   private final static Range range = new Range(-10, 10); 
   
   /**
    * Creates a Shubert fitness function 
    *
    * @param parameters How many parameters (dimensions) the problem
    * should be in 
    */
   public Shubert(int parameters)
   {
      this.parameters = parameters;
   }

   
   /**
    * @see org.seatiger.pso.FitnessFunction#getNumVariables()
    */
   public int getNumVariables()
   {
      return parameters;
   }

   /**
    * @see org.seatiger.pso.FitnessFunction#getFitness(double[])
    */
   public double getFitness(double[] location)
   {
      double result = 1;
      
      for (int index = 0; index < location.length; index++)
      {
         double sum = 0;
         
         for (int i = 1; i <= 5; i++)
            sum += i * Math.cos((i + 1) * location[index] + i);
         
         result *= sum;
      }
         
      
      return -result;
   }

   /**
    * Returns the preferred range of this fitness function, 
    * which is [-10, 10] for all parameters
    * 
    * @return The preferred range
    * @see org.seatiger.pso.FitnessFunction#getPreferredRanges()
    */
   public Range[] getPreferredRanges()
   {
      Range[] result = new Range[parameters];
      Arrays.fill(result, range);
      
      return result;
   }

}
