package org.seatiger.pso;

import java.util.Arrays;

/**
 * Calculates a sigmoid regression: <pre>
 *      d
 * ------------
 * a + be^(-cx) 
 * </pre>
 * 
 * Where a, b, c and d are constants
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class SigmoidRegression extends RegressionAdapter
{
   /** The default range for the a, b and d constants */
   private final static Range DEFAULT_RANGE_A_B_D = new Range(-1E6, 1E6);
   /** The default range for the c constant */
   private final static Range DEFAULT_RANGE_C = new Range(-100, 100);
   
   /**
    * Creates a new SigmoidRegression
    */
   public SigmoidRegression()
   {
      setDefaultRange(DEFAULT_RANGE_A_B_D);
   }
   
   public int getNumConstants()
   {
      return 4;
   }

   public int getNumVariables()
   {
      return 1;
   }

   public double calculate(double[] location)
   {
      double[] constants = getConstants();
      double x = location[0];
      double a = constants[0];
      double b = constants[1];
      double c = constants[2];
      double d = constants[3];
      
      return d / (a + b * Math.pow(Math.E, -c * x));
   }
   
   @Override
   public void setPreferredRanges(Range[] preferredRanges)
   {
      // This function is only needed because we want c
      // to have a different range to a, b and d. 
      // If we want them to all have the same range, just call
      // setDefaultRange as is done in the constructor
      
      if (preferredRanges == null)
      {
         preferredRanges = new Range[getNumConstants()];
         Arrays.fill(preferredRanges, getDefaultRange());

         // Limit the constant c to values that don't produce infinities
         preferredRanges[2] = DEFAULT_RANGE_C;
      }
      
      super.setPreferredRanges(preferredRanges);
   }
}
