package org.seatiger.pso;

import java.util.*;

/** Holds useful information about a species */
public class Species
{
   /** The swarm the particles belong to */
   private Swarm swarm;
   /** When the species was born */
   private int born;
   /** When the seed last moved */
   private int lastMovement;
   /** The location of the species seed */
   private double[] seed;
   /** The seed particle */
   private Particle seedP;
   /** The particles in the species */
   private ArrayList<Particle> particles = new ArrayList<Particle>();
   /** True if this species is extinct */
   private boolean extinct = false;

   /**
    * Creates a new Species
    * 
    * @param particles The particles that belong to the species
    * @param swarm The swarm the particles belong to
    */
   public Species(Swarm swarm, Collection<Particle> particles)
   {
      this.swarm = swarm;
      born = swarm.getCurrTick();
      lastMovement = swarm.getCurrTick();
      seed = new double[swarm.getNumVariables()];
      
      setParticles(particles);
   }
   
   /**
    * Sets the particles in the species. The seed particle will be set
    * to the particle with the fittest pBest
    *  
    * @param particles The particles in the species
    */
   public void setParticles(Collection<Particle> particles)
   {
      if (isExtinct())
         throw new IllegalStateException("Species is extinct");
      
      this.particles.clear();
      this.particles.addAll(particles);
      
      updateSeedPos();
   }
   
   /**
    * Returns the seed particle of the species
    * 
    * @return The seed particle
    */
   public Particle getSeedP()
   {
      if (isExtinct())
         throw new IllegalStateException("Species is extinct");
      
      return seedP;
   }
   
   /**
    * Returns the tick on which the species was born
    * 
    * @return The tick on which the species was born
    */
   public int getBorn()
   {
      return born;
   }
   
   /**
    * Returns the tick on which the species seed location last changed
    * 
    * @return The tick of the last movement. 
    */
   public int getLastMovement()
   {
      return lastMovement;
   }
   
   /**
    * Returns the location of this species' seed
    * 
    * @return The location of the seed
    */
   public double[] getSeed()
   {
      if (isExtinct())
         throw new IllegalStateException("Species is extinct");
      
      double[] result = new double[seed.length];
      System.arraycopy(seed, 0, result, 0, seed.length);
      
      return result;
   }
   
   /**
    * Returns the particles in this species
    * 
    * @return The particles in the species
    */
   public Collection<Particle> getParticles()
   {
      if (isExtinct())
         throw new IllegalStateException("Species is extinct");
      
      return Collections.unmodifiableCollection(particles);
   }
   
   /**
    * Returns true if this species is extinct 
    * 
    * @return True if the species is extinct
    */
   public boolean isExtinct()
   {
      return extinct;
   }
   
   /**
    * Returns the fitness of the seed particle
    * 
    * @return The fitness of the seed particle
    */
   public double[] getSeedFitness()
   {
      return seedP.getPersonalBestFitness();
   }
   
   /**
    * Updates the location of the seed particle
    */
   public void updateSeedPos()
   {
      if (isExtinct())
         throw new IllegalStateException("Species is extinct");
      
      seedP = Collections.max(particles, new FitnessComparator(true));
      
      boolean moved = false;
      double[] pBest = seedP.getPersonalBestLocation();
      for (int i = 0; i < seed.length; i++)
      {
         if (pBest[i] != seed[i])
            moved = true;
      }
      
      if (moved)
      {
         lastMovement = swarm.getCurrTick();
         System.arraycopy(seedP.getPersonalBestLocation(), 0, seed, 0, seed.length);
      }
   }
   
   /**
    * Makes the species extinct
    */
   void die()
   {
      extinct = true;
      particles.clear();
      seedP = null;
      seed = null;
   }
}