/**
 * 
 */
package org.seatiger.pso;

import java.util.*;

/**
 * Represents a neighborhood model with multiple species. 
 * 
 * The actual species are determined by a class implementing the
 * SpeciesClassifier interface. Observers are notified whenever the
 * species have been rebuilt. 
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */

public class SpeciesNeighborhood extends Observable
   implements Neighborhood 
{
   /** Tracks the species as they move around */
   private SpeciesTracker tracker;
   
   /** The Neighborhood to use for unspecied particles */
   private Neighborhood unspeciedNeighborhood;
   
   private Swarm swarm;
   
   /** The Particles being managed */
   private ArrayList<Particle> particles = new ArrayList<Particle>();
   
   /** The Particles that don't belong to any species */
   private HashSet<Particle> unspeciedParticles = new HashSet<Particle>();
   
   /** The species each particle belongs to */
   private HashMap<Particle, LinkedHashSet<Particle>> species =
      new HashMap<Particle, LinkedHashSet<Particle>>();
   
   /** The last time the species was built */
   private int lastSpeciesBuild = 0;
   
   /**
    * Creates a new SpeciesNeighborhood
    * 
    * @param classifier The classifier to generate the species
    * @param swarm The swarm the particles belong to
    * @param unspeciedNeighborhood The neighborhood to use for the
    * unspecied particles
    */
   public SpeciesNeighborhood(SpeciesClassifier classifier, Swarm swarm, 
         Neighborhood unspeciedNeighborhood)
   {
      tracker = new SpeciesTracker(classifier, swarm);
      this.unspeciedNeighborhood = unspeciedNeighborhood;
      this.swarm = swarm;
   }

   /* (non-Javadoc)
    * @see org.seatiger.pso.Neighborhood#getNeighborhoodOf(org.seatiger.pso.Particle)
    */
   public Set<Particle> getNeighborhoodOf(Particle p)
   {
      if (swarm.getCurrTick() != lastSpeciesBuild)
         buildSpecies();
      
      if (species.containsKey(p))
         return Collections.unmodifiableSet(species.get(p));
      else
         return unspeciedNeighborhood.getNeighborhoodOf(p);
   }

   /* (non-Javadoc)
    * @see org.seatiger.pso.Neighborhood#register(org.seatiger.pso.Particle)
    */
   public void register(Particle p)
   {
      particles.add(p);
      unspeciedNeighborhood.register(p);
   }

   /* (non-Javadoc)
    * @see org.seatiger.pso.Neighborhood#deregister(org.seatiger.pso.Particle)
    */
   public void deregister(Particle p)
   {
      particles.remove(p);
      
      HashSet<Particle> particleSpecies = species.get(p);
      if (particleSpecies == null)
      {
         unspeciedNeighborhood.deregister(p);
         unspeciedParticles.remove(p);
      }
      else
      {
         particleSpecies.remove(p);
         species.remove(p);
      }
   }

   /* (non-Javadoc)
    * @see org.seatiger.pso.Neighborhood#deregisterAll()
    */
   public void deregisterAll()
   {
      particles.clear();
      species.clear();
      unspeciedNeighborhood.deregisterAll();
      unspeciedParticles.clear();
   }

   /**
    * Builds the species. It is not usually necessary to call this
    * method - the SpeciesNeighborhood automatically rebuilds
    * the species after each step of the Swarm. Call this method
    * to manually trigger a rebuild. Any Observers of this instance
    * will be notified after the build
    */
   public void buildSpecies()
   {
      Map<Particle, Integer> speciesMap = tracker.classify(
            Collections.unmodifiableCollection(particles));
      
      species.clear();
      HashMap<Integer, LinkedHashSet<Particle>> reverseSpeciesMap = new 
            HashMap<Integer, LinkedHashSet<Particle>>();
      
      for (Particle p: particles)
      {
         Integer speciesNum = speciesMap.get(p);
         
         if (speciesNum == null)
         {
            if (!unspeciedParticles.contains(p))
            {
               unspeciedNeighborhood.register(p);
               unspeciedParticles.add(p);
            }
         }
         else
         {
            LinkedHashSet<Particle> currSpecies = reverseSpeciesMap.get(speciesNum);
            
            if (currSpecies == null)
            {
               currSpecies = new LinkedHashSet<Particle>();
               reverseSpeciesMap.put(speciesNum, currSpecies);
            }
            
            currSpecies.add(p);
            species.put(p, currSpecies);
            
            if (unspeciedParticles.contains(p))
            {
               unspeciedNeighborhood.deregister(p);
               unspeciedParticles.remove(p);
            }
         }
      }
      
      lastSpeciesBuild = swarm.getCurrTick();
      
      setChanged();
      notifyObservers();
   }

//   /**
//    * Returns the particles currently in each species. 
//    * 
//    * @return A Set of Particle Sets, where each Particle Set represents
//    * the Particles in an individual species. 
//    */
//   public Set<Set<Particle>> getAllSpecies()
//   {
//      Set<Set<Particle>> result = new HashSet<Set<Particle>>();
//      
//      HashSet<LinkedHashSet<Particle>> speciesSets = 
//            new HashSet<LinkedHashSet<Particle>>(species.values());
//
//      for (LinkedHashSet<Particle> currSet : speciesSets)
//         result.add(Collections.unmodifiableSet(currSet));
//      
//      return result;
//   }
   
   /**
    * Returns the set of particles that do not belong to a species
    * 
    * @return The set of particles that do not belong to a species
    */
   public Set<Particle> getUnspeciedParticles()
   {
      return Collections.unmodifiableSet(unspeciedParticles);
   }
   
   /**
    * Returns the classifier used to determine species
    * 
    * @return The classifier used to determine species
    */
   public SpeciesClassifier getClassifier()
   {
      return tracker.getDelegate();
   }
   
   /**
    * Returns all of the current species 
    * 
    * @return The current species
    */
   public List<Species> getAllSpecies()
   {
      return tracker.getAllSpecies();
   }
   
   /**
    * Returns teh SpeciesTracker being used
    * 
    * @return The SpeciesTracker being used
    */
   public SpeciesTracker getTracker()
   {
      return tracker;
   }
}
