package org.seatiger.pso;

import java.awt.Color;
import java.util.*;

/**
 * Monitors a SpeciesNeighborhood and can returns which species are on
 * global optima.
 * This class can also be given a {@link SwarmDisplay2D} and will
 * colour the particles based on whether they are in a species or
 * not.  
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class SpeciesNeighborhoodMonitor 
{
   /** The swarm being monitored */
   private Swarm swarm;
   /** The neighbourhood being monitored */
   private SpeciesNeighborhood neighborhood;
   /** The display to colour the particles on */
   private SwarmDisplay2D display;
   /** The best fitness of a species must be within this value 
    * of the optima's fitness for it to be considered to have 
    * found an optima */
   private double requiredFitnessAccuracy = 0.1;
   /** Specifies the fitness of the global optimum, if known */
   private double optimalFitness = Double.NaN;
   
   /**
    * Creates a new SpeciesNeighborhoodMonitor
    *
    * @param swarm The swarm to monitor
    * @param neighborhood The neighbourhood to monitor
    */
   public SpeciesNeighborhoodMonitor(Swarm swarm, 
         SpeciesNeighborhood neighborhood)
   {
      this(swarm, neighborhood, null);
   }

   /**
    * Creates a new SpeciesNeighborhoodMonitor
    *
    * @param swarm The swarm to monitor
    * @param neighborhood The neighbourhood to monitor
    * @param display The display to colour the particles on (can be null)
    */
   public SpeciesNeighborhoodMonitor(Swarm swarm, 
         SpeciesNeighborhood neighborhood, SwarmDisplay2D display)
   {
      this.swarm = swarm;
      this.neighborhood = neighborhood;
      this.display = display;
   }
   
   
   /**
    * Returns the requiredFitnessAccuracy 
    *
    * @return The requiredFitnessAccuracy.
    */
   public double getRequiredFitnessAccuracy()
   {
      return this.requiredFitnessAccuracy;
   }

   /**
    * Sets the requiredFitnessAccuracy 
    *
    * @param requiredFitnessAccuracy The requiredFitnessAccuracy to set.
    */
   public void setRequiredFitnessAccuracy(double requiredFitnessAccuracy)
   {
      this.requiredFitnessAccuracy = requiredFitnessAccuracy;
   }

   
   /**
    * Builds the species on optima set
    * 
    * @param speciesDistance The minimum distance two species must be
    * apart before they are considered to represent different optima.
    * To allow multiple species to represent the same optimum, set
    * this to a negative value. 
    * @return The species that are currently on global optima
    */
   public Set<Set<Particle>> findSpeciesOnOptima(double speciesDistance)
   {
      Set<Set<Particle>> species = new HashSet<Set<Particle>>();
      
      for (Species currS : neighborhood.getAllSpecies())
      {
         Set<Particle> currSParticles = new HashSet<Particle>();
         currSParticles.addAll(currS.getParticles());
         
         species.add(currSParticles);
      }
      
      HashMap<Particle, Set<Particle>> speciesBests = 
         new HashMap<Particle, Set<Particle>>();
    
      Set<Set<Particle>> speciesOnOptima = new HashSet<Set<Particle>>();
      
      double bestFitness = Double.NEGATIVE_INFINITY;
      
      for (Set<Particle> currSpecies : species)
      {
         ArrayList<Particle> csList = new ArrayList<Particle>(currSpecies);
         Collections.sort(csList, new FitnessComparator(true));
         Particle best = csList.get(csList.size() - 1);
         
         speciesBests.put(best, currSpecies);
         
         double speciesBestFitness = 
            csList.get(csList.size() - 1).getPersonalBestFitness()[0];
         
         if (speciesBestFitness > bestFitness)
            bestFitness = speciesBestFitness;
      }

      // We still need the code above to work out the species bests
      if (!Double.isNaN(optimalFitness))
         bestFitness = optimalFitness;

      BranchNode rootNode = new BranchNode(swarm.getVariableRange());
      
      for (Particle p : speciesBests.keySet())
      {
         if (bestFitness - p.getPersonalBestFitness()[0] 
               < requiredFitnessAccuracy)
         {
            if (speciesDistance < 0)
            {
               speciesOnOptima.add(speciesBests.get(p));
            }
            else
            {
               Set<LeafNode> closeOptima = rootNode.getAllLeafNodesNear(
                     p.getLocation(), speciesDistance);
               
               if (closeOptima.size() == 0)
               {
                  speciesOnOptima.add(speciesBests.get(p));
                  rootNode.add(p.getLocation());
               }
            }
         }
      }
      
      return speciesOnOptima;
   }
   
   /**
    * Returns the display being updated
    * 
    * @return Returns the display.
    */
   public SwarmDisplay2D getDisplay()
   {
      return display;
   }

   /**
    * Sets the display to update 
    * 
    * @param display The display to set.
    */
   public void setDisplay(SwarmDisplay2D display)
   {
      this.display = display;
   }

   /**
    * Updates the SwarmDisplay2D
    */
   public void updateDisplay()
   {
      if (display == null)
         return;
      
      Set<Set<Particle>> speciesOnOptima = findSpeciesOnOptima(-1);
      
      for (Particle p : swarm.getParticles())
         display.setParticleColor(p, null);
         
      for (Species species : neighborhood.getAllSpecies())
      {
         boolean onOptima = speciesOnOptima.contains(species);
         
         for (Particle p : species.getParticles())
         {
            if (onOptima)
               display.setParticleColor(p, Color.GREEN);
            else   
               display.setParticleColor(p, Color.CYAN);
         }
      }
   }

   /** 
    * Returns the fitness of the global optimum. Note the fitness is not
    * calculated; it is specified by calling {@link #setOptimalFitness(double)}.
    * 
    * @return The fitness of the global optimum.
    */
   public double getOptimalFitness()
   {
      return optimalFitness;
   }

   /** 
    * Sets the fitness of the global optimum
    * 
    * @param optimalFitness The fitness of the global optimum, or {@link Double#NaN} 
    * if unknown
    */
   public void setOptimalFitness(double optimalFitness)
   {
      this.optimalFitness = optimalFitness;
   }
}
