package org.seatiger.pso;

import java.util.*;

/**
 * Tracks species movements across iterations
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class SpeciesTracker implements SpeciesClassifier
{
   /** The Swarm we are running under */
   private Swarm swarm;
   /** The delegate SpeciesClassifier */
   private SpeciesClassifier classifier;
   /** The current list of species */
   private ArrayList<Species> species = new ArrayList<Species>();
      
   /**
    * Creates a new SpeciesTracker
    * 
    * @param swarm The swarm we're tracking particles within
    * @param classifier The classifier that is determining the species
    */
   public SpeciesTracker(SpeciesClassifier classifier, Swarm swarm)
   {
      if (swarm == null)
         throw new NullPointerException("Swarm cannot be null");
      if (classifier == null)
         throw new NullPointerException("Classifier cannot be null");
      
      this.swarm = swarm;
      this.classifier = classifier;
   }


   public Map<Particle, Integer> classify(Collection<Particle> particles)
   {
      buildSpecies(particles);
      
      HashMap<Particle, Integer> result = 
            new HashMap<Particle, Integer>(particles.size());
      int speciesNum = 0;
      for (Species currSpecies : species)
      {
         Integer value = speciesNum++;
         
         for (Particle p : currSpecies.getParticles())
            result.put(p, value);
      }
      
      return result;
   }
   
   /**
    * Returns the species
    * 
    * @return The species
    */
   public List<Species> getAllSpecies()
   {
      Iterator<Species> iterator = species.iterator();
      while (iterator.hasNext())
      {
         if (iterator.next().isExtinct())
            iterator.remove();
      }
      
      return Collections.unmodifiableList(species);
   }
   
   /**
    * Builds the new species list
    * 
    * @param particles The current set of particles
    */
   private void buildSpecies(Collection<Particle> particles)
   {
      Map<Integer, Set<Particle>> classification = 
            Swarm.reverseMap(classifier.classify(particles));
      
      ArrayList<Species> currSpecies = new ArrayList<Species>();
      for (Set<Particle> speciesSet : classification.values())
         currSpecies.add(new Species(swarm, speciesSet));
      
      trackSpecies(currSpecies);
   }
   
   /**
    * Compares the species at the last tick with the species at the current tick.
    * Finds the species that have moved and matches their new position to the old 
    * 
    * @param thisTick The current list of species
    */
   private void trackSpecies(List<Species> thisTick)
   {
      final int thisTickSpecies = thisTick.size();
      final int lastTickSpecies = species.size();
      final double[][] distances = new double[lastTickSpecies][thisTickSpecies];
      
      ArrayList<Species> result = new ArrayList<Species>(thisTickSpecies);
      boolean oldSpeciesUsed[] = new boolean[lastTickSpecies];
      boolean newSpeciesUsed[] = new boolean[thisTickSpecies];
      
      if (lastTickSpecies > 0)
      {
         for (int i = 0; i < lastTickSpecies; i++)
         {
            if (species.get(i).isExtinct())
            {
               oldSpeciesUsed[i] = true;
               continue;
            }
            
            for (int j = 0; j < thisTickSpecies; j++)
            {
               distances[i][j] = VectorMath.distance(
                     species.get(i).getSeed(), thisTick.get(j).getSeed());
            }
         }
               
         int totalDists = lastTickSpecies * thisTickSpecies;
         ArrayList<Integer> refs = new ArrayList<Integer>(totalDists);
         for (int i = 0; i < totalDists; i++)
            refs.add(i);
         
         // Sort all the elements in the distances array
         Collections.sort(refs, new Comparator<Integer>() {
            public int compare(Integer a, Integer b)
            {
               double distA = distances[a / thisTickSpecies][a % thisTickSpecies];
               double distB = distances[b / thisTickSpecies][b % thisTickSpecies];
               
               if (distA < distB)
                  return -1;
               else if (distA > distB)
                  return 1;
               else
                  return 0;
            }
         });
         
         int j = 0;
         for (int i = Math.min(lastTickSpecies, thisTickSpecies); i > 0; i--)
         {
            while (true)
            {
               int ref = refs.get(j++);
               int oldSpeciesNum = ref / thisTickSpecies;
               int newSpeciesNum = ref % thisTickSpecies;
               
               if (!oldSpeciesUsed[oldSpeciesNum] && !newSpeciesUsed[newSpeciesNum])
               {
                  Species existing = species.get(oldSpeciesNum);
                  existing.setParticles(thisTick.get(newSpeciesNum).getParticles());
                  result.add(existing);
                  oldSpeciesUsed[oldSpeciesNum] = true;
                  newSpeciesUsed[newSpeciesNum] = true;
                  
                  break;
               }
            }
         }
      }
      
      for (int i = 0; i < newSpeciesUsed.length; i++)
      {
         if (!newSpeciesUsed[i])
            result.add(thisTick.get(i));
      }
      
      for (int i = 0; i < oldSpeciesUsed.length; i++)
      {
         if (!oldSpeciesUsed[i])
            species.get(i).die();
      }
      
      species = result;
   }
   
   /**
    * Returns the delegate classifier
    * 
    * @return The delegate classifier
    */
   public SpeciesClassifier getDelegate()
   {
      return classifier;
   }
}
 