package org.seatiger.pso;

import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Represents a star neighbourhood, where all individuals are linked
 * to a single "focal" individual
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class StarNeighborhood implements Neighborhood
{
   /** The particles registered to the neighbourhood */
   private LinkedHashSet<Particle> particles = new LinkedHashSet<Particle>();
   /** The focal particle of the neighbourhood */
   private Particle focal = null;
   
   /**
    * @see org.seatiger.pso.Neighborhood#getNeighborhoodOf(org.seatiger.pso.Particle)
    */
   public Set<Particle> getNeighborhoodOf(Particle p)
   {
      if (!particles.contains(p))
         return null;
      
      LinkedHashSet<Particle> result = new LinkedHashSet<Particle>();
      
      if (p != focal)
      {
         result.add(p);
         result.add(focal);
      }
      else
         result.addAll(particles);
      
      return result;
   }

   /**
    * Registers the particle with the neighbourhood. If there are no
    * other particles registered, this particle will be set as the focal
    * particle. 
    * 
    * @see org.seatiger.pso.Neighborhood#register(org.seatiger.pso.Particle)
    */
   public void register(Particle p)
   {
      if (particles.isEmpty())
         focal = p;
      
      particles.add(p);
   }

   /**
    * Deregisters the particle from the neighbourhood. If <code>p</code>
    * is the focal particle, another particle will be set as the focal.
    * 
    * @see org.seatiger.pso.Neighborhood#deregister(org.seatiger.pso.Particle)
    */
   public void deregister(Particle p)
   {
      particles.remove(p);
      
      if (particles.isEmpty())
         focal = null;
      else if (focal == p)
      {
         for (Particle temp : particles)
         {
            focal = temp;
            break;
         }
      }
   }
   
   /**
    * @see Neighborhood#deregisterAll()
    */
   public void deregisterAll()
   {
      particles.clear();
   }
   
   /**
    * Returns the focal particle of this neighbourhood. 
    * 
    * @return The focal particle, or null if there are no particles
    * in this neighbourhood
    */
   public Particle getFocalParticle()
   {
      return focal;
   }
   
   /**
    * Sets <code>p</code> to be the focal particle. If <code>p</code>
    * is not currently registered with this neighbourhood, this method
    * does nothing
    * 
    * @param p The particle to set as the focal particle
    */
   public void setFocalParticle(Particle p)
   {
      if (particles.contains(p))
         focal = p;
   }

}
