package org.seatiger.pso;

import java.util.*;

import javax.swing.event.EventListenerList;

import uchicago.src.sim.util.Random;

/**
 * Represents a particle swarm
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class Swarm
{
   /** Doubles closer than this amount are considered the same */	
   public final static double EPSILON = 0.0000001;
	
   /** Represents NONE. This value is used by various other classes */
   public final static int NONE = -1;
   
   /** The minimum number of particles the swarm will have unless otherwise set */
   public final static int DEFAULT_MIN_PARTICLES = 20;
   
   /** Whether we can still add fitness functions. This becomes false once
    * particles are added. Other method calls may also cause this variable
    * to become false - it is important to add all of the fitness functions 
    * before doing anything else. 
    */
   private boolean canAddFitnessFunctions = true;
   
   /** Records the number of ticks since the start of the simulation */
   private int currTick = 0;
   
   /** All of the fitness functions that are being used. For single objective
    * systems this will only contain one fitness function
    */
   private ArrayList<FitnessFunction> fitnessFunctions;
   
   /** All of the particles in the simulation */
   private ArrayList<Particle> particles;
   
   /** The allowable range for each dimension */
   private Range[] ranges;
   
   /** The number of dimensions in the search problem */
   private int numVariables;

   /** The minimum number of particles in the simulation. If when the simulation
    * is started there are fewer than <code>minParticles</code> particles, 
    * extra particles will be created automatically. 
    */
   private int minParticles = DEFAULT_MIN_PARTICLES;
   
   /** The movement model to use to create particles */
   private MovementModel defaultMovementModel = null;
   
   /** A list of all the movement models in the swarm */
   private ArrayList<MovementModel> movementModels;
   
   /** Registered listeners for events */
   private EventListenerList listeners = new EventListenerList();
   
   /**
    * Creates a new Swarm without any fitness functions
    */
   public Swarm()
   {
      fitnessFunctions = new ArrayList<FitnessFunction>();
      particles = new ArrayList<Particle>();
   }
   
   /**
    * Sets the default MovementModel. This model is used when 
    * creating new particles. 
    * 
    * @param model The model to use by default when creating new
    * particles 
    */
   public void setDefaultMovementModel(MovementModel model)
   {
      defaultMovementModel = model;
   }
   
   /**
    * Returns the default MovementModel
    * 
    * @return The default MovementModel
    */
   public MovementModel getDefaultMovementModel()
   {
      return defaultMovementModel;
   }
   
   /**
    * Adds a single particle to the swarm
    * 
    * @param p The particle to add
    * @throws IllegalStateException if step() has already been called (ie the 
    * simulation has been started)
    */
   public void addParticle(Particle p)
   {
      if (ranges == null)
         calculateRanges();
      
      particles.add(p);
   }

   /**
    * Adds a single, randomly located particle with the given MovementModel
    * 
    * @param model The MovementModel for the particle
    */
   public void addParticle(MovementModel model)
   {
      addParticles(1, model);
   }
   
   /**
    * Adds the given number of particles to the swarm
    * 
    * @param numToAdd The number of particles to add
    * @throws NullPointerException if the default movement model hasn't been 
    * set or there are no defined fitness functions
    * @throws IllegalStateException if step() has already been called (ie the 
    * simulation has been started)
    */
   public void addParticles(int numToAdd)
   {
      if (defaultMovementModel == null)
         throw new NullPointerException("Default Movement Model has not been "
               + "set");
            
      addParticles(numToAdd, defaultMovementModel);
   }
   
   /**
    * Adds the given number of particles to the swarm
    * 
    * @param numToAdd The number of particles to add
    * @param movementModel The movement model to use for the particles
    * @throws NullPointerException if there are no defined fitness functions
    * @throws IllegalStateException if step() has already been called (ie the 
    * simulation has been started)
    */
   public void addParticles(int numToAdd, MovementModel movementModel)
   {
      if (fitnessFunctions.size() == 0)
         throw new NullPointerException("No fitness functions have been defined");
      
      canAddFitnessFunctions = false;
      
      if (ranges == null)
         calculateRanges();
      
      for (int i = 0; i < numToAdd; i++)
      {
         addParticle(new Particle(getRandomLocation(), getRandomVelocity(), 
               this, movementModel));
      }
   }
   
   /**
    * Returns a random location
    * 
    * @return A random location
    */
   public double[] getRandomLocation()
   {
      double[] location = new double[numVariables];
      
      for (int dimension = 0; dimension < numVariables; dimension++)
      {
         Range range = ranges[dimension];
         
         location[dimension] = range.getMin() + 
               Random.uniform.nextDouble() * 
               (range.getMax() - range.getMin());
      }
      
      return location;
   }
   
   /**
    * Returns a random point within maxDistance of seedPoint
    * 
    * @param seedPoint The point to generate a point near
    * @param maxDistance The maximum allowable distance from seedPoint
    * @return A poing within maxDistance of seedPoint
    */
   public double[] getRandomLocation(double[] seedPoint, double maxDistance)
   {
      double r = Random.uniform.nextDouble() * maxDistance;
      double[] point = new double[seedPoint.length];
      Arrays.fill(point, r);
      
      for (int i = 1; i < point.length; i++)
      {
         double angle = Random.uniform.nextDouble() * Math.PI * 2;
         
         for (int j = 0; j < i; j++)
            point[j] *= Math.cos(angle);
         
         point[i] *= Math.sin(angle);
      }
      
      for (int i = 0; i < point.length; i++)
         point[i] += seedPoint[i];
      
      return sanitizeLocation(point);
   }
   
   /**
    * Returns a random velocity 
    * 
    * @return A random velocity
    */
   public double[] getRandomVelocity()
   {
      double[] velocity = new double[numVariables];
      
      for (int dimension = 0; dimension < numVariables; dimension++)
      {
         Range range = ranges[dimension];
         
         velocity[dimension] = range.getSize() - 
               Random.uniform.nextDouble() * range.getSize() * 2; 
      }
      
      return velocity;
   }
   
   /**
    * Returns a random velocity 
    * 
    * @param The maximum allowed velocity
    * @return A random velocity
    */
   public double[] getRandomVelocity(double maxVelocity)
   {
      double r = Random.uniform.nextDouble() * maxVelocity;
      double[] velocity = new double[getNumVariables()];
      Arrays.fill(velocity, r);
      
      for (int i = 1; i < velocity.length; i++)
      {
         double angle = Random.uniform.nextDouble() * Math.PI * 2;
         
         for (int j = 0; j < i; j++)
            velocity[j] *= Math.cos(angle);
         
         velocity[i] *= Math.sin(angle);
      }
      
      return velocity;
   }
   
   /**
    * Removes the given particle from the swarm. 
    * 
    * @param p The particle to remove
    */
   public void removeParticle(Particle p)
   {
      if (particles.contains(p))
      {
         p.getMovementModel().deregister(p);
         particles.remove(p);
      }
   }

   /**
    * Returns the minimum number of particles in the swarm
    *
    * @return The minParticles.
    */
   public int getMinParticles()
   {
      return minParticles;
   }

   /**
    * Returns the number of particles currently in the swarm. Until 
    * {@link #step()} is first called, this number may be less than 
    * the number returned by {@link #getMinParticles()}. The extra
    * particles will be created when step() is first called
    * 
    * @return The number of particles currently in the swarm
    */
   public int getNumParticles()
   {
      return particles.size();
   }

   /**
    * Sets the minimum number of particles in the swarm. If there are fewer
    * than <code>minParticles</code> in the swarm, extra particles will be
    * created using the default movement model.  
    *
    * @param minParticles The minimum number of particles in the swarm
    * @throws IllegalStateException if step() has already been called (ie
    * the simulation has been started) 
    */
   public void setMinParticles(int minParticles)
   {
      if (currTick > 0)
         throw new IllegalStateException("Cannot set the number of particles "
               + "after the simulation has started");
      
      this.minParticles = minParticles;
   }


   /**
    * Adds a fitness function to the simulation. Any number of fitness functions
    * can be added, allowing multi-objective simulations to be performed. 
    * Fitness functions can only added when the Swarm is first created. Calls 
    * to most methods will prevent further fitness functions from being added. 
    * 
    * @param func The fitness function to add
    * @throws IllegalStateException if step() has already been called (ie
    * the simulation has been started), particles have already been added or
    * some other method has been called that prevents further fitness functions
    * from being registered. 
    * @throws InvalidFitnessFunctionException if <code>func</code> requires
    * a different number of parameters to the previously added fitness functions 
    */
   public void addFitnessFunction(FitnessFunction func)
   {
      if (!canAddFitnessFunctions)
         throw new IllegalStateException("Fitness functions must be added " +
                "before any other methods are called.");
      
      if (fitnessFunctions.size() > 0 && func.getNumVariables() != numVariables)
         throw new InvalidFitnessFunctionException("Existing fitness functions "
               + "use a " + numVariables + " dimensional decision space, "
               + "added function uses a " + func.getNumVariables() + " dimensional"
               + "decision space");
      
      fitnessFunctions.add(func);
      
      if (fitnessFunctions.size() == 1)
         numVariables = func.getNumVariables();
   }
   
   /**
    * Returns the number of ticks since the start of the simulation 
    *
    * @return The number of ticks since the start of the simulation
    */
   public int getCurrTick()
   {
      return currTick;
   }
   
   /** 
    * Initialises the swarm. This creates the movement model list from 
    * the movement model of each particle. It also notifies  
    * {@link SwarmListener#initialized(SwarmEvent)} on each registered 
    * SwarmListener.
    */
   private void initialise()
   {
      movementModels = new ArrayList<MovementModel>();
      
      if (particles.size() < minParticles)
         addParticles(particles.size() - minParticles);
      
      for (Particle p : particles)
      {
         MovementModel currModel = p.getMovementModel();
         
         if (!movementModels.contains(currModel))
            movementModels.add(currModel);
      }

      SwarmEvent event = new SwarmEvent(this);
      for (SwarmListener listener : listeners.getListeners(SwarmListener.class))
         listener.initialized(event);
   }
   
   /** 
    * Runs the swarm for one step
    */
   public void step()
   {
      if (currTick == 0)
         initialise();
      
      currTick++;
      
      Iterator<Particle> particleIterator = particles.iterator();
      while (particleIterator.hasNext())
      {
         Particle currParticle = particleIterator.next();
         currParticle.move();
      }

      SwarmEvent e = new SwarmEvent(this);
      for (SwarmListener listener : listeners.getListeners(SwarmListener.class))
         listener.step(e);
   }
   

   /**
    * Calculate the valid ranges for each dimension. The valid range is the
    * smallest single range that includes every fitness function's preferred
    * range
    * 
    * @throws IllegalStateException if no fitness functions have yet been 
    * defined
    */
   private void calculateRanges()
   {
      canAddFitnessFunctions = false;

      if (fitnessFunctions.size() == 0)
         throw new IllegalStateException("No fitness functions have been " +
                "defined");
      
      ranges = new Range[numVariables];
      Range[][] allRanges = new Range[fitnessFunctions.size()][];
      
      for (int i = 0; i < fitnessFunctions.size(); i++)
         allRanges[i] = fitnessFunctions.get(i).getPreferredRanges();
      
      for (int dimension = 0; dimension < numVariables; dimension++)
      {
         double min = allRanges[0][dimension].getMin();
         double max = allRanges[0][dimension].getMax();
         
         for (int functionNum = 1; functionNum < fitnessFunctions.size(); 
               functionNum++)
         {
            if (allRanges[functionNum][dimension].getMin() < min)
               min = allRanges[functionNum][dimension].getMin();
            if (allRanges[functionNum][dimension].getMax() > max)
               max = allRanges[functionNum][dimension].getMax();
         }
         
         ranges[dimension] = new Range(min, max);
      }
   }
   
   /**
    * Returns the fitnesses of the given location. The fitnesses are found
    * by asking each fitness function in turn for its evaluation of the
    * location
    * 
    * @param location The position to check 
    * @return An array with one element per fitness function holding the fitness
    * of location as given by each fitness function
    */
   public double[] calculateFitness(double[] location)
   {
      double[] fitnesses = new double[fitnessFunctions.size()];
      
      for (int i = 0; i < fitnesses.length; i++)
      {
         fitnesses[i] = fitnessFunctions.get(i).getFitness(location);
      }
      
      return fitnesses;
   }
   
   /**
    * Returns all of the particles in the Swarm
    * 
    * @return The particles in the swarm
    */
   public List<Particle> getParticles()
   {
      return Collections.unmodifiableList(particles);
   }
   
   /**
    * Returns a list containing all of the currently non-dominated particles
    * 
    * @return a list containing all of the currently non-dominated particles
    */
   public List<Particle> getNonDominatedParticles()
   {
      return getNonDominatedParticles(particles);
   }
   
   /**
    * Returns the number of objectives this particle swarm is trying
    * to optimise
    * 
    * @return The number of objectives that have been defined for the swarm. 
    */
   public int getNumObjectives()
   {
      return fitnessFunctions.size();
   }
   
   /**
    * Returns the bounds of the problem space. Calling this function will 
    * prevent further fitness functions being added. 
    * 
    * @return The bounds of the problem space. 
    * @throws IllegalStateException if no fitness functions have yet been 
    * defined
    */
   public Range[] getVariableRange()
   {
      if (ranges == null)
         calculateRanges();
      
      Range[] result = new Range[ranges.length];
      System.arraycopy(ranges, 0, result, 0, ranges.length);

      return result;
   }
   
   /**
    * Returns the number of decision variables 
    * 
    * @return The number of decision variables
    */
   public int getNumVariables()
   {
      return numVariables;
   }
   
   /**
    * Registers a listener interested in receiving SwarmEvents
    * 
    * @param l The interested listener
    */
   public void addSwarmListener(SwarmListener l)
   {
      listeners.add(SwarmListener.class, l);
   }
   
   /**
    * Stops further notifications to the listener of future SwarmEvents 
    * 
    * @param l The listener to deregister
    */
   public void removeSwarmListener(SwarmListener l)
   {
      listeners.remove(SwarmListener.class, l);
   }
   
   /**
    * Returns true if the Swarm is already running, that is {@link #step()}
    * has been called
    * 
    * @return True if the Swarm is running
    */
   public boolean isRunning()
   {
      return currTick > 0;
   }
   
   /**
    * Converts a double array to a String 
    * 
    * @param array The array to convert
    * @return The array formatted as a String
    */
   public static String toString(double[] array)
   {
      StringBuffer result = new StringBuffer();
      
      if (array == null)
         result.append("null");
      else
      {
         result.append('{');
         
         for (int i = 0; i < array.length; i++)
         {
            if (i > 0)
               result.append(", ");
            
            result.append(array[i]);
         }
         
         result.append('}');
      }
      
      return result.toString();
   }

   /**
    * Returns a list containing only the non-dominated particles in <code>
    * particles</code>
    * 
    * @param particles The list to search for non-dominated particles
    * @return The non-dominated particles within <code>particles</code>
    */
   public static List<Particle> getNonDominatedParticles(List<Particle> particles)
   {
      List<Particle> nonDominatedParticles = new ArrayList<Particle>();
      Iterator<Particle> particleIterator = particles.iterator();
      
      while (particleIterator.hasNext())
         addIfNonDominated(nonDominatedParticles, particleIterator.next());
      
      return nonDominatedParticles;
   }
   
/**
    * Adds the solution to the list of non-dominated solutions if it is
    * not dominated by any of the solutions in the list. Any solutions 
    * in the list that are dominated by this solution are removed. 
    * 
    * @param nonDominatedSolns The list of non-dominated solutions
    * @param solution The solution to add
    */
   private static void addIfNonDominated(List<Particle> nonDominatedSolns,
         Particle solution)
   {
      boolean solnNonDominated = true;
      
      // We try to cut down the non-dominated solution list up here so
      // we have fewer things to check against the global list
      Iterator<Particle> solutionIterator = nonDominatedSolns.iterator();
      while (solnNonDominated && solutionIterator.hasNext())
      {
         Particle testingSoln = solutionIterator.next();
         
         if (solution.dominates(testingSoln))
            solutionIterator.remove();
         else if (testingSoln.dominates(solution))
            solnNonDominated = false;
      }
      
      if (solnNonDominated)
         nonDominatedSolns.add(solution);
   }

   /**
    * Returns true if the point is within the decision space
    * 
    * @param testPoint The point to test
    * @return True if the point is within the decision space
    */
   public boolean pointWithinBounds(double[] testPoint)
   {
      if (testPoint.length != getNumVariables())
         throw new IllegalArgumentException("Invalid number of dimensions");
      
      for (int i = getNumVariables() - 1; i >= 0; i--)
      {
         if (!ranges[i].contains(testPoint[i]))
            return false;
      }
      
      return true;
   }

   /**
    * Returns a location within the decision space that is in some way close
    * to <code>location</code>. If <code>location</code> is already within the
    * decision space, the returned result will be identical
    *  
    * @param location The location to sanitise
    * @return A new location that is within the decision space
    */
   public double[] sanitizeLocation(double[] location)
   {
      double[] result = new double[location.length];
      System.arraycopy(location, 0, result, 0, location.length);
      
      for (int i = result.length - 1; i >= 0; i--)
      {
            // Clipping
   //            if (location[i] < decisionRange[i].getMin())
   //               location[i] = decisionRange[i].getMin();
   //            else
   //               location[i] = decisionRange[i].getMax();
            
            // Bouncing
         if (result[i] < ranges[i].getMin())
         {
            double excess = ranges[i].getMin() - result[i];
            int bounces = 1 + (int) Math.floor(excess / ranges[i].getSize());
            
            if (bounces % 2 == 1) // last bounce is off left
               result[i] = ranges[i].getMin() + excess % ranges[i].getSize();
            else // last bounce is off right
               result[i] = ranges[i].getMax() - excess % ranges[i].getSize();
         }
         
         if (result[i] > ranges[i].getMax())
         {
            double excess = result[i] - ranges[i].getMax();
            int bounces = 1 + (int) Math.floor(excess / ranges[i].getSize());
            
            if (bounces % 2 == 0) // last bounce is off left
               result[i] = ranges[i].getMin() + excess % ranges[i].getSize();
            else // last bounce is off right
               result[i] = ranges[i].getMax() - excess % ranges[i].getSize();
         }
         
         if (!ranges[i].contains(result[i]))
            throw new RuntimeException(ranges[i] + " not in " + ranges[i].toString());
      }
      
      return result;
   }
   
   /**
    * Returns the pareto fronts as a list
    * 
    * @param fitnesses The fitnesses to place into fronts
    * @return The fitnesses organised into a list of non-dominated
    * collections. The list is ordered from the most dominant to the 
    * least dominant front. 
    */
   public static List<Collection<double[]>> findParetoFronts(
         Collection<double[]> fitnesses)
   {
      ArrayList<ArrayList<Integer>> dominationList = 
         new ArrayList<ArrayList<Integer>>(fitnesses.size());
      
      // An array of Integer objects to avoid boxing costs
      Integer[] ints = new Integer[fitnesses.size()];

      for (int i = 0; i < ints.length; i++)
      {
         ints[i] = new Integer(i);
         dominationList.add(new ArrayList<Integer>());
      }
      
      double[][] individuals = fitnesses.toArray(new double[0][]);
      int[] dominationCounts = new int[fitnesses.size()];
      
      ArrayList<Collection<Integer>> fronts = new ArrayList<Collection<Integer>>();
      fronts.add(new ArrayList<Integer>());
      
      for (int p = 0; p < fitnesses.size(); p++)
      {
         for (int q = 0; q < fitnesses.size(); q++)
         {
            if (p == q)
               continue;
            
            int dominates = Swarm.compareDomination(individuals[p], individuals[q]);
            
            if (dominates < 0) // p dominates q
               dominationList.get(p).add(ints[q]);
            else if (dominates > 0) // q dominates p
               dominationCounts[p]++;
         }
         
         if (dominationCounts[p] == 0)
            fronts.get(0).add(ints[p]);
      }
      
      for (int frontNum = 0; fronts.get(frontNum).size() > 0; frontNum++)
      {
         fronts.add(new ArrayList<Integer>());
         
         Iterator<Integer> iterator = fronts.get(frontNum).iterator();
         while (iterator.hasNext())
         {
            int p = iterator.next();
            
            for (int q : dominationList.get(p))
            {
               if (--dominationCounts[q] == 0)
                  fronts.get(frontNum + 1).add(ints[q]);
            }
         }
      }
      
      ArrayList<Collection<double[]>> result = 
         new ArrayList<Collection<double[]>>(fronts.size());
      
      for (int frontNum = 0; frontNum < fronts.size() - 1; frontNum++)
      {
         ArrayList<double[]> currFront = 
            new ArrayList<double[]>(fronts.get(frontNum).size());
         
         for (int p : fronts.get(frontNum))
            currFront.add(individuals[p]);

         result.add(currFront);
      }
      
      return result;
   }

   /**
    * Returns an integer depending on whether fitness1 dominates, is 
    * dominated by or is non-dominant of fitness2 
    *  
    * @param fitness1 The fitness to compare
    * @param fitness2 The fitness to compare
    * @return Negative, 0 or positive if fitness1 dominates, is non-dominated
    * or dominated by fitness2 respectively. 
    */
   public static int compareDomination(double[] fitness1, double[] fitness2)
   {
      boolean fitness1Dominates = false;
      boolean fitness2Dominates = false;
      
      if (fitness1.length != fitness2.length)
         throw new IllegalArgumentException("Fitnesses must be the same length");
      
      for (int i = 0; i < fitness1.length; i++)
      {
         if (fitness1[i] > fitness2[i])
            fitness1Dominates = true;
         else if (fitness2[i] > fitness1[i])
            fitness2Dominates = true;
      }
      
      if (fitness1Dominates && !fitness2Dominates)
         return -1;
      else if (fitness2Dominates && !fitness1Dominates)
         return 1;
      else 
         return 0;
   }
   
   /**
    * Performs a non-dominated sort of the given fitnesses
    * 
    * @param fitnesses The fitnesses to sort
    * @return The fitnesses sorted by domination, the non-dominated fitnesses being in
    * the first element of the return List.  
    */
   public static List<Collection<double[]>> nonDomSort(Collection<double[]> fitnesses)
   {
      ArrayList<Collection<double[]>> result = new ArrayList<Collection<double[]>>();
      LinkedList<DomInfo> domInfos = new LinkedList<DomInfo>();
      for (double[] fitness : fitnesses)
         domInfos.add(new DomInfo(fitness));
      
      for (DomInfo infoA : domInfos)
      {
         for (DomInfo infoB : domInfos)
         {
            if (dominates(infoA.fitness, infoB.fitness))
               infoA.dominates.add(infoB);
            else if (dominates(infoB.fitness, infoA.fitness))
               infoA.numDominatedBy++;
         }
      }
      
      while (!domInfos.isEmpty())
      {
         LinkedList<DomInfo> currLevelInfos = new LinkedList<DomInfo>();
         
         Iterator<DomInfo> iterator = domInfos.iterator();
         while (iterator.hasNext())
         {
            DomInfo currInfo = iterator.next();
            
            if (currInfo.numDominatedBy == 0)
            {
               currLevelInfos.add(currInfo);
               iterator.remove();
            }
         }
         
         ArrayList<double[]> currLevelResult = new ArrayList<double[]>();
         result.add(currLevelResult);
         
         for (DomInfo currInfo : currLevelInfos)
         {
            for (DomInfo otherInfo : currInfo.dominates)
               otherInfo.numDominatedBy--;
            
            currLevelResult.add(currInfo.fitness);
         }
      }
      
      return result;
   }
   
   /**
    * Returns true if fitness1 dominates fitness2
    * 
    * @param fitness1 The first fitness to compare
    * @param fitness2 The second fitness to compare
    * @return True if fitness1 dominates fitness2
    */
   public static boolean dominates(double[] fitness1, double[] fitness2)
   {
      return compareDomination(fitness1, fitness2) < 0;
   }

   /**
    * Clears the personal best memory of all particles
    */
   public void clearPersonalBests()
   {
	   for (Particle p : particles)
		   p.clearPBestMemory();
   }
   
   /**
    * Reverses a key->value map to a value->Set(key) map. This method was
    * written to convert the result returned by 
    * {@link SpeciesClassifier#classify(Collection)} to a 
    * Map<Integer, Set<Particle>>, however it has other uses as well. 
    * 
    * @param <T> The key type in the original map
    * @param <U> The value type in the original map
    * @param input The input map
    * @return A map with the keys and values reversed
    */
   public static <T, U> Map<U, Set<T>> reverseMap(Map<T, U> input)
   {
      HashMap<U, Set<T>> result = new HashMap<U, Set<T>>();
      
      for (T key : input.keySet())
      {
         U value = input.get(key);
         
         Set<T> keySet = result.get(value);
         if (keySet == null)
         {
            keySet = new HashSet<T>();
            result.put(value, keySet);
         }
         
         keySet.add(key);
      }
   
      return result;
   }
   
   /**
    * Tracks info about which fitnesses dominate which other fitnesses.
    * Used by {@link #nonDomSort(Collection)}
    */
   private static class DomInfo
   {
      /** The fitness being stored */
      public double[] fitness;
      /** The number of fitnesses this one is currently dominated by */
      public int numDominatedBy;
      /** The fitesses this fitness dominates */
      public LinkedList<DomInfo> dominates;
 
      /**
       * Creates a new DomInfo for the given fitness
       * 
       * @param fitness The fitness
       */
      public DomInfo(double[] fitness)
      {
         this.fitness = fitness;
         numDominatedBy = 0;
         dominates = new LinkedList<DomInfo>();
      }
   }
}
