package org.seatiger.pso;

import java.awt.*;
import java.awt.geom.Point2D;
import java.util.*;

import uchicago.src.sim.gui.*;

/**
 * Displays a swarm that has two decision variables. 
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class SwarmDisplay2D implements Displayable, Probeable, SwarmListener
{
   /** The size of each particle when drawn on the panel */
   private static final int PARTICLE_SIZE = 7;
   
   /** The Swarm to display */
   protected Swarm swarm;
   
   /** A map of particles to their set colour. If a colour hasn't been set
    * for a particular colour, it will be drawn in the default colour (blue)
    */   
   protected HashMap<Particle, Color>particleColors = 
      new HashMap<Particle, Color>();
   /** The particles already seen by this swarm display. This is used to 
    * highlight recently added particles */
   protected HashSet<Particle> existingParticles = 
      new HashSet<Particle>();
   

   /** The decision space range to display on the screen */
   protected Range[] displayRange = null;
   
   /** The width of the display in pixels */
   protected int width = 0;
   /** The height of the display in pixels */
   protected int height = 0;
      
   /**
    * Creates a new SwarmDisplay2D
    *
    * @param swarm The Swarm to display
    */
   public SwarmDisplay2D(Swarm swarm)
   {
      this.swarm = swarm;
      swarm.addSwarmListener(this);
   }
   
   /**
    * Paints the display onto the given Graphics instance
    * 
    * @param simGraphics The SimGraphics to paint on 
    */
   public void drawDisplay(SimGraphics simGraphics)
   {
      if (!swarm.isRunning())
         return;
      
      if (displayRange == null)
         displayRange = swarm.getVariableRange();
      
      width = simGraphics.getDisplayWidth();
      height = simGraphics.getDisplayHeight();
      
      Graphics2D g = simGraphics.getGraphics();
      
      for (Particle p : swarm.getParticles())
      {
         Color particleColor = particleColors.get(p);
         if (particleColor == null)
         {
            if (existingParticles.contains(p))
               particleColor = Color.BLUE;
            else
               particleColor = Color.YELLOW;
         }

         
         double[] location = p.getLocation();
         double[] velocity = p.getVelocity();
         double[] lastLocation = VectorMath.subtract(location, velocity);
         double[] pBest = p.getPersonalBestLocation();
         int offset = PARTICLE_SIZE / 2;
         
         Point screenLoc = convertToScreen(location[0], location[1]);
         Point lastScreenLoc = convertToScreen(
                  lastLocation[0], lastLocation[1]);
         Point pBestLoc = convertToScreen(pBest[0], pBest[1]);

         if (swarm.getDefaultMovementModel() instanceof GravityPSO)
         {
            GravityPSO gravityPSO = (GravityPSO) swarm.getDefaultMovementModel();
            
//            for (Particle other : swarm.getParticles())
//            {
//               g.setColor(Color.CYAN);
//               double[] force = gravityPSO.getForce(p, other);
//               
//               Point forceLoc = convertToScreen(location[0] + force[0], location[1] + force[1]);
//            
//            }
            
            g.setColor(Color.GREEN);
            double[] nextVel = gravityPSO.getForce(p);
            
            Point nextLoc = convertToScreen(location[0] + nextVel[0], location[1] + nextVel[1]);
            g.drawLine(screenLoc.x, screenLoc.y, nextLoc.x, nextLoc.y);
         }         
         g.setColor(particleColor);
         g.drawLine(screenLoc.x, screenLoc.y, lastScreenLoc.x, lastScreenLoc.y);
         g.fillOval(screenLoc.x - offset, screenLoc.y - offset, 
               PARTICLE_SIZE, PARTICLE_SIZE);
         
         g.drawLine(pBestLoc.x - offset, pBestLoc.y - offset, 
               pBestLoc.x + offset, pBestLoc.y + offset);
         g.drawLine(pBestLoc.x - offset, pBestLoc.y + offset, 
               pBestLoc.x + offset, pBestLoc.y - offset);
         


      }      
   }
   
 
   /**
    * @see uchicago.src.sim.gui.Probeable#getObjectsAt(int, int)
    */
   public ArrayList getObjectsAt(int x, int y)
   {
      ArrayList<Object> result = new ArrayList<Object>();
      
      for (Particle particle : swarm.getParticles())
      {
         Point point = convertToScreen(
               particle.getLocation()[0], particle.getLocation()[1]);
         
         double distanceX = point.getX() - x;
         double distanceY = point.getY() - y;
         
         if (Math.sqrt(distanceX * distanceX + distanceY * distanceY) 
               < PARTICLE_SIZE)
            result.add(particle);
      }
      
      return result;
   }
   

   /**
    * @see org.seatiger.pso.SwarmListener#initialized(org.seatiger.pso.SwarmEvent)
    */
   public void initialized(SwarmEvent e) 
   {
      step(e);
   }

   /**
    * @see org.seatiger.pso.SwarmListener#step(org.seatiger.pso.SwarmEvent)
    */
   public void step(SwarmEvent e)
   {
      existingParticles.clear();
      existingParticles.addAll(swarm.getParticles());
   }

   /* (non-Javadoc)
    * @see uchicago.src.sim.gui.Probeable#setMoveableXY(uchicago.src.sim.gui.Moveable, int, int)
    */
   public void setMoveableXY(Moveable moveable, int x, int y) { }

   /* (non-Javadoc)
    * @see uchicago.src.sim.gui.Displayable#getDisplayableInfo()
    */
   public ArrayList getDisplayableInfo()
   {
      return new ArrayList();
   }

   /* (non-Javadoc)
    * @see uchicago.src.sim.gui.Displayable#viewEventPerformed(uchicago.src.sim.gui.ViewEvent)
    */
   public void viewEventPerformed(ViewEvent evt) { }

   /**
    * @see uchicago.src.sim.gui.Displayable#getSize()
    */
   public Dimension getSize()
   {
      return new Dimension(400, 400);
   }
   
   /**
    * Converts a model point to screen coordinates
    * 
    * @param x1 The point in the model in dimension 1
    * @param x2 The point in the model in dimension 2
    * @return The location on screen of the specified model coordinate, or 
    * null if {@link #drawDisplay(SimGraphics)} has not been called
    */
   protected Point convertToScreen(double x1, double x2)
   {
      if (displayRange == null)
         return null;
      
      double[] coordinates = {x1, x2};
      Range[] visualRange = {new Range(0, width), new Range(0, height)};
      
      coordinates = convert(displayRange, coordinates, visualRange);
      
      return new Point((int) coordinates[0], (int) coordinates[1]);
   }
   
   /** 
    * Converts a screen point to model coordinates
    * 
    * @param x The x screen coordinate
    * @param y The y screen coordinate
    * @return The location in the model of the specified coordinate, or
    * null if {@link #drawDisplay(SimGraphics)} has not been called
    */
   protected Point2D.Double convertToModel(int x, int y)
   {
      if (displayRange == null)
         return null;
      
      double[] coordinates = {x, y};
      Range[] visualRanges = {new Range(0, width), new Range(0, height)};
      
      coordinates = convert(visualRanges, coordinates, displayRange);
      
      return new Point2D.Double(coordinates[0], coordinates[1]);
   }
   
   /**
    * Converts a set of coordinates from one range to another
    * 
    * @param inputRanges The range of <code>coordinates</code>
    * @param coordinates The coordinates to convert
    * @param outputRanges The range to convert <code>coordinates</code> into
    * @return <code>coordinates</code> converted from <code>inputRanges</code>
    * to <code>outputRanges</code>
    * 
    * @throws IllegalArgumentException if <code>inputRanges</code>, 
    * <code>coordinates</code> and <code>outputRanges</code> do not have
    * the same length
    */
   public static double[] convert(Range[] inputRanges, 
         double[] coordinates, Range[] outputRanges)
   {
      if (inputRanges.length != coordinates.length || 
            outputRanges.length != coordinates.length)
         throw new IllegalArgumentException("All parameter arrays must have" +
                " the same length");
      
      double[] result = new double[coordinates.length];
      for (int i = coordinates.length - 1; i >= 0; i--)
      {
         double inputRange = inputRanges[i].getMax() - inputRanges[i].getMin();
         double outputRange = outputRanges[i].getMax() - 
               outputRanges[i].getMin();
         result[i] = (coordinates[i] - inputRanges[i].getMin()) / inputRange
               * outputRange + outputRanges[i].getMin();
      }
      
      return result;
   }
   
   /**
    * Sets the colour the given particle is drawn in 
    * 
    * @param p The particle to set
    * @param c The colour to draw the particle in
    */
   public void setParticleColor(Particle p, Color c)
   {
      if (c == null)
         particleColors.remove(p);
      else
         particleColors.put(p, c);
   }
}
