package org.seatiger.pso;

/**
 * Utility functions to perform operations on vectors
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class VectorMath
{
   /**
    * Adds several vectors together, returning the result
    * 
    * @param vector1 The first vector
    * @param vectors The vectors to add
    * @return The sum of all the vectors
    * @throws IllegalArgumentException if the vectors do not have the same number
    * of dimensions
    */
   public static double[] add(double[] vector1, double[]... vectors)
   {
      int dimensions = vector1.length;
      
      for (double[] currVector : vectors)
      {
         if (currVector.length != dimensions)
            throw new IllegalArgumentException("Vectors must have the same number " +
                  "of dimensions");
      }
      
      double[] result = new double[vector1.length];
      
      for (int i = result.length - 1; i >= 0; i--)
      {
         result[i] = vector1[i];
         
         for (int j = vectors.length - 1; j >= 0; j--)
            result[i] += vectors[j][i];
      }
      
      return result;
   }

   /**
    * Subtracts several vectors from the first, returning the result
    * 
    * @param vector1 The first vector
    * @param vectors The vectors to subtract from <code>vector1</code>
    * @return The result of the subtraction
    * @throws IllegalArgumentException if the vectors do not have the same number
    * of dimensions
    */
   public static double[] subtract(double[] vector1, double[]... vectors)
   {
      int dimensions = vector1.length;
      
      for (double[] currVector : vectors)
      {
         if (currVector.length != dimensions)
            throw new IllegalArgumentException("Vectors must have the same number " +
                  "of dimensions");
      }
      
      double[] result = new double[vector1.length];
      
      for (int i = result.length - 1; i >= 0; i--)
      {
         result[i] = vector1[i];
         
         for (int j = vectors.length - 1; j >= 0; j--)
            result[i] -= vectors[j][i];
      }
      
      return result;
   }
   
   /**
    * Performs a pointwise vector multiplication
    * 
    * @param vector1 The first vector
    * @param vectors The vectors to multiply with <code>vector1</code>
    * @return The result of the multiplication
    * @throws IllegalArgumentException if the vectors do not have the same number
    * of dimensions
    */
   public static double[] multiply(double[] vector1, double[]... vectors)
   {
      int dimensions = vector1.length;
      
      for (double[] currVector : vectors)
      {
         if (currVector.length != dimensions)
            throw new IllegalArgumentException("Vectors must have the same number " +
                  "of dimensions");
      }
      
      double[] result = new double[vector1.length];
      
      for (int i = result.length - 1; i >= 0; i--)
      {
         result[i] = vector1[i];
         
         for (int j = vectors.length - 1; j >= 0; j--)
            result[i] *= vectors[j][i];
      }
      
      return result;
   }
   
   /**
    * Performs a pointwise vector division
    * 
    * @param vector1 The first vector
    * @param vectors The vectors to divide <code>vector1</code> by
    * @return The result of the division
    * @throws IllegalArgumentException if the vectors do not have the same number
    * of dimensions
    */
   public static double[] divide(double[] vector1, double[]... vectors)
   {
      int dimensions = vector1.length;
      
      for (double[] currVector : vectors)
      {
         if (currVector.length != dimensions)
            throw new IllegalArgumentException("Vectors must have the same number " +
                  "of dimensions");
      }
      
      double[] result = new double[vector1.length];
      
      for (int i = result.length - 1; i >= 0; i--)
      {
         result[i] = vector1[i];
         
         for (int j = vectors.length - 1; j >= 0; j--)
            result[i] /= vectors[j][i];
      }
      
      return result;
   }
   
   /**
    * Multiplies the vector by the coefficient
    * 
    * @param coefficient The scalar number to multiply the vector by
    * @param vector The vector
    * @return The result of multiplying the vector and coefficient
    */
   public static double[] multiply(double coefficient, double[] vector)
   {
      double[] result = new double[vector.length];
      
      for (int i = result.length - 1; i >= 0; i--)
         result[i] = coefficient * vector[i];
      
      return result;
   }
   
   /**
    * Returns the scalar length of a vector
    * 
    * @param vector The vector to get the length of
    * @return The scalar length
    */
   public static double length(double[] vector)
   {
      double sum = 0;
      
      for (int i = vector.length - 1; i >= 0; i--)
         sum += vector[i] * vector[i];
      
      return Math.sqrt(sum);
   }
   
   /**
    * Returns the distance between two vectors
    * 
    * @param vector1 The first vector
    * @param vector2 The second vector
    * @return The scalar distance between the two vectors
    */
   public static double distance(double[] vector1, double[] vector2)
   {
      return length(subtract(vector1, vector2));
   }
   
   /**
    * Returns the unit vector 
    * 
    * @param vector The vector to get the unit vector of
    * @return The unit vector for <code>vector</code>
    * @throws ArithmeticException if the vector's length is 0
    */
   public static double[] unitVector(double[] vector)
   {
      return multiply(1 / length(vector), vector);
   }
   
   /**
    * Returns the dot product of the two vectors
    * 
    * @param vector1 The first vector
    * @param vector2 The second vector
    * @return The dot product of the two vectors
    * @throws IllegalArgumentException if the vectors have a different number
    * of dimensions
    */
   public static double dotProduct(double[] vector1, double[] vector2) 
   {
      if (vector1.length != vector2.length)
         throw new IllegalArgumentException("Vectors must have the same number" +
                "of dimensions");
      
      double sum = 0;
      
      for (int i = vector1.length - 1; i >= 0; i--)
         sum += vector1[i] * vector2[i];      
      
      return sum;
   }
}
