package org.seatiger.pso;

import java.awt.Point;
import java.util.*;

/**
 * Represents the Von Neumann neighbourhood topology where each particle
 * is connected in a grid to the particles left, right, above and below it
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class VonNeumannNeighborhood implements Neighborhood
{
   /** The particles being managed by this neighbourhood */
   private ArrayList<Particle> particles = new ArrayList<Particle>();
   /** A map of each particle to its index in {@link #network} */
   private LinkedHashMap<Particle, Point> particleLocs
         = new LinkedHashMap<Particle, Point>();
   /** The network of particles */
   private Particle[][] network = new Particle[0][0];
   
   /**
    * Returns the particles in the neighbourhood of <code>p</code>
    * 
    * @see org.seatiger.pso.Neighborhood#getNeighborhoodOf(org.seatiger.pso.Particle)
    */
   public Set<Particle> getNeighborhoodOf(Particle p)
   {
      Point particleLoc = particleLocs.get(p);
      int numParticles = particles.size();
      
      if (particleLoc == null)
         return null;

      HashSet<Particle> result = new HashSet<Particle>();
      result.add(p);
      
      int x = particleLoc.x;
      int y = particleLoc.y;
      int height = (int) Math.ceil(numParticles / (double) network.length);
      
      //Left
      if (x == 0 && y == numParticles / network.length)
         result.add(network[numParticles % network.length - 1][y]);
      else if (x == 0)
         result.add(network[network.length - 1][y]);
      else
         result.add(network[x - 1][y]);
      
      //Right
      if (x == network.length - 1 || network[x + 1][y] == null)
         result.add(network[0][y]);
      else
         result.add(network[x + 1][y]);
      
      //Up
      if (y == 0 && network[x][height - 1] == null)
         result.add(network[x][height - 2]);
      else if (y == 0)
         result.add(network[x][height - 1]);
      else
         result.add(network[x][y - 1]);
      
      //Down
      if (y == height - 1 || network[x][y + 1] == null)
         result.add(network[x][0]);
      else
         result.add(network[x][y + 1]);
      
      return result;
   }

   /**
    * @see org.seatiger.pso.Neighborhood#register(org.seatiger.pso.Particle)
    */
   public void register(Particle p)
   {
      if (!particleLocs.containsKey(p))
      {
         int particleNum = particles.size();
         particles.add(p);
         
         if (particleNum < network.length * network.length)
         {
            int x = particleNum % network.length;
            int y = particleNum / network.length;
            
            network[x][y] = p;
            particleLocs.put(p, new Point(x, y));   
         }
         else
            buildNetwork();
      }
   }

   /**
    * @see org.seatiger.pso.Neighborhood#deregister(org.seatiger.pso.Particle)
    */
   public void deregister(Particle p)
   {
      Point particleLoc = particleLocs.get(p);
      
      if (particleLoc == null)
         return;

      particleLocs.remove(p);
      particles.remove(p);
      buildNetwork();
   }
   
   /**
    * @see Neighborhood#deregisterAll()
    */
   public void deregisterAll()
   {
      particles.clear();
      particleLocs.clear();
      network = new Particle[0][0];
   }

   /**
    * Rebuilds the network
    */
   private void buildNetwork()
   {
      int width = (int) Math.ceil(Math.sqrt(particles.size()));
      int height = width;
      
      network = new Particle[width][height];
      Iterator<Particle> iterator = particles.iterator(); 
      
      for (int y = 0; y < height; y++)
      {
         for (int x = 0; x < width && iterator.hasNext(); x++)
         {
            Particle p = iterator.next();
            network[x][y] = p;
            particleLocs.put(p, new Point(x, y));
         }
      }
   }
}
