package org.seatiger.pso.test;

import org.seatiger.pso.*;

import junit.framework.TestCase;

public class RegressionHeuristicTest extends TestCase
{
   private static double EPSILON = 0.00001;

   public RegressionHeuristicTest(String blah)
   {
      super(blah);
   }
   


   public void testEvaluate()
   {
      // y = 8x^3 - 4x^2 - x + 5
      double[] coeff1 = {5, -1, -4, 8};
      
      assertEquals(5, RegressionHeuristic.evaluate(coeff1, 0), EPSILON);
      assertEquals(64 - 16 - 2 + 5, RegressionHeuristic.evaluate(coeff1, 2), EPSILON);
   }

   public void testGetDerivative()
   {
      double[] cubic = {5, -1, -4, 8}; // f(x) = 8x^3 - 4x^2 - x + 5 
      double[] quadratic = {3, 4, 5};  // f(x) = 5x^2 + 4x + 3

      double[] cubicExpected = {-1, -8, 24}; // f'(x) = 24x^2 - 8x - 1
      double[] quadraticExpected = {4, 10};  // f'(x) = 10x + 4
      
      assertEquals(0, VectorMath.distance(cubicExpected, 
            RegressionHeuristic.getDerivative(cubic)), EPSILON);
      assertEquals(0, VectorMath.distance(quadraticExpected, 
            RegressionHeuristic.getDerivative(quadratic)), EPSILON);
   }
   
   public void testFindMaximum()
   {
      double[] quadratic = {3, 4, -5};  // f(x) = -5x^2 + 4x + 3
      double[] cubic = {5, -1, -4, 8}; // f(x) = 8x^3 - 4x^2 - x + 5 
      
      // Calculated by:
      // http://library.thinkquest.org/C004647/calculator/polynomial/polynomial.html
      double quadraticMaximum = 0.40000;
      double cubicMaximum = -0.09686;
      
      assertEquals(quadraticMaximum, RegressionHeuristic.findMaximum(quadratic), EPSILON);
      assertEquals(cubicMaximum, RegressionHeuristic.findMaximum(cubic), EPSILON);
      
      // These polynomials have no local maximum
      double[] quadNoMax = {3, 4, 5};
      double[] cubicNoLocalMax = {0, 0, 0, 1};
      
      assertTrue(Double.isNaN(RegressionHeuristic.findMaximum(quadNoMax)));
      assertTrue(Double.isNaN(RegressionHeuristic.findMaximum(cubicNoLocalMax)));
   }

   public void testFindSurfaceMaximum1D()
   {
      double[] xFunc = {5, 4, -3};
      
      double[][] points = {{1}, {2}, {3}};
      
      double[] values = new double[points.length];
      
      for (int i = 0; i < points.length; i++)
         values[i] = RegressionHeuristic.evaluate(xFunc, points[i][0]);

      double[] expectedPeak = {0.666667};
      
      assertEquals(0, VectorMath.distance(expectedPeak, 
            RegressionHeuristic.findSurfaceMaximum(2, points, values)), EPSILON);
   }
   
   public void testFindSurfaceMaximum3D()
   {
      double[] xFunc = {5, 4, -3};
      double[] yFunc = {12, -5, -1};
      double[] zFunc = {-6, 23, -8};
      
      double[][] points = {{1, 2, 3}, {6, -1, 5}, {3, 6, 1}, {9, 8, 7},
            {-5, -5, -5}, {-2, 3, -3}, {9, 5, 1}, {18, 5, 4}};
      
      double[] values = new double[points.length];
      
      for (int i = 0; i < points.length; i++)
      {
         values[i] = RegressionHeuristic.evaluate(xFunc, points[i][0]) +
               RegressionHeuristic.evaluate(yFunc, points[i][1]) +
               RegressionHeuristic.evaluate(zFunc, points[i][2]);
      }

      double[] expectedPeak = {0.666667, -2.5, 1.4375};
      
      assertEquals(0, VectorMath.distance(expectedPeak, 
            RegressionHeuristic.findSurfaceMaximum(2, points, values)), EPSILON);
   }
   
   public void testOverfit()
   {
      // Maximum @ (0, 0)
      double[] xFunc = {0, 0, -1};
      double[][] points = {{-3}, {-2}, {1}, {5}};
      double[] fitnesses = new double[points.length];
      
      for (int i = 0; i < points.length; i++)
         fitnesses[i] = RegressionHeuristic.evaluate(xFunc, points[i][0]);
      
      double[] expectedPeak = {0};
      
      assertEquals(0, VectorMath.distance(expectedPeak, 
            RegressionHeuristic.findSurfaceMaximum(2, points, fitnesses)), EPSILON);
   }
}
