package org.seatiger.pso.test;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;

import org.seatiger.pso.*;


/**
 * @author Stefan Bird <stbird@seatiger.org>
 *
 */
@SuppressWarnings("all")
public class TreeDisplay extends JPanel implements MouseListener, 
      MouseMotionListener, ActionListener
{
   private Range[] dataRange = new Range[] {new Range(0, 1), new Range(0, 1)};
   private BranchNode rootNode = new BranchNode(dataRange);
   private Point pressPoint = null;
   private Point currPoint;
   private HashSet<LeafNode> touchedLeaves = new HashSet<LeafNode>();
   
   public TreeDisplay()
   {
      this.addMouseListener(this);
      this.addMouseMotionListener(this);
      
      double[] point = {0.5, 0.5};
      rootNode.add(point);
      rootNode.add(point);
      
   }
   
   @Override
   public void paintComponent(Graphics g)
   {
      super.paintComponent(g);
      
      Range[] visualArea = 
         new Range[] {new Range(0, getWidth()), new Range(0, getHeight())};
      
      paint(g, visualArea, rootNode);

      Set<LeafNode> blueNodes = rootNode.getAllLeafNodes();
      Set<LeafNode> yellowNodes = new HashSet<LeafNode>();
      Set<LeafNode> redNodes = new HashSet<LeafNode>();

      if (currPoint != null)
      {
         double[] dCPoint = SwarmDisplay2D.convert(visualArea, 
               new double[] {currPoint.getX(), currPoint.getY()}, dataRange);
         touchedLeaves.clear();
         
         if (pressPoint != null)
         {
            // Highlight all points within the dragged area
            double[] dPPoint = SwarmDisplay2D.convert(visualArea, 
                  new double[] {pressPoint.getX(), pressPoint.getY()}, dataRange);
            double radius = VectorMath.length(VectorMath.subtract(dPPoint, dCPoint));
            
            double[] circTopLeft = SwarmDisplay2D.convert(dataRange, 
                  new double[] { dPPoint[0] - radius, dPPoint[1] - radius }, visualArea);
            double[] circBottomRight = SwarmDisplay2D.convert(dataRange, 
                  new double[] { dPPoint[0] + radius, dPPoint[1] + radius }, visualArea);
            double[] circSize = {circBottomRight[0] - circTopLeft[0], 
                  circBottomRight[1] - circTopLeft[1]};
            
            g.setColor(Color.RED);
            g.drawOval((int) circTopLeft[0], (int) circTopLeft[1], 
                  (int) circSize[0], (int) circSize[1]); 
   
            long start = System.currentTimeMillis();
            redNodes = rootNode.getAllLeafNodesNear(dPPoint, radius);
            System.out.println("Finding: " + (System.currentTimeMillis()-start) + "ms (" + redNodes.size() + " node(s) found)");
            yellowNodes = new LinkedHashSet<LeafNode>(touchedLeaves);
         }
         else
         {
            // Highlight the closest point to the mouse cursor
            NearestPointQuery query = new NearestPointQuery(dCPoint, rootNode);
            yellowNodes = new LinkedHashSet<LeafNode>(touchedLeaves);
            if (query.hasNext())
               redNodes.add(query.next());
            
         }
      }
      
      blueNodes.removeAll(yellowNodes);
      yellowNodes.removeAll(redNodes);
      
      g.setColor(Color.BLUE);
      for (LeafNode currNode : blueNodes)
         paint(g, visualArea, currNode);

      g.setColor(Color.YELLOW);
      for (LeafNode currNode : yellowNodes)
         paint(g, visualArea, currNode);

      g.setColor(Color.RED);
      for (LeafNode currNode : redNodes)
         paint(g, visualArea, currNode);
   }
   
   public void paint(Graphics g, Range[] visualArea, LeafNode node)
   {
      
      double[] point = SwarmDisplay2D.convert(dataRange, node.getPoint(), visualArea);
      g.fillOval((int) point[0] - 3, (int) point[1] - 3, 7, 7); 
   }
   
   public void paint(Graphics g, Range[] visualArea, BranchNode node)
   {
      Range[] dataArea = node.getExtents();
      double[] tlData = {dataArea[0].getMin(), dataArea[1].getMin()};
      double[] brData = {dataArea[0].getMax(), dataArea[1].getMax()};
      
      double[] topLeft = SwarmDisplay2D.convert(dataRange, tlData, visualArea);
      double[] bottomRight = SwarmDisplay2D.convert(dataRange, brData, visualArea);
      double[] middle = {(topLeft[0] + bottomRight[0]) / 2, 
            (topLeft[1] + bottomRight[1]) / 2};
      
      g.setColor(Color.GREEN);
      g.drawRect((int) topLeft[0], (int) topLeft[1], 
            (int) (bottomRight[0] - topLeft[0]), 
            (int) (bottomRight[1] - topLeft[1]));
      g.drawLine((int) middle[0], (int) topLeft[1], 
            (int) middle[0], (int) bottomRight[1]);
      g.drawLine((int) topLeft[0], (int) middle[1],  
            (int) bottomRight[0], (int) middle[1]);
      
      for (int i = 0; i < 4; i++)
      {
         TreeNode currNode = node.getChild(i);
         
         if (currNode instanceof BranchNode)
            paint(g, visualArea, (BranchNode) currNode);
      }
      
      
 
   }
   
   public static void main(String[] args)
   {
      TreeDisplay treeDisplay = new TreeDisplay();
      JFrame frame = new JFrame();
      frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      frame.getContentPane().setLayout(new BorderLayout());
      frame.getContentPane().add(treeDisplay, BorderLayout.CENTER);
      
      JButton button = new JButton("Add points");
      frame.getContentPane().add(button, BorderLayout.SOUTH);
      button.addActionListener(treeDisplay);
      
      frame.setSize(new Dimension(400, 400));
      frame.setVisible(true);
   }

   /* (non-Javadoc)
    * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
    */
   public void mouseClicked(MouseEvent e)
   {
      Range[] visualArea = {new Range(0, getWidth()), new Range(0, getHeight())};
      double[] point = {e.getPoint().x, e.getPoint().y};

      if (e.getButton() == MouseEvent.BUTTON3)
      {         
         NearestPointQuery query = new NearestPointQuery(
               SwarmDisplay2D.convert(visualArea, point, dataRange), rootNode);
         LeafNode nearestNode = query.next();
         
         nearestNode.getParent().remove(nearestNode);
      }
      else
      {
         RecordingLeafNode node = new RecordingLeafNode(
               SwarmDisplay2D.convert(visualArea, point, dataRange));
         rootNode.add(node);
      }
      
      repaint();
   }

   public void mouseEntered(MouseEvent e) { }
   public void mouseExited(MouseEvent e) { }
   
   public void mouseReleased(MouseEvent e) 
   { 
      if (e.getButton() == MouseEvent.BUTTON3 
            && currPoint != null && pressPoint != null)
      {
         Range[] visualArea = 
            new Range[] {new Range(0, getWidth()), new Range(0, getHeight())};
         
         double[] dCPoint = SwarmDisplay2D.convert(visualArea, 
               new double[] {currPoint.getX(), currPoint.getY()}, dataRange);
         
         // Highlight all points within the dragged area
         double[] dPPoint = SwarmDisplay2D.convert(visualArea, 
               new double[] {pressPoint.getX(), pressPoint.getY()}, dataRange);
         double radius = VectorMath.length(VectorMath.subtract(dPPoint, dCPoint));
         
         for (LeafNode currNode : rootNode.getAllLeafNodesNear(dPPoint, radius))
            currNode.getParent().remove(currNode);
      }
      
      pressPoint = null;
      repaint();
   }

   public void mousePressed(MouseEvent e) 
   { 
      pressPoint = e.getPoint();
   }
   
   public void mouseDragged(MouseEvent e)
   {
      currPoint = e.getPoint();
      repaint();
   }

   public void mouseMoved(MouseEvent e) 
   { 
      currPoint = e.getPoint();
      repaint();
   }
   
   private class RecordingLeafNode extends LeafNode
   {
      /**
       * @param point
       */
      public RecordingLeafNode(double[] point)
      {
         super(point);
      }

      @Override
      public double[] getPoint()
      {
         touchedLeaves.add(this);
         return super.getPoint();
      }
      
      
   }
   
   public void actionPerformed(ActionEvent e)
   {
      long start = System.currentTimeMillis();
      for (int i = 0; i < 1000; i++)
      {
         RecordingLeafNode node = new RecordingLeafNode(
               new double[] {Math.random(), Math.random()});
               
         rootNode.add(node);
      }
      System.out.println("Adding 1000 nodes: " + (System.currentTimeMillis()-start) + "ms");
      repaint();
   }
}
