package org.seatiger.pso.traffic;

import java.util.*;

/**
 * Represents a single intersection in the traffic simulation
 * 
 * @author Stefan Bird <stbird@seatiger.org>
 */
public class Intersection
{
   /** The default amount of time in seconds a sequence will be green for */
   public final static double DEFAULT_GREEN_TIME = 30;
   
   /** The road each car enters after exiting a given road */
   private HashMap<Road, Road> exits = new HashMap<Road, Road>();
   
   /** A mapping of sequence number -> List<Road> saying which roads
    * get the green at each point in the sequence */
   private ArrayList<ArrayList<Road>> roadSeq = 
      new ArrayList<ArrayList<Road>>();
   
   /** The amount of time in seconds all roads will see a red light */
   private double allRedTime = 5;
   
   /** The amount of time in seconds the light will be yellow */
   private double yellowTime = 5;
   
   /** The amount of time in seconds the light will be green for each sequence */
   private double[] greenTimes = new double[0];
   
   /** The current sequence */
   private int currSeqNum = 0;
   
   /** The number of seconds until red */
   private double timeTilRed = 0;
   
   /**
    * Adds a road to the intersection
    * 
    * @param entry The road the cars will enter from
    * @param exit The road the cars will exit to
    * @param sequenceNum When the light sequence number equals this,
    * the road will have a green light. 
    */
   public void addRoad(int sequenceNum, Road entry, Road exit)
   {
      exits.put(entry, exit);
      
      while (roadSeq.size() <= sequenceNum)
         roadSeq.add(new ArrayList<Road>());
      
      roadSeq.get(sequenceNum).add(entry);
      
      if (greenTimes.length <= sequenceNum)
      {
         double[] temp = greenTimes;
         greenTimes = new double[sequenceNum];
         System.arraycopy(temp, 0, greenTimes, 0, temp.length);
         
         for (int i = temp.length; i < greenTimes.length; i++)
            greenTimes[i] = DEFAULT_GREEN_TIME;
      }
      
      entry.setExit(this);
   }

   /**
    * Returns the amount of time between each sequence where all the lights
    * will be red
    * 
    * @return The amount of time in seconds where all the lights are red
    */
   public double getAllRedTime()
   {
      return allRedTime;
   }

   /**
    * Sets the amount of time between each sequence where all the lights
    * will be red
    * 
    * @param allRedTime The amount of time in seconds where all the lights are red
    */
   public void setAllRedTime(double allRedTime)
   {
      this.allRedTime = allRedTime;
   }

   /**
    * Returns the amount of time the given sequence will be green
    * 
    * @param sequenceNum The sequence number
    * @return The amount of time the given sequence is green
    */
   public double getGreenTime(int sequenceNum)
   {
      return greenTimes[sequenceNum];
   }

   /**
    * Sets the amount of time the given sequence will be green
    * 
    * @param sequenceNum The sequence number to set
    * @param greenTime The amount of time the given sequence will be green in seconds
    */
   public void setGreenTime(int sequenceNum, double greenTime)
   {
      this.greenTimes[sequenceNum] = greenTime;
   }

   /**
    * Returns the amount of time before the end of each sequence the lights will
    * be yellow. 
    * 
    * @return The amount of time the lights will be yellow in seconds. 
    */
   public double getYellowTime()
   {
      return yellowTime;
   }

   /**
    * Sets hte amount fo time the lights will be yellow before the end of each sequence
    * 
    * @param yellowTime The amount of time the lights will be yellow in seconds
    */
   public void setYellowTime(double yellowTime)
   {
      this.yellowTime = yellowTime;
   }

   /**
    * Returns the road a car will exit on after entering from the given road
    * 
    * @param entry The road the car will enter on
    * @return The road the car will exit on 
    */
   public Road getExit(Road entry)
   {
      return exits.get(entry);
   }
   
   public int getCurrSeqNum()
   {
      return currSeqNum;
   }
   
   public double getTimeTilRed()
   {
      if (timeTilRed < 0)
         return 0;
      else
         return timeTilRed;
   }
   
   public double getTimeTilGreen()
   {
      if (timeTilRed > 0)
         return 0;
      else
         return -timeTilRed;
   }
   
   public void incrementTime(double amount)
   {
      timeTilRed -= amount;
      
      if (timeTilRed <= -allRedTime)
      {
         currSeqNum = (currSeqNum + 1) % greenTimes.length;
         timeTilRed += greenTimes[currSeqNum] + allRedTime; 
      }
   }
   
   /**
    * Returns whether the light is currently green or yellow for the given road
    * 
    * @param road The road
    * @return True if the road currently has a green or yellow light
    */
   public boolean canProceed(Road road)
   {
      return timeTilRed > 0 && roadSeq.get(currSeqNum).contains(road);
   }
}
