package peaks;

import java.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;

/**
 * <p>Title: PSOAgent.java</p>
 * <p>Description: Represent a particel agent for SPSO</p>
 * <p>Copyright: Copyright (c) 27/10/2005</p>
 * <p>Company: RMIT CS</p>
 * @author Xiaodong Li
 * @version 1.0
 */


// If an agent, or any other object is to be displayed in a display it
// needs to implement at least the drawable interface which has one method:
// draw(SimGraphics g)
public class PSOAgent implements Drawable {

  // ---------------------------------------------------------------- Members

  /** Two coordindates (x, y) represent this agent on the 2-dimensional lattice **/
  protected int x, y;

  /** An array contains the value for each dimension of the current position of the
   * agent
   */
  protected double[] present;

  /** An array represents the best location so far of the agant **/
  protected double[] pbest;

  /** An array represent the current velocity vector of the agent **/
  protected double[] v;

  /** The current fitness value **/
  protected double fitnessValue;

  /** The best fitness value in history of this agent **/
  protected double pbestValue;

  /** Flag to indicate whether the agent is the species seed **/
  protected boolean SEED = false;
  
  /** Flag to indicate whether the agent is processed when determing specie seeds **/
  protected boolean PROCESSED = false;  
  
  /** The species seed array for the current agent **/
  protected double[] seed;

  /** The model associated with this agent **/
  protected PSOModel model;

  /** The Euclidean distance to a comparing particle **/
  protected double euDist, seedFit;
  
  /** Flag to indicate whether the agent is a Quantum particle or not **/
  protected boolean QUANTUMP = false;  
  
 
  public PSOAgent() {

  }

  public PSOAgent(PSOModel model) {
    // Copy the space and model
    this.model = model;

    // Initalize the arrays
    present = new double[model.getDimension()];
    pbest = new double[model.getDimension()];
    v = new double[model.getDimension()];    
    seed = new double[model.getDimension()];
    
    // Randomly initialize the velocity and pbest
    initialization();
    
  }


  /**
   * Set the xy-coordinates of the agent
   * @param x
   * @param y
   */
  public void setXY(int x, int y) {
   this.x = x;
   this.y = y;
 
  }

  /**
   * Return x-coordinate
   * @return
   */
  public int getX() {
    return x;
  }

  /**
   * Return y-coordinate
   * @return
   */
  public int getY() {
    return y;
  }
  
   /* set the flag PROESSED to indicate it is done or not */
   public void setProcessed(boolean prd) {
       this.PROCESSED = prd;
   }
 
   public boolean getProcessed() {       
       return PROCESSED;
   }

  
   /* set the flag QUANTUMP to indicate it is a Quantum particle */
   public void setQuantump(boolean qp) {
       this.QUANTUMP = qp;
   }
 
   public boolean getQuantump() {       
       return QUANTUMP;
   }


  /**
   * step() is called by the scheduler once per tick.
   */
  public void step() {
 
    // first get the fitness calculated
    fitnessValue = getFitness();
    
    // If this is the first iteration, then pbestValue is assigned to the current value
    if (model.getTickCount() == 1) {
        this.pbestValue = fitnessValue;
    }
    
    // update the pbest
    if (fitnessValue > pbestValue) {
       pbestValue = fitnessValue;
       for (int i=0; i < present.length; i++) 
          pbest[i] = present[i];
    }
 }
  
  protected double getFitness() {

   // Calculate fitness value
    if (model.getFunction().equals("movPeaks")) {
       fitnessValue = movPeaksFunction() ;
    }

    return fitnessValue;
 }
 
 
 /**
   * Update the current position of each agent in the popoluation
   * using the neutral update rules (ie., stardard constriction PSO)
   */
protected void neutralUpdate() {

   // do the normal constriction PSO stuff with species seeds as lbests
   for (int k=0; k < model.getDimension(); k++) 
   {
         v[k] = 0.729843788*(v[k] + 
                2.05 * model.getRandomUniform()      
                *( pbest[k] - present[k]) 
                + 2.05 * model.getRandomUniform()  
                * (seed[k] - present[k]));   // adjusting according to a particle's seed     
              
         
         
         // Adjust the velocity vector so that the agent cannot go
         // beyond the search space, preventing the population from explosion
         if (v[k] > model.getVelocityMax()) {
              v[k] = model.getVelocityMax();
         }
         else if (v[k] < -model.getVelocityMax()) {
              v[k] = - model.getVelocityMax();
         }

         // Based on the new velocity, update the agent's position
         present[k] = present[k] + v[k];
         
         // Obtain right range and left range from PSOParameters object
         double rightRange = model.getRightRange();
         double leftRange = model.getLeftRange();
            
         // Adjust the position vector so that the agent's new location varies
         // It is hoped the method below increases the overall population diversity
         if (present[k] > rightRange) {
            //present[k] = rightRange;
            present[k] = (present[k] + rightRange)/2.0;
         }
         else if (present[k] < leftRange) {
            //present[k] = leftRange;
            present[k] = (present[k] + leftRange)/2.0;
         }
            
   }//for loop1
   
//   System.out.println("v: " + Arrays.toString(v));
//   System.out.println("x: " + Arrays.toString(present));
//   System.out.println("seed: " + Arrays.toString(seed));
//   System.out.println("pBest: " + Arrays.toString(pbest));
   
}


 /**
   * Update the current position of each agent in the popoluation
   * using quantum update rules (Blackwell and Branke, 2005)
   */
protected void quantumUpdate() {
	System.out.println("BANG");
    double[] posInCloud = new double[model.getDimension()];
    posInCloud = getPositionInCloud();    
    // do the normal PSO stuff with species seeds as lbests
    for (int k=0; k < model.getDimension(); k++) 
    {
        present[k] = seed[k]+ posInCloud[k]; 
    }
}
 

/* get the cloud for quantum particles */
protected double[] getPositionInCloud() 
{
	System.out.println("BANG");
    double[] cloudX = new double[model.getDimension()];
    double sumPresent=0.0, length=0.0;
    /*
     * Knuth Vol II section 3.4.1E(5) algorithm for generating points on a
     * unit hypersphere is to generate d gaussian random variables from N(0,
     * 1) and then normalise, x = d / |d|
    */

    for(int j = 0; j < model.getDimension(); j++) 
        cloudX[j] = model.getGaussian();

    double[] origin = new double[model.getDimension()];
    
    for(int k = 0; k < model.getDimension(); k++) 
        sumPresent += Math.pow(cloudX[k] - 0.0, 2);
    length = Math.sqrt(sumPresent);
    
    if (length < 1.0e-50)
       length = 1.0e-50;

    // We need a point on the surface of a hypersphere of radius r ~
    // [SMALLVALUE, perceptionLimit]
    //double scale = model.getGaussian() * model.getRadius() / length;
    double scale = model.getGaussian() * 1.0 / length; // set r_cloud to 1.0

    for (int j = 0; j < model.getDimension(); j++) {
        cloudX[j] *= scale;
    }

    return cloudX;
}



/**
  * movPeaks function where each solution gets evaluated
  * also the number of evaluations is counted here (from eval_movpeaks(.))
  **/

  private double movPeaksFunction() {
    int i;
    double dummy;  
    //double[] present = { 51.0, 30.0, 94.0, 70.0, 49.0 };
    
    dummy = model.FktnLib.eval_movpeaks(present);
    //dummy = model.FktnLib.eval_movpeaks(pbest);
    //System.out.println(dummy + "\n");

    return dummy;
  }


  private double square(double x) {
    return x * x;
  }


  /**
   * Drawable implementationdraw() is called whenever the display is updated,
   * assuming the agent is part of what is being displayed.
   */
  public void draw(SimGraphics g) {
    g.drawFastRoundRect(java.awt.Color.red);
  }


  /**
   * Initialize the current position
   */
  private void initialization() {
  
     // Obtain right range and left range from PSOParameters object
     double rightRange = model.getRightRange();
     double leftRange = model.getLeftRange();

     for (int i=0; i < model.getDimension(); i++) 
     {
         present[i] = (rightRange - leftRange) * model.getRandomUniform() + leftRange;
         pbest[i] = present[i];  
         seed[i] = present[i];
          
         // 50% of v[i] are in an opposite direction
         v[i] = (rightRange - leftRange) * model.getRandomUniform() + leftRange;
         if (model.getRandomUniform()  <= 0.5)
            v[i] = -v[i];
     }
     
     pbestValue = -1.7976931348623157E308;
     seedFit = -1.7976931348623157E308;
     euDist = 1.7976931348623157E308;
     
  }
}
