package peaks;

/**
 * <p>Title: PSOModel.java </p>
 * <p>Description: A species-based PSO for tracking optima in a dynamic environment</p>
 * <p>Copyright: Copyright (c) 27/10/2005</p>
 * <p>Company: RMIT CS</p>
 * @author Xiaodong Li (xiaodong@cs.rmit.edu.au)
 * @version 1.0
 */
 
import java.util.Vector;
import java.util.ArrayList;
//import java.util.Random;
import java.util.Iterator;
import java.util.ListIterator;
import java.awt.Color;
import java.util.Collections;
import java.io.*;
import java.math.*;

import cern.jet.random.*;
import uchicago.src.sim.util.Random;
import uchicago.src.sim.util.*;
import uchicago.src.sim.space.*;
import uchicago.src.sim.engine.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.analysis.*;
import uchicago.src.reflector.*;

import org.seatiger.pso.*;


// Simulation models should extend SimModelImpl which does some basic setup
// and allows a controller easy access to a simulations initial parameters.
public class PSOModel extends SimModelImpl {

  // -------------------------------------------------------------- Members

  /** A schedule associated with this model **/
  protected Schedule schedule;

  /** A list of all existing agents **/
  protected ArrayList agentList;
  
  /** A list of species seeds from the current population **/
  protected ArrayList seedList;

  /** A list of all existing agents' personal bests **/
  protected ArrayList agentPerList;
  
  /** A list of species seeds based on personal bests **/
  protected ArrayList seedPerList;

  /** A DataRecorder allows data generated by a simulation to be written to a file **/
  protected DataRecorder recorder;

  /** A Plot to display the fitness for each Iteration **/
  protected Plot aPlot;

  /** The array stores the loccation of the global best **/
  private double[] gBest;

  /** The value of the global best **/
  private double gBestValue;

  protected OpenSequenceGraph graph; 
  
  /** Number of run **/
  protected int run = 0;

  /** create an instance of movpeaks **/  
  protected MovingPeaks FktnLib;
  
  /** population diversity **/
  protected double diversity;
  
  /* parameters accessor methods */
  private int numParticles = 100; 
  private double leftRange = 0;
  private double rightRange = 100;
  private double velocityMax = 100;
  private int maxIteration = 5000;
  private int maxNumEvals = 500000;
  private int dimension = 5;  // movPeaks Scenario2
  private double radius = 30; 
  private double aThreshold = 0.1; //particle diversity threshold (distance from the seed to the furthest particle)
  private double dThreshold = 0.5; // diversity threshold
  private double decayScale = 1.0; // scale for decaying species radius
  private String function = "movPeaks";
  // max number of particles allowed in a species
  private int speciesMax = 10; 
  // model variant chosen to redistribute particles at each iteration
  private String method = "BaselineSPSO"; 
  private boolean batch = false;  // Batch or graph mode
  private boolean originalSPSO = true; //original SPSO sorting on x_i fitness instead of p_i.
  private int numPeaks = 10; // used for tests in varying the number of peaks
  
  private XDMovingPeaksMetric mpmetric;
  
  // ---------------------------------------------------------- Constructor(s)

  public PSOModel() {
    
    Vector v = new Vector();
    v.add("BaselineSPSO");
    v.add("ReplaceWSpecies");
    v.add("NeutralQuantum1");
    v.add("NeutralQuantum2");
    v.add("NeutralQuantum3"); 
    v.add("QuantumSPSO");
    ListPropertyDescriptor pd = new ListPropertyDescriptor("Method", v);
    descriptors.put("Method", pd);
        
  }


  // ------------------------------------------------------------- Methods

  /** Every model must have begin() and setup() methods (required for implementing
    * the SimModel inteface).
    * begin() should intialize the model for the start of a run. Consequently,
    * the build* methods are called here, and any displays are displayed. build()
    * is called whenever the start button (or the step button if the run has
    * not yet started) is clicked.
    */
  public void begin() {
  
    // initialize Random.uniform
    Random.createUniform();
    Random.createNormal(0, 1);   

    // Create an array list contains all agents;
    agentList = new ArrayList();         
    // Create an array list containing all species seeds
    seedList = new ArrayList();  

    // Create an array list contains all agents' personal bests;
    agentPerList = new ArrayList();
    // Create an array list containing all personal bests as species seeds
    seedPerList = new ArrayList();  

    // Initialization the gBest[] and gBest value
    gBest = new double[dimension];
    gBestValue = -1.7976931348623157E308;
     
    /** create an instance of movpeaks **/
    // FktnLib = new movpeaks(); 
    FktnLib = new MovingPeaks(numPeaks, dimension);
    mpmetric = new XDMovingPeaksMetric(this, FktnLib);
    
    /* initialize peaks at beginning */
//    FktnLib.init_peaks();  
  
    // build the model
    buildModel();

    // build the display
    buildDisplay();

    // build the schedule
    buildSchedule();

    if (!batch)
    {
    // Display the plot
    	aPlot.display();
    
    	// Display the graph
    	graph.display();
    }
  }

  /**
   * setup() prepares the model for another run. Called whenver the setup button
   * is clicked. Setup should set any objects that are created over the course
   * of the run to null, and dispose of any DisplaySurfaces or graphs. While
   * not strictly necessary this should some prevent memory leaks and calling
   * System.gc() helps too. The initial parameters should be set to whatever
   * defaults the user wants to see initially.
   */
  public void setup() {
    // Initialize all the variables    
    schedule = null;
    agentList = new ArrayList();
   
    if (aPlot != null)
      aPlot.dispose();
    aPlot = null;
    
    if (graph != null) {
      graph.dispose();
      graph = null;
    } 

    // Create the plot
    aPlot = new Plot("Decision space plot");
    aPlot.addLegend(1, "Seed", Color.RED, aPlot.CROSS);
    aPlot.addLegend(2, "Peak", Color.BLUE, aPlot.SQUARE);
    aPlot.addLegend(3, "Maximum Peak", Color.GREEN, aPlot.CROSS);
    aPlot.addLegend(4, "Quantum", Color.BLACK, aPlot.CIRCLE);
    aPlot.addLegend(5, "pBest", Color.MAGENTA, aPlot.DIAMOND);
    // new sequence graph
    graph = new OpenSequenceGraph("Error Graph", this);
    this.registerMediaProducer("Error Graph", graph);
     
    // create a schedule with an interval of one.
    schedule = new Schedule(); 
    
    System.gc();
    
   
  }

  /**
   * A required method, returns the schedule associated with this model
   */
  public Schedule getSchedule() {
    return schedule;
  }

  /**
   * a required method - displayed on the Controller toolbar.
   */
  public String getName() {
    return "PSO";
  }


  /**
   * The typical way to code a simulation is to divide up the creation of the
   * simulation into three methods - buildModel(), buildDisplay(),
   * buildSchedule(). This division is not strictly necessary, but it does
   * make the creation conceptually clearer.
   *
   * The buildModel() method is responsible for creating those parts of the
   * simulation that represent what is being modeled. Consequently, the agents
   * and their environment are typically created here together with any
   * optional data collection objects. Of course, this method may call other
   * methods to help build the model.
   */
  public void buildModel() {
    System.out.println("Build Model");           
    
    gBestValue = -1.7976931348623157E308;
    
    // We start off the simulation by adding new agent into the model
    for (int i = 0; i < numParticles; i++) 
    {   
        PSOAgent agent = new PSOAgent(this);
        // Add this agent into the agent list
        agentList.add(agent);
       
        // half the species are marked as quantum, the remaining neutral
        //if ((Random.uniform.nextDoubleFromTo(0, 1) <= 0.5)  && method.equals("NeutralQuantum1")) 
        //    agent.setQuantump(true);
    }
    
    // create a new DataRecorder and write the data to ./sugar_data.txt
    recorder = new DataRecorder("./data.txt", this); 

    // Adds a data source that counts the agents all the agents alive
    // in the simulation. It does this by getting the size of the agentList.
    recorder.addNumericDataSource("offline_error", new offlineErrorSource());
    recorder.addNumericDataSource("numSpecies", new NumSpeciesSource());
    
    aPlot.clear(0);
    // Display a plot to for all the particles at the 1st iteration
    aPlot.setConnected(false);
    for (int i = 0; i < agentList.size(); i++) {   //step through each particle
            PSOAgent p = (PSOAgent)agentList.get(i);
            aPlot.plotPoint(p.present[0], p.present[1], 0);
    }
    aPlot.fillPlot();
    aPlot.updateGraph();   
        
    System.out.println("Method used : "+method);
    System.out.println("numPeaks : "+numPeaks);
    System.out.println("aThreshold : "+aThreshold);
    //System.out.println("dThreshold : "+dThreshold);    
  }


  // data source for offline_error
  class offlineErrorSource implements NumericDataSource {
    public double execute() {
      return FktnLib.get_offline_error();      
    }
  }
  
  // data source for numSpecies
  class NumSpeciesSource implements NumericDataSource {
    public double execute() {
      return seedList.size();      
    }
  }
 
  /**
   * buildDisplay() builds those parts of the simulation that have to do with
   * displaying the simulation to a user.
   */
  protected void buildDisplay() {

    // Create a display to display the agentGrid on the screen.
    // This will display any Drawable contained within the Object2DGrid
    // In a non-batch simulation the agents in a simulation will typically
    // implement the Drawable interface.

    graph.setXRange(0, 100);
    graph.setYRange(0, 50);
    graph.setAxisTitles("Iteration", "Current Error");
    
    // have the graph create the sequence for us from the getNumTriggered method.
    
    graph.addSequence("Offline Error", new Sequence() {
        public double getSValue() {
         return FktnLib.get_offline_error();
    }
    });
    graph.addSequence("Current Error", new Sequence() {
        public double getSValue() {
         return FktnLib.get_current_error();
    }
    });
    graph.addSequence("Peak cover", new Sequence() {
        public double getSValue() {
         return mpmetric.getPCP();
    }});
    graph.addSequence("BKPE", new Sequence() {
        public double getSValue() {
         return mpmetric.getBKPE();
    }});
/*    
    graph.addSequence("Current Error", new Sequence() {
        public double getSValue() {
         return FktnLib.get_current_error();
    }
    });


    graph.addSequence("Average Error", new Sequence() {
        public double getSValue() {
         return FktnLib.get_avg_error();
    }
    });
    
    
    graph.addSequence("Diversity", new Sequence() {
        public double getSValue() {
         return diversity;
    }
    });
*/
  }



  /**
   * buildSchedule builds the schedule that changes the simulations state.
   * Under this scheme, a simulation is a state machine where all transitions
   * between states are the result of actions initiated by a schedule.
   */
  public void buildSchedule() {
     // For each iteration (tick) of the simulation, call the step() method
     // of this PSOModel object to update the update the position,
     // pbest, gBest of the particle swarm
     
     schedule.scheduleActionBeginning(0, this, "step");
     if (batch == true)
     {
        schedule.scheduleActionAtEnd(recorder, "record");
        schedule.scheduleActionAtEnd(recorder, "writeToFile");
     }
     schedule.scheduleActionAtEnd(this, "writeStat");
  }

 
  public void writeStat() {
    System.out.println("offline_error : "+FktnLib.get_offline_error());
  }
 
  /**
   * This step() method implements the PSO algorithm as stated below
   *
   * (a):  v[] = ki * (v[] + c1 * rand() * (pbest[] - present[])
   *               + c2 * rand() * (seedBest[] - present[]))
   * (b):   present[] = persent[] + v[]
   *
   *    Do
   *      For each particle
   *      Calculate fitness value
   *        If the fitness value is better than the best fitness value (pBest) in history
   *        set current value as the new pBest
   *   End
   *
   *   Choose the particle with the best fitness value of all the particles as the gBest
   *
   *   For each particle
   *        Calculate particle velocity according equation (a)
   *        Update particle position according equation (b)
   *   End
   *  While maximum iterations or minimum error criteria is not attained
   *
   */
  public void step() {

    // If hit the maximum genereration then stop
    // if (this.getTickCount() > maxIteration) {
    if (FktnLib.get_number_of_evals() > maxNumEvals) 
        this.stop();
 
    // print the offline error
    if (batch == false)
    {  
        //System.out.println("iteration : "+this.getTickCount());
        //System.out.println("Method used : "+method);
        //System.out.println("offline error : "+FktnLib.get_offline_error());
        //System.out.println("radius : "+radius);
    }
    
    // radius decreasing from 51 to a minimum of 1.0
    //radius = 1 + (double)50*(500000-FktnLib.get_number_of_evals())/500000 ;

    // decaying equation (10% per iteration) 
    //radius = 50 * Math.exp(-0.000007824*decayScale*FktnLib.get_number_of_evals());
    //if (radius < 1.0)
    //    radius = 1.0;
    
    int numQuantum = 0;
  
    // Iterate through the list of agents and do the following
    for (int i = 0; i < agentList.size(); i++) 
    {
        PSOAgent agent = (PSOAgent) agentList.get(i);
        
        if (agent.getQuantump() == true)
            numQuantum++;
        // Calculate the agent fitness value as well as update the
        // pbest in history of the agent by calling the agent's step() method
        agent.step();      
                
        //System.out.println("number_of_evals : "+FktnLib.get_number_of_evals());
     
        // Choose the particle with the best fitness value of
        // all the particles as the gBest
        if (agent.fitnessValue > gBestValue)
        {
            // Update the gBest value
            gBestValue = agent.fitnessValue;
            
            // Update the gBest location
            for (int j=0; j < dimension; j++) 
                gBest[j] = agent.present[j];
        } 
    }
    
    // sort the agentList in fitness decreasing order
    if (originalSPSO == true)
    {
        agentList = sortAListFit(agentList);
        //System.out.println("originalSPSO is used..");
    }
    else
        agentList = sortAListPFit(agentList);
        
    //System.out.println("agentList size : "+agentList.size());
    ArrayList agentListCopy = new ArrayList();
    for(int i=0; i<numParticles; i++)
        agentListCopy.add(agentList.get(i));
    
    agentList.clear();
    agentList.addAll(agentListCopy);
    // first clear the species seed list, then get the species seeds
    seedList.clear();
    seedList = getSeeds(agentList, seedList); 
    
    // check to see if a change occurs
    boolean changeOccurred = false;
    
    // below used for tests in varying the number of peaks
    int numDetected = 0;
    if (seedList.size() < 5)
        numDetected = seedList.size();
    else
        numDetected = 5;    
    
    for (int i=0; i < numDetected; i++)
    {
        PSOAgent s = (PSOAgent)seedList.get(i);
        //double pbestValueNew = FktnLib.dummy_eval(s.pbest); 
        double pbestValueNew = FktnLib.eval_movpeaks(s.pbest);
        if (pbestValueNew < s.pbestValue)
        {
            //for (int k=0; k < s.present.length; k++) 
            //    s.pbest[k] = s.present[k];       
            changeOccurred = true;
            break;
        }    
    }
    
    if (changeOccurred == true)    
    {
        for (int ii = 0; ii < agentList.size(); ii++) 
        {                
            PSOAgent p = (PSOAgent) agentList.get(ii);
            //p.fitnessValue = FktnLib.dummy_eval(p.present);                
            p.pbestValue = p.fitnessValue; 
            for (int k=0; k < p.present.length; k++) 
                p.pbest[k] = p.present[k];       
                
            //System.out.println("agent refreshed ");
        }

        seedList.clear();
        seedList = getSeeds(agentList, seedList); 

    }
    
    // make sure each species not exceeding its cap
    agentList = setSpeciesCap(agentList, seedList);
        
    //System.out.println("numQuantum : "+numQuantum); 
    // first get the population diversity
    diversity = getDiversity(agentList);
    //System.out.println("radius : "+radius+"\n");

    // replace the worst species with randomly generated particles 
    if (method.equals("ReplaceWSpecies") && (diversity <= dThreshold))
    {
    	System.out.println("BANG");
    
           agentList = replaceWSpecies(agentList);
    }
           
 
    // when a particle is within the acceptable close distance to its species seed 
    // the particle is repelled in the opposite direction as it would be done normally 
    //if (method.equals("Repell") && diversity < 1.0) 
    //    agentList = repell(agentList);   
    
    // starts the initial pop with half neutral and half quantum 
    // neutral do not interact directly with quantum. Both subject to clearing procedure
    if (method.equals("NeutralQuantum1"))
    {
    	System.out.println("BANG");
        agentList = neutralQuantum1(agentList, seedList);
        //agentList = localDiverse(agentList, seedList);  
        if (diversity <= dThreshold)
           agentList = replaceWSpecies(agentList); 

        for (int i = 0; i < agentList.size(); i++)
        {
            PSOAgent agent = (PSOAgent) agentList.get(i);
            if (agent.getQuantump() == false)
                agent.neutralUpdate();       
            else if (agent.getQuantump() == true)
                agent.quantumUpdate();   
        }             
        
    }
    
    
    // generate two quantum particles for each species seed, then use neutralUpdate()     
    if (method.equals("NeutralQuantum2")  )
    {
    	System.out.println("BANG");
        agentList = twoQuantum(agentList, seedList);        
        //agentList = localDiverse(agentList, seedList);
        if (diversity <= dThreshold)
           agentList = replaceWSpecies(agentList); 
        
        for (int i = 0; i < agentList.size(); i++)
        {
            PSOAgent agent = (PSOAgent) agentList.get(i);
            //if (FktnLib.get_number_of_evals()%5000 == 0)
            
            if (agent.getQuantump() == false)
                agent.neutralUpdate();       
            else if (agent.getQuantump() == true)
                agent.quantumUpdate();
        }                               
    }
    
    
    // generate a quantum cloud if a species' diversity is below "aThreshold" (locally diversifying)
    if (method.equals("NeutralQuantum3")  )
    {
    	System.out.println("BANG");
        agentList = localDiverse(agentList, seedList);
        //if (diversity <= dThreshold)
        //   agentList = replaceWSpecies(agentList); 
        
        for (int i = 0; i < agentList.size(); i++)
        {
            PSOAgent agent = (PSOAgent) agentList.get(i);
            if (agent.getQuantump() == false)
                agent.neutralUpdate();       
            else if (agent.getQuantump() == true)
                agent.quantumUpdate();
        }                               
    }


    // generate two quantum particles for each species seed, then use quantumUpdate()
    // the particles act as local hill-climbers  
    if (method.equals("QuantumSPSO")  )
    {
    	System.out.println("BANG");
        agentList = twoQuantum(agentList, seedList);
        //agentList = localDiverse(agentList, seedList);
        if (diversity <= dThreshold)
           agentList = replaceWSpecies(agentList); 
                
        for (int i = 0; i < agentList.size(); i++)
        {
            PSOAgent agent = (PSOAgent) agentList.get(i);
            agent.quantumUpdate();
        }   
        
    }


    // check to see if expected acuracy is reached, and if so, stop the run
    //checkAcuracy();
     
    // Update the particles (for "BaselineSPSO" and "ReplaceWSpecies")
    if (method.equals("BaselineSPSO") || method.equals("ReplaceWSpecies"))
    {
        for (int i = 0; i < agentList.size(); i++)
        {
            PSOAgent agent = (PSOAgent) agentList.get(i);
            agent.neutralUpdate();          
        }
    }
    
    mpmetric.updateMetrics();
    
    if (batch == false)
    {    
        if (this.getTickCount()%10 == 0)
        //    System.out.println(this.getTickCount()+"\t"+numQuantum);
        //System.out.println("numQuantum : "+numQuantum);
        //System.out.println("diversity : "+diversity+"\n"); 
        // draw the fitness plot
        graph.step();
    
        // Clear the plot first
        aPlot.clear(0);
        aPlot.clear(1);
        aPlot.clear(2);   
        aPlot.clear(3);
        aPlot.clear(4);
        aPlot.clear(5);
         
        // Display a plot for all the particles at each Iteration
        aPlot.setConnected(false);
        for (int i = 0; i < agentList.size(); i++) {   //step through each particle
            PSOAgent p = (PSOAgent)agentList.get(i);
            if (p.getQuantump() != true)
                aPlot.plotPoint(p.present[0], p.present[1], 0);
            else
                aPlot.plotPoint(p.present[0], p.present[1], 4);
            
            aPlot.plotPoint(p.pbest[0], p.pbest[1], 5);
        }
 
        // Disply a plot for all the seeds
        for (int i=0; i < seedList.size(); i++)
        {
            PSOAgent p = (PSOAgent)seedList.get(i);
            //aPlot.plotPoint(p.present[0], p.present[1], 1);
            aPlot.plotPoint(p.seed[0], p.seed[1], 1);
        }
    
        aPlot.setConnected(true);
        // Display all the peaks on the plot     
        int maxPosition = FktnLib.getMaximumPeak();
        double[][] positions = FktnLib.getPeakPositions();
        for(int i = 0; i < FktnLib.number_of_peaks; i++) 
        {
            for(int j=0; j < 2; j++)
            {
                aPlot.plotPoint(positions[i][j], positions[i][j+1], 2);   
                j++;
            }             
        }     

        aPlot.setConnected(false); 
        for(int i =0; i < FktnLib.number_of_peaks; i++)
        {
            if (i == maxPosition)
            for(int j=0; j < 2; j++)                                                   
            {                                                             
                aPlot.plotPoint(positions[i][j], positions[i][j+1], 3);   
                j++;                                                      
            }                                                                                               
        }         

        aPlot.fillPlot();
        aPlot.updateGraph();
    }
     
    //System.out.println("Maximum number of peaks : "+FktnLib.getMaximumPeak());
       
  }



 /**
   * Sort the population based on fitness in decreasing order of fitness
   * from the best-fit to the lest-fit individual
   */
   private ArrayList sortAListFit(ArrayList aList)
   {
        ArrayList sortedList = new ArrayList();
        sortedList.addAll(aList);         
                
        double a=0, b=0;
        //sort particles in decreasing order according to fitnessValues
        try 
        {               
            int i,j;
            i = 0;
            while (i < (sortedList.size() - 1))       
            {
                j = i + 1;                
                while (j < sortedList.size())
                {
                    a = ((PSOAgent)sortedList.get(i)).fitnessValue;
                    b = ((PSOAgent)sortedList.get(j)).fitnessValue;
                                    
                    if (a < b)                     
                    {
                        PSOAgent temp = (PSOAgent)sortedList.get(i);
                        sortedList.set(i, sortedList.get(j));
                        sortedList.set(j, temp);
                    }
                    j++;
                }
                i++;
            }
        } 
        catch (Exception e) 
        {
            e.printStackTrace (System.err);
        }   
        
        return sortedList;
   }


 /**
   * Sort the population based on Personal best of each particle in decreasing order of values
   * from the highest to the lowest one
   */
   private ArrayList sortAListPFit(ArrayList aListPBest)
   {
        ArrayList sortedList = new ArrayList();
        sortedList.addAll(aListPBest);         
                
        double a=0, b=0;
        //sort particles in decreasing order according to pbestValue
        try 
        {               
            int i,j;
            i = 0;
            while (i < (sortedList.size() - 1))       
            {
                j = i + 1;                
                while (j < sortedList.size())
                {
                    a = ((PSOAgent)sortedList.get(i)).pbestValue;
                    b = ((PSOAgent)sortedList.get(j)).pbestValue;
                    
                    // according to pbestValue in decreasing order
                    if (a < b)                     
                    {
                        PSOAgent temp = (PSOAgent)sortedList.get(i);
                        sortedList.set(i, sortedList.get(j));
                        sortedList.set(j, temp);
                    }
                    j++;
                }
                i++;
            }
        } 
        catch (Exception e) 
        {
            e.printStackTrace (System.err);
        }   
        
        return sortedList;
   }


   

 /**
   * Clearing procedure to determine the species seeds in the population
   * Please refer to the species-based PSO (SPSO) (Li, 2004). 
   */
  protected ArrayList getSeeds(ArrayList aList, ArrayList seedList) {
  
    boolean FOUND;
    double euDistPresent = 0.0;
    int seedListSize;
  
    ArrayList sortedList = new ArrayList();
    sortedList.addAll(aList);   
           
    //System.out.println("sortedList size "+sortedList.size());
    Iterator sortedIter = sortedList.iterator();    
   
    while (sortedIter.hasNext())
    {
         PSOAgent p = (PSOAgent) sortedIter.next();
         //p.PROCESSED = true;        
         FOUND = false;
                
         seedListSize = seedList.size();
         //System.out.println("current seedlist size: "+seedListSize);
         for (int i=0; i < seedListSize; i++)
         {
            double sumPresent = 0;
            PSOAgent seed = (PSOAgent)seedList.get(i);
            for (int j=0; j < dimension; j++)
                if (originalSPSO == true)
                    sumPresent += Math.pow(p.present[j] - seed.present[j], 2);
                else
                    sumPresent += Math.pow(p.present[j] - seed.pbest[j], 2);
            //System.out.println("seed "+i+":"+seed.present[0]);
            euDistPresent = Math.sqrt(sumPresent);
            //System.out.println("euDistPresent: "+euDistPresent);
            
            // if Euclidean distance is less than the radius
            if (euDistPresent < radius)  
            {
               for (int k=0; k < dimension; k++) 
               {
            	   /*** SCB 16/1/7 ***/
            	   if (originalSPSO)
            		   p.seed[k] = seed.present[k];
            	   else	   
            		   p.seed[k] = seed.pbest[k];
            	   /*** End SCB ***/
                  //System.out.println("seed "+i+":"+seed.present[k]);
               }
               p.seedFit = seed.pbestValue;

               FOUND = true;
               //System.out.println("found a similar one!"+"\n");
               break;   // break the for loop
            }
         }
         
         if ((FOUND != true) && (p.fitnessValue != -100))
         {
             seedList.add(p);
             /*** SCB 16/1/07 ***/
             /*for (int x = 0; x < p.seed.length; x++)
             {
            	 if (originalSPSO)
            		 p.seed[x] = p.present[x];
            	 else	 
            		 p.seed[x] = p.pbest[x];
            	 
             }*/
			 /*** End SCB ***/
             
         }
    }
    //System.out.println("current seedlist size: "+seedList.size());
    return seedList;
 }   
 
 
 
   
 
/**
   * Set a maximum for the numbers of particles allowed in each species. 
   * Excess particles will be replaced by random particles injected into the search space.
   */
 protected ArrayList setSpeciesCap(ArrayList aList, ArrayList seedList)  {
  
    double sumPresent=0.0, euDist=0.0, sumEuDist;
    int numQParticles = 0, numNParticles = 0;
    int numParticles = 0;
    ArrayList sortedList = new ArrayList(); 
    sortedList.addAll(aList);     
          
    //set all PROCESSED flags to false for all particles first
    for(int i=0; i<sortedList.size(); i++)
    {
        PSOAgent ps = (PSOAgent)sortedList.get(i);
        ps.setProcessed(false); 
    }

    Iterator seedIter = seedList.iterator();   // only seeds                 
    while (seedIter.hasNext())
    {
        PSOAgent s = (PSOAgent) seedIter.next();
        euDist = 0.0;   
        numParticles = 0; 
       
        Iterator sortedIter = sortedList.iterator();
        while (sortedIter.hasNext())
        {
            PSOAgent p = (PSOAgent)sortedIter.next();
               
            sumPresent = 0.0;
            for (int j=0; j < dimension; j++)
                if (originalSPSO == true)
                    sumPresent += Math.pow(p.present[j] - s.present[j], 2);
                else
                    sumPresent += Math.pow(p.present[j] - s.pbest[j], 2); 
            euDist = Math.sqrt(sumPresent);
            //System.out.println("euDist : "+euDist);   
            
            // if Euclidean distance is less than the radius, then count the particle 
            if ((euDist < radius) && (p.getProcessed() != true))   
            {
                p.setProcessed(true);  
                numParticles++;                   
                                
                if ((numParticles) > speciesMax)
                {
                    int position = sortedList.indexOf(p);
                    PSOAgent child = new PSOAgent(this);
                    //System.out.println("one child created..");                 
                    //if (Random.uniform.nextDoubleFromTo(0, 1) < 0.5)
                    if ((Random.uniform.nextDoubleFromTo(0, 1) <= 0.5)  && method.equals("NeutralQuantum1")) 
                        child.setQuantump(true);
                    child.setProcessed(false);  
                    sortedList.set(position, child);                  
                }               
            }    
        } 
                  
     }              
     return sortedList;            
}  
         
          
            
 /**
   * Replace the worst species with randomly generated particles
   * to allow better exploration of the search space.
   * Similar to the idea of Blackwell and Branke (2005)
   */
 protected ArrayList replaceWSpecies(ArrayList aList)  {
  
    double sumPresent, euDist;
    ArrayList agentList = new ArrayList();
    
    //Iterator seedIter = seedList.iterator();   // only seeds
    agentList.addAll(aList);   
         
    //while (seedIter.hasNext())
    //{
        // get the last species seed
        PSOAgent s = (PSOAgent) seedList.get(seedList.size() - 1);
        
        // replace all particles in the last species
        for(int i=0; i < agentList.size(); i++)
        {
            PSOAgent p = (PSOAgent)agentList.get(i);
             
            sumPresent = 0.0;
            for (int j=0; j < dimension; j++)
                if (originalSPSO == true)
                    sumPresent += Math.pow(p.present[j] - s.present[j], 2);
                else
                    sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);
            euDist = Math.sqrt(sumPresent);
              
            // if Euclidean distance is less than the radius, then replace the particle with a new one
            if (euDist < radius) 
            {
                PSOAgent agent = new PSOAgent(this);
                // repeat until the newly generated particle is NOT within this seed's radius
                //euDist = 0.0;
                //while (euDist < radius)
                {
                    // generate a new particle
                    //agent = new PSOAgent(this);
                }
                //System.out.println("one replaced ");  
                agentList.set(i,agent);
            }
        }
   // }// while loop
             
    return agentList;
 }   




 /**
   * Repell a particle if the population diversity is too low
   * Followed the idea by Riget, J. and Vesterstrom, J.S. (2002)
   * "A Diversity-Guided Particle Swarm Optimizer - the ARPSO"
   */
 protected ArrayList repell(ArrayList aList)  {
  
    double sumPresent, euDist;
    ArrayList agentList = new ArrayList();
    
    Iterator seedIter = seedList.iterator();   // only seeds
    agentList.addAll(aList);   
   
    for(int i=0; i < agentList.size(); i++)
    {
        PSOAgent agent = new PSOAgent(this);
        sumPresent = 0.0;
        int sign;
        for (int k=0; k < dimension; k++)
        {
            agent.v[k] =  0.729844 * ( agent.v[k] - 2.05 * Random.uniform.nextDoubleFromTo(0, 1)        
              *( agent.pbest[k] - agent.present[k]) - 2.05 * Random.uniform.nextDoubleFromTo(0, 1)
              * (agent.seed[k] - agent.present[k]) );           
      
            // Adjust the velocity vector so that the agent cannot go
            // beyond the search space, preventing the population from explosion
            if (agent.v[k] > velocityMax) {
               agent.v[k] = velocityMax;
            }
            else if (agent.v[k] < -velocityMax) {
               agent.v[k] = -velocityMax;
            }                 
                        
            // Based on the new velocity, update the agent's position
            agent.present[k] = agent.present[k] + agent.v[k];
               
            // Adjust the position vector so that the agent cannot go
            // beyond the search space, preventing the population from explosion
            if (agent.present[k] > rightRange) {
                agent.present[k] = (agent.present[k] + rightRange)/2.0;
            }
            else if (agent.present[k] < leftRange) {
                agent.present[k] = (agent.present[k] + leftRange)/2.0;
            }
        }
        agentList.set(i,agent); 
    }//for
    return agentList;
}  


/**
   * Since particles in the initial popluation are tagged with 50% neutral and 50% quantum
   * quantum particles are identified in each species and then updated following the 
   * the quantumUpdate() rule. At meantime, the neutral particles follow the neutralUpdate()
   * rules as standard PSO. Both quantum and neutral particles are subject to the 
   * clearing procedure, where the best-fit from both types of particles will have their
   * opportunities to be selected as species seeds.
   */
 protected ArrayList neutralQuantum1(ArrayList aList, ArrayList seedList)  
 {
  
    double sumPresent=0.0, euDist=0.0, sumEuDist;
    ArrayList sortedList = new ArrayList(); 
    sortedList.addAll(aList);     
    
    //set all PROCESSED flags to false for all particles first
    for(int i=0; i<sortedList.size(); i++)
    {
        PSOAgent ps = (PSOAgent)sortedList.get(i);
        ps.setProcessed(false); 
    }

    Iterator seedIter = seedList.iterator();   // only seeds                 
    while (seedIter.hasNext())
    {
        PSOAgent s = (PSOAgent) seedIter.next();
        euDist = 0.0;            
                
        Iterator sortedIter = sortedList.iterator();
        while (sortedIter.hasNext())
        {
            PSOAgent p = (PSOAgent)sortedIter.next();
            //if (Random.uniform.nextDoubleFromTo(0, 1) < 0.5)
            //    p.setQuantump(true);
            sumPresent = 0.0;
            for (int j=0; j < dimension; j++)
                if (originalSPSO == true)
                    sumPresent += Math.pow(p.present[j] - s.present[j], 2);
                else
                    sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);
            euDist = Math.sqrt(sumPresent);
            //System.out.println("euDist : "+euDist);   
            // if Euclidean distance is less than the radius, then count the particle 
            if ((euDist < radius) && (p.getProcessed() != true))   
            {
                p.setProcessed(true); 
                for (int k=0; k < dimension; k++)
                   if (originalSPSO == true)
                        p.seed[k] = s.present[k];
                   else
                        p.seed[k] = s.pbest[k];                                  
            }
            
        }
     }        
     //sortedList = sortAListFit(sortedList);           
     return sortedList;            
}  



/**
   * Quantum swarm using a cloud distribution (Blackwell and Branke 2004) in their Multi-swarms
   * r_cloud of 1.0 is chosen, which is the amount of shift (displacement in the moving peak)
   * Below two quantum particles are created locally for each species seed (like "mutation")
   * the best-fit one is selected to replace its seed in the population.
   * Two types of update rules, neutralUpdate() and quantumUpdate(), are applied to the population
   */
 protected ArrayList twoQuantum(ArrayList aList, ArrayList seedList)  {

    double sumPresent=0.0, euDist=0.0;
    ArrayList sortedList = new ArrayList(); 
    sortedList.addAll(aList);  
    int numQuantum = 2;    
    double[] originalPbest = new double[dimension];

    //set all PROCESSED flags to false for all particles first
    for(int i=0; i<sortedList.size(); i++)
    {
        PSOAgent ps = (PSOAgent)sortedList.get(i);
        ps.setProcessed(false); 
    }

    Iterator seedIter = seedList.iterator();   // only seeds                 
    while (seedIter.hasNext())
    {
        PSOAgent s = (PSOAgent) seedIter.next();
        int seedPos = sortedList.indexOf(s);
        //System.out.println("seedPos :"+seedPos);
        
        for (int j=0; j < dimension; j++) 
            if (originalSPSO == true)
                originalPbest[j] = s.present[j];
            else
                originalPbest[j] = s.pbest[j];   
                  
        for(int i=0; i < numQuantum; i++)
        {
            PSOAgent p = new PSOAgent(this);      
            for (int k=0; k < dimension; k++) 
            { 
                //p.seed[k] = s.seed[k];
                p.seed[k] = s.pbest[k];
            }
            p.quantumUpdate();
            //p.fitnessValue = FktnLib.dummy_eval(p.present);   
            p.fitnessValue = FktnLib.eval_movpeaks(p.present);    
            if (p.fitnessValue > s.pbestValue) 
            {
                s.pbestValue = p.fitnessValue;
                for (int j=0; j < dimension; j++) 
                    s.pbest[j] = p.present[j];
            }
        }
        
        // updating all particles' seed in each species
        euDist = 0.0; 
                
        Iterator sortedIter = sortedList.iterator();
        while (sortedIter.hasNext())
        {
            PSOAgent p = (PSOAgent)sortedIter.next();
               
            sumPresent = 0.0;
            for (int j=0; j < dimension; j++)
                //sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);
                sumPresent += Math.pow(p.present[j] - originalPbest[j], 2);
            euDist = Math.sqrt(sumPresent);
            //if (FktnLib.get_number_of_evals() > 150000) 
            //    System.out.println("euDist : "+euDist);   
            // if Euclidean distance is less than the radius, then count the particle 
            if ((euDist < radius) && (p.getProcessed() != true))   
            {
                p.setProcessed(true); 
                for (int k=0; k < dimension; k++)
                    p.seed[k] = s.pbest[k];                                                
            }            
        }
        sortedList.set(seedPos, s);  //replacing with the updated seed  
    }    
    return sortedList;
}


/**
   * Quantum swarm using a cloud distribution (Blackwell and Branke 2004) in their Multi-swarms
   * r_cloud of 1.0 is chosen, which is the amount of shift (displacement in the moving peak)
   * Below two quantum particles are created locally for each species seed (like "mutation")
   * then added to the agentList.
   * Two types of update rules, neutralUpdate() and quantumUpdate(), are applied to the population
   */
 protected ArrayList addQuantums(ArrayList aList, ArrayList seedList)  {

    double sumPresent=0.0, euDist=0.0;
    ArrayList sortedList = new ArrayList(); 
    sortedList.addAll(aList);  
    int numQuantum = 5;    
    ArrayList sortedListNew = new ArrayList();
    sortedListNew.addAll(aList);

    //set all PROCESSED flags to false for all particles first
    for(int i=0; i<sortedList.size(); i++)
    {
        PSOAgent ps = (PSOAgent)sortedList.get(i);
        ps.setProcessed(false); 
    }

    Iterator seedIter = seedList.iterator();   // only seeds                 
    while (seedIter.hasNext())
    {
        PSOAgent s = (PSOAgent) seedIter.next();
        int seedPos = sortedList.indexOf(s);
        //System.out.println("seedPos :"+seedPos);
             
        for(int i=0; i < numQuantum; i++)
        {
            PSOAgent p = new PSOAgent(this);      
            for (int k=0; k < dimension; k++) 
            { 
                p.seed[k] = s.present[k];
                //p.v[k] = 0.0;  // set the velocity to 0.0, to stop the osillation       
                //p.v[k] = s.v[k]; // seed's velocity??
            }
            p.quantumUpdate();
            p.setQuantump(false);
            sortedListNew.add(p);
        }                
    } 
    return sortedListNew;
}




  /**
   * Find the converged swarms. Substitute the particles that are similar
   * to the species seed, in order to increase the local diversity
   */
 protected ArrayList localDiverse(ArrayList aList, ArrayList seedList)  
 {
	 System.out.println("BANGBANG");
	 
    double sumPresent=0.0, euDist=0.0;
    ArrayList sortedList = new ArrayList(); 
    sortedList.addAll(aList);      
    double speciesSize = 0.0;
    int countParticles = 0;
    int worstPos = 0;

    //set all PROCESSED flags to false for all particles first
    for(int i=0; i<sortedList.size(); i++)
    {
        PSOAgent ps = (PSOAgent)sortedList.get(i);
        ps.setProcessed(false); 
    }

    Iterator seedIter = seedList.iterator();   // only seeds                 
    while (seedIter.hasNext())
    {
        PSOAgent s = (PSOAgent) seedIter.next();
        
        speciesSize = 0;
        euDist = 0.0;   
        countParticles = 0;
       
        Iterator sortedIter = sortedList.iterator();
        while (sortedIter.hasNext())
        {
            PSOAgent p = (PSOAgent)sortedIter.next();
               
            sumPresent = 0.0;
            for (int j=0; j < getDimension(); j++)
                if (originalSPSO == true)
                    sumPresent += Math.pow(p.present[j] - s.present[j], 2);                
                else
                    sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);
            euDist = Math.sqrt(sumPresent);            
            //if (FktnLib.get_number_of_evals() > 150000) 
            //    System.out.println("euDist : "+euDist);   
            // if Euclidean distance is less than the radius, then count the particle 
            if ((euDist < radius) && (p.getProcessed() != true))   
            {                                              
                if (euDist >= speciesSize)
                {
                    speciesSize = euDist;
                    //worstPos = sortedList.indexOf(p);            
                }
                
                //p.setProcessed(true);  
                countParticles++;                        
            }
        }//while loop
        //System.out.println("speciesSize : "+speciesSize);      

        if ((speciesSize <= aThreshold)) // && (countParticles > 1))
        {                                           
            //System.out.println("countParticles .......................: "+countParticles);  
            //System.out.println("speciesSize : "+speciesSize); 
            int counts = 0;     
            Iterator sortedIter2 = sortedList.iterator();
            while (sortedIter2.hasNext())
            {
                PSOAgent p = (PSOAgent)sortedIter2.next(); 
                PSOAgent p2 = new PSOAgent(this);
                int pos = 0;
                sumPresent = 0.0;    

                for (int j=0; j < getDimension(); j++)      
                    if (originalSPSO == true)
                        sumPresent += Math.pow(p.present[j] - s.present[j], 2);                
                    else                                             
                        sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);                    
                euDist = Math.sqrt(sumPresent);                                              
                                                                                          
                // if Euclidean distance is less than the radius, then create a quantum cloud around the seed    
                if ((euDist < radius) && (countParticles > 1))  //(p.pbestValue != s.pbestValue))                                  
                {
                    pos = sortedList.indexOf(p);   
                    
                    for (int k=0; k < dimension; k++)                                                       
                    {
                        p2.seed[k] = s.seed[k];  
                        p2.v[k] = s.v[k];
                    }
                    
                    //if (Random.uniform.nextDoubleFromTo(0, 1) < 0.5)
                        p2.setQuantump(true);
                    
                    if (Random.uniform.nextDoubleFromTo(0, 1) < 0.5)
                    {
                        p2.quantumUpdate();
                        p2.setQuantump(false);
                    }
                                             
                    sortedList.set(pos, p2);  
                    p.setProcessed(true); 
                    //counts++;                                     
                }
                
            }  
       }//if              

     }//while loop              
     return sortedList;   
 }
 


//  /**
//   * Find the converged swarms. Substitute the particles that are similar
//   * to the species seed, in order to increase the local diversity
//   */
// protected ArrayList localDiverse2(ArrayList aList, ArrayList seedList)  
// {
//    double sumPresent=0.0, euDist=0.0;
//    ArrayList sortedList = new ArrayList(); 
//    sortedList.addAll(aList);      
//    double speciesSize = 0.0;
//    int countParticles = 0;
//    int worstPos = 0;
//
//    //set all PROCESSED flags to false for all particles first
//    for(int i=0; i<sortedList.size(); i++)
//    {
//        PSOAgent ps = (PSOAgent)sortedList.get(i);
//        ps.setProcessed(false); 
//    }
//
//    Iterator seedIter = seedList.iterator();   // only seeds                 
//    while (seedIter.hasNext())
//    {
//        PSOAgent s = (PSOAgent) seedIter.next();
//        
//        speciesSize = 0;
//        euDist = 0.0;   
//        countParticles = 0;
//       
//        Iterator sortedIter = sortedList.iterator();
//        while (sortedIter.hasNext())
//        {
//            PSOAgent p = (PSOAgent)sortedIter.next();
//               
//            sumPresent = 0.0;
//            for (int j=0; j < getDimension(); j++)
//                sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);
//            euDist = Math.sqrt(sumPresent);            
//
//            // if Euclidean distance is less than the radius AND aThreshold, then count the particle 
//            if ((euDist < radius) && (p.getProcessed() != true))   
//            {                                              
//                if (euDist < aThreshold) 
//                    countParticles++;
//                                
//                if (countParticles > speciesMax)
//                {             
//                    PSOAgent p2 = new PSOAgent(this);
//                    if (Random.uniform.nextDoubleFromTo(0, 1) < 0.5)
//                        p2.setQuantump(true);
//                    int pos = sortedList.indexOf(p); 
//                    sortedList.set(pos, p2);             
//                }
//                p.setProcessed(true);                        
//            }                        
//        }//while loop
//  
//     }//while loop              
//     return sortedList;   
// }
// 



  /**
   * Calculate the population diversity. Regardless the number of species,
   * if all species have converged, then this diversity value should be 0.0
   */
 protected double getDiversity(ArrayList aList)  
 {
    double sumPresent=0.0, euDist=0.0, sumEuDist, speciesDiversity=0.0, diversity=0.0;
    ArrayList sortedList = new ArrayList(); 
    sortedList.addAll(aList);      
    
    //set all PROCESSED flags to false for all particles first
    for(int i=0; i<sortedList.size(); i++)
    {
        PSOAgent ps = (PSOAgent)sortedList.get(i);
        ps.setProcessed(false); 
    }

    Iterator seedIter = seedList.iterator();   // only seeds                 
    while (seedIter.hasNext())
    {
        PSOAgent s = (PSOAgent) seedIter.next();
        
        speciesDiversity = 0.0;
        sumEuDist = 0.0;
        euDist = 0.0;    
        
        Iterator sortedIter = sortedList.iterator();
        while (sortedIter.hasNext())
        {
            PSOAgent p = (PSOAgent)sortedIter.next();
               
            sumPresent = 0.0;
            for (int j=0; j < getDimension(); j++)
                if (originalSPSO == true)
                    sumPresent += Math.pow(p.present[j] - s.present[j], 2);
                else
                    sumPresent += Math.pow(p.present[j] - s.pbest[j], 2);
            euDist = Math.sqrt(sumPresent);
            //System.out.println("euDist : "+euDist);   
            // if Euclidean distance is less than the radius, then count the particle 
            if ((euDist < radius) && (p.getProcessed() != true))   
            {
                //sumEuDist +=euDist;
                
                if (euDist >= speciesDiversity)
                {
                    speciesDiversity = euDist;
                    //worstPos = sortedList.indexOf(p);            
                }
 
                p.setProcessed(true);                              
            }
        }
        //System.out.println("specisDiversity : "+speciesDiversity); 
        diversity += speciesDiversity/(double)seedList.size();   
     }              
     return diversity;   
 }
 
 
    
  public int getNumParticles() {
    return numParticles;
  }

  public void setNumParticles(int num) {
    numParticles = num;
  }

  public double getLeftRange() {
    return leftRange;
  }

  public void setLeftRange(double lr) {
    leftRange = lr;
  }
 
  public double getRightRange() {
    return rightRange;
  }

  public void setRightRange(double rr) {
    rightRange = rr;
  }

  public double getVelocityMax() {
    return velocityMax;
  }

  public void setVelocityMax(double v) {
    velocityMax = v;
  }

  public int getMaxIteration() {
       return maxIteration;
   }

  public void setMaxIteration(int gen)  {
       maxIteration = gen;
  }

  public int getMaxNumEvals() {
       return maxNumEvals;
   }

  public void setMaxNumEvals(int mne)  {
       maxNumEvals = mne;
  }

  public int getDimension() {
      return dimension;
  }

  public void setDimension(int dim) {
      dimension = dim;
  }
  
  public double getRadius() {
    return radius;
  }

  public void setRadius(double r) {
    radius = r;
  }

  public String getFunction() {
       return function;
   }

   public void setFunction(String f) {
       function = f;
   }
 
   public double getaThreshold() {
       return aThreshold;
   }

   public void setaThreshold(double ath) {
       aThreshold = ath;
   }

   public int getRun() {
       return run;
   }

   public void setRun(int r) {
       run = r;
   }

   public int getSpeciesMax() {
       return speciesMax;
   }

   public void setSpeciesMax(int sm) {
       speciesMax = sm;
   }
   
    public double getDThreshold() {
       return dThreshold;
   }

   public void setDThreshold(double dt) {
       dThreshold = dt;
   }
   
   public double getDecayScale() {
       return decayScale;
   }

   public void setDecayScale(double dc) {
       decayScale = dc;
   }
   
  public boolean getBatch() {
    return batch;
  }

  public void setBatch(boolean b) {
     batch = b;
  }

  public String getMethod() {
    return method;
  }

  public void setMethod(String m) {
     method = m;
  }
  
  public double getRandomUniform() {
    return Random.uniform.nextDoubleFromTo(0, 1);
  }

  public double getGaussian() {
    return Random.normal.nextDouble();
  }
  
  public int getNumPeaks() {
    return numPeaks;
  }

  public void setNumPeaks(int np) {
    numPeaks = np;
  }
 
 
  /**
   * Get init parameters
   * @return
   */
  public String[] getInitParam() {
    String[] params =  {"NumParticles", "VelocityMax", "MaxIteration", "MaxNumEvals", "dThreshold",
            "Dimension", "Radius", "Run", "Batch", "Function", "aThreshold", "Method", "SpeciesMax", "DecayScale", "NumPeaks"};
    return params;
  }

   public static void main (String[] args) {
    SimInit init = new SimInit();
   
    PSOModel model = new PSOModel();
    init.loadModel(model,"",false);
  }

}
