package peaks;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Observable;
import java.util.Observer;

import org.seatiger.pso.ConstrictionPSO;
import org.seatiger.pso.DeJongF1;
import org.seatiger.pso.MovingPeaks;
import org.seatiger.pso.Particle;
import org.seatiger.pso.Swarm;
import org.seatiger.pso.VectorMath;

public class XDMovingPeaksMetric
{
	/** The moving peaks instance */
	private MovingPeaks peaks;
	/** The swarm being analysed */
	private PSOModel swarm;

	/** Whether BKPE should return fitness error (false = distance error) */
	private boolean useFitness = false;
	/** The cumulative percentage of known, non-hidden peaks */
	private double totalPercentKnownPeaks = 0;
	/** The amount of error to each peak since the last move */
	private double[] errsSinceLastMove;
	/** Whether each peak was known at every tick since the last move */
	private boolean[] knownPeaks;
//	/** The rank of each peak, with 0 being the highest peak */
//	private Integer[] peakRanks;
	/** The cumulative error for the best known peak error metric */
	private double bestKnownPeakTotalErr = 0;
	/** The distance between the best peak and the nearest particle to it */
	private double offlineDistErr = 0;
	/** The tick that the peaks last moved on */
	private int lastMoveTick = 0;
	/** The number of ticks for which at least one peak was known for the
	 * entire time between movements */
	private int numTicks = 0;

	/**
	 * Creates a new MovingPeaksMetric
	 *
	 * @param swarm The swarm being analysed
	 * @param peaks The moving peaks instance
	 */
	public XDMovingPeaksMetric(PSOModel swarm, MovingPeaks peaks)
	{
		this.swarm = swarm;
		this.peaks = peaks;
		AgentParticle.initialise(swarm.getDimension());

		Listener listener = new Listener();
		peaks.getPeakChangeObservable().addObserver(listener);

		errsSinceLastMove = new double[peaks.getNumPeaks()];
		knownPeaks = new boolean[peaks.getNumPeaks()];
//		peakRanks = new Integer[peaks.getNumPeaks()];

		Arrays.fill(knownPeaks, true);

//		for (int i = 0; i < peakRanks.length; i++)
//		peakRanks[i] = new Integer(i);
//		calculatePeakRanks();
	}




	/**
	 * Returns whether the BKPE metric uses the fitness error instead of the positional
	 * error
	 * 
	 * @return True if the BKPE metric is using the fitness error
	 */
	public boolean isUseFitness()
	{
		return useFitness;
	}

	/** 
	 * Controls whether the BKPE metric uses the fitness error instead of the positional
	 * error
	 * 
	 * @param useFitness True to use the fitness error instead of the positional error
	 */ 
	public void setUseFitness(boolean useFitness)
	{
		this.useFitness = useFitness;
	}




	/**
	 * Returns the offline error
	 * 
	 * @return The offline error
	 */
	public double getOfflineError()
	{
		return peaks.get_offline_error();
	}

	/**
	 * Returns the current error
	 * 
	 * @return The current error
	 */
	public double getCurrentError()
	{
		return peaks.get_current_error();
	}

	/**
	 * Returns the offline distance error, that is the average distance
	 * between the best peak and the closest particle to it
	 * 
	 * @return The offline distance error
	 */
	public double getOfflineDistanceError()
	{
		return offlineDistErr / (int) swarm.getTickCount();
	}

	/**
	 * Returns the average number of peaks known about
	 * 
	 * @return THe average number of peaks known about
	 */
	public double getPCP()
	{
		return totalPercentKnownPeaks / (int) swarm.getTickCount();
	}

	/**
	 * Returns the average error for the best known peak. This measures the
	 * convergence speed of the algorithm
	 * 
	 * @return The average error for the best known peak
	 */
	public double getBKPE()
	{
		return bestKnownPeakTotalErr / numTicks;
	}

	/**
	 * Updates the {@link #totalPercentKnownPeaks} variable
	 * 
	 * @param distances The distance between each peak and the closest pBest in its
	 * catchment area
	 */
	private void updatePercentKnownPeaks(double[] distances)
	{
		int numKnownPeaks = 0;
		int numVisiblePeaks = 0;

		for (double dist : distances)
		{
			if (dist == MovingPeaks.HIDDEN)
				continue;

			numVisiblePeaks++;
			if (dist != MovingPeaks.NO_PARTICLES)
				numKnownPeaks++;
		}

		if (numVisiblePeaks == 0)
			totalPercentKnownPeaks++;
		else
			totalPercentKnownPeaks += numKnownPeaks / (double) numVisiblePeaks;
	}

//	/**
//	* Recalculates the height rank of each peak
//	*/
//	private void calculatePeakRanks()
//	{
//	final double[] peakHeights = peaks.getPeakHeights();

//	Arrays.sort(peakRanks, new Comparator<Integer>() {
//	public int compare(Integer a, Integer b)
//	{
//	// We want the highest peaks ranked first, so reverse
//	// the normal comparison order
//	if (peakHeights[a] > peakHeights[b])
//	return -1;
//	else if (peakHeights[a] < peakHeights[b])
//	return 1;
//	else
//	return 0;
//	}  
//	});
//	}

	/**
	 * Updates the metrics
	 */
	public void updateMetrics()
	{
		ArrayList<Particle> particles = new ArrayList<Particle>(swarm.agentList.size());
		for (Object agent : swarm.agentList)
			particles.add(new AgentParticle((PSOAgent) agent));
		
		
		double[] peakErrors;

		if (useFitness)
			peakErrors = peaks.getFitnessDeltasToPeaks(particles);
		else
			peakErrors = peaks.getDistancesToPeaks(particles);

		updatePercentKnownPeaks(peakErrors);
		for (int i = 0; i < peakErrors.length; i++)
		{
			if (peakErrors[i] == MovingPeaks.NO_PARTICLES)
				knownPeaks[i] = false;
			else
				errsSinceLastMove[i] += peakErrors[i];
		}

		Particle closestParticle = peaks.getClosestToBestPeak(particles);
		offlineDistErr += VectorMath.distance(closestParticle.getPersonalBestLocation(), 
				peaks.getPeakPositions()[peaks.getMaximumPeak()]);
	}

	/**
	 * Returns the current error to the given peak
	 * 
	 * @param peakNum The index of the peak to get the error of
	 * @return The error to the given peak
	 */
	public double getCurrError(int peakNum)
	{		
		ArrayList<Particle> particles = new ArrayList<Particle>(swarm.agentList.size());
		for (Object agent : swarm.agentList)
			particles.add(new AgentParticle((PSOAgent) agent));
		
		if (useFitness)
			return peaks.getFitnessDeltasToPeaks(particles)[peakNum];
		else
			return peaks.getDistancesToPeaks(particles)[peakNum];
	}

	/** 
	 * Notification that the peaks have moved
	 */
	private void peaksMoved()
	{
		int bestPeak = getBestKnownPeak();

		if (bestPeak != -1)
		{
			bestKnownPeakTotalErr += errsSinceLastMove[bestPeak];
			numTicks += (int) swarm.getTickCount() - lastMoveTick;
		}

		lastMoveTick = (int) swarm.getTickCount();

		Arrays.fill(errsSinceLastMove, 0);
		Arrays.fill(knownPeaks, true);
		//calculatePeakRanks();
	}

	/**
	 * Returns the highest peak continually covered since the last peak movement
	 * 
	 * @return The index of the highest known peak, or -1 if there were no continually
	 * covered peaks
	 */
	public int getBestKnownPeak()
	{
		int bestPeak = -1;
		double bestFitness = Double.NEGATIVE_INFINITY;
		
		ArrayList<Particle> particles = new ArrayList<Particle>(swarm.agentList.size());
		for (Object agent : swarm.agentList)
			particles.add(new AgentParticle((PSOAgent) agent));
		
//		double[] peakHeights = peaks.getPeakHeights();
//		for (int i = 0; i < peakHeights.length; i++)
//		{
//		if (knownPeaks[i] && !peaks.isPeakHidden(i) && 
//		(bestPeak == -1 || peakHeights[i] > peakHeights[bestPeak]))
//		bestPeak = i;
//		}
		for (Particle p : particles)
		{
			int peakNum = peaks.getPeakAt(p.getPersonalBestLocation());
			if (!knownPeaks[peakNum] && peaks.isPeakHidden(peakNum))
				continue;

			double fitness = peaks.dummy_eval(p.getPersonalBestLocation());

			if (fitness > bestFitness)
			{
				bestFitness = fitness;
				bestPeak = peakNum;
			}
		}

		return bestPeak;
	}

	/**
	 * Listens for external events
	 * 
	 * @author Stefan Bird <stbird@seatiger.org>
	 */
	private class Listener implements Observer
	{
		public void update(Observable o, Object arg)
		{
			peaksMoved();
		}
	}
	
	/**
	 * Makes an PSOAgent look like a Particle
	 * 
	 * @author scbird
	 */
	private static class AgentParticle extends Particle
	{
		private static final Swarm swarm;
		
		private PSOAgent delegate;
		
		static
		{
			swarm = new Swarm();
			swarm.setDefaultMovementModel(new ConstrictionPSO(swarm));
		}
		
		public static void initialise(int numPeaks)
		{
			swarm.addFitnessFunction(new DeJongF1(numPeaks));
		}
		
		public AgentParticle(PSOAgent delegate)
		{
			super(delegate.present, delegate.v, swarm, swarm.getDefaultMovementModel());
			this.delegate = delegate;
			swarm.removeParticle(this);
		}

		@Override
		public double[] getFitness() {
			return new double[] {delegate.fitnessValue};
		}

		@Override
		public double[] getLocation() {
			return delegate.present;
		}

		@Override
		public double[] getPersonalBestFitness() {
			return new double[] {delegate.pbestValue};
		}

		@Override
		public double[] getPersonalBestLocation() {
			return delegate.pbest;
		}

		@Override
		public double[] getVelocity() {
			return delegate.v;
		}
	}
}



